#include "albert.h"

const GRAPH_RGBCOLOR rgb_black   = {0.0,0.0,0.0};
const GRAPH_RGBCOLOR rgb_white   = {1.0,1.0,1.0};
const GRAPH_RGBCOLOR rgb_red     = {1.0,0.0,0.0};
const GRAPH_RGBCOLOR rgb_green   = {0.0,1.0,0.0};
const GRAPH_RGBCOLOR rgb_blue    = {0.0,0.0,1.0};
const GRAPH_RGBCOLOR rgb_yellow  = {1.0,1.0,0.0};
const GRAPH_RGBCOLOR rgb_magenta = {1.0,0.0,1.0};
const GRAPH_RGBCOLOR rgb_cyan    = {0.0,1.0,1.0};
const GRAPH_RGBCOLOR rgb_grey50  = {0.5,0.5,0.5};

#include <GL/glx.h>
#include <GL/gl.h>
#include <unistd.h>

struct ogl_window
{
  Display      *dpy;
  Window       win;
  GLXContext   cx;
  unsigned int width, height;
};
typedef struct ogl_window OGL_WINDOW;
static Bool       opengl_available = True;

/****************************************************************************/

static Bool WaitForNotify(Display *d, XEvent *e, char *arg)
{
  return (e->type == MapNotify) && (e->xmap.window == (Window)arg);
}

/****************************************************************************/

static Bool OGL_set_std_window(OGL_WINDOW *ogl_win)
{
  if (ogl_win) 
    return(glXMakeCurrent(ogl_win->dpy, ogl_win->win, ogl_win->cx));
  else
    return(false);
}

static OGL_WINDOW *OGL_create_window(const char *name, const char *geometry)
{
  FUNCNAME("OGL_create_window");
  static Display     *dpy = nil;
  static XVisualInfo *vi  = nil;
  static Colormap    cmap;
  static GLXContext  cx;
  static char        *default_geometry = "=300x300+0+0";
  static int         attributeList[] = { GLX_RGBA, None };
  XSetWindowAttributes swa;
  OGL_WINDOW         *ogl_win;
  Window             win, root;
  XSizeHints         size_hints;
  XEvent             event;
  unsigned int       bwidth, depth;
  int                status, x, y;
  
  if (!dpy) /* get a connection */
  {
    if (!(dpy = XOpenDisplay(nil))) 
    {
      MSG("cannot connect to display.\n");
      return(nil);
    }

    opengl_available = glXQueryExtension(dpy, NULL, NULL);
    if (!opengl_available)
    {
      MSG("no OpenGL available...\n");
      return(nil);
    }
  }

  
  if (!vi) /* get an appropriate visual */
  {
    vi = glXChooseVisual(dpy, DefaultScreen(dpy), attributeList);
    if (!vi) 
    {
      MSG("no visual\n");
      return(nil);
    }
    cmap = XCreateColormap(dpy, RootWindow(dpy, vi->screen),
			   vi->visual, AllocNone);
    if (!cmap) 
    {
      MSG("no color map\n");
      return(nil);
    }
  }
  

  size_hints.flags = PResizeInc | PMinSize ;
  size_hints.width_inc  =  1;
  size_hints.height_inc =  1;
  size_hints.min_width  = 10;
  size_hints.min_height = 10;

  if (geometry) 
  {
    status = XGeometry(dpy, vi->screen, geometry, default_geometry,
		       2, 1, 1, 0, 0, 
		       &(size_hints.x), &(size_hints.y),
		       &(size_hints.width), &(size_hints.height));
    if (status & (XValue|YValue)) {
      size_hints.flags |= USPosition;
      size_hints.flags &= ~PPosition;
    }
    if (status & (WidthValue|HeightValue)) 
    {
      size_hints.flags |= USSize;
      size_hints.flags &= ~PSize;
    }
  }

  if (size_hints.flags & USSize) 
  {
    size_hints.flags |= PAspect;
    size_hints.min_aspect.x = size_hints.width;
    size_hints.min_aspect.y = size_hints.height;
    size_hints.max_aspect.x = size_hints.width;
    size_hints.max_aspect.y = size_hints.height;
  }

  /* create a window */
  swa.colormap = cmap;
  swa.border_pixel = 0;
  swa.event_mask = StructureNotifyMask;
  win = XCreateWindow(dpy, RootWindow(dpy, vi->screen),
		      size_hints.x, size_hints.y,
		      size_hints.width, size_hints.height,
		      0, vi->depth, InputOutput, vi->visual,
		      CWBorderPixel|CWColormap|CWEventMask, &swa);
  if (!win) 
  {
    MSG("no window\n");
    return(nil);
  }

  XSetStandardProperties(dpy, win, name, name, None, NULL, 0, &size_hints);

  XMapWindow(dpy, win);
  XIfEvent(dpy, &event, WaitForNotify, (char*)win);


  /* create a GLX context */
  cx = glXCreateContext(dpy, vi, 0, GL_TRUE);
  if (!cx) 
  {
    MSG("no GLX context\n");
    return(nil);
  }

  /* connect the context to the window */
  glXMakeCurrent(dpy, win, cx);

  /* clear the buffer */
  glClearColor(0,0,0,1);
  glClear(GL_COLOR_BUFFER_BIT);
  glFlush();

  ogl_win = MEM_ALLOC(1, OGL_WINDOW);
  ogl_win->dpy = dpy;
  ogl_win->win = win;
  ogl_win->cx  = cx;
  XGetGeometry(dpy, win, &root, &x, &y,
	       &ogl_win->width, &ogl_win->height,
	       &bwidth, &depth);
  
  return(ogl_win);
}

/****************************************************************************/

static void OGL_destroy_window(OGL_WINDOW *ogl_win)
{
  if (!ogl_win) return;

  glXDestroyContext(ogl_win->dpy, ogl_win->cx);

  XDestroyWindow(ogl_win->dpy, ogl_win->win);
  XFlush(ogl_win->dpy);

  MEM_FREE(ogl_win, 1, OGL_WINDOW);

  return;
}

/****************************************************************************/

static void OGL_clear_window(OGL_WINDOW *ogl_win, const GRAPH_RGBCOLOR c)
{
  if (!ogl_win) return;

  if (OGL_set_std_window(ogl_win))
  {
    Window       root;
    unsigned int w, h, depth, bwidth;
    int          x, y;

/*--- has the window size changed?                                       ---*/
    XGetGeometry(ogl_win->dpy, ogl_win->win, &root, &x, &y, &w, &h, &bwidth,
		 &depth);
 
    if (w != ogl_win->width || h != ogl_win->height)
    {
      ogl_win->width  = w;
      ogl_win->height = h;
      XResizeWindow(ogl_win->dpy, ogl_win->win, w, h);
    }

    glXMakeCurrent(ogl_win->dpy, ogl_win->win, ogl_win->cx);

    glViewport(0, 0, w, h);
    if (c)
      glClearColor(c[0], c[1], c[2], 1.0);
    else
      glClearColor(0.0, 0.0, 0.0, 1.0);

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glXWaitGL();
  }
  return;
}

void graph_close_window(GRAPH_WINDOW win)
{
  OGL_destroy_window((OGL_WINDOW *)win);
  return;
}

/*--------------------------------------------------------------------------*/
/*---  get min and max coordinates                                       ---*/
/*--------------------------------------------------------------------------*/

#if DIM_OF_WORLD > 1
static float xmin[DIM_OF_WORLD], xmax[DIM_OF_WORLD], diam[DIM_OF_WORLD];
#else
static float xmin[2], xmax[2];
#endif

static void xminmax_fct(const EL_INFO *el_info)
{
  FUNCNAME("xminmax_fct");
  int        i, j;
  PARAMETRIC *parametric = el_info->mesh->parametric;
  
  if (parametric && parametric->init_element(el_info, parametric)) 
  {
    const REAL lambda[N_VERTICES][DIM+1] = {{1,0},{0,1}};
    REAL_D     world[N_VERTICES];

    parametric->coord_to_world(el_info, nil, N_VERTICES, lambda, world);

    for (i = 0; i < N_VERTICES; i++) 
    {
      for (j = 0; j < DIM_OF_WORLD; j++)
      {
	xmin[j] = MIN(xmin[j], world[i][j]);
	xmax[j] = MAX(xmax[j], world[i][j]);
      }
    }
  }
  else 
  {
    TEST_FLAG(FILL_COORDS, el_info);

    for (i = 0; i < N_VERTICES; i++)
    {
      for (j = 0; j < DIM_OF_WORLD; j++)
      {
	xmin[j] = MIN(xmin[j], el_info->coord[i][j]);
	xmax[j] = MAX(xmax[j], el_info->coord[i][j]);
      }    
    }
  }

  return;
}

GRAPH_WINDOW graph_open_window(const char *title, const char *geom, 
			       REAL *world, MESH *mesh)
{
  OGL_WINDOW *winO = nil;
  int        i;
  char       geometry[16];
#if DIM_OF_WORLD > 1
  float      xdiam[DIM_OF_WORLD];
#else
  float      xdiam[2];
#endif

  if (world)
  {
    xmin[0] = world[0];
    xmax[0] = world[1];
    xmin[1] = world[2];
    xmax[1] = world[3];
    xdiam[0] = MAX(xmax[0] - xmin[0], 1.E-10);
    xdiam[1] = MAX(xmax[1] - xmin[1], 1.E-10);
  }
  else if (mesh)
  {
    for (i = 0; i < DIM_OF_WORLD; i++)
      xmax[i] = -(xmin[i] = 1.0E10);
    mesh_traverse(mesh, -1, CALL_LEAF_EL|FILL_COORDS, xminmax_fct);

#if DIM_OF_WORLD == 1
    xmin[1] = -1.0;
    xmax[1] = 1.0;
#endif

    for (i = 0; i < DIM_OF_WORLD; i++)
    {
      xdiam[i] = MAX(xmax[i] - xmin[i], 1.E-10);
      xmin[i] -= 0.1*xdiam[i];
      xmax[i] += 0.1*xdiam[i];
      xdiam[i] *= 1.2;
    }
  }
  else
  {
    for (i = 0; i < DIM_OF_WORLD; i++)
    {
      xmin[i] = -1.0;
      xmax[i] =  1.0;
      xdiam[i] =  2.0;
    }
  }

  if (!title) title = "ALBERT graphics";

  if (!geom)
  {
    int size = 400, xsize, ysize;

    if (xdiam[0] >= xdiam[1])
    {
      xsize = size;
      ysize = size*xdiam[1]/xdiam[0];
    }
    else
    {
      xsize = size*xdiam[0]/xdiam[1];
      ysize = size;
    }
    sprintf(geometry, "%dx%d", xsize, ysize);
    geom = geometry;
  }

  if ((winO = OGL_create_window(title, geom)))
  {
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glOrtho(xmin[0], xmax[0], xmin[1], xmax[1], -1.0, 1.0);

    return((GRAPH_WINDOW) winO);
  }

  return(nil);
}

void graph_clear_window(GRAPH_WINDOW win, const GRAPH_RGBCOLOR c)
{
  OGL_clear_window((OGL_WINDOW *)win, c);
  return;
}

void graph_mesh(GRAPH_WINDOW win, MESH *mesh, const GRAPH_RGBCOLOR c, FLAGS flag)
{
  TRAVERSE_STACK  *stack;
  EL_INFO         *el_info;
  double          coord[2];
  int             n;
  const float     *color;

  if (!win || !mesh)  return;

  OGL_set_std_window((OGL_WINDOW *) win);

#if DIM_OF_WORLD == 1
  coord[1] = 0.0;
#endif

  glLineWidth(1);
  glPointSize(3);

  stack  = get_traverse_stack();
  el_info = traverse_first(stack, mesh, -1, CALL_LEAF_EL|FILL_COORDS|FILL_BOUND);
  while (el_info)
  {
    color = c ? c : rgb_black;
    if (flag & GRAPH_MESH_ELEMENT_MARK)
    {
      if (el_info->el->mark > 0)
	color = rgb_red;
      else if (el_info->el->mark < 0)
	color = rgb_blue;
    }

    glColor3fv(color);
    glBegin(GL_LINE_STRIP);
    for (n = 0; n < DIM_OF_WORLD; n++)
      coord[n] = el_info->coord[0][n];
    glVertex2dv(coord);
    for (n = 0; n < DIM_OF_WORLD; n++)
      coord[n] = el_info->coord[1][n];
    glVertex2dv(coord);
    glEnd();

/*--------------------------------------------------------------------------*/
/*--- draw element boundaries                                            ---*/
/*--------------------------------------------------------------------------*/

    if (el_info->bound[0])
    {
      for (n = 0; n < DIM_OF_WORLD; n++)
	coord[n] = el_info->coord[0][n];

      if (el_info->bound[0] > 0)
	glColor3fv(rgb_blue);
      else if (el_info->bound[0] < 0)
	glColor3fv(rgb_red);

      glPointSize(5);
      glBegin(GL_POINTS);
      glVertex2dv(coord);
      glEnd();
    }

    for (n = 0; n < DIM_OF_WORLD; n++)
      coord[n] = el_info->coord[1][n];

    if (el_info->bound[1] > 0)
      glColor3fv(rgb_blue);
    else if (el_info->bound[1] < 0)
      glColor3fv(rgb_red);
    else
      glColor3fv(c ? c : rgb_black);

    glBegin(GL_POINTS);
    glVertex2dv(coord);
    glEnd();

    el_info = traverse_next(stack, el_info);
  }
  free_traverse_stack(stack);

  glFlush();
  return;
}

void graph_drv(GRAPH_WINDOW win, const DOF_REAL_VEC *uh, 
	       const GRAPH_RGBCOLOR c, REAL min, REAL max)
{
  TRAVERSE_STACK  *stack;
  EL_INFO         *el_info;
  MESH            *mesh;
  double          coord[2], scal;
  const REAL      *(*get_uh_el)(const EL *el, const DOF_REAL_VEC *, REAL *);

  if (!win || !uh)  return;

  if (min >= max)
  {
    min = dof_min(uh);
    max = dof_max(uh);
  }
  min = ABS(min);
  max = ABS(max);

  scal = min == max ? 0.9 : 0.9/MAX(min,max);

  OGL_set_std_window((OGL_WINDOW *) win);

  glLineWidth(1);
  glColor3fv(c ? c : rgb_black);

  glBegin(GL_LINE_STRIP);
  mesh = uh->fe_space->mesh;
  get_uh_el = uh->fe_space->bas_fcts->get_real_vec;

  stack  = get_traverse_stack();
  el_info = traverse_first(stack, mesh, -1, CALL_LEAF_EL|FILL_COORDS);
  while (el_info)
  {
    const REAL *uh_el = (*get_uh_el)(el_info->el, uh, nil);

    coord[0] = el_info->coord[0][0];
    coord[1] = scal*uh_el[0];
    glVertex2dv(coord);
    coord[0] = el_info->coord[1][0];
    coord[1] = scal*uh_el[1];
    glVertex2dv(coord);

    el_info = traverse_next(stack, el_info);
  }
  free_traverse_stack(stack);
  glEnd();
  glFlush();
  return;
}
