/****************************************************************************/
/* ALBERT:   an Adaptive multi Level finite element toolbox using           */
/*           Bisectioning refinement and Error control by Residual          */
/*           Techniques                                                     */
/*                                                                          */
/* file:                                                                    */
/*                                                                          */
/*                                                                          */
/* description:                                                             */
/*                                                                          */
/*                                                                          */
/* history:                                                                 */
/*                                                                          */
/*                                                                          */
/****************************************************************************/
/*                                                                          */
/*        authors: Alfred Schmidt and Kunibert G. Siebert                   */
/*                 Institut f"ur Angewandte Mathematik                      */
/*                 Albert-Ludwigs-Universit"at Freiburg                     */
/*                 Hermann-Herder-Str. 10                                   */
/*                 79104 Freiburg                                           */
/*                 Germany                                                  */
/*                                                                          */
/*        email:   alfred,kunibert@mathematik.uni-freiburg.de               */
/*                                                                          */
/*     (c) by A. Schmidt and K.G. Siebert (1996,97)                         */
/*                                                                          */
/****************************************************************************/
#include "albert.h"

/****************************************************************************/
/*  calculates the transformation of the DIM_OF_WORLD x DIM_OF_WORLD matrix */
/*  B to the standard simplex: A^t B A of size DIM+1 x DIM+1;               */
/*  the matrix A^t B A is additionally scaled by a factor of 2*|T|, thus no */
/*  additional factor will appear in the quadruature!                       */
/****************************************************************************/

REAL get_transformed_el_mat(const EL_INFO *el_info, REAL mat[DIM+1][DIM+1])
{
  REAL       e1[DIM_OF_WORLD], e2[DIM_OF_WORLD];
  REAL       a11, a12, a22, det, det2;
  const REAL *v0;
  int        i;

/****************************************************************************/
/*  A = transformation from unit triangle to element triangle               */
/*  compute A^T*A in (a11,a12,a22)                                          */
/****************************************************************************/
  a11 = a12 = a22 = 0.0;
  v0 = el_info->coord[0];
  for (i = 0; i < DIM_OF_WORLD; i++)
  {
    e1[i] = el_info->coord[1][i] - v0[i];
    e2[i] = el_info->coord[2][i] - v0[i];
    a11 += e1[i] * e1[i];
    a12 += e1[i] * e2[i];
    a22 += e2[i] * e2[i];
  }

#if DIM_OF_WORLD==2
  det = e1[0]*e2[1] - e1[1]*e2[0];
  det = ABS(det);
#else
#if DIM_OF_WORLD==3
  {
    REAL n[3];

    MSG(" not tested up to now...\n");

    n[0] = e1[1]*e2[2] - e1[2]*e2[1];
    n[1] = e1[2]*e2[0] - e1[0]*e2[2];
    n[2] = e1[0]*e2[1] - e1[1]*e2[0];
    det = sqrt(n[0]*n[0] + n[1]*n[1] + n[1]*n[1]);
  }
#else
  TEST_EXIT(0)("not yet for DIM_OF_WORLD = %d",DIM_OF_WORLD);
#endif
#endif

  det2 = 1.0 /det;

  mat[1][1] = a22 * det2;
  mat[1][2] = mat[2][1] = -a12 * det2;
  mat[2][2] = a11 * det2;
  
  mat[0][1] = mat[1][0] = -(mat[1][1] + mat[2][1]);
  mat[0][2] = mat[2][0] = -(mat[1][2] + mat[2][2]);
  mat[0][0] = -(mat[0][1] + mat[0][2]);
  
  return(det);
}

/****************************************************************************/
/*  compute volume of an element                                            */
/****************************************************************************/

REAL el_det(const EL_INFO *el_info)
{
  FUNCNAME("el_det");
  REAL        e1[DIM_OF_WORLD], e2[DIM_OF_WORLD], det;
  const REAL  *v0;
  int         i;

  TEST_FLAG(FILL_COORDS, el_info);

  v0 = el_info->coord[0];
  for (i = 0; i < DIM_OF_WORLD; i++)
  {
    e1[i] = el_info->coord[1][i] - v0[i];
    e2[i] = el_info->coord[2][i] - v0[i];
  }

#if DIM_OF_WORLD==2
  det = e1[0]*e2[1] - e1[1]*e2[0];
  det = ABS(det);
#else
#if DIM_OF_WORLD==3
  {
    REAL n[3];

    MSG("not tested yet\n");
    n[0] = e1[1]*e2[2] - e1[2]*e2[1];
    n[1] = e1[2]*e2[0] - e1[0]*e2[2];
    n[2] = e1[0]*e2[1] - e1[1]*e2[0];
    det = sqrt(n[0]*n[0] + n[1]*n[1] + n[1]*n[1]);
  }
#else
  TEST_EXIT(0)("not yet for DIM_OF_WORLD = %d",DIM_OF_WORLD);
#endif
#endif

  return(det);
}

#define VOL_SHAT 0.5

REAL el_volume(const EL_INFO *el_info)
{
  return(VOL_SHAT*el_det(el_info));
}

/****************************************************************************/
/*  compute gradients of basis functions on element; return the absulute    */
/*  value of the determinante from the transformation to the reference      */
/*  element                                                                 */
/****************************************************************************/

REAL el_grd_lambda(const EL_INFO *elinfo, REAL grd_lam[DIM+1][DIM_OF_WORLD])
{
  FUNCNAME("el_grd_lambda");
  int         i, j;
  REAL        e1[DIM_OF_WORLD], e2[DIM_OF_WORLD], det, adet;
  const REAL  *v0;
  REAL        a11, a12, a21, a22;

  TEST_FLAG(FILL_COORDS, elinfo);

  v0 = elinfo->coord[0];
  for (i = 0; i < DIM_OF_WORLD; i++) {
    e1[i] = elinfo->coord[1][i] - v0[i];
    e2[i] = elinfo->coord[2][i] - v0[i];
  }

#if DIM_OF_WORLD == 2
  det = e1[0] * e2[1] - e1[1] * e2[0];
  adet = ABS(det);
  if (adet < 1.0E-25) 
  {
    MSG("abs(det) = %lf\n", adet);
    for (i = 0; i <= DIM; i++)
      for (j = 0; j < DIM_OF_WORLD; j++)
	grd_lam[i][j] = 0.0;
  }
  else
  {
    det = 1.0 / det;
    a11 =  e2[1] * det;       /* (a_ij) = A^{-T} */
    a21 = -e2[0] * det;
    a12 = -e1[1] * det;
    a22 =  e1[0] * det;

    grd_lam[1][0] = a11;
    grd_lam[1][1] = a21;
    grd_lam[2][0] = a12;
    grd_lam[2][1] = a22;
    grd_lam[0][0] = - grd_lam[1][0] - grd_lam[2][0];
    grd_lam[0][1] = - grd_lam[1][1] - grd_lam[2][1];
  }

#if 0
  MSG("el %d : vol = %8.6lf ----------------\n",
      elinfo->el->index, vol);
  for (i = 0; i < N_VERTICES; i++)
    MSG("vertex     %1d = (%8.4lf, %8.4lf)\n",i,
	elinfo->coord[i][0], elinfo->coord[i][1]);
  for (i = 0; i < N_VERTICES; i++)
    MSG("grd_lam %1d = (%8.4lf, %8.4lf)\n",i,
	grd_lam[i][0], grd_lam[i][1]);
  
#endif

#else
  ERROR_EXIT("not yet for DIM_OF_WORLD = %d\n", DIM_OF_WORLD);
#endif

  return(adet);
}

#if DIM_OF_WORLD == 3
/****************************************************************************/
/*  vector product in R^3                                                   */
/****************************************************************************/

void vector_product(const REAL_D x, const REAL_D y, REAL_D z)
{
  FUNCNAME("vector_product");

  z[0] = x[1]*y[2] - x[2]*y[1];
  z[1] = x[2]*y[0] - x[0]*y[2];
  z[2] = x[0]*y[1] - x[1]*y[0];

  return;
}
#endif

/****************************************************************************/
/*  calculate a facenormal of edge side of a traingle with coordinates      */
/*  coord; return the absulute value of the determinante from the           */
/*  transformation to the reference element                                 */
/****************************************************************************/

REAL get_face_normal(const REAL_D coord[N_VERTICES], int side, REAL *normal)
{
  FUNCNAME("get_face_normal");
  REAL   det;
  int    i0, i1;

#if DIM_OF_WORLD == 2
  i0 = (side+1)%3;
  i1 = (side+2)%3;
  normal[0] = coord[i1][1] - coord[i0][1];
  normal[1] = coord[i0][0] - coord[i1][0];
  det = sqrt(SQR(normal[0])+SQR(normal[1]));
  TEST_EXIT(det > 1.e-30)("det = 0 on face %d\n", side);
  normal[0] /= det;
  normal[1] /= det;
#else 
  MSG("not implemented for DIM_OF_WORLD = %d in 2d\n", DIM_OF_WORLD);
  for (i0 = 0; i0 < DIM_OF_WORLD; i0++)  normal[i0] = 0.0;
#endif

  return(det);
}

/*--------------------------------------------------------------------------*/
/* orient the vertices of edges/faces                                       */
/* used by estimator for the jumps => same quadrature nodes from both sides!*/
/*--------------------------------------------------------------------------*/

const int *sort_face_indices(const EL *el, int face, int *vec)
{
  static int  sorted[3][2][2] = {
    {{1,2}, {2,1}}, {{2,0},{0,2}}, {{0,1}, {1,0}}};
  static int  vertex_of_face[3][2] = {{1,2}, {2,0}, {0,1}};
  int    i, no, *val  = nil;
  int    *vof = vertex_of_face[face];
  DOF    **dof = el->dof;

  if (dof[vof[0]][0] < dof[vof[1]][0])
    no = 0;
  else
    no = 1;


  if (vec)
  {
    val = vec;
    for (i = 0; i < 2; i++)
      val[i] = sorted[face][no][i];
  }
  else
    val = sorted[face][no];
 
  return((const int *) val);
}
