/****************************************************************************/
/* ALBERT:   an Adaptive multi Level finite element toolbox using           */
/*           Bisectioning refinement and Error control by Residual          */
/*           Techniques                                                     */
/*                                                                          */
/* file:                                                                    */
/*                                                                          */
/*                                                                          */
/* description:                                                             */
/*                                                                          */
/*                                                                          */
/* history:                                                                 */
/*                                                                          */
/*                                                                          */
/****************************************************************************/
/*                                                                          */
/*        authors: Alfred Schmidt and Kunibert G. Siebert                   */
/*                 Institut f"ur Angewandte Mathematik                      */
/*                 Albert-Ludwigs-Universit"at Freiburg                     */
/*                 Hermann-Herder-Str. 10                                   */
/*                 79104 Freiburg                                           */
/*                 Germany                                                  */
/*                                                                          */
/*        email:   alfred,kunibert@mathematik.uni-freiburg.de               */
/*                                                                          */
/*     (c) by A. Schmidt and K.G. Siebert (1996,97)                         */
/*                                                                          */
/****************************************************************************/
/****************************************************************************/
/* to_do:                                                                   */
/*   use memcopy() for stack copy in traverse_neighbour()		    */
/*                                                      		    */
/****************************************************************************/

#include "albert.h"
#include "traverse_common.c"

static int coarse_nb[3][3] = {{-2,-2,-2}, {2,-1,1}, {-1,2,0}};
                    /* father.neigh[coarse_nb[i][j]] == child[i-1].neigh[j] */


/****************************************************************************/
/*   traverse_neighbour:                             		            */
/*   -------------------                                		    */
/*   walk through hierarchy tree and look for a neighbour		    */
/****************************************************************************/


EL_INFO *traverse_neighbour(EL_INFO *elinfo_old, int neighbour)
{
  FUNCNAME("traverse_neighbour");
  EL      *el, *el2, *sav_el;
  EL_INFO *old_elinfo, *elinfo, *elinfo2;
  int     i, nb, opp_vertex, stack2_used;

  int     sav_index, sav_neighbour = neighbour;

  TEST_EXIT(stack_used > 0)("no current element");
  TEST_EXIT(traverse_fill_flag & CALL_LEAF_EL)
    ("invalid traverse_fill_flag=%d", traverse_fill_flag);

  TEST_EXIT(elinfo_old == elinfo_stack+stack_used)("invalid old elinfo");

  TEST_FLAG(FILL_NEIGH, (elinfo_stack+stack_used));
  el = elinfo_stack[stack_used].el;
  sav_index = INDEX(el);
  sav_el    = el;

  /* first, goto to leaf level, if necessary... */
  if ((el->child[0]) && (neighbour < 2)) {
    if (stack_used >= stack_size-1) enlarge_traverse_stack();
    i = 1 - neighbour;
    fill_elinfo(i, elinfo_stack+stack_used, elinfo_stack+stack_used+1);
    info_stack[stack_used] = i + 1;
    stack_used++;
    neighbour = 2;
  }


  /* save information about current element and its position in the tree */
  save_traverse_mel = traverse_mel;
  save_stack_used   = stack_used;
  for (i=0; i<=stack_used; i++) save_info_stack[i]   = info_stack[i];
  for (i=0; i<=stack_used; i++) save_elinfo_stack[i] = elinfo_stack[i];
  old_elinfo = save_elinfo_stack+stack_used;
  opp_vertex = OPP_VERTEX(old_elinfo->el,old_elinfo)[neighbour];


/****************************************************************************/
/* First phase: go up in tree until we can go down again.                   */
/*                                                                          */
/* During this first phase, nb is the neighbour index which points from an  */
/* element of the OLD hierarchy branch to the NEW branch                    */
/****************************************************************************/

  nb = neighbour;

  while (stack_used > 1)
  {
    stack_used--;
    nb = coarse_nb[info_stack[stack_used]][nb];
    if (nb == -1) break;
    TEST_EXIT(nb >= 0)("invalid coarse_nb %d\n",nb);
  }

/****************************************************************************/
/* Now, goto neighbouring element at the local hierarchy entry              */
/* This is either a macro element neighbour or the other child of parent.   */
/* initialize nb for second phase (see below)                               */
/****************************************************************************/

  if (nb >= 0) {                        /* go to macro element neighbour */

    if ((nb < 2) && (save_stack_used > 1)) {
      stack2_used = 2;           /* go down one level in OLD hierarchy */
    }
    else {
      stack2_used = 1;
    }
    elinfo2 = save_elinfo_stack+stack2_used;
    el2 = elinfo2->el;

    i = traverse_mel->opp_vertex[nb];
    traverse_mel = traverse_mel->neigh[nb];
    if (traverse_mel == nil)  return(nil);
    nb = i;
    
    stack_used = 1;
    fill_macro_info(traverse_mel, elinfo_stack+stack_used);
    info_stack[stack_used] = 0;
    
  }
  else {                                               /* goto other child */

    stack2_used = stack_used + 1;
    if (save_stack_used > stack2_used) {
      stack2_used++;               /* go down one level in OLD hierarchy */
    }
    elinfo2 = save_elinfo_stack+stack2_used;
    el2 = elinfo2->el;


    if (stack_used >= stack_size-1) enlarge_traverse_stack();
    i = 2 - info_stack[stack_used];
    info_stack[stack_used] = i+1;
    fill_elinfo(i, elinfo_stack+stack_used,
		elinfo_stack+stack_used+1);
    stack_used++;
    nb = 1-i;

  }

/****************************************************************************/
/* Second phase: go down in a new hierarchy branch until leaf level.        */
/* Now, nb is the neighbour index which points from an element of the       */
/* NEW hierarchy branch to the OLD branch.                                  */
/****************************************************************************/

  elinfo = elinfo_stack+stack_used;
  el = elinfo->el;

  while(el->child[0]) {

    if (nb < 2) {   /* go down one level in hierarchy */
      if (stack_used >= stack_size-1) enlarge_traverse_stack();
      fill_elinfo(1-nb, elinfo_stack+stack_used, elinfo_stack+stack_used+1);
      info_stack[stack_used] = 2-nb;
      stack_used++;
      nb = 2;
    }

    if (save_stack_used > stack2_used) {        /* `refine' both el and el2 */
      TEST_EXIT(el->child[0])("invalid NEW refinement?");

                              /* use child i, neighbour of el2->child[nb-1] */
      i = 2 - save_info_stack[stack2_used];
      TEST_EXIT(i < 2)("invalid OLD refinement?");
      info_stack[stack_used] = i+1;
      fill_elinfo(i, elinfo_stack+stack_used, elinfo_stack+stack_used+1);
      stack_used++;
      nb = i;

      elinfo = elinfo_stack+stack_used;
      el = elinfo->el;

      stack2_used++;
      if (save_stack_used > stack2_used) {
	stack2_used++;                /* go down one level in OLD hierarchy */
      }
      elinfo2 = save_elinfo_stack+stack2_used;
      el2 = elinfo2->el;

    }
    else {   /* now we're done... */
      elinfo = elinfo_stack+stack_used;
      el = elinfo->el;
    }
  }


  if (NEIGH(el,elinfo)[opp_vertex] != old_elinfo->el) {
    MSG(" looking for neighbour %d of element %d at %8X\n",
	neighbour, INDEX(old_elinfo->el), old_elinfo->el);
    MSG(" originally: neighbour %d of element %d at %8X\n",
	sav_neighbour, sav_index, sav_el);
    MSG(" got element %d at %8X with opp_vertex %d neigh %d\n",
	INDEX(elinfo->el), elinfo->el, opp_vertex,
	INDEX(NEIGH(el,elinfo)[opp_vertex]));
    TEST_EXIT(NEIGH(el,elinfo)[opp_vertex] == old_elinfo->el)
      ("didn't succeed !?!?!?");
  }
  if (elinfo->el->child[0])
  {
    MSG(" looking for neighbour %d of element %d at %8X\n",
	neighbour, INDEX(old_elinfo->el), old_elinfo->el);
    MSG(" originally: neighbour %d of element %d at %8X\n",
	sav_neighbour, sav_index, sav_el);
    MSG(" got element %d at %8X with opp_vertex %d neigh %d\n",
	INDEX(elinfo->el), elinfo->el, opp_vertex,
	INDEX(NEIGH(el,elinfo)[opp_vertex]));
    MSG("got no leaf element\n");
    WAIT_REALLY;
  }

  return(elinfo);

}


/****************************************************************************/
/*   fill_macro_info:                                		            */
/*   ----------------                                   		    */
/*   fill EL_INFO structure for a macro element         		    */
/****************************************************************************/

void fill_macro_info(const MACRO_EL *mel, EL_INFO *elinfo)
{
  FUNCNAME("fill_macro_info");
  EL       *nb;
  MACRO_EL *mnb;
  int      i, j, k;
#if !NEIGH_IN_EL
  FLAGS    fill_opp_coords;
#endif  

  elinfo->macro_el = mel;
  elinfo->el       = mel->el;
  elinfo->parent   = nil;
  elinfo->level    = 0;

  if (elinfo->fill_flag & FILL_COORDS) {

    for (i=0; i<N_VERTICES; i++) {
      if (mel->coord[i]) {
	for (j=0;j<DIM_OF_WORLD;j++) elinfo->coord[i][j] = mel->coord[i][j];
      }
      else {
	TEST_EXIT(0)("no mel->coord[%d]\n",i);
      }
    }
  }

#if NEIGH_IN_EL
  if (elinfo->fill_flag & FILL_OPP_COORDS) {

    for (i = 0; i < N_NEIGH; i++) {
      if ((mnb = mel->neigh[i])) {
	nb = mel->neigh[i]->el;
	k  = mel->opp_vertex[i];
	if (nb->child[0] && (k != 2)) {         /* make nb nearest el. */
	  if (nb->new_coord)
	  {
	    for (j = 0; j < DIM_OF_WORLD; j++)
	      elinfo->opp_coord[i][j] = nb->new_coord[j];
	  }
	  else
	  {
	    for (j = 0; j < DIM_OF_WORLD; j++)
	      elinfo->opp_coord[i][j] = 
		(mnb->coord[0][j] + mnb->coord[1][j]) / 2;
	  }
	}
	else {
	  for (j = 0; j < DIM_OF_WORLD; j++)
	  {
	    elinfo->opp_coord[i][j] = mnb->coord[k][j];
	  }
	}
      }
    }
  }

#else  /* ! NEIGH_IN_EL */

  if (elinfo->fill_flag & (FILL_OPP_COORDS | FILL_NEIGH)) {

    fill_opp_coords = (elinfo->fill_flag & FILL_OPP_COORDS);
    for (i = 0; i < N_NEIGH; i++) {
      if ((mnb = mel->neigh[i])) {
	nb = elinfo->neigh[i]      = mel->neigh[i]->el;
	k  = elinfo->opp_vertex[i] = mel->opp_vertex[i];
	if (nb->child[0] && (k != 2)) {   /*make nb nearest el.*/
	  nb = elinfo->neigh[i]      = nb->child[1-k];
	  k  = elinfo->opp_vertex[i] = 2;
	  if  (fill_opp_coords) {
	    if (nb->new_coord)
	    {
	      for (j = 0; j < DIM_OF_WORLD; j++)
		elinfo->opp_coord[i][j] = nb->new_coord[j];
	    }
	    else
	    {
	      for (j = 0; j < DIM_OF_WORLD; j++) 
		elinfo->opp_coord[i][j] = 
		  (mnb->coord[0][j] + mnb->coord[1][j]) / 2;
	    }
	  }
	}
	else {
	  if  (fill_opp_coords) {
	    for (j = 0; j < DIM_OF_WORLD; j++)
	      elinfo->opp_coord[i][j] = mnb->coord[k][j];
	  }
	}
      }
      else {
	elinfo->neigh[i] = nil;
      }
    }
  }
#endif

  if (elinfo->fill_flag & FILL_BOUND) {

    for (i = 0; i < N_VERTICES; i++)
      elinfo->bound[i] = mel->bound[i];

    for (i=0;i<N_NEIGH; i++)
      elinfo->boundary[i] = mel->boundary[i];

  }

}


/****************************************************************************/
/*   fill_elinfo:                                    		            */
/*   ------------                                       		    */
/*   fill EL_INFO structure for one child of an element   		    */
/****************************************************************************/

void fill_elinfo(int ichild, const EL_INFO *elinfo_old, EL_INFO *elinfo)
{
  FUNCNAME("fill_elinfo");
  int     j;
  EL      *nb;
  EL      *el = elinfo_old->el;
  FLAGS   fill_flag = elinfo_old->fill_flag;
#if !NEIGH_IN_EL
  FLAGS   fill_opp_coords;
#endif

  TEST_EXIT(el->child[0])("no children?\n");
  TEST_EXIT((elinfo->el = el->child[ichild]))("missing child %d?\n", ichild);

  elinfo->macro_el  = elinfo_old->macro_el;
  elinfo->fill_flag = fill_flag;
  elinfo->mesh      = elinfo_old->mesh;
  elinfo->parent    = el;
  elinfo->level     = elinfo_old->level + 1;


  if (fill_flag & FILL_COORDS)
  {
    if (el->new_coord) {
      for (j = 0; j < DIM_OF_WORLD; j++)
	elinfo->coord[2][j] = el->new_coord[j];
    }
    else {
      for (j = 0; j < DIM_OF_WORLD; j++)
	elinfo->coord[2][j] = 
	  0.5 * (elinfo_old->coord[0][j] + elinfo_old->coord[1][j]);
    }

    if (ichild==0) {
      for (j = 0; j < DIM_OF_WORLD; j++) {
	elinfo->coord[0][j] = elinfo_old->coord[2][j];
	elinfo->coord[1][j] = elinfo_old->coord[0][j];
      }
    }
    else {
      for (j = 0; j < DIM_OF_WORLD; j++) {
	elinfo->coord[0][j] = elinfo_old->coord[1][j];
	elinfo->coord[1][j] = elinfo_old->coord[2][j];
      }
    }
  }



#if NEIGH_IN_EL

  if (fill_flag & FILL_OPP_COORDS)
  {
    if (ichild==0) {
      if (el->neigh[1]) {
	for (j=0; j<DIM_OF_WORLD; j++)
	  elinfo->opp_coord[2][j] = elinfo_old->opp_coord[1][j];
      }

      if (el->child[0]  &&  el->child[1]->child[0]  &&
	  el->child[1]->child[1]) 
      {  
	if (el->child[1]->new_coord)
	  for (j=0; j<DIM_OF_WORLD; j++)
	    elinfo->opp_coord[1][j] = el->child[1]->new_coord[j];
	else
	  for (j=0; j<DIM_OF_WORLD; j++)
	    elinfo->opp_coord[1][j] = 
	      0.5 * (elinfo_old->coord[1][j] + elinfo_old->coord[2][j]);
      }
      else {
	for (j=0; j<DIM_OF_WORLD; j++)
	  elinfo->opp_coord[1][j] = elinfo_old->coord[1][j];
      }

      if ((nb = el->neigh[2]))
      {
	TEST_EXIT(nb->child[0])("missing child?\n");
	nb = nb->child[1];
	if (nb->child[0]) {
	  if (nb->new_coord) 
	    for (j=0; j<DIM_OF_WORLD; j++)
	      elinfo->opp_coord[0][j] = nb->new_coord[j];
	  else
	    for (j=0; j<DIM_OF_WORLD; j++)
	      elinfo->opp_coord[0][j] =
		0.5 * (elinfo_old->opp_coord[2][j] + elinfo_old->coord[0][j]);
	}
	else {
	  for (j=0; j<DIM_OF_WORLD; j++)
	    elinfo->opp_coord[0][j] = elinfo_old->opp_coord[2][j];
	}
      }
    }
    else { /* ichild==1 */

      if (el->neigh[0]) {
	for (j=0; j<DIM_OF_WORLD; j++)
	  elinfo->opp_coord[2][j] = elinfo_old->opp_coord[0][j];
      }

      if (el->child[0]->child[0]) {
	if (el->child[0]->new_coord)
	  for (j=0; j<DIM_OF_WORLD; j++)
	    elinfo->opp_coord[0][j] = el->child[0]->new_coord[j];
	else
	  for (j=0; j<DIM_OF_WORLD; j++)
	    elinfo->opp_coord[0][j] = 
	      0.5 * (elinfo_old->coord[0][j] + elinfo_old->coord[2][j]);
      }
      else {
	for (j=0; j<DIM_OF_WORLD; j++)
	  elinfo->opp_coord[0][j] = elinfo_old->coord[0][j];
      }

      if ((nb = el->neigh[2]))
      {
	TEST((nb = nb->child[0]))("missing child?\n");
	if (nb->child[0]) 
	{
	  if (nb->new_coord)
	    for (j=0; j<DIM_OF_WORLD; j++)
	      elinfo->opp_coord[1][j] = nb->new_coord[j];
	  else
	    for (j=0; j<DIM_OF_WORLD; j++)
	      elinfo->opp_coord[1][j] =
		0.5 * (elinfo_old->opp_coord[2][j] + elinfo_old->coord[1][j]);
	}
	else {
	  for (j=0; j<DIM_OF_WORLD; j++)
	    elinfo->opp_coord[1][j] = elinfo_old->opp_coord[2][j];
	}
      }
    }
  }

#else  /* ! NEIGH_IN_EL */

  if (fill_flag & (FILL_NEIGH | FILL_OPP_COORDS))
  {
    fill_opp_coords = (fill_flag & FILL_OPP_COORDS);
    if (ichild==0) {
      if ((elinfo->neigh[2] = elinfo_old->neigh[1])) {
	if (fill_opp_coords) {
	  for (j=0; j<DIM_OF_WORLD; j++)
	    elinfo->opp_coord[2][j] = elinfo_old->opp_coord[1][j];
	}
      }
      elinfo->opp_vertex[2] = elinfo_old->opp_vertex[1];

/*       if (el->child[1]->child[0]  &&  el->child[1]->child[1]) */
      if (el->child[0]  &&  el->child[1]->child[0]  &&  
	  el->child[1]->child[0])
      {
	TEST_EXIT((elinfo->neigh[1] = el->child[1]->child[1]))
	  ("el->child[1]->child[0]!=nil, but el->child[1]->child[1]=nil\n");
	elinfo->opp_vertex[1] = 2;
	if (fill_opp_coords) 
	{
	  if (el->child[1]->new_coord)
	    for (j=0; j<DIM_OF_WORLD; j++)
	      elinfo->opp_coord[1][j] = el->child[1]->new_coord[j];
	  else
	    for (j=0; j<DIM_OF_WORLD; j++)
	      elinfo->opp_coord[1][j] = 
		0.5 * (elinfo_old->coord[1][j] + elinfo_old->coord[2][j]);
	}
      }
      else 
      {
	TEST_EXIT((elinfo->neigh[1] = el->child[1]))
	  ("el->child[0] != nil, but el->child[1] = nil\n");
	elinfo->opp_vertex[1] = 0;
	if (fill_opp_coords) {
	  for (j=0; j<DIM_OF_WORLD; j++)
	    elinfo->opp_coord[1][j] = elinfo_old->coord[1][j];
	}
      }

      if ((nb = elinfo_old->neigh[2]))
      {
	TEST(elinfo_old->opp_vertex[2] == 2)("invalid neighbour\n"); 
	TEST_EXIT(nb->child[0])("missing children?\n");
	TEST_EXIT((nb = nb->child[1]))("missing child[1]?\n");

	if (nb->child[0]) 
	{
	  elinfo->opp_vertex[0] = 2;
	  if (fill_opp_coords)
	  {
	    if (nb->new_coord)
	    {
	      for (j=0; j<DIM_OF_WORLD; j++)
		elinfo->opp_coord[0][j] = nb->new_coord[j];
	    }
	    else
	    {
	      for (j=0; j<DIM_OF_WORLD; j++)
		elinfo->opp_coord[0][j] =
		  0.5*(elinfo_old->opp_coord[2][j] + elinfo_old->coord[0][j]);
	    }
	  }	    
	  nb = nb->child[0];
	}
	else 
	{
	  elinfo->opp_vertex[0] = 1;
	  if (fill_opp_coords) 
	  {
	    for (j=0; j<DIM_OF_WORLD; j++)
	      elinfo->opp_coord[0][j] = elinfo_old->opp_coord[2][j];
	  }
	}
      }
      elinfo->neigh[0] = nb;
    }
    else   /* ichild==1 */
    { 
      if ((elinfo->neigh[2] = elinfo_old->neigh[0]))
      {
	if (fill_opp_coords)
	{
	  for (j=0; j<DIM_OF_WORLD; j++)
	    elinfo->opp_coord[2][j] = elinfo_old->opp_coord[0][j];
	}
      }
      elinfo->opp_vertex[2] = elinfo_old->opp_vertex[0];

      if (el->child[0]->child[0])
      {
	elinfo->neigh[0] = el->child[0]->child[0];
	elinfo->opp_vertex[0] = 2;
	if (fill_opp_coords)
	{
	  if (el->child[0]->new_coord)
	  {
	    for (j=0; j<DIM_OF_WORLD; j++)
	      elinfo->opp_coord[0][j] = el->child[0]->new_coord[j];
	  }
	  else
	  {
	    for (j=0; j<DIM_OF_WORLD; j++)
	      elinfo->opp_coord[0][j] = 
		0.5 * (elinfo_old->coord[0][j] + elinfo_old->coord[2][j]);
	  }
	}
      }
      else 
      {
	elinfo->neigh[0] = el->child[0];
	elinfo->opp_vertex[0] = 1;
	if (fill_opp_coords)
	{
	  for (j=0; j<DIM_OF_WORLD; j++)
	    elinfo->opp_coord[0][j] = elinfo_old->coord[0][j];
	}
      }

      if ((nb = elinfo_old->neigh[2]))
      {
	TEST(elinfo_old->opp_vertex[2] == 2)("invalid neighbour\n"); 
	TEST((nb = nb->child[0]))("missing child?\n");
	if (nb->child[0])
	{
	  elinfo->opp_vertex[1] = 2;
	  if (fill_opp_coords) 
	  {
	    if (nb->new_coord)
	      for (j=0; j<DIM_OF_WORLD; j++)
		elinfo->opp_coord[1][j] = nb->new_coord[j];
	    else
	      for (j=0; j<DIM_OF_WORLD; j++)
		elinfo->opp_coord[1][j] =
		  0.5*(elinfo_old->opp_coord[2][j] + elinfo_old->coord[1][j]);
	  }
	  nb = nb->child[1];
	}
	else 
	{
	  elinfo->opp_vertex[1] = 0;
	  if (fill_opp_coords) {
	    for (j=0; j<DIM_OF_WORLD; j++)
	      elinfo->opp_coord[1][j] = elinfo_old->opp_coord[2][j];
	  }
	}
      }
      elinfo->neigh[1] = nb;
    }
  }
#endif


  if (fill_flag & FILL_BOUND)
  {
    if (elinfo_old->boundary[2])
      elinfo->bound[2] = elinfo_old->boundary[2]->bound;
    else
      elinfo->bound[2] = INTERIOR;

    if (ichild==0) {
      elinfo->bound[0] = elinfo_old->bound[2];
      elinfo->bound[1] = elinfo_old->bound[0];
      elinfo->boundary[0] = elinfo_old->boundary[2];
      elinfo->boundary[1] = nil;
      elinfo->boundary[2] = elinfo_old->boundary[1];
    }
    else {
      elinfo->bound[0] = elinfo_old->bound[1];
      elinfo->bound[1] = elinfo_old->bound[2];
      elinfo->boundary[0] = nil;
      elinfo->boundary[1] = elinfo_old->boundary[2];
      elinfo->boundary[2] = elinfo_old->boundary[0];
    }
  }

}



