/****************************************************************************/
/* ALBERT:   an Adaptive multi Level finite element toolbox using           */
/*           Bisectioning refinement and Error control by Residual          */
/*           Techniques                                                     */
/*                                                                          */
/* file:                                                                    */
/*                                                                          */
/*                                                                          */
/* description:                                                             */
/*                                                                          */
/*                                                                          */
/* history:                                                                 */
/*                                                                          */
/*                                                                          */
/****************************************************************************/
/*                                                                          */
/*        authors: Alfred Schmidt and Kunibert G. Siebert                   */
/*                 Institut f"ur Angewandte Mathematik                      */
/*                 Albert-Ludwigs-Universit"at Freiburg                     */
/*                 Hermann-Herder-Str. 10                                   */
/*                 79104 Freiburg                                           */
/*                 Germany                                                  */
/*                                                                          */
/*        email:   alfred,kunibert@mathematik.uni-freiburg.de               */
/*                                                                          */
/*     (c) by A. Schmidt and K.G. Siebert (1996,97)                         */
/*                                                                          */
/****************************************************************************/
#include "albert.h"

#if 0
#define SKP_3D(x, y)   (x[0]*y[0] + x[1]*y[1] + x[2]*y[2]) 

/****************************************************************************/
/*  calculates the transformation of the DIM_OF_WORLD x DIM_OF_WORLD matrix */
/*  B to the standard simplex: A^t B A of size DIM+1 x DIM+1;               */
/*  the matrix A^t B A is additionally scaled by a factor of 6*|T|, thus no */
/*  additional factor will appear in the quadruature!                       */
/****************************************************************************/

REAL get_transformed_el_mat(const EL_INFO *el_info, REAL mat[DIM+1][DIM+1])
{
  FUNCNAME("fill_el_mat");
  const REAL  *v0;

  double det, det6;
  double a11, a12, a13, a22, a23, a33, b11, b12, b13, b22, b23, b33;
  double e1[DIM_OF_WORLD], e2[DIM_OF_WORLD], e3[DIM_OF_WORLD];
  int i;

  TEST_FLAG(FILL_COORDS, el_info);

#if DIM_OF_WORLD != DIM
  ERROR_EXIT("not yet for DIM_OF_WORLD = %d", DIM_OF_WORLD);
#endif

  v0 = el_info->coord[0];

  for (i = 0; i < DIM_OF_WORLD; i++)
  {
    e1[i] = el_info->coord[1][i] - v0[i];
    e2[i] = el_info->coord[2][i] - v0[i];
    e3[i] = el_info->coord[3][i] - v0[i];
  }

/****************************************************************************/
/*  det DF_T = abs(e1 * (e2 x e3))                                          */
/****************************************************************************/

  det = e1[0]*(e2[1]*e3[2] - e2[2]*e3[1]) + e1[1]*(e2[2]*e3[0] - e2[0]*e3[2])
    + e1[2]*(e2[0]*e3[1] - e2[1]*e3[0]);
  det = ABS(det);

/****************************************************************************/
/*  Calculate:                                                              */
/*     /                                                                    */
/*     |                                                                    */
/*     | grad phi_i grad phi_j =                                            */
/*     /                                                                    */
/*      T                                                                   */
/*                                                                          */
/*                        /                                                 */
/*                        |  ^    ^                          ^   ^          */
/*            |det DF_T|  | grad phi_i (DF_T^t DF_T)^(-1)  grad phi_j       */
/*                        /^                                                */
/*                         T                                                */
/*          ^                                         ^    ^                */
/*    where T is the reference simplex in R^3. Thus, grad phi_i is the      */
/*    i-th unit vector in R^3. Hence, the element stiffness matrix is       */
/*    equal to 1/6 |det DF_T|  (DF_T^t DF_T)^(-1).                          */
/*    The inverse of (DF_T^t DF_T) is calculated by Cramer's rule and       */
/*                                                                          */
/*                     -             -                                      */
/*                     | a11 a12 a13 |                                      */
/*     (DF_T^t DF_T) = | a12 a22 a23 |                                      */
/*                     | a13 a23 a33 |                                      */
/*                     -             -                                      */
/****************************************************************************/

  a11 = SKP_3D(e1,e1);
  a12 = SKP_3D(e1,e2);
  a13 = SKP_3D(e1,e3);

  a22 = SKP_3D(e2,e2);
  a23 = SKP_3D(e2,e3);

  a33 = SKP_3D(e3,e3);

  b11 = a22*a33 - a23*a23;
  b12 = a23*a13 - a12*a33;
  b13 = a12*a23 - a22*a13;

  b22 = a11*a33 - a13*a13;
  b23 = a12*a13 - a11*a23;

  b33 = a11*a22 - a12*a12;

#if 0
  det6 = 1.0/(det*6.0);
#else
  det6 = 1.0/det;
#endif

  mat[1][1] = b11*det6;
  mat[2][2] = b22*det6;
  mat[3][3] = b33*det6;

  mat[1][2] = mat[2][1] = b12*det6;
  mat[1][3] = mat[3][1] = b13*det6;
  mat[2][3] = mat[3][2] = b23*det6;

  mat[0][1] = mat[1][0] = -(mat[1][1] + mat[2][1] + mat[3][1]);
  mat[0][2] = mat[2][0] = -(mat[1][2] + mat[2][2] + mat[3][2]);
  mat[0][3] = mat[3][0] = -(mat[1][3] + mat[2][3] + mat[3][3]);
  mat[0][0] = -(mat[0][1] + mat[0][2] + mat[0][3]);

  return(det);
}
#endif


/****************************************************************************/
/*  compute volume of an element                                            */
/****************************************************************************/

REAL el_det(const EL_INFO *el_info)
{
  FUNCNAME("el_det");
  REAL        e1[DIM_OF_WORLD], e2[DIM_OF_WORLD], e3[DIM_OF_WORLD], det;
  int         i;
  const REAL  *v0;

  TEST_FLAG(FILL_COORDS, el_info);

  v0 = el_info->coord[0];
  for (i = 0; i < DIM_OF_WORLD; i++)
  {
    e1[i] = el_info->coord[1][i] - v0[i];
    e2[i] = el_info->coord[2][i] - v0[i];
    e3[i] = el_info->coord[3][i] - v0[i];
  }

#if DIM_OF_WORLD==3
  det =   e1[0] * (e2[1]*e3[2] - e2[2]*e3[1])
        - e1[1] * (e2[0]*e3[2] - e2[2]*e3[0])
        + e1[2] * (e2[0]*e3[1] - e2[1]*e3[0]);
  det = ABS(det);
#else
  TEST_EXIT(0)("not yet for DIM_OF_WORLD = %d\n",DIM_OF_WORLD);
#endif

  return(det);
}

#define VOL_THAT  1.0/6.0

REAL el_volume(const EL_INFO *el_info)
{
  return(VOL_THAT*el_det(el_info));
}

/****************************************************************************/
/*  compute gradients of basis functions on element                         */
/*  - returns abs(determinante)                                             */
/****************************************************************************/

REAL el_grd_lambda(const EL_INFO *el_info, REAL grd_lam[DIM+1][DIM_OF_WORLD])
{
  FUNCNAME("el_grd_lambda");
  int         i, j;
  REAL        e1[DIM_OF_WORLD], e2[DIM_OF_WORLD], e3[DIM_OF_WORLD];
  const REAL  *v0;
  REAL        det, adet;
  REAL        a11, a12, a13, a21, a22, a23, a31, a32, a33;

  TEST_FLAG(FILL_COORDS, el_info);

  v0 = el_info->coord[0];
  for (i = 0; i < DIM_OF_WORLD; i++) {
    e1[i] = el_info->coord[1][i] - v0[i];
    e2[i] = el_info->coord[2][i] - v0[i];
    e3[i] = el_info->coord[3][i] - v0[i];
  }

#if DIM_OF_WORLD == 3
  det =   e1[0] * (e2[1]*e3[2] - e2[2]*e3[1])
        - e1[1] * (e2[0]*e3[2] - e2[2]*e3[0])
        + e1[2] * (e2[0]*e3[1] - e2[1]*e3[0]);
  adet = ABS(det);
  if (adet < 1.0E-25)
  {
    MSG("abs(det) = %lf\n",adet);
    for (i = 0; i < N_VERTICES; i++)
      for (j = 0; j < DIM_OF_WORLD; j++)
	grd_lam[i][j] = 0.0;
  }
  else 
  {
    det = 1.0 / det;
    a11 = (e2[1]*e3[2] - e2[2]*e3[1]) * det;    /* (a_ij) = A^{-T} */
    a12 = (e2[2]*e3[0] - e2[0]*e3[2]) * det;
    a13 = (e2[0]*e3[1] - e2[1]*e3[0]) * det;
    a21 = (e1[2]*e3[1] - e1[1]*e3[2]) * det;
    a22 = (e1[0]*e3[2] - e1[2]*e3[0]) * det;
    a23 = (e1[1]*e3[0] - e1[0]*e3[1]) * det;
    a31 = (e1[1]*e2[2] - e1[2]*e2[1]) * det;
    a32 = (e1[2]*e2[0] - e1[0]*e2[2]) * det;
    a33 = (e1[0]*e2[1] - e1[1]*e2[0]) * det;

    grd_lam[1][0] = a11;
    grd_lam[1][1] = a12;
    grd_lam[1][2] = a13;
    grd_lam[2][0] = a21;
    grd_lam[2][1] = a22;
    grd_lam[2][2] = a23;
    grd_lam[3][0] = a31;
    grd_lam[3][1] = a32;
    grd_lam[3][2] = a33;

    grd_lam[0][0] = -grd_lam[1][0] -grd_lam[2][0] -grd_lam[3][0];
    grd_lam[0][1] = -grd_lam[1][1] -grd_lam[2][1] -grd_lam[3][1];
    grd_lam[0][2] = -grd_lam[1][2] -grd_lam[2][2] -grd_lam[3][2];
  }


#else
  ERROR_EXIT("not yet for DIM_OF_WORLD = %d", DIM_OF_WORLD);
#endif

  return(adet);
}

#if DIM_OF_WORLD == 3
/****************************************************************************/
/*  vector product in R^3                                                   */
/****************************************************************************/

void vector_product(const REAL_D x, const REAL_D y, REAL_D z)
{
  z[0] = x[1]*y[2] - x[2]*y[1];
  z[1] = x[2]*y[0] - x[0]*y[2];
  z[2] = x[0]*y[1] - x[1]*y[0];

  return;
}
#endif

/****************************************************************************/
/*  calculate a facenormal to triangular face side of a tetrahedron with    */
/*  coordinates coord; return the absulute value of the determinante from   */
/*  the transformation to the reference element                             */
/****************************************************************************/

REAL get_face_normal(const REAL_D coord[N_VERTICES], int face, REAL *normal)
{
  FUNCNAME("get_face_normal");
  REAL   det;
  int    i, i0, i1, i2;
  REAL_D e0, e1, e2;

#if DIM_OF_WORLD == 3
  i0 = (face+1)%4;
  i1 = (face+2)%4;
  i2 = (face+3)%4;

  for (i = 0; i < DIM_OF_WORLD; i++)
  {
    e0[i] = coord[i1][i] - coord[i0][i];
    e1[i] = coord[i2][i] - coord[i0][i];
    e2[i] = coord[face][i] - coord[i0][i];
  }

  vector_product(e0, e1, normal);

  if (SCP_DOW(e2,normal) < 0.0)
    for (i = 0; i < DIM_OF_WORLD; i++)
      normal[i] = -normal[i];

  det = NORM_DOW(normal);
  TEST_EXIT(det > 1.e-30)("det = 0 on face %d\n", face);

  normal[0] /= det;
  normal[1] /= det;
  normal[2] /= det;
#else
  MSG("not implemented for DIM_OF_WORLD = %d in 3d\n", DIM_OF_WORLD);
#endif

  return(det);
}

/*--------------------------------------------------------------------------*/
/* orient the vertices of edges/faces                                       */
/* used by estimator for the jumps => same quadrature nodes from both sides!*/
/*--------------------------------------------------------------------------*/

const int *sort_face_indices(const EL *el, int face, int *vec)
{
  FUNCNAME("sort_face_indices");
  static int  sorted[4][7][3] = 
      {{{0,0,0}, {1,3,2}, {2,1,3}, {1,2,3}, {3,2,1}, {3,1,2}, {2,3,1}},
       {{0,0,0}, {0,3,2}, {2,0,3}, {0,2,3}, {3,2,0}, {3,0,2}, {2,3,0}},
       {{0,0,0}, {0,3,1}, {1,0,3}, {0,1,3}, {3,1,0}, {3,0,1}, {1,3,0}},
       {{0,0,0}, {0,2,1}, {1,0,2}, {0,1,2}, {2,1,0}, {2,0,1}, {1,2,0}}};
  static int  vertex_of_face[4][3] = {{1,2,3}, {0,2,3}, {0,1,3}, {0,1,2}};
  int    i, no, *val  = nil;
  int    *vof = vertex_of_face[face];
  DOF    **dof = el->dof;

  no = 0;
  if (dof[vof[0]][0] < dof[vof[1]][0])
    no++;
  if (dof[vof[1]][0] < dof[vof[2]][0])
    no += 2;
  if (dof[vof[2]][0] < dof[vof[0]][0])
    no += 4;
  
  if (no >= 1  &&  no <= 6)
  {
    if (vec)
    {
      val = vec;
      for (i = 0; i < 3; i++)
	val[i] = sorted[face][no][i];
    }
    else
      val = sorted[face][no];
  }
  else
  {
    MSG("can not sort face indices of element %d at face %d\n", 
	INDEX(el), face);
  }
 
  return((const int *) val);
}
