/****************************************************************************/
/* ALBERT:   an Adaptive multi Level finite element toolbox using           */
/*           Bisectioning refinement and Error control by Residual          */
/*           Techniques                                                     */
/*                                                                          */
/* file:                                                                    */
/*                                                                          */
/*                                                                          */
/* description:                                                             */
/*                                                                          */
/*                                                                          */
/* history:                                                                 */
/*                                                                          */
/*                                                                          */
/****************************************************************************/
/*                                                                          */
/*        authors: Alfred Schmidt and Kunibert G. Siebert                   */
/*                 Institut f"ur Angewandte Mathematik                      */
/*                 Albert-Ludwigs-Universit"at Freiburg                     */
/*                 Hermann-Herder-Str. 10                                   */
/*                 79104 Freiburg                                           */
/*                 Germany                                                  */
/*                                                                          */
/*        email:   alfred,kunibert@mathematik.uni-freiburg.de               */
/*                                                                          */
/*     (c) by A. Schmidt and K.G. Siebert (1996,97)                         */
/*                                                                          */
/****************************************************************************/
#ifndef _ALBERT_UTIL_H_
#define _ALBERT_UTIL_H_

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <float.h>

#ifndef true
#define true          1
#define false         0
#endif

#ifndef nil
#define nil           NULL
#endif

#ifndef MAX
#define MAX(a, b)     ((a) > (b) ? (a) : (b))
#define MIN(a, b)     ((a) < (b) ? (a) : (b))
#define ABS(a)        ((a) >= 0 ? (a) : -(a))
#endif

#ifndef SQR
#define SQR(a)        ((a)*(a))
#endif

#ifndef M_E
#define M_E		2.7182818284590452354
#endif
#ifndef M_LOG2E
#define M_LOG2E		1.4426950408889634074
#endif
#ifndef M_LOG10E
#define M_LOG10E	0.43429448190325182765
#endif
#ifndef M_LN2
#define M_LN2		0.69314718055994530942
#endif
#ifndef M_LN10
#define M_LN10		2.30258509299404568402
#endif
#ifndef M_PI
#define M_PI		3.14159265358979323846
#endif
#ifndef M_PI_2
#define M_PI_2		1.57079632679489661923
#endif
#ifndef M_PI_4
#define M_PI_4		0.78539816339744830962
#endif
#ifndef M_1_PI
#define M_1_PI		0.31830988618379067154
#endif
#ifndef M_2_PI
#define M_2_PI		0.63661977236758134308
#endif
#ifndef M_2_SQRTPI
#define M_2_SQRTPI	1.12837916709551257390
#endif
#ifndef M_SQRT2
#define M_SQRT2		1.41421356237309504880
#endif
#ifndef M_SQRT1_2
#define M_SQRT1_2	0.70710678118654752440
#endif

/****************************************************************************/
/*  Definition of basic data types      				    */
/****************************************************************************/

#ifndef REAL_DEFINED
#define REAL_DEFINED

#define USE_DOUBLE_AS_REAL  1

#if USE_DOUBLE_AS_REAL

typedef double              REAL;

#define REAL_MANT_DIG     DBL_MANT_DIG     
#define REAL_EPSILON      DBL_EPSILON      
#define REAL_DIG          DBL_DIG          
#define REAL_MIN_EXP      DBL_MIN_EXP      
#define REAL_MIN          DBL_MIN          
#define REAL_MIN_10_EXP   DBL_MIN_10_EXP   
#define REAL_MAX_EXP      DBL_MAX_EXP      
#define REAL_MAX          DBL_MAX          
#define REAL_MAX_10_EXP   DBL_MAX_10_EXP   

#else  /*  USE_DOUBLE_AS_REAL  */

typedef float             REAL;

#define REAL_MANT_DIG     FLT_MANT_DIG     
#define REAL_EPSILON      FLT_EPSILON      
#define REAL_DIG          FLT_DIG          
#define REAL_MIN_EXP      FLT_MIN_EXP      
#define REAL_MIN          FLT_MIN          
#define REAL_MIN_10_EXP   FLT_MIN_10_EXP   
#define REAL_MAX_EXP      FLT_MAX_EXP      
#define REAL_MAX          FLT_MAX          
#define REAL_MAX_10_EXP   FLT_MAX_10_EXP   

#endif  /*  USE_DOUBLE_AS_REAL  */

#ifndef LARGE
#define LARGE         REAL_MAX
#endif

#endif  /*  REAL_DEFINED  */

#ifndef U_CHAR_DEFINED
#define U_CHAR_DEFINED

typedef unsigned char       U_CHAR;
typedef signed char         S_CHAR;
typedef unsigned long       FLAGS;
#endif

#ifndef WORKSPACE_DEFINED
#define WORKSPACE_DEFINED
typedef struct workspace    WORKSPACE;

struct workspace
{
  size_t  size;
  void    *work;
};
#endif

#ifndef VOID_LIST_ELEMENT_DEFINED
#define VOID_LIST_ELEMENT_DEFINED
typedef struct void_list_element VOID_LIST_ELEMENT;

struct void_list_element
{
  void               *data;
  VOID_LIST_ELEMENT  *next;
};
VOID_LIST_ELEMENT *get_void_list_element(void);
void free_void_list_element(VOID_LIST_ELEMENT *);
#endif

/****************************************************************************/
/*  Macros for messages							    */
/****************************************************************************/

void print_msg(const char *format, ...);
void print_error_msg(const char *format, ...);
void print_error_msg_exit(const char *format, ...);
void print_error_funcname(const char *name, const char *file, int line);
void print_warn_msg(const char *format, ...);
void print_warn_funcname(const char *name, const char *file, int line);
void print_funcname(const char *name);
void print_int_vec(const char *s, const int *vec, int no);
void print_real_vec(const char *s, const REAL *vec, int no);

void change_msg_out(FILE *fp);
void open_msg_file(const char *filename, const char *type);
void change_error_out(FILE *fp);
void open_error_file(const char *filename, const char *type);
const char *generate_filename(const char *, const char *, int);

#define FUNCNAME(nn)    const char *funcName = nn

#define ERROR           print_error_funcname(funcName, __FILE__, __LINE__),\
                        print_error_msg
#define ERROR_EXIT      print_error_funcname(funcName, __FILE__, __LINE__),\
                        print_error_msg_exit
#define WARNING         print_warn_funcname(funcName, __FILE__, __LINE__),\
                        print_warn_msg

#define TEST(test)      if ((test));else ERROR
#define TEST_EXIT(test) if ((test));else ERROR_EXIT

#define MSG             print_funcname(funcName), print_msg

#define INFO(info,noinfo)\
 if (!(msg_info&&(MIN(msg_info,(info))>=(noinfo))));else MSG

#define PRINT_INFO(info,noinfo)\
 if (!(msg_info&&(MIN(msg_info,(info))>=(noinfo))));else print_msg

#define PRINT_INT_VEC   print_funcname(funcName), print_int_vec
#define PRINT_REAL_VEC  print_funcname(funcName), print_real_vec

extern void albert_wait(const char *, int);
#define WAIT            albert_wait(funcName, msg_wait)
#define WAIT_REALLY     albert_wait(funcName, 1)

/****************************************************************************/
/*  Definition of extern variables      				    */
extern int           msg_wait;                       /* declared in util.c  */
extern const char    *funcName;                      /* declared in util.c  */
extern int           msg_info;                       /* declared in util.c  */

/****************************************************************************/
/*  Macros for memory allocation					    */
/****************************************************************************/

void *albert_alloc(size_t, const char *, const char *,int);
void *albert_realloc(void *, size_t, size_t, const char *, const char *, int);
void *albert_calloc(size_t, size_t, const char *, const char *,int);
void albert_free(void *, size_t);
void print_mem_use(void);

void **albert_matrix(size_t, size_t, size_t, const char *, const char *, int);
void free_albert_matrix(void **, size_t, size_t, size_t);

WORKSPACE *get_workspace(size_t, const char *, const char *, int);
#define GET_WORKSPACE(s) get_workspace(s,funcName,__FILE__,__LINE__)

WORKSPACE  *realloc_workspace(WORKSPACE *ws, size_t new_size, 
			      const char *fct, const char *file, int line);
#define REALLOC_WORKSPACE(ws,size)\
 (realloc_workspace(ws,size,funcName,__FILE__,__LINE__))

void clear_workspace(WORKSPACE *ws);
#define CLEAR_WORKSPACE(ws) clear_workspace(ws)
void free_workspace(WORKSPACE *ws);
#define FREE_WORKSPACE(ws) free_workspace(ws)

#define MEM_ALLOC(n,type)\
   ((type*)albert_alloc((n)*sizeof(type),funcName,__FILE__,__LINE__))
#define MEM_REALLOC(ptr,old_n,new_n,type)\
   ((type*)albert_realloc((void *)ptr,(old_n)*sizeof(type),\
    (new_n)*sizeof(type),funcName,__FILE__,__LINE__))
#define MEM_CALLOC(n,type)\
   ((type*)albert_calloc((n), sizeof(type),funcName,__FILE__,__LINE__))
#define MEM_FREE(ptr,n,type) albert_free((void *)ptr,(n)*sizeof(type))

#define MAT_ALLOC(nr,nc,type)\
  ((type**)albert_matrix((nr),(nc),sizeof(type),funcName,__FILE__, __LINE__))
#define MAT_FREE(ptr,nr,nc,type)\
    free_albert_matrix((void **)ptr,(nr),(nc),sizeof(type))

#define NAME(struct)\
   ((struct) ? ((struct)->name ? (struct)->name : #struct "->name unknown") :\
   #struct " pointer to nil")

#define GET_STRUCT(struct, from)\
   TEST_EXIT((from)  && (struct = (from)->struct))\
     ((from) ? "GET_STRUCT: %s->%s == nil\n" : "GET_STRUCT: %s == nil%s\n",\
      #from, (from) ? #struct : "")

/****************************************************************************/
/* functions for handling parameters                                        */
/****************************************************************************/

void add_parameter(int p, const char *key, const char *par);
void Add_parameter(int p, const char *key, const char *par,
		   const char *fname, const char *file, int line);
#define ADD_PARAMETER(p,key,param)\
   Add_parameter(p,key,param,funcName, __FILE__, __LINE__)

void init_parameters(int p, const char *filename);
void init_parameters_cpp(int p, const char *fn, const char *flags);

int get_parameter(int flag, const char *key, const char *format, ...);
int init_param_func_name(const char *, const char *, int call_line);

#define GET_PARAMETER  !init_param_func_name(funcName, __FILE__, __LINE__)?\
                       0 : get_parameter
void save_parameters(const char *file, int info);

#endif  /* !_ALBERT_UTIL_H_ */
