/****************************************************************************/
/* ALBERT:   an Adaptive multi Level finite element toolbox using           */
/*           Bisectioning refinement and Error control by Residual          */
/*           Techniques                                                     */
/*                                                                          */
/* file:     Common/cg.c                                                    */
/*                                                                          */
/*                                                                          */
/* description:                                                             */
/*                                                                          */
/*                                                                          */
/* history:                                                                 */
/*                                                                          */
/*                                                                          */
/****************************************************************************/
/*                                                                          */
/*        authors: Alfred Schmidt and Kunibert G. Siebert                   */
/*                 Institut f"ur Angewandte Mathematik                      */
/*                 Albert-Ludwigs-Universit"at Freiburg                     */
/*                 Hermann-Herder-Str. 10                                   */
/*                 79104 Freiburg                                           */
/*                 Germany                                                  */
/*                                                                          */
/*        email:   alfred,kunibert@mathematik.uni-freiburg.de               */
/*                                                                          */
/*     (c) by A. Schmidt and K.G. Siebert (1996-1999)                       */
/*                                                                          */
/****************************************************************************/
/****************************************************************************/
/* Conjugate gradient solver                                                */
/* (uses W. Doerler's ofmlib routines)                                      */
/****************************************************************************/

#include "albert.h"

int cghs_(void mat_vec(double *ausg, double *eing, int *n),
	  double *uh, double *fh, int *n, int *icon,
	  int *info, int *iflag, double *work, int *mw);
int cgn_(void mat_vec(double *ausg, double *eing, int *n),
	 void matT_vec(double *ausg, double *eing, int *n),
	  double *uh, double *fh, int *n, int *icon,
	  int *info, int *iflag, double *work, int *mw);


static const DOF_MATRIX *a;
static const S_CHAR     *bound_vec;

static void mat_vec(double *ausg, double *eing, int *size)
{
  FUNCNAME("mat_vec");
  static DOF_REAL_VEC dof_ausg = {nil, nil, "ausg", 0, nil, nil, nil};
  static DOF_REAL_VEC dof_eing = {nil, nil, "eing", 0, nil, nil, nil};

  dof_eing.fe_space = dof_ausg.fe_space = a->fe_space;
  dof_ausg.size = *size;
  dof_ausg.vec = ausg;
  dof_eing.size = *size;
  dof_eing.vec = eing;

  dof_mv(NoTranspose, a, &dof_eing, &dof_ausg);

  if (bound_vec)
    FOR_ALL_DOFS(a->fe_space->admin,
		 if (bound_vec[dof] > 0) ausg[dof] = eing[dof]);
}


static void matT_vec(double *ausg, double *eing, int *size)
{
  FUNCNAME("matT_vec");
  static DOF_REAL_VEC dof_ausg = {nil, nil, "ausg", 0, nil, nil, nil};
  static DOF_REAL_VEC dof_eing = {nil, nil, "eing", 0, nil, nil, nil};

  dof_eing.fe_space = dof_ausg.fe_space = a->fe_space;
  dof_ausg.size = *size;
  dof_ausg.vec = ausg;
  dof_eing.size = *size;
  dof_eing.vec = eing;

  dof_mv(Transpose, a, &dof_eing, &dof_ausg);

  if (bound_vec)
    FOR_ALL_DOFS(a->fe_space->admin,
		 if (bound_vec[dof] > 0) ausg[dof] = eing[dof] );
}

/****************************************************************************/

void precon_(double *h, int *ic, int *n, int *it)
{
  FUNCNAME("precon_");
  const DOF_ADMIN *admin = a->fe_space->admin;
  MATRIX_ROW **row = a->matrix_row;

  if (*ic == 1)          /*                        diagonal preconditioning */
  {
    if (bound_vec)
    {
      FOR_ALL_DOFS(admin,
		   if ((bound_vec[dof] <= 0) && row[dof]) 
		      h[dof] /= row[dof]->entry[0]);
    }
    else
    {
      FOR_ALL_DOFS(admin, h[dof] /= row[dof]->entry[0]);
    }
  }
}

/****************************************************************************/
/* Fortran Common block /ABBR/                                              */

struct {
  double akrit, akriti;
  int    irestart, itmax, itmaxi;
} abbr_;

static int    mw = 0;
static double *work = nil;

/****************************************************************************/

int cghs(const DOF_MATRIX *mat, DOF_REAL_VEC *u, const DOF_REAL_VEC *f,
	 const DOF_SCHAR_VEC *b, REAL tol, int icon, int max_iter, int info)
{
  FUNCNAME("cghs");
  int       iflag = 0, n, iter;

  a = mat;
  bound_vec = b ? b->vec : nil;

  n = mat->fe_space->admin->size_used;

  if (mw < 3*(n+1)) {
    work = MEM_REALLOC(work, mw, 3*(n+1), double);
    mw = 3*(n+1);
  }
  abbr_.akrit = abbr_.akriti = tol;
  abbr_.irestart = 0; 
  abbr_.itmax = abbr_.itmaxi = max_iter; 

  if (icon > 1) {
    MSG("Precon %d not yet implemented.\n", icon);
    icon = 1;
  }

  iter = cghs_(mat_vec, u->vec, f->vec, &n, &icon, &info, &iflag, work, &mw);

  bound_vec = nil;
  a = nil;

  return(iter);
}

/****************************************************************************/

int cgn(const DOF_MATRIX *mat, DOF_REAL_VEC *u, const DOF_REAL_VEC *f,
	 const DOF_SCHAR_VEC *b, REAL tol, int icon, int max_iter, int info)
{
  FUNCNAME("cgn");
  int       iflag = 0, n, mw_needed, iter;

  a = mat;
  bound_vec = b ? b->vec : nil;

  n = mat->fe_space->admin->size_used;

  mw_needed = 4*(n+1);

  if (mw < mw_needed) {
    work = MEM_REALLOC(work, mw, mw_needed, double);
    mw = mw_needed;
  }
  abbr_.akrit = abbr_.akriti = tol;
  abbr_.irestart = 0; 
  abbr_.itmax = abbr_.itmaxi = max_iter; 

  if (icon) 
  {
    MSG("Precon %d not yet implemented.\n", icon);
    icon = 0;
  }

  iter = cgn_(mat_vec, matT_vec, u->vec, f->vec, &n, &icon, &info, &iflag,
	      work, &mw);

  bound_vec = nil;
  a = nil;
  
  return(iter);
}

