/****************************************************************************/
/* ALBERT:   an Adaptive multi Level finite element toolbox using           */
/*           Bisectioning refinement and Error control by Residual          */
/*           Techniques                                                     */
/*                                                                          */
/* file:                                                                    */
/*                                                                          */
/*                                                                          */
/* description:                                                             */
/*                                                                          */
/*                                                                          */
/* history:                                                                 */
/*                                                                          */
/*                                                                          */
/****************************************************************************/
/*                                                                          */
/*        authors: Alfred Schmidt and Kunibert G. Siebert                   */
/*                 Institut f"ur Angewandte Mathematik                      */
/*                 Albert-Ludwigs-Universit"at Freiburg                     */
/*                 Hermann-Herder-Str. 10                                   */
/*                 79104 Freiburg                                           */
/*                 Germany                                                  */
/*                                                                          */
/*        email:   alfred,kunibert@mathematik.uni-freiburg.de               */
/*                                                                          */
/*     (c) by A. Schmidt and K.G. Siebert (1996-1999)                       */
/*                                                                          */
/****************************************************************************/
#include "albert.h"

/****************************************************************************/
/*  calculates the L2 scalar product of the function f with each basis      */
/*  function of the mesh belonging to the dof admin of fh; the scalar       */
/*  products are stored in fh;                                              */
/*                                                                          */
/*    fh->vec[j] =  (f, phi_j)_\Omega                                       */
/*                                                                          */
/*  where phi_j are the basis functions belonging to fh->fe_space->admin    */
/*                                                                          */
/****************************************************************************/

void l2skp_fct_bas(REAL (*f)(const REAL_D), const QUAD *quad, 
		   DOF_REAL_VEC *fh)
{
  FUNCNAME("l2skp_fct_bas");
  static WORKSPACE ws = {0, nil};
  TRAVERSE_STACK  *stack = get_traverse_stack();
  EL_INFO         *el_info;
  MESH            *mesh;
  const QUAD_FAST *quad_fast;
  const REAL      *x;
  REAL            *wf_qp, *f_vec, **phi;
  REAL            val, det;
  int             iq, j, n_phi;
  const DOF       *(*get_dof)(const EL *, const DOF_ADMIN *, DOF *);
  const DOF       *dof;

  TEST_EXIT(fh)("no DOF_REAL_VEC fh\n");
  if (!f) return;

  TEST_EXIT(fh->fe_space)("no fe_space in DOF_REAL_VEC %s\n", NAME(fh));

  GET_STRUCT(mesh, fh->fe_space);
  GET_DOF_VEC(f_vec, fh);

  get_dof = fh->fe_space->bas_fcts->get_dof_indices;
  n_phi   = fh->fe_space->bas_fcts->n_bas_fcts;

  if (!quad)
    quad = get_quadrature(DIM, 2*fh->fe_space->bas_fcts->degree-2);
  quad_fast = get_quad_fast(fh->fe_space->bas_fcts, quad, INIT_PHI);
  phi = quad_fast->phi;

  REALLOC_WORKSPACE(&ws, quad->n_points*sizeof(REAL));
  wf_qp = ws.work;

  el_info = traverse_first(stack, mesh, -1, CALL_LEAF_EL|FILL_COORDS);
  while(el_info)
  {
    dof = get_dof(el_info->el, fh->fe_space->admin, nil);
    det = el_det(el_info);

    for (iq = 0; iq < quad->n_points; iq++)
    {
      x = coord_to_world(el_info, quad->lambda[iq], nil);
      wf_qp[iq] = quad->w[iq]*(*f)(x);
    }

    for (j = 0; j < n_phi; j++)
    {
      for (val = iq = 0; iq < quad->n_points; iq++)
	val += phi[iq][j]*wf_qp[iq];
      f_vec[dof[j]] += val*det;
    }
    el_info = traverse_next(stack, el_info);
  }
  free_traverse_stack(stack);

  return;
}

void l2skp_fct_bas_d(const REAL *(*f)(const REAL_D, REAL_D),
		     const QUAD *quad, DOF_REAL_D_VEC *fh)
{
  FUNCNAME("l2skp_fct_bas_d");
  static WORKSPACE ws = {0, nil};
  TRAVERSE_STACK  *stack = get_traverse_stack();
  EL_INFO         *el_info;
  MESH            *mesh;
  const QUAD_FAST *quad_fast;
  const REAL      *x;
  REAL_D          *wf_qp, *f_vec;
  REAL            val, det, **phi;
  int             iq, j, n, n_phi;
  const DOF       *(*get_dof)(const EL *, const DOF_ADMIN *, DOF *);
  const DOF       *dof;

  TEST_EXIT(fh)("no DOF_REAL_D_VEC fh\n");
  if (!f) return;

  TEST_EXIT(fh->fe_space)("no fe_space in DOF_REAL_VEC %s\n", NAME(fh));

  GET_STRUCT(mesh, fh->fe_space);
  GET_DOF_VEC(f_vec, fh);

  get_dof = fh->fe_space->bas_fcts->get_dof_indices;
  n_phi   = fh->fe_space->bas_fcts->n_bas_fcts;

  if (!quad)
    quad = get_quadrature(DIM, 2*fh->fe_space->bas_fcts->degree-2);
  quad_fast = get_quad_fast(fh->fe_space->bas_fcts, quad, INIT_PHI);
  phi = quad_fast->phi;

  REALLOC_WORKSPACE(&ws, quad->n_points*sizeof(REAL_D));
  wf_qp = ws.work;

  el_info = traverse_first(stack, mesh, -1, CALL_LEAF_EL|FILL_COORDS);
  while(el_info)
  {
    dof = get_dof(el_info->el, fh->fe_space->admin, nil);
    det = el_det(el_info);

    for (iq = 0; iq < quad->n_points; iq++)
    {
      x = coord_to_world(el_info, quad->lambda[iq], nil);
      (*f)(x,  wf_qp[iq]);
      for (n = 0; n < DIM_OF_WORLD; n++)
	wf_qp[iq][n] *= quad->w[iq];
    }

    for (j = 0; j < n_phi; j++)
    {
      for (n = 0; n < DIM_OF_WORLD; n++)
      {
	for (val = iq = 0; iq < quad->n_points; iq++)
	  val += phi[iq][j]*wf_qp[iq][n];
	f_vec[dof[j]][n] += val*det;
      }
    }
    el_info = traverse_next(stack, el_info);
  }
  free_traverse_stack(stack);

  return;
}

/****************************************************************************/
/*  sets dirichlet boundary values g() in the vectors fh and uh and fills   */
/*  the vector boundary with the boundary type of the corresponding DOF     */
/****************************************************************************/

void dirichlet_bound(REAL (*g)(const REAL_D), DOF_REAL_VEC *fh,
		     DOF_REAL_VEC *uh, DOF_SCHAR_VEC *bound)
{
  FUNCNAME("dirichlet_bound");
  static WORKSPACE ws = {0, nil};
  TRAVERSE_STACK  *stack = get_traverse_stack();
  EL_INFO         *el_info;
  MESH            *mesh;
  const DOF       *(*get_dof)(const EL *, const DOF_ADMIN *, DOF *);
  const DOF       *dof;
  DOF             *bound_dof;
  const S_CHAR    *(*get_bound)(const EL_INFO *, S_CHAR *);
  const S_CHAR    *el_bound;
  S_CHAR          *bound_vec;
  const REAL      *(*interpol)(const EL_INFO *, int, const int *,
			       REAL (*)(const REAL_D), REAL *);
  const REAL      *dirichlet;
  REAL            *f_vec = nil, *u_vec = nil;
  int             j, j_dof, n_phi, n_dirichlet;

  if (!fh && !uh && !bound) return;

  if (fh && uh)
    TEST_EXIT(fh->fe_space == uh->fe_space)("fe_spaces in fh and uh differ\n");
  else if (!fh && uh)
  {
    fh = uh;
    uh = nil;
  }

  if (fh)
  {
    GET_STRUCT(mesh, fh->fe_space);
    GET_DOF_VEC(f_vec, fh);
    if (uh) GET_DOF_VEC(u_vec, uh);

    get_bound = fh->fe_space->bas_fcts->get_bound;
    get_dof   = fh->fe_space->bas_fcts->get_dof_indices;
    n_phi     = fh->fe_space->bas_fcts->n_bas_fcts;
    mesh      = fh->fe_space->mesh;
  }
  else if (bound)
  {
    get_bound = bound->fe_space->bas_fcts->get_bound;
    get_dof   = bound->fe_space->bas_fcts->get_dof_indices;
    n_phi     = bound->fe_space->bas_fcts->n_bas_fcts;
    mesh      = bound->fe_space->mesh;
    bound_vec = bound->vec;

    el_info = traverse_first(stack, mesh, -1, CALL_LEAF_EL|FILL_BOUND);
    while(el_info)
    {
      dof      = get_dof(el_info->el, bound->fe_space->admin, nil);
      el_bound = get_bound(el_info, nil);

      for (j = 0; j < n_phi; j++)
	bound_vec[dof[j]] = el_bound[j];

      el_info = traverse_next(stack, el_info);
    }
    free_traverse_stack(stack);
    return;
  }

  if (bound)
  {
    TEST_EXIT(fh->fe_space == bound->fe_space)
      ("fh->fe_space, bound->fe_space differ!\n");

    bound_vec = bound->vec;
    FOR_ALL_DOFS(bound->fe_space->admin, bound_vec[dof] = 0);
  }
  else
    bound_vec = nil;

  REALLOC_WORKSPACE(&ws, n_phi*sizeof(DOF));
  bound_dof = ws.work;

  if (g)
    interpol = fh->fe_space->bas_fcts->interpol;

  el_info = traverse_first(stack, mesh, -1, 
			   CALL_LEAF_EL|FILL_COORDS|FILL_BOUND);
  while(el_info)
  {
    dof      = get_dof(el_info->el, fh->fe_space->admin, nil);
    el_bound = get_bound(el_info, nil);
    n_dirichlet = 0;

    if (bound_vec)
    {
      for (j = 0; j < n_phi; j++)
      {
	if (el_bound[j] >= DIRICHLET  &&  bound_vec[dof[j]] <= INTERIOR)
	{
/*--- Dirichlet boundary for this DOF not set! Set it now -------------------*/
	  bound_dof[n_dirichlet++] = j;
	}
	bound_vec[dof[j]] = el_bound[j];
      }
    }
    else
    {
      for (j = 0; j < n_phi; j++)
      {
	if (el_bound[j] >= DIRICHLET)
	{
/*--- set Dirichlet value for this DOF ------------------------------------*/
	  bound_dof[n_dirichlet++] = j;
	}
      }
    }
      
    if (n_dirichlet && g)
    {
/*--- compute coefficients for all Dirichlet nodes ------------------------*/
      dirichlet = (*interpol)(el_info, n_dirichlet, bound_dof, g, nil);
      
      if (u_vec)
      {
	for (j = 0; j < n_dirichlet; j++)
	{
	  j_dof = dof[bound_dof[j]];
	  u_vec[j_dof] = f_vec[j_dof] = dirichlet[j];
	}
      }
      else
      {
	for (j = 0; j < n_dirichlet; j++)
	{
	  j_dof = dof[bound_dof[j]];
	  f_vec[j_dof] = dirichlet[j];
	}
      }
    }
    else if (n_dirichlet)
    {
      if (u_vec)
      {
	for (j = 0; j < n_dirichlet; j++)
	{
	  j_dof = dof[bound_dof[j]];
	  u_vec[j_dof] = f_vec[j_dof] = 0.0;
	}
      }
      else
      {
	for (j = 0; j < n_dirichlet; j++)
	{
	  j_dof = dof[bound_dof[j]];
	  f_vec[j_dof] = 0.0;
	}
      }
    }

    el_info = traverse_next(stack, el_info);
  }
  free_traverse_stack(stack);

  return;
}

/****************************************************************************/
/*  sets dirichlet boundary values g() in the vectors fh and uh and fills   */
/*  the vector boundary with the boundary type of the corresponding DOF     */
/*  for vector valued problems                                              */
/****************************************************************************/

void dirichlet_bound_d(const REAL *(*g)(const REAL_D, REAL_D),
		       DOF_REAL_D_VEC *fh, DOF_REAL_D_VEC *uh,
		       DOF_SCHAR_VEC *bound)
{
  static WORKSPACE ws = {0, nil};
  TRAVERSE_STACK  *stack = get_traverse_stack();
  EL_INFO         *el_info;
  MESH            *mesh;
  const DOF       *(*get_dof)(const EL *, const DOF_ADMIN *, DOF *);
  const DOF       *dof;
  DOF             *bound_dof;
  const S_CHAR    *(*get_bound)(const EL_INFO *, S_CHAR *);
  const S_CHAR    *el_bound;
  S_CHAR          *bound_vec;
  const REAL_D    *(*interpol)(const EL_INFO *, int, const int *b_no,
			       const REAL *(*)(const REAL_D, REAL_D), 
			       REAL_D *);
  const REAL_D    *dirichlet;
  REAL_D          *f_vec = nil, *u_vec = nil;
  int             n, j, j_dof, n_phi, n_dirichlet;

  if (!fh && !uh && !bound) return;

  if (fh && uh)
    TEST_EXIT(fh->fe_space == uh->fe_space)("fe_spaces in fh and uh differ\n");
  else if (!fh && uh)
  {
    fh = uh;
    uh = nil;
  }

  if (fh)
  {
    GET_STRUCT(mesh, fh->fe_space);
    GET_DOF_VEC(f_vec, fh);
    if (uh) GET_DOF_VEC(u_vec, uh);

    get_bound = fh->fe_space->bas_fcts->get_bound;
    get_dof   = fh->fe_space->bas_fcts->get_dof_indices;
    n_phi     = fh->fe_space->bas_fcts->n_bas_fcts;
    mesh      = fh->fe_space->mesh;
  }
  else if (bound)
  {
    get_bound = bound->fe_space->bas_fcts->get_bound;
    get_dof   = bound->fe_space->bas_fcts->get_dof_indices;
    n_phi     = bound->fe_space->bas_fcts->n_bas_fcts;
    mesh      = bound->fe_space->mesh;
    bound_vec = bound->vec;

    el_info = traverse_first(stack, mesh, -1, CALL_LEAF_EL|FILL_BOUND);
    while(el_info)
    {
      dof      = get_dof(el_info->el, bound->fe_space->admin, nil);
      el_bound = get_bound(el_info, nil);

      for (j = 0; j < n_phi; j++)
	bound_vec[dof[j]] = el_bound[j];

      el_info = traverse_next(stack, el_info);
    }
    free_traverse_stack(stack);
    return;
  }

  if (bound)
  {
    TEST_EXIT(fh->fe_space == bound->fe_space)
      ("fh->fe_space, bound->fe_space differ!\n");

    bound_vec = bound->vec;
    FOR_ALL_DOFS(bound->fe_space->admin, bound_vec[dof] = 0);
  }
  else
    bound_vec = nil;

  REALLOC_WORKSPACE(&ws, n_phi*sizeof(DOF));
  bound_dof = ws.work;

  if (g)
    interpol = fh->fe_space->bas_fcts->interpol_d;

  el_info = traverse_first(stack, mesh, -1, 
			   CALL_LEAF_EL|FILL_COORDS|FILL_BOUND);
  while(el_info)
  {
    dof      = get_dof(el_info->el, fh->fe_space->admin, nil);
    el_bound = get_bound(el_info, nil);
    n_dirichlet = 0;

    if (bound_vec)
    {
      for (j = 0; j < n_phi; j++)
      {
	if (el_bound[j] >= DIRICHLET  &&  bound_vec[dof[j]] <= INTERIOR)
	{
/*--- Dirichlet boundary for this DOF not set! Set it now -------------------*/
	  bound_dof[n_dirichlet++] = j;
	}
	bound_vec[dof[j]] = el_bound[j];
      }
    }
    else
    {
      for (j = 0; j < n_phi; j++)
      {
	if (el_bound[j] >= DIRICHLET)
	{
/*--- set Dirichlet value for this DOF ------------------------------------*/
	  bound_dof[n_dirichlet++] = j;
	}
      }
    }
      
    if (n_dirichlet && g)
    {
/*--- compute coefficients for all Dirichlet nodes ------------------------*/
      dirichlet = (*interpol)(el_info, n_dirichlet, bound_dof, g, nil);
      
      if (u_vec)
      {
	for (j = 0; j < n_dirichlet; j++)
	{
	  j_dof = dof[bound_dof[j]];
	  for (n = 0; n < DIM_OF_WORLD; n++)
	    u_vec[j_dof][n] = f_vec[j_dof][n] = dirichlet[j][n];
	}
      }
      else
      {
	for (j = 0; j < n_dirichlet; j++)
	{
	  j_dof = dof[bound_dof[j]];
	  for (n = 0; n < DIM_OF_WORLD; n++)
	    f_vec[j_dof][n] = dirichlet[j][n];
	}
      }
    }
    else if (n_dirichlet)
    {
      if (u_vec)
      {
	for (j = 0; j < n_dirichlet; j++)
	{
	  j_dof = dof[bound_dof[j]];
	  for (n = 0; n < DIM_OF_WORLD; n++)
	    u_vec[j_dof][n] = f_vec[j_dof][n] = 0.0;
	}
      }
      else
      {
	for (j = 0; j < n_dirichlet; j++)
	{
	  j_dof = dof[bound_dof[j]];
	  for (n = 0; n < DIM_OF_WORLD; n++)
	    f_vec[j_dof][n] = 0.0;
	}
      }
    }

    el_info = traverse_next(stack, el_info);
  }
  free_traverse_stack(stack);

  return;
}
