/****************************************************************************/
/* ALBERT:   an Adaptive multi Level finite element toolbox using           */
/*           Bisectioning refinement and Error control by Residual          */
/*           Techniques                                                     */
/*                                                                          */
/* file:                                                                    */
/*                                                                          */
/*                                                                          */
/* description:                                                             */
/*                                                                          */
/*                                                                          */
/* history:                                                                 */
/*                                                                          */
/*                                                                          */
/****************************************************************************/
/*                                                                          */
/*        authors: Alfred Schmidt and Kunibert G. Siebert                   */
/*                 Institut f"ur Angewandte Mathematik                      */
/*                 Albert-Ludwigs-Universit"at Freiburg                     */
/*                 Hermann-Herder-Str. 10                                   */
/*                 79104 Freiburg                                           */
/*                 Germany                                                  */
/*                                                                          */
/*        email:   alfred,kunibert@mathematik.uni-freiburg.de               */
/*                                                                          */
/*     (c) by A. Schmidt and K.G. Siebert (1996,97)                         */
/*                                                                          */
/****************************************************************************/
/****************************************************************************/
/* to_do:                                                                   */
/*   free memory of blocks                               		    */
/*   be careful if mesh->max_edge_neigh changes (rc_lists...)	    */
/*                                                      		    */
/****************************************************************************/

#include "albert.h"

const DOF_ADMIN *get_dof_admin(MESH *mesh, const char *name, 
				      const int n_dof[DIM+1]);

/****************************************************************************/
/*  some routines for allocation and deallocation                           */
/****************************************************************************/

typedef struct free_mem       FREE_MEM;
typedef struct mem_info       MEM_INFO;
typedef struct mem_block_info MEM_BLOCK_INFO;

struct free_mem
{
  FREE_MEM  *next;
};

struct mem_info
{
  void      *begin, *end;
  MEM_INFO  *next;
};

struct mem_block_info
{
  FREE_MEM  *free;
  MEM_INFO  *info;

  int       free_count;
  int       used_count;
  int       unit_size;
  int       block_size;
};



static void *get_mem(MEM_BLOCK_INFO *mbi)
{
  FUNCNAME("get_mem");
  void     *vptr;

  TEST_EXIT(mbi)("mem_block_info == nil\n");

  if (!mbi->free)
  {
    char     *ptr;
    MEM_INFO *info;
    int      j, size;

    if (mbi->unit_size < sizeof(FREE_MEM *))
    {
      MSG("unit_size = %d of mem_block_info too small, changing to %d\n",
	  mbi->unit_size, sizeof(FREE_MEM *));
      mbi->unit_size = sizeof(FREE_MEM *);
    }
    size = mbi->unit_size * mbi->block_size;
    ptr  = MEM_ALLOC(size, char);

/****************************************************************************/
/* information about allocated memory block                                 */
/****************************************************************************/
    info = MEM_ALLOC(1, MEM_INFO);
    info->next = mbi->info;
    mbi->info  = info;
    info->begin = (void *)(mbi->free);
    info->end   = (void *)(ptr + size);

/****************************************************************************/
/* build list of units out of memory block                                  */
/****************************************************************************/
    mbi->free = (FREE_MEM *) ptr;
    for (j = 0; j < mbi->block_size - 1; j++)
    {
      ((FREE_MEM *)ptr)->next = (FREE_MEM *)(ptr + mbi->unit_size);
      ptr += mbi->unit_size;
    }
    ((FREE_MEM *)ptr)->next = nil;
    mbi->free_count += mbi->block_size;
  }

  vptr = (void *) mbi->free;
  mbi->free = mbi->free->next;

  mbi->free_count--;
  mbi->used_count++;

  return(vptr);
}

static void free_mem(void *mem, MEM_BLOCK_INFO *mbi)
{
  FUNCNAME("free_mem");

  TEST_EXIT(mbi)("mem_block_info == nil\n");

  ((FREE_MEM *)mem)->next = mbi->free;
  mbi->free = (FREE_MEM *) mem;

  mbi->free_count++;
  mbi->used_count--;
}



/****************************************************************************/
/*  allocate memory for a new mesh, and build a simply linked list of these */
/****************************************************************************/

typedef struct mesh_info MESH_INFO;

struct mesh_info
{
  MESH       *mesh;
  MESH_INFO  *next;
};

static MESH_INFO  *first_mesh = nil;


MESH *get_mesh(const char *name, void (*init_dof_admins)(MESH *),
	       void (*init_leaf_data)(LEAF_DATA_INFO *))
{
  FUNCNAME("get_mesh");
  MESH_INFO       *new_mesh;
  MESH            *mesh;

  new_mesh = MEM_CALLOC(1,MESH_INFO);
  mesh =     MEM_CALLOC(1,MESH);

  new_mesh->mesh = mesh;
  new_mesh->next = first_mesh;
  first_mesh = new_mesh;

  mesh->name = strdup(name);

  if (init_leaf_data)
    init_leaf_data(mesh->leaf_data_info);

  if (init_dof_admins)
    init_dof_admins(mesh);
  else
    WARNING("no dof admin specified; adding dofs at vertices\n");

  if (mesh->n_dof[VERTEX] == 0)
  {
    static int  n_dof[4] = {1,0,0,0};
    get_dof_admin(mesh, "vertex dofs", n_dof);
  }

  return(mesh);
}

void free_mesh(MESH *mesh)
{
  FUNCNAME("free_mesh");
  MESH_INFO *mesh_info, **prev;

  prev = &first_mesh;
  for (mesh_info = first_mesh; mesh_info; mesh_info = mesh_info->next)
  {
    if (mesh_info->mesh == mesh)
    {
      *prev = mesh_info->next;
      MEM_FREE(mesh_info, 1, MESH_INFO);

      /* free all elements/dofs/... ??? */

      MEM_FREE(mesh->dof_admin, mesh->n_dof_admin, DOF_ADMIN);

      MEM_FREE(mesh, 1, MESH);
      break;
    }
    prev = &(mesh_info->next);
  }
  TEST(mesh_info)("mesh not found; mesh not freed\n");

  return;
}

/****************************************************************************/
/*  memory management for DOF admin structures                              */
/****************************************************************************/

const DOF_ADMIN *get_dof_admin(MESH *mesh, const char *name, 
			       const int n_dof[DIM+1])
{
  FUNCNAME("get_dof_admin");
  static DOF_ADMIN  dof_admins[5];
  static DOF_ADMIN  *next_admin = dof_admins;
  static int        used;
  DOF_ADMIN         *admin;
  int               i;

  extern void add_dof_admin_to_mesh(DOF_ADMIN *, MESH *);

  if (used == 5)
  {
    next_admin =  MEM_CALLOC(5, DOF_ADMIN);
    used = 0;
  }
  admin = next_admin++;
  used++;
  admin->mesh = mesh;
  admin->name = strdup(name);

  admin->dof_free         = nil;
  admin->dof_free_size    = admin->first_hole     = 0;

  for (i = 0; i < DIM+1; i++)
    admin->n_dof[i] = n_dof[i];

  add_dof_admin_to_mesh(admin, mesh);

  return(admin);
}

#if 0
void free_dof_admin(DOF_ADMIN *admin, MESH *mesh)
{
  FUNCNAME("free_dof_admin");
  int     i, j, n;

  TEST_EXIT(admin && mesh)("no admin or mesh\n");

  n = mesh->n_dof_admin;
  if ((n > 0) & (mesh->dof_admin == nil))
    ERROR_EXIT("no mesh->dof_admin but n_dof_admin=%d\n", n);
  if ((n <= 0) & (mesh->dof_admin != nil))
    ERROR_EXIT("found mesh->dof_admin but n_dof_admin=%d\n", n);

  for (i = 0; i < n; i++)
  {
    if (admin == mesh->dof_admin[i])
    {
      mesh->n_dof_admin--;                 /* remove it from dof_admin list */
      for (j = i+1; j < n; j++)
	mesh->dof_admin[j-1] = mesh->dof_admin[j];

      if (admin->dof_int_vec || admin->dof_dof_vec || admin->int_dof_vec
	  || admin->dof_uchar_vec || admin->dof_schar_vec
	  || admin->dof_real_vec || admin->dof_real_d_vec
	  || admin->dof_matrix)
      {
	MSG("WARNING: admin still contains dof_vecs or dof_matrices\n");
      }
      else
      {
	MEM_FREE(admin->dof_free, admin->dof_free_size, DOF_FREE_UNIT);
	MEM_FREE(admin, 1, DOF_ADMIN);
      }
      break;
    }
  }
  TEST_EXIT(i < n)("admin not found in mesh->dof_admin\n");

  return;
}
#endif


/****************************************************************************/
/*  memory management for FE_SPACE structures                               */
/****************************************************************************/

const FE_SPACE *get_fe_space(MESH *mesh, const char *name,
			     const int n_dof[DIM+1],
			     const BAS_FCTS *bas_fcts)
{
  FUNCNAME("get_fe_space");
  const DOF_ADMIN   *admin = nil;
  FE_SPACE          *fe_space;
  int               i, j;
  const int         *ndof;

  extern void add_dof_admin_to_mesh(DOF_ADMIN *, MESH *);

  fe_space = MEM_ALLOC(1, FE_SPACE);
  fe_space->name = strdup(name);

  if (bas_fcts)  ndof = bas_fcts->n_dof;
  else           ndof = n_dof;
  TEST_EXIT(ndof)("no n_dof or bas_fcts->n_dof\n");

  for (i = 0; i < mesh->n_dof_admin; i++) {
    admin = mesh->dof_admin[i];
    for (j = 0; j <= DIM; j++) {
      if (admin->n_dof[j] != ndof[j]) break;
    }
    if (j > DIM) break;
    admin = nil;
  }

  if (!admin) {
    admin = get_dof_admin(mesh, name, ndof);
  }

  fe_space->admin = admin;
  fe_space->bas_fcts = bas_fcts;
  fe_space->mesh = mesh;

  TEST_EXIT((bas_fcts==nil) || (bas_fcts->init_element == 0))
    ("Use of bas_fcts->init_element is NOT IMPLEMENTED YET !!!\n");

  return(fe_space);
}


/****************************************************************************/
/*  memory management for DOF pointers                                      */
/****************************************************************************/

#define DOF_PTRS_BLOCK  1000

#if DIM == 2
#define MAX_DOF_PTRS       7
static MEM_BLOCK_INFO *dof_ptrs_block_infos[MAX_DOF_PTRS+1];
#else
#define MAX_DOF_PTRS      15
static MEM_BLOCK_INFO *dof_ptrs_block_infos[MAX_DOF_PTRS+1];
#endif

static DOF **get_dof_ptrs(MESH *mesh)
{
  FUNCNAME("get_dof_ptrs");
  int i, n;
  DOF **ptrs;

  TEST_EXIT(mesh)("mesh=nil\n");
  n = mesh->n_node_el;
  if (n <= 0)
    return(nil);

  TEST_EXIT(n <= MAX_DOF_PTRS)("too many nodes: %d > %d\n", n, MAX_DOF_PTRS);

  if (!(dof_ptrs_block_infos[n]))
  {
    dof_ptrs_block_infos[n] = MEM_ALLOC(1, MEM_BLOCK_INFO);
    dof_ptrs_block_infos[n]->free       = nil;
    dof_ptrs_block_infos[n]->info       = nil;
    dof_ptrs_block_infos[n]->free_count = 0;
    dof_ptrs_block_infos[n]->used_count = 0;
    dof_ptrs_block_infos[n]->unit_size  = n * sizeof(DOF*);
    dof_ptrs_block_infos[n]->block_size = DOF_PTRS_BLOCK;
  }

  ptrs = (DOF **)get_mem(dof_ptrs_block_infos[n]);
  for (i = 0; i < n; i++)
    ptrs[i] = nil;

  return(ptrs);
}


static void free_dof_ptrs(DOF **ptrs, MESH *mesh)
{
  FUNCNAME("free_dof_ptrs");
  int      n;

  TEST_EXIT(ptrs)("ptrs=nil\n");
  TEST_EXIT(mesh)("mesh=nil\n");
  n = mesh->n_node_el;
  if (n <= 0)
    return;

  TEST_EXIT(n <= MAX_DOF_PTRS)("too many nodes: %d > %d\n", n, MAX_DOF_PTRS);
  
  free_mem((void *)ptrs, dof_ptrs_block_infos[n]);
  return;
}


/****************************************************************************/
/*  memory management for DOFs                                              */
/****************************************************************************/

#define DOF_BLOCK 1000

#define MAX_DOF    100                             /* hope that's enough */

static MEM_BLOCK_INFO *dof_block_infos[MAX_DOF+1];

DOF *get_dof(MESH *mesh, int position)
{
  FUNCNAME("get_dof");
  DOF_ADMIN *admin;
  DOF     *dof;
  int     i, j, n, n0, ndof;

  TEST_EXIT(mesh)("mesh=nil\n");

  TEST_EXIT(position >= VERTEX && position <= CENTER)
    ("unknown position %d\n",position);

  ndof = mesh->n_dof[position];
  if (ndof <= 0) return(nil);

  TEST_EXIT(ndof <= MAX_DOF)
    ("ndof too big: ndof=%d, MAX_DOF=%d\n",ndof,MAX_DOF);

  if (!(dof_block_infos[ndof]))
  {
    dof_block_infos[ndof] = MEM_ALLOC(1, MEM_BLOCK_INFO);
    dof_block_infos[ndof]->free       = nil;
    dof_block_infos[ndof]->info       = nil;
    dof_block_infos[ndof]->free_count = 0;
    dof_block_infos[ndof]->used_count = 0;
    dof_block_infos[ndof]->unit_size  = sizeof(DOF)*ndof;
    dof_block_infos[ndof]->block_size = DOF_BLOCK;
  }
  dof = (DOF *)get_mem(dof_block_infos[ndof]);

  for (i = 0; i < mesh->n_dof_admin; i++)
  {
    admin = mesh->dof_admin[i];
    TEST_EXIT(admin)("no dof_admin[%d]\n", i);

    n  = admin->n_dof[position];
    n0 = admin->n0_dof[position];

    TEST_EXIT(n+n0 <= ndof)("n=%d, n0=%d too large: ndof=%d\n", n, n0, ndof);

    for (j = 0; j < n; j++)
    {
      dof[n0+j] = get_dof_index(admin);
    }
  }
  
  return(dof);
}

void free_dof(DOF *dof, MESH *mesh, int position)
{
  FUNCNAME("free_dof");
  DOF_ADMIN *admin;
  int     i, j, n, n0, ndof;

  TEST_EXIT(mesh)("mesh=nil\n");

  TEST_EXIT(position >= VERTEX && position <= CENTER)
    ("unknown position %d\n",position);

  ndof = mesh->n_dof[position];
  if (ndof) 
  {
    if (!dof)
    {
      MSG("dof = nil, but ndof=%d\n", ndof);
      return;
    }
  }
  else
  {
    if (dof)
    {
      MSG("dof != nil, but ndof=0\n");
    }
    return;
  }

  TEST_EXIT(ndof <= MAX_DOF)
    ("ndof too big: ndof=%d, MAX_DOF=%d\n",ndof,MAX_DOF);
  TEST_EXIT(dof_block_infos[ndof])
    ("no dof_block found for ndof=%d\n",ndof);

  for (i = 0; i < mesh->n_dof_admin; i++)
  {
    admin = mesh->dof_admin[i];
    TEST_EXIT(admin)("no dof_admin[%d]\n", i);

    n  = admin->n_dof[position];
    n0 = admin->n0_dof[position];

    TEST_EXIT(n+n0 <= ndof)("n=%d, n0=%d too large: ndof=%d\n", n, n0, ndof);

    for (j = 0; j < n; j++)
    {
      free_dof_index(admin, dof[n0+j]);
    }
  }

  free_mem((void *)dof, dof_block_infos[ndof]);
  return;
}

/****************************************************************************/
/*  memory management for leaf_data                                         */
/****************************************************************************/

#define LEAF_DATA_BLOCK  1000

void *get_leaf_data(MESH *mesh)
{
  FUNCNAME("get_leaf_data");
  int      n;
  void     *leaf_data;
  MEM_BLOCK_INFO  *ldbi;

  TEST_EXIT(mesh)("mesh=nil\n");
  n = mesh->leaf_data_info->leaf_data_size;
  if (n <= 0)
    return(nil);

  n = MAX(n, sizeof(void *));
  ldbi = (MEM_BLOCK_INFO *) mesh->leaf_data_info->leaf_data_block_info;

  if (!ldbi)
  {
    ldbi = MEM_ALLOC(1, MEM_BLOCK_INFO);
    ldbi->free       = nil;
    ldbi->info       = nil;
    ldbi->free_count = 0;
    ldbi->used_count = 0;
    ldbi->unit_size  = n;
    ldbi->block_size = LEAF_DATA_BLOCK;

    mesh->leaf_data_info->leaf_data_block_info = (const void *) ldbi;
  }

  TEST_EXIT(n == ldbi->unit_size)
    ("size %d of leaf data changes; old size = %d\n", n, ldbi->unit_size);
  leaf_data = (void *) get_mem(ldbi);

  return(leaf_data);
}


void free_leaf_data(void *leaf_data, MESH *mesh)
{
  FUNCNAME("free_leaf_data");
  int             n;
  MEM_BLOCK_INFO  *ldbi;

  if (!leaf_data)  return;

  TEST_EXIT(mesh)("mesh = nil\n");
  n = mesh->leaf_data_info->leaf_data_size;
  n = MAX(n, sizeof(void *));
  ldbi = (MEM_BLOCK_INFO *) mesh->leaf_data_info->leaf_data_block_info;
  if (!ldbi  || n <= 0) return;

  TEST_EXIT(n == ldbi->unit_size)
    ("size %d of leaf data changes; old size = %d\n", n, ldbi->unit_size);
  
  free_mem((void *) leaf_data, ldbi);
  return;
}

#if 0  /*  not used at the moment              */
/****************************************************************************/
/*  memory management for leaf_data                                         */
/****************************************************************************/

#define QUAD_VEC_BLOCK  100
#define MAX_QUAD_INFO     3

static MEM_BLOCK_INFO *quad_vec_block_infos[MAX_QUAD_INFO+1];

REAL *get_quad_vec(int dim)
{
  FUNCNAME("get_quad_vec");
  int             i, n;
  REAL            *quad_vec;
  MEM_BLOCK_INFO  *gvbi;

  TEST_EXIT(dim >= 1  &&  dim  <= 3)("dim = %d not allowed\n");

  gvbi = quad_vec_block_infos[dim];

  if (!gvbi)
  {
    gvbi = MEM_ALLOC(1, MEM_BLOCK_INFO);
    gvbi->free       = nil;
    gvbi->info       = nil;
    gvbi->free_count = 0;
    gvbi->used_count = 0;
    gvbi->unit_size  = get_max_no_quad_point(dim)*sizeof(REAL);
    gvbi->block_size = QUAD_VEC_BLOCK;

    quad_vec_block_infos[dim] = gvbi;
  }

  quad_vec = (REAL *) get_mem(gvbi);

  return(quad_vec);
}

void free_quad_vec(REAL *quad_vec, int dim)
{
  FUNCNAME("free_quad_vec");
  int             n;
  MEM_BLOCK_INFO  *gvbi;

  if (!quad_vec)  return;

  TEST_EXIT(dim >= 1  &&  dim  <= 3)("dim = %d not allowed\n");

  gvbi = quad_vec_block_infos[dim];

  if (!gvbi) return;
  
  free_mem((void *) quad_vec, gvbi);
  return;
}

static MEM_BLOCK_INFO *quad_vec_d_block_infos[MAX_QUAD_INFO+1];

REAL_D *get_quad_vec_d(int dim)
{
  FUNCNAME("get_quad_vec_d");
  int      i, n;
  REAL_D          *quad_vec_d;
  MEM_BLOCK_INFO  *gvdbi;

  TEST_EXIT(dim >= 1  &&  dim  <= 3)("dim = %d not allowed\n");

  gvdbi = quad_vec_d_block_infos[dim];

  if (!gvdbi)
  {
    gvdbi = MEM_ALLOC(1, MEM_BLOCK_INFO);
    gvdbi->free       = nil;
    gvdbi->info       = nil;
    gvdbi->free_count = 0;
    gvdbi->used_count = 0;
    gvdbi->unit_size  = get_max_no_quad_point(dim)*sizeof(REAL_D);
    gvdbi->block_size = QUAD_VEC_BLOCK;

    quad_vec_d_block_infos[dim] = gvdbi;
  }

  quad_vec_d = (REAL_D *) get_mem(gvdbi);

  return(quad_vec_d);
}

void free_quad_vec_d(REAL_D *quad_vec_d, int dim)
{
  FUNCNAME("free_quad_vec_d");
  int             n;
  MEM_BLOCK_INFO  *gvdbi;

  if (!quad_vec_d)  return;

  TEST_EXIT(dim >= 1  &&  dim  <= 3)("dim = %d not allowed\n");

  gvdbi = quad_vec_d_block_infos[dim];

  if (!gvdbi) return;
  
  free_mem((void *) quad_vec_d, gvdbi);
  return;
}
#endif

/****************************************************************************/
/*  memory management for elements                                          */
/****************************************************************************/

#define EL_BLOCK   1000

static MEM_BLOCK_INFO el_block_info = { nil, nil, 0, 0, sizeof(EL), EL_BLOCK};
static int el_index = 0;

EL  *get_element(MESH *mesh)
{
  EL       *el;
  FREE_MEM *free;

  if (el_block_info.free)
    el = (EL *)get_mem(&el_block_info);
  else {
    el = (EL *)get_mem(&el_block_info);
#if EL_INDEX
    el->index     = el_index++;
    for (free = el_block_info.free; free; free = free->next)
      ((EL *)free)->index = el_index++;
#endif
  }

  el->child[0]  = nil;
  el->child[1]  = (EL *) get_leaf_data(mesh);

  el->dof       = get_dof_ptrs(mesh);
#if DIM > 1
  el->new_coord = nil;
#endif

  return(el);
}

void free_element(EL *el, MESH *mesh)
{
  free_dof_ptrs(el->dof, mesh);
#if DIM > 1
  if (el->new_coord)
  {
    free_real_d(mesh, el->new_coord);
    el->new_coord = nil;
  }
#endif
  if (el->child[1])  free_leaf_data((void *) el->child[1], mesh);

  free_mem((void *)el, &el_block_info);
}


#if DIM == 3
/****************************************************************************/
/*  some routines for allocation and deallocation of list for collacating   */
/*  neighbours at the refinement edge in 3 dimensions                       */
/****************************************************************************/

#define LIST_BLOCK 20

static MEM_BLOCK_INFO rc_list_block_info = {nil, nil, 0, 0,  0, LIST_BLOCK};

RC_LIST_EL *get_rc_list(MESH *mesh)
{
  FUNCNAME("get_rc_list");

  if (rc_list_block_info.unit_size == 0)
  {
    rc_list_block_info.unit_size = 
      mesh->max_edge_neigh*sizeof(RC_LIST_EL);
  }
  else
  {
    TEST_EXIT(rc_list_block_info.unit_size 
	        == (mesh->max_edge_neigh * sizeof(RC_LIST_EL)))
      ("mesh->max_edge_neigh changed\n");
  }

  return((RC_LIST_EL *)get_mem(&rc_list_block_info));
}


void free_rc_list(MESH *mesh, RC_LIST_EL *list)
{
  free_mem((void *)list, &rc_list_block_info);
}

#endif

/****************************************************************************/
/*  memory management for matrix rows                                       */
/****************************************************************************/

#define REAL_D_BLOCK   1000

static MEM_BLOCK_INFO real_d_block_info = 
                  { nil, nil, 0, 0, sizeof(REAL_D), REAL_D_BLOCK};

REAL *get_real_d(MESH *mesh)
{
  return((REAL *)get_mem(&real_d_block_info));
}

void free_real_d(MESH *mesh, REAL *ptr)
{
  free_mem((void *)ptr, &real_d_block_info);
  return;
}

/****************************************************************************/
/*  memory management for matrix rows                                       */
/****************************************************************************/

#define ROW_BLOCK   1000

static MEM_BLOCK_INFO row_block_info = 
                  { nil, nil, 0, 0, sizeof(MATRIX_ROW), ROW_BLOCK};

MATRIX_ROW  *get_matrix_row(void)
{
  MATRIX_ROW *row;
  int        j;

  row = (MATRIX_ROW *)get_mem(&row_block_info);
  row->next = nil;
  for (j=0; j<ROW_LENGTH; j++) 
    row->col[j] = NO_MORE_ENTRIES;

  return(row);
}

void free_matrix_row(MATRIX_ROW *row)
{
  free_mem((void *)row, &row_block_info);
}

/****************************************************************************/
/*  memory management for matrices                                          */
/****************************************************************************/

#define MAT_BLOCK   10

static MEM_BLOCK_INFO mat_block_info = 
                  { nil, nil, 0, 0, sizeof(DOF_MATRIX), MAT_BLOCK};

DOF_MATRIX  *get_dof_matrix(const char *name, const FE_SPACE *fe_space)
{
  DOF_MATRIX *mat;
  void  add_dof_matrix_to_admin(DOF_MATRIX *, DOF_ADMIN *);

  mat = (DOF_MATRIX *)get_mem(&mat_block_info);

  mat->next   = nil;
  mat->fe_space = fe_space;
  mat->name   = strdup(name);
  mat->matrix_row = nil;
  mat->size   = 0;

  mat->refine_interpol = mat->coarse_restrict = nil;

  if (fe_space && fe_space->admin)
    add_dof_matrix_to_admin(mat, (DOF_ADMIN *)fe_space->admin);

  return(mat);
}

void free_dof_matrix(DOF_MATRIX *mat)
{
  void remove_dof_matrix_from_admin(DOF_MATRIX *);

  if (mat->fe_space && mat->fe_space->admin)
    remove_dof_matrix_from_admin(mat);

  clear_dof_matrix(mat);  /* free all matrix_rows */
  MEM_FREE(mat->matrix_row, mat->size, MATRIX_ROW *);
  mat->matrix_row = nil;
  mat->size = 0;
  free_mem((void *)mat, &mat_block_info);
}


/****************************************************************************/
/*  memory management for dof_vectors                                       */
/****************************************************************************/

#define VEC_BLOCK   10

static MEM_BLOCK_INFO dof_int_vec_block_info = 
                  { nil, nil, 0, 0, sizeof(DOF_INT_VEC), VEC_BLOCK};

static MEM_BLOCK_INFO dof_dof_vec_block_info = 
                  { nil, nil, 0, 0, sizeof(DOF_DOF_VEC), VEC_BLOCK};

static MEM_BLOCK_INFO dof_uchar_vec_block_info = 
                  { nil, nil, 0, 0, sizeof(DOF_UCHAR_VEC), VEC_BLOCK};

static MEM_BLOCK_INFO dof_schar_vec_block_info = 
                  { nil, nil, 0, 0, sizeof(DOF_SCHAR_VEC), VEC_BLOCK};

static MEM_BLOCK_INFO dof_real_vec_block_info = 
                  { nil, nil, 0, 0, sizeof(DOF_REAL_VEC), VEC_BLOCK};

static MEM_BLOCK_INFO dof_real_d_vec_block_info = 
                  { nil, nil, 0, 0, sizeof(DOF_REAL_D_VEC), VEC_BLOCK};

DOF_INT_VEC  *get_dof_int_vec(const char *name, const FE_SPACE *fe_space)
{
  DOF_INT_VEC *vec;

  void add_dof_int_vec_to_admin(DOF_INT_VEC *, DOF_ADMIN *);

  vec = (DOF_INT_VEC *)get_mem(&dof_int_vec_block_info);

  vec->next  = nil;
  vec->fe_space = fe_space;
  vec->name  = strdup(name);
  vec->size  = 0;
  vec->vec   = nil;
  vec->refine_interpol = vec->coarse_restrict = nil;

  if (fe_space && fe_space->admin)
    add_dof_int_vec_to_admin(vec, (DOF_ADMIN *)fe_space->admin);

  return(vec);
}

void free_dof_int_vec(DOF_INT_VEC *vec)
{
  void remove_dof_int_vec_from_admin(DOF_INT_VEC *);

  if (vec->fe_space && vec->fe_space->admin)
    remove_dof_int_vec_from_admin(vec);

  MEM_FREE(vec->vec, vec->size, int);
  vec->vec = nil;
  vec->size = 0;
  free_mem((void *)vec, &dof_int_vec_block_info);
}

DOF_DOF_VEC  *get_dof_dof_vec(const char *name, const FE_SPACE *fe_space)
{
  DOF_DOF_VEC *vec;

  void add_dof_dof_vec_to_admin(DOF_DOF_VEC *, DOF_ADMIN *);

  vec = (DOF_DOF_VEC *)get_mem(&dof_dof_vec_block_info);

  vec->next  = nil;
  vec->fe_space = fe_space;
  vec->name  = strdup(name);
  vec->size  = 0;
  vec->vec   = nil;
  vec->refine_interpol = vec->coarse_restrict = nil;

  if (fe_space && fe_space->admin)
    add_dof_dof_vec_to_admin(vec, (DOF_ADMIN *)fe_space->admin);

  return(vec);
}

void free_dof_dof_vec(DOF_DOF_VEC *vec)
{
  void remove_dof_dof_vec_from_admin(DOF_DOF_VEC *);

  if (vec->fe_space && vec->fe_space->admin)
    remove_dof_dof_vec_from_admin(vec);

  MEM_FREE(vec->vec, vec->size, int);
  vec->vec = nil;
  vec->size = 0;
  free_mem((void *)vec, &dof_dof_vec_block_info);
}

DOF_DOF_VEC  *get_int_dof_vec(const char *name, const FE_SPACE *fe_space)
{
  DOF_DOF_VEC *vec;

  void add_int_dof_vec_to_admin(DOF_DOF_VEC *, DOF_ADMIN *);

  vec = (DOF_DOF_VEC *)get_mem(&dof_dof_vec_block_info);

  vec->next  = nil;
  vec->fe_space = fe_space;
  vec->name  = strdup(name);
  vec->size  = 0;
  vec->vec   = nil;
  vec->refine_interpol = vec->coarse_restrict = nil;

  if (fe_space && fe_space->admin)
    add_int_dof_vec_to_admin(vec, (DOF_ADMIN *)fe_space->admin);

  return(vec);
}

void free_int_dof_vec(DOF_DOF_VEC *vec)
{
  void remove_int_dof_vec_from_admin(DOF_DOF_VEC *);

  if (vec->fe_space && vec->fe_space->admin)
    remove_int_dof_vec_from_admin(vec);

  MEM_FREE(vec->vec, vec->size, int);
  vec->vec = nil;
  vec->size = 0;
  free_mem((void *)vec, &dof_dof_vec_block_info);
}

DOF_UCHAR_VEC  *get_dof_uchar_vec(const char *name, const FE_SPACE *fe_space)
{
  DOF_UCHAR_VEC *vec;

  void add_dof_uchar_vec_to_admin(DOF_UCHAR_VEC *, DOF_ADMIN *);

  vec = (DOF_UCHAR_VEC *)get_mem(&dof_uchar_vec_block_info);

  vec->next = nil;
  vec->fe_space = fe_space;
  vec->name = strdup(name);
  vec->size = 0;
  vec->vec = nil;
  vec->refine_interpol = vec->coarse_restrict = nil;

  if (fe_space && fe_space->admin)
    add_dof_uchar_vec_to_admin(vec, (DOF_ADMIN *)fe_space->admin);

  return(vec);
}

void free_dof_uchar_vec(DOF_UCHAR_VEC *vec)
{
  void remove_dof_uchar_vec_from_admin(DOF_UCHAR_VEC *);

  if (vec->fe_space && vec->fe_space->admin)
    remove_dof_uchar_vec_from_admin(vec);

  MEM_FREE(vec->vec, vec->size, U_CHAR);
  vec->vec = nil;
  vec->size = 0;
  free_mem((void *)vec, &dof_uchar_vec_block_info);
}


DOF_SCHAR_VEC  *get_dof_schar_vec(const char *name, const FE_SPACE *fe_space)
{
  DOF_SCHAR_VEC *vec;

  void add_dof_schar_vec_to_admin(DOF_SCHAR_VEC *, DOF_ADMIN *);

  vec = (DOF_SCHAR_VEC *)get_mem(&dof_schar_vec_block_info);

  vec->next = nil;
  vec->fe_space = fe_space;
  vec->name = strdup(name);
  vec->size = 0;
  vec->vec = nil;
  vec->refine_interpol = vec->coarse_restrict = nil;

  if (fe_space && fe_space->admin)
    add_dof_schar_vec_to_admin(vec, (DOF_ADMIN *)fe_space->admin);

  return(vec);
}

void free_dof_schar_vec(DOF_SCHAR_VEC *vec)
{
  void remove_dof_schar_vec_from_admin(DOF_SCHAR_VEC *);

  if (vec->fe_space && vec->fe_space->admin)
    remove_dof_schar_vec_from_admin(vec);

  MEM_FREE(vec->vec, vec->size, S_CHAR);
  vec->vec = nil;
  vec->size = 0;
  free_mem((void *)vec, &dof_schar_vec_block_info);
}


DOF_REAL_VEC  *get_dof_real_vec(const char *name, const FE_SPACE *fe_space)
{
  DOF_REAL_VEC *vec;

  void add_dof_real_vec_to_admin(DOF_REAL_VEC *, DOF_ADMIN *);

  vec = (DOF_REAL_VEC *)get_mem(&dof_real_vec_block_info);

  vec->next = nil;
  vec->fe_space = fe_space;
  vec->name = strdup(name);
  vec->size = 0;
  vec->vec = nil;
  vec->refine_interpol = vec->coarse_restrict = nil;

  if (fe_space && fe_space->admin)
    add_dof_real_vec_to_admin(vec, (DOF_ADMIN *)fe_space->admin);

  return(vec);
}

void free_dof_real_vec(DOF_REAL_VEC *vec)
{
  void remove_dof_real_vec_from_admin(DOF_REAL_VEC *);

  if (vec->fe_space && vec->fe_space->admin)
    remove_dof_real_vec_from_admin(vec);

  MEM_FREE(vec->vec, vec->size, REAL);
  vec->vec = nil;
  vec->size = 0;
  free_mem((void *)vec, &dof_real_vec_block_info);
}


DOF_REAL_D_VEC  *get_dof_real_d_vec(const char *name, const FE_SPACE *fe_space)
{
  DOF_REAL_D_VEC *vec;

  void add_dof_real_d_vec_to_admin(DOF_REAL_D_VEC *, DOF_ADMIN *);

  vec = (DOF_REAL_D_VEC *)get_mem(&dof_real_d_vec_block_info);

  vec->next = nil;
  vec->fe_space = fe_space;
  vec->name = strdup(name);
  vec->size = 0;
  vec->vec = nil;
  vec->refine_interpol = vec->coarse_restrict = nil;

  if (fe_space && fe_space->admin)
    add_dof_real_d_vec_to_admin(vec, (DOF_ADMIN *)fe_space->admin);

  return(vec);
}

void free_dof_real_d_vec(DOF_REAL_D_VEC *vec)
{
  void remove_dof_real_d_vec_from_admin(DOF_REAL_D_VEC *);

  if (vec->fe_space && vec->fe_space->admin)
    remove_dof_real_d_vec_from_admin(vec);

  MEM_FREE(vec->vec, vec->size, REAL_D);
  vec->vec = nil;
  vec->size = 0;
  free_mem((void *)vec, &dof_real_d_vec_block_info);
}





