#include "albert.h"

#define SPARSE_MG 1
#define RESTRICT_MATRICES 1
#define TEST_RESTRICT 0
#define MG_GRAPHICS 0           /* show intermediate values          */
#define GRAPH_REFINE 3          /* recursive refinement for graphics */


#include <time.h>
#include <string.h>

static LIN_SYS_INFO mg_lin_sys_info[1]
   = {{nil, nil, nil, nil, nil, 0, 0.0, 0, 0}};

#if MG_GRAPHICS
static GRAPH_WINDOW mg_win = 0;

void graph_values_mg(GRAPH_WINDOW win, const DOF_REAL_VEC *v, int refine,
		     int mg_level, const FE_SPACE *fe_space,
		     const int *sort_dof_invers);
#endif

/****************************************************************************/
/*  sort_dof[ sorted dof ]          = unsorted dof                          */
/*  sort_dof_invers[ unsorted dof ] = sorted dof                            */
/****************************************************************************/

typedef struct
{
  const FE_SPACE   *fe_space;
  int              size;                          /* current size of vectors*/
  int              mg_levels;                     /* current no. of levels  */
  LIN_SYS_INFO     *lin_sys_info;                 /* one for each level     */
  DOF_MATRIX       **matrix;                      /* one for each level     */
  DOF_REAL_VEC     **f_h;                         /* one for each level     */
  DOF_REAL_VEC     **u_h;                         /* one for each level     */
  DOF_REAL_VEC     **r_h;                         /* one for each level     */
  DOF_UCHAR_VEC    *dof_level;
  int              cycle;                         /* 1=V-cycle, 2=W-cycle   */
  int              presmooth, postsmooth;
  int              smoother, coarse_solver;
  int              exact_level;                   /* level for exact solver */
  REAL             tol;                           /* tol. for inner solver  */

  int              *dofs_per_level;               /* count dofs per level   */
  int              sort_size;                     /* size of sort vectors   */
  int              *sort_dof;                     /* dofs in order of levels*/
  int              *sort_dof_invers;              /* inverse sort list      */
  DOF_DOF_VEC      *(dof_parent[2]);              /* (for linear elements)  */
  int              *tmp_per_level;                /* work area              */

} MULTI_GRID_INFO;



MULTI_GRID_INFO multi_grid_info[1] 
   = {{nil, 0, 0, mg_lin_sys_info,
       nil, nil, nil, nil, nil, 1, 1, 1, 0, 0, 0, 1.0,
       nil, 0, nil, nil, {nil,nil}, nil }};

#define FINE_GRID_OMEGA 0
#if FINE_GRID_OMEGA
static REAL fine_grid_omega = 1.0;
#endif

#if SPARSE_MG
static DOF *sort_dof        = nil;
static DOF *sort_dof_invers = nil;
static int *dofs_per_level  = nil;
static DOF *(dof_parent[2]) = {nil, nil};
#endif

#if SPARSE_MG
/****************************************************************************/
/* routines for sparse dof_matrix and dof_vector handling:                  */
/*                                                                          */
/*  dof_copy_to_sparse(DOF_ADMIN *admin, MULTI_GRID_INFO *mg_info,          */
/*       	       DOF_REAL_VEC *x, DOF_REAL_VEC *y)                    */
/*  dof_copy_from_sparse(DOF_ADMIN *admin, MULTI_GRID_INFO *mg_info,        */
/*			 DOF_REAL_VEC *x, DOF_REAL_VEC *y)                  */
/*  dof_copy_sparse(MULTI_GRID_INFO *mg_info, int mg_level,                 */
/*		    DOF_REAL_VEC *x, DOF_REAL_VEC *y)                       */
/*  dof_set_sparse(MULTI_GRID_INFO *mg_info, int mg_level,                  */
/*		    REAL alpha, DOF_REAL_VEC *x)                            */
/*  REAL dof_nrm2_sparse(MULTI_GRID_INFO *mg_info, int mg_level,            */
/*			    DOF_REAL_VEC *x)                                */
/*  dof_gemv_sparse(MULTI_GRID_INFO *mg_info, int mg_level,                 */
/*		    MatrixTranspose transpose, REAL alpha,                  */
/*		    DOF_MATRIX *a, DOF_REAL_VEC *x,                         */
/*		    REAL beta, DOF_REAL_VEC *y)                             */
/****************************************************************************/

static void dof_copy_to_sparse(MULTI_GRID_INFO *mg_info, 
			       DOF_REAL_VEC *x, DOF_REAL_VEC *y)
{
  FUNCNAME("dof_copy_to_sparse");
  int             j, jmax, *sort_dof_invers;
  REAL            *xvec, *yvec;
  const DOF_ADMIN *admin;

  TEST_EXIT(mg_info && y)
    ("no mg_info or y\n");
  TEST_EXIT(x && x->fe_space && (admin=x->fe_space->admin))
    ("no x or x->fe_space or x->fe_space->admin\n");

  xvec = x->vec;
  yvec = y->vec;
  jmax = mg_info->dofs_per_level[mg_info->mg_levels-1];
  sort_dof_invers = mg_info->sort_dof_invers;

  FOR_ALL_DOFS(admin,
	       j = sort_dof_invers[dof];
	       TEST_EXIT(j < jmax)("j=%d too big; dof=%d, max+1=%d\n",
				   j, dof, jmax);
	       yvec[j] = xvec[dof];
    );
}


static void dof_copy_from_sparse(MULTI_GRID_INFO *mg_info, 
				 DOF_REAL_VEC *x, DOF_REAL_VEC *y)
{
  FUNCNAME("dof_copy_from_sparse");
  int             i, j, jmax, *sort_dof, ysize;
  REAL            *xvec, *yvec;
  const DOF_ADMIN *admin;

  TEST_EXIT(mg_info && x)
    ("no mg_info or x\n");
  TEST_EXIT(y && y->fe_space && (admin=y->fe_space->admin))
    ("no y or y->fe_space or y->fe_space->admin\n");

  sort_dof = mg_info->sort_dof;
  jmax = mg_info->dofs_per_level[mg_info->mg_levels-1];
  xvec = x->vec;
  yvec = y->vec;
  ysize = y->size;

  for (j = 0; j < jmax; j++)
  {
    i = sort_dof[j];
    TEST_EXIT(i < ysize)("i=%d too big; j=%d, y->size=%d\n", i, j, ysize);
    yvec[i] = xvec[j];
  }
}

static void dof_copy_sparse(MULTI_GRID_INFO *mg_info, int mg_level,
			    DOF_REAL_VEC *x, DOF_REAL_VEC *y)
{
  FUNCNAME("dof_copy_sparse");
  int   i, last;
  REAL  *xvec = x->vec;
  REAL  *yvec = y->vec;

  TEST_EXIT(mg_level < mg_info->mg_levels)
    ("mg_level %d >= mg_info->mg_levels %d\n", mg_level, mg_info->mg_levels);
  last = mg_info->dofs_per_level[mg_level];

  for (i = 0; i < last; i++) yvec[i] = xvec[i];
}


void dof_set_sparse(MULTI_GRID_INFO *mg_info, int mg_level,
		    REAL alpha, DOF_REAL_VEC *x)
{
  FUNCNAME("dof_set_sparse");
  int     i, last;
  REAL    *xvec;

  TEST_EXIT(mg_level < mg_info->mg_levels)
    ("mg_level %d >= mg_info->mg_levels %d\n", mg_level, mg_info->mg_levels);
  TEST_EXIT(mg_info && x)("pointer is nil: %8X, %8X", mg_info, x);

  last = mg_info->dofs_per_level[mg_level];
  TEST_EXIT(x->size >= last)
    ("x->size = %d too small: dofs_per_level = %d", x->size, last);

  xvec = x->vec;
  for (i = 0; i < last; i++) xvec[i] = alpha;
}


static REAL dof_nrm2_sparse(MULTI_GRID_INFO *mg_info, int mg_level,
			    DOF_REAL_VEC *x)
{
  FUNCNAME("dof_nrm2_sparse");
  int     i, last;
  REAL    nrm, *xvec;

  TEST_EXIT(mg_info && x)("pointer is nil: %8X, %8X", mg_info, x);
  TEST_EXIT(mg_level < mg_info->mg_levels)("mg_level %d too big\n",mg_level);
  last = mg_info->dofs_per_level[mg_level];
  TEST_EXIT(x->size >= last)
    ("x->size = %d too small: dofs_per_level = %d", x->size, last);

  xvec = x->vec;
  nrm = 0.0;
  for (i = 0; i < last; i++) {
    nrm += *xvec * *xvec;
    xvec++;
  }

  return(sqrt(nrm));
}

static void clear_dof_matrix_sparse(MULTI_GRID_INFO *mg_info, int mg_level,
				    DOF_MATRIX *matrix)
{
  FUNCNAME("clear_dof_matrix_sparse");
  int  i, j, last;
  MATRIX_ROW *row, *next;

  for (i = 0; i < matrix->size; i++) {
    for (row = matrix->matrix_row[i]; row; row = next) {
      next = row->next;
      free_matrix_row(row);
    }
    matrix->matrix_row[i] = nil;
  }

  return;
}


static void dof_gemv_sparse(MULTI_GRID_INFO *mg_info, int mg_level,
			    MatrixTranspose transpose, REAL alpha,
			    DOF_MATRIX *a, DOF_REAL_VEC *x,
			    REAL beta, DOF_REAL_VEC *y)
{
  FUNCNAME("dof_gemv_sparse");
  int        i,j,jcol,last;
  REAL       sum, ax;
  REAL       *xvec, *yvec;
  MATRIX_ROW *row;
  
  TEST_EXIT(mg_info && a && x && y)("pointer is nil: %8X, %8X, %8X, %8X",
				    mg_info, a,x,y);

  TEST_EXIT(mg_level < mg_info->mg_levels)
    ("mg_level %d >= mg_info->mg_levels %d\n", mg_level, mg_info->mg_levels);
  last = mg_info->dofs_per_level[mg_level];

  
  TEST_EXIT(x->size >= last)
    ("x->size = %d too small: dofs_per_level = %d", x->size, last);
  TEST_EXIT(y->size >= last)
    ("y->size = %d too small: dofs_per_level = %d", y->size, last);
  TEST_EXIT(a->size >= last)
    ("a->size = %d too small: dofs_per_level = %d", a->size, last);

  xvec = x->vec;
  yvec = y->vec;

  if (transpose == NoTranspose) {
    for (i = 0; i < last; i++) {
      sum = 0.0;
      for (row = a->matrix_row[i]; row; row = row->next) {
	for (j=0; j<ROW_LENGTH; j++) {
	  jcol = row->col[j];
	  if (ENTRY_USED(jcol)) {
	    sum += row->entry[j] * xvec[jcol];
	  }
	  else {
	    if (jcol == NO_MORE_ENTRIES)
	      break;
	  }
	}
      }
      yvec[i] *= beta;
      yvec[i] += alpha * sum;
    }
  }

  else if (transpose == Transpose) {
    for (i = 0; i < last; i++) {
      yvec[i] *= beta;
    }

    for (i = 0; i < last; i++) {
      ax = alpha * xvec[i];
      for (row = a->matrix_row[i]; row; row = row->next) {
	for (j=0; j<ROW_LENGTH; j++) {
	  jcol = row->col[j];
	  if (ENTRY_USED(jcol)) {
	    yvec[jcol] += ax * row->entry[j];
	  }
	  else {
	    if (jcol == NO_MORE_ENTRIES)
	      break;
	  }
	}
      }
    }
  }

  else {
    ERROR_EXIT("transpose=%d", transpose);
  }
}

/****************************************************************************/

static void update_sparse_dof_matrix(DOF_MATRIX *matrix, REAL sign, int dim, 
			      const REAL **el_mat, const DOF *el_dof)
{
  FUNCNAME("update_sparse_dof_matrix");
  int        i,j,k,irow,jcol;
  int        free_col;
  MATRIX_ROW *row, *free_row;

  TEST_EXIT(matrix)("no matrix");
  TEST_EXIT(el_mat && el_dof)("no el_mat or el_dof");

  for (i=0; i < dim; i++) {
    irow = sort_dof_invers[el_dof[i]];
    TEST_EXIT(irow < matrix->size)
      ("irow = %d, mat.size = %d, dof = %d, lev = %d",
       irow, matrix->size, el_dof[i], 
       multi_grid_info->dof_level->vec[el_dof[i]]);
  
    if (matrix->matrix_row[irow] == nil) {
      row = matrix->matrix_row[irow] = get_matrix_row();
      row->col[0] = irow;           /* first entry is diagonal element */
      row->entry[0] = 0.0;
    }
    for (j=0; j < dim; j++) {
      jcol = sort_dof_invers[el_dof[j]];
      TEST_EXIT(jcol < matrix->size)
	("jcol = %d, matrix->size = %d\n", jcol, matrix->size);

      row = matrix->matrix_row[irow];
      free_row = nil;
      do {
	for (k=0; k<ROW_LENGTH; k++) {
	  if (row->col[k] == jcol) {
	    row->entry[k] += sign * el_mat[i][j];
	    break;
	  }
	  if (ENTRY_NOT_USED(row->col[k])) {
	    free_col = k;
	    free_row = row;
	    if (row->col[k] == NO_MORE_ENTRIES) {
	      k = ROW_LENGTH;
	      break;
	    }
	  }
	}
	if (k < ROW_LENGTH) break;               /* done? */
	if (row->next || free_row) {
	  row = row->next;
	}
	else {
	  free_row = row->next = get_matrix_row();
	  free_col = 0;
	  row = nil;
	}
      } while (row);

      if (k >= ROW_LENGTH) {                     /* not done? */
	TEST_EXIT(free_row)("no free_row");
	TEST_EXIT(sign > 0.0)("new entry (%d,%d) in matrix %s with sign=%lf",
			      irow, jcol, matrix->name, sign);
	free_row->col[free_col]   = jcol;
	free_row->entry[free_col] = sign * el_mat[i][j];
      }

    }
  }
}

/****************************************************************************/
/* new_size(): for enlargement of sparse multigrid matrices/vectors         */

static int new_size(int level, int *dofs_per_level)
{
  FUNCNAME("new_size");
  int size0, size;

  size0 = (dofs_per_level[0] - (int)(sqrt(dofs_per_level[0])))
    * (int)(pow(2.0, (double)(level*DIM)));
  if (size0 > dofs_per_level[level] + level*50)
    size = dofs_per_level[level] + level*50;
  else
    size = MAX(size0, dofs_per_level[level]);

#if 0
  MSG("level %2d: ndofs=%5d, size0=%5d, size=%5d\n",
      level, dofs_per_level[level], size0, size);
#endif

  return(size);
}


/****************************************************************************/

static void init_sparse_mg(MESH *mesh, LIN_SYS_INFO *lin_sys_info, int levels)
{
  FUNCNAME("init_sparse_mg");
  U_CHAR  *dof_level;
  int     *tmp_per_level;
  int     i, j, k, size;
  const DOF_ADMIN  *admin;
  DOF_MATRIX       *mat;
  DOF_REAL_VEC     *drv;

  TEST_EXIT(lin_sys_info && lin_sys_info->matrix
	    && lin_sys_info->matrix->fe_space
	    && (admin = lin_sys_info->matrix->fe_space->admin))
    ("no lin_sys_info or matrix or fe_space or admin\n");

  if (levels > multi_grid_info->size)
  {
    multi_grid_info->dofs_per_level = multi_grid_info->dofs_per_level ?
      MEM_REALLOC(multi_grid_info->dofs_per_level,
		  multi_grid_info->size, levels, int)
      : MEM_ALLOC(levels, int);
    multi_grid_info->tmp_per_level = multi_grid_info->tmp_per_level ?
      MEM_REALLOC(multi_grid_info->tmp_per_level,
		  multi_grid_info->size, levels, int)
      : MEM_ALLOC(levels, int);
  }
  multi_grid_info->mg_levels = levels;

  if (admin->used_count > multi_grid_info->sort_size)
  {
    multi_grid_info->sort_dof = multi_grid_info->sort_dof ?
      MEM_REALLOC(multi_grid_info->sort_dof,
		  multi_grid_info->sort_size,
		  admin->used_count, int)
      : MEM_ALLOC(admin->used_count, int);
  }
  if (admin->size > multi_grid_info->sort_size)
  {
    multi_grid_info->sort_dof_invers = multi_grid_info->sort_dof_invers ?
      MEM_REALLOC(multi_grid_info->sort_dof_invers,
		  multi_grid_info->sort_size,
		  admin->size, int)
      : MEM_ALLOC(admin->size, int);
  }

  dof_level       = multi_grid_info->dof_level->vec;
  sort_dof        = multi_grid_info->sort_dof;
  sort_dof_invers = multi_grid_info->sort_dof_invers;
  dofs_per_level  = multi_grid_info->dofs_per_level;
  tmp_per_level   = multi_grid_info->tmp_per_level;

  dof_parent[0]   = multi_grid_info->dof_parent[0]->vec;
  dof_parent[1]   = multi_grid_info->dof_parent[1]->vec;

/****************************************************************************/
/* count dofs per multigrid level                                           */

  for (i = 0; i < levels; i++) dofs_per_level[i] = 0;

  FOR_ALL_DOFS(admin, dofs_per_level[dof_level[dof]]++);

  if (lin_sys_info->info > 2) {
    MSG("dofs_per_level:");
    for (i = 0; i < levels; i++) print_msg(" %d", dofs_per_level[i]);
    print_msg("\n");
  }

  for (i = 1; i < levels; i++) {
    tmp_per_level[i]   = dofs_per_level[i-1];
    dofs_per_level[i] += dofs_per_level[i-1];
  }
  tmp_per_level[0] = 0;            /* pointers for filling the sort vectors */


  if (lin_sys_info->info > 2) {
    MSG("dofs_per_level accumulated:");
    for (i = 0; i < levels; i++) print_msg(" %d", dofs_per_level[i]);
    print_msg("\n");
  }


#if 0
  for (i = 0; i < dofs_per_level[levels-1]; i++)
    MSG("dof_parent[%3d] = (%3d,%3d), lev=%2d (%2d,%2d)\n",
	i, dof_parent[0][i], dof_parent[1][i],
	dof_level[i], dof_level[dof_parent[0][i]],
	dof_level[dof_parent[1][i]]);
#endif

/****************************************************************************/
/* build sort_dof[] and sort_dof_invers[] vectors                           */

  FOR_ALL_DOFS(admin,
	       j = dof_level[dof];
	       k = tmp_per_level[j]++;
	       sort_dof[k] = dof;
	       sort_dof_invers[dof] = k;
    );

#if 0
  for (i = 0; i < dofs_per_level[levels-1]; i++)
  {
    j = sort_dof[i];
    MSG("sort[%3d]: dof=%3d, lev=%2d; invers[%3d]=%3d\n",
	i, j, dof_level[j], j, sort_dof_invers[j]);
  }
  WAIT;
#endif

/****************************************************************************/
/* enlarge vectors and matrices, if necessary                               */
/****************************************************************************/

  for (i = 0; i < levels; i++)
  {
    mat = multi_grid_info->matrix[i];
    if (mat->size < dofs_per_level[i])
    {
      size = new_size(i, dofs_per_level);

      mat->matrix_row = mat->matrix_row ?
	MEM_REALLOC(mat->matrix_row, mat->size, size, MATRIX_ROW *)
	: MEM_ALLOC(size, MATRIX_ROW *);
      for (j = mat->size; j < size; j++)          /* clear new row entries! */
	mat->matrix_row[j] = nil;
      mat->size = size;
    }

    drv = multi_grid_info->f_h[i];
    if (drv->size < dofs_per_level[i])
    {
      size = new_size(i, dofs_per_level);

      drv->vec = drv->vec ?
	MEM_REALLOC(drv->vec, drv->size, size, REAL)
	: MEM_ALLOC(size, REAL);
      drv->size = size;
    }

    drv = multi_grid_info->u_h[i];
    if (drv->size < dofs_per_level[i])
    {
      size = new_size(i, dofs_per_level);

      drv->vec = drv->vec ?
	MEM_REALLOC(drv->vec, drv->size, size, REAL)
	: MEM_ALLOC(size, REAL);
      drv->size = size;
    }

    drv = multi_grid_info->r_h[i];
    if (drv->size < dofs_per_level[i])
    {
      size = new_size(i, dofs_per_level);

      drv->vec = drv->vec ?
	MEM_REALLOC(drv->vec, drv->size, size, REAL)
	: MEM_ALLOC(size, REAL);
      drv->size = size;
    }

  }
  return;
}


static void diff_sparse_nonsparse_matrices(MULTI_GRID_INFO *mg_info, int level,
					   DOF_MATRIX *nonsparse)
{
  FUNCNAME("diff_sparse_nonsparse_matrices");
  DOF_MATRIX *sparse = mg_info->matrix[level];
  MATRIX_ROW **Srows = sparse->matrix_row, *Srow;
  MATRIX_ROW **NSrows = nonsparse->matrix_row, *NSrow;
  int    i, j, k, size, i2, j2, k2, k3;
  int    *sort_dof = mg_info->sort_dof;
  int    flag;
  REAL   diff, max_diff = 0.0;

  size = mg_info->dofs_per_level[level];
  TEST(size == nonsparse->fe_space->admin->used_count)
    ("different sizes: %d  %d\n", size,
     nonsparse->fe_space->admin->used_count);

  for (i = 0; i < size; i++) {
    for (Srow = Srows[i]; Srow; Srow = Srow->next) {
      for (j = 0; j < ROW_LENGTH; j++) {
	k = Srow->col[j];
	if (ENTRY_USED(k)) {
	  flag = false;
	  k3 = sort_dof[k];
	  i2 = sort_dof[i];
	  for (NSrow = NSrows[i2]; NSrow; NSrow = NSrow->next)
	  {
	    for (j2 = 0; j2 < ROW_LENGTH; j2++) {
	      k2 = NSrow->col[j2];
	      if (ENTRY_USED(k2)) {
		if (k2==k3) {
		  diff = Srow->entry[j] - NSrow->entry[j2];
		  diff = ABS(diff);
		  TEST(diff < 1.E-5)
		    ("entries differ: (%d,%d)=%le, (%d,%d)=%le\n",
		     i, k, Srow->entry[j], i2, k2, NSrow->entry[j2]);
		  max_diff = MAX(max_diff, diff);
		  flag = true;
		}
	      }
	      else {
		if (k2 == NO_MORE_ENTRIES) break;
	      }
	    }
	  }
	  TEST(flag)("entry (%d,%d) not in nonsparse matrix\n", i2, k3);
	}
	else {
	  if (k==NO_MORE_ENTRIES) break;
	}
      }
    }
  }
  if (max_diff >= 1.E-5) MSG("matrices differ !!!\n");
  MSG("max_diff = %le\n", max_diff);
  WAIT;
}


#endif /* SPARSE_MG */


/****************************************************************************/
/* init_multi_grid(): setup multi_grid_info and all matrices/vectors        */
/****************************************************************************/

static int max_level = 0, max_dof_level = 0;
static U_CHAR *dof_level = nil;
static DOF    *dof_parent0 = nil, *dof_parent1 = nil;

static void max_level_fct(const EL_INFO *el_info)
{
  FUNCNAME("max_level_fct");
  int dof, dof0, dof1, level = (int)(el_info->level);
  EL  *el   = el_info->el;

  max_level = MAX(level, max_level);

  if (!(IS_LEAF_EL(el))) {
    dof   = el->child[0]->dof[N_VERTICES-1][0];
    dof0  = dof_parent0[dof] = el->dof[0][0];
    dof1  = dof_parent1[dof] = el->dof[1][0];
    level = dof_level[dof]   = 1 + MAX(dof_level[dof0], dof_level[dof1]);
    max_dof_level = MAX(level, max_dof_level);
  }
}
/****************************************************************************/

static void init_multi_grid(MESH *mesh, LIN_SYS_INFO *lin_sys_info)
{
  FUNCNAME("init_multi_grid");
  int          levels, i;
  int          el_dim;
  void         (*fill_el_mat)(EL_INFO *, DOF_MATRIX *);
  char         name[64];
  static const REAL zero = 0.0;
  const FE_SPACE *fe_space;
#if SPARSE_MG
  int          j, k;
  int          size, size0, i1;
  DOF_MATRIX   *mat;
  DOF_REAL_VEC *drv;
#endif

  multi_grid_info->presmooth = 1;
  GET_PARAMETER(1, "MG presmooth", "%d", &multi_grid_info->presmooth);
  multi_grid_info->postsmooth = 1;
  GET_PARAMETER(1, "MG postsmooth", "%d", &multi_grid_info->postsmooth);
  multi_grid_info->cycle = 1;
  GET_PARAMETER(1, "MG cycle", "%d", &multi_grid_info->cycle);
  multi_grid_info->exact_level = 0; 
  GET_PARAMETER(1, "MG exact_level", "%d", &multi_grid_info->exact_level);
  lin_sys_info->omega = 1.0;
  GET_PARAMETER(1, "MG omega", "%e", &lin_sys_info->omega);

  TEST_EXIT(lin_sys_info)("no lin_sys_info");
  TEST_EXIT(lin_sys_info->matrix)("no lin_sys_info->matrix");
  fe_space = lin_sys_info->matrix->fe_space;
  multi_grid_info->fe_space = fe_space;

  TEST_EXIT(fe_space->bas_fcts)("no bas_fcts\n");
  if (strcmp(fe_space->bas_fcts->name, "lagrange1")) {
    ERROR_EXIT("not for bas_fcts <%s>, only for <lagrange1>\n",
	       fe_space->bas_fcts->name);
  }

  if (!(multi_grid_info->dof_level)) {
    multi_grid_info->dof_level = get_dof_uchar_vec("dof_level", fe_space);
    multi_grid_info->dof_parent[0] = get_dof_dof_vec("dof_parent0", fe_space);
    multi_grid_info->dof_parent[1] = get_dof_dof_vec("dof_parent1", fe_space);
  }

  dof_level = multi_grid_info->dof_level->vec;
  dof_parent0 = multi_grid_info->dof_parent[0]->vec;
  dof_parent1 = multi_grid_info->dof_parent[1]->vec;
  FOR_ALL_DOFS(fe_space->admin,
	       dof_level[dof] = 0;
	       dof_parent0[dof] = dof_parent1[dof] = 0; );
  /* max_level = (int)mesh->max_level; */
  max_level = max_dof_level = 0;
  mesh_traverse(mesh, -1, CALL_EVERY_EL_PREORDER, max_level_fct);
  levels = (max_level + DIM-1) / DIM + 1;
  TEST_EXIT(levels == (max_dof_level+1))
    ("levels %d != max_dof_level %d + 1\n", levels, max_dof_level);
  
  if (levels > multi_grid_info->size)
  {
    multi_grid_info->matrix = multi_grid_info->matrix ?
      MEM_REALLOC(multi_grid_info->matrix,
		  multi_grid_info->size, levels, DOF_MATRIX *)
      : MEM_ALLOC(levels, DOF_MATRIX *);
    multi_grid_info->f_h = multi_grid_info->f_h ?
      MEM_REALLOC(multi_grid_info->f_h,
		  multi_grid_info->size, levels, DOF_REAL_VEC *)
      : MEM_ALLOC(levels, DOF_REAL_VEC *);
    multi_grid_info->u_h = multi_grid_info->u_h ?
      MEM_REALLOC(multi_grid_info->u_h,
		  multi_grid_info->size, levels, DOF_REAL_VEC *)
      : MEM_ALLOC(levels, DOF_REAL_VEC *);
    multi_grid_info->r_h = multi_grid_info->r_h ?
      MEM_REALLOC(multi_grid_info->r_h,
		  multi_grid_info->size, levels, DOF_REAL_VEC *)
      : MEM_ALLOC(levels, DOF_REAL_VEC *);

#if SPARSE_MG
    multi_grid_info->dofs_per_level = multi_grid_info->dofs_per_level ?
      MEM_REALLOC(multi_grid_info->dofs_per_level,
		  multi_grid_info->size, levels, int)
      : MEM_ALLOC(levels, int);
    multi_grid_info->tmp_per_level = multi_grid_info->tmp_per_level ?
      MEM_REALLOC(multi_grid_info->tmp_per_level,
		  multi_grid_info->size, levels, int)
      : MEM_ALLOC(levels, int);
#endif


#if SPARSE_MG
#define MG_FE_SPACE nil
#else
#define MG_FE_SPACE fe_space
#endif
    for (i = multi_grid_info->size; i < levels; i++)
    {
      sprintf(name, "mg matrix level %d", i);
      multi_grid_info->matrix[i] = get_dof_matrix(strdup(name), MG_FE_SPACE);

      sprintf(name, "mg rhs level %d", i);
      multi_grid_info->f_h[i] = get_dof_real_vec(strdup(name), MG_FE_SPACE);

      sprintf(name, "mg soln level %d", i);
      multi_grid_info->u_h[i] = get_dof_real_vec(strdup(name), MG_FE_SPACE);

      sprintf(name, "mg resid level %d", i);
      multi_grid_info->r_h[i] = get_dof_real_vec(strdup(name), MG_FE_SPACE);
    }
#undef MG_FE_SPACE

    multi_grid_info->mg_levels = multi_grid_info->size = levels;
  }
  else
  {
    multi_grid_info->mg_levels = levels;
  }

  mg_lin_sys_info->matrix   = multi_grid_info->matrix[levels-1];
  mg_lin_sys_info->f_h      = multi_grid_info->f_h[levels-1];
  mg_lin_sys_info->u_h      = multi_grid_info->u_h[levels-1];
  mg_lin_sys_info->bound    = lin_sys_info->bound;
  mg_lin_sys_info->max_iter = lin_sys_info->max_iter;
  mg_lin_sys_info->tol      = lin_sys_info->tol;
  mg_lin_sys_info->solver   = lin_sys_info->solver;
  mg_lin_sys_info->precon   = lin_sys_info->precon;
  mg_lin_sys_info->omega    = lin_sys_info->omega;
  mg_lin_sys_info->info     = lin_sys_info->info;

  
#if SPARSE_MG == 0
  for (i = 0; i < levels-1; i++)
  {
    dof_set(zero, multi_grid_info->u_h[i]);
    dof_set(zero, multi_grid_info->f_h[i]);
  }
#endif


#if SPARSE_MG
  init_sparse_mg(mesh, lin_sys_info, levels);

  dof_copy_to_sparse(multi_grid_info,
		     lin_sys_info->u_h, mg_lin_sys_info->u_h);
  dof_copy_to_sparse(multi_grid_info,
		     lin_sys_info->f_h, mg_lin_sys_info->f_h);
#else
  dof_copy(lin_sys_info->u_h, mg_lin_sys_info->u_h);
  dof_copy(lin_sys_info->f_h, mg_lin_sys_info->f_h);
#endif

  return;
}


/****************************************************************************/
/*  restriction of fine-level matrix to coarser level matrices              */
/****************************************************************************/
#if RESTRICT_MATRICES

static void add_sparse_dof_entry(DOF_MATRIX *matrix, REAL sign,
				 int irow, int jcol, REAL entry)
{
  FUNCNAME("add_sparse_dof_entry");
  int        i,j,k;
  int        free_col;
  MATRIX_ROW *row, *free_row;

  TEST_EXIT(matrix)("no matrix");
  
  TEST_EXIT(irow < matrix->size)
    ("irow = %d, mat.size = %d\n", irow, matrix->size);

  if (matrix->matrix_row[irow] == nil) {
    row = matrix->matrix_row[irow] = get_matrix_row();
    row->col[0] = irow;           /* first entry is diagonal element */
    row->entry[0] = 0.0;
  }

  TEST_EXIT(jcol < matrix->size)
    ("jcol = %d, matrix->size = %d\n", jcol, matrix->size);

  row = matrix->matrix_row[irow];
  free_row = nil;
  do {
    for (k=0; k<ROW_LENGTH; k++) {
      if (row->col[k] == jcol) {
	row->entry[k] += sign * entry;;
	break;
      }
      if (ENTRY_NOT_USED(row->col[k])) {
	free_col = k;
	free_row = row;
	if (row->col[k] == NO_MORE_ENTRIES) {
	  k = ROW_LENGTH;
	  break;
	}
      }
    }
    if (k < ROW_LENGTH) break;               /* done? */
    if (row->next || free_row) {
      row = row->next;
    }
    else {
      free_row = row->next = get_matrix_row();
      free_col = 0;
      row = nil;
    }
  } while (row);

  if (k >= ROW_LENGTH) {                     /* not done? */
    TEST_EXIT(free_row)("no free_row");
    TEST_EXIT(sign > 0.0)("new entry (%d,%d) in matrix %s with sign=%lf",
			  irow, jcol, matrix->name, sign);
    free_row->col[free_col]   = jcol;
    free_row->entry[free_col] = sign * entry;
  }
}

/****************************************************************************/

static REAL matrix_entry(DOF_MATRIX *matrix, int irow, int jcol)
{
  FUNCNAME("matrix_entry");
  int        k;
  MATRIX_ROW *row;

  TEST_EXIT(matrix)("no matrix");
  
  TEST_EXIT(irow < matrix->size)
    ("irow = %d, mat.size = %d\n", irow, matrix->size);

  for (row = matrix->matrix_row[irow]; row; row = row->next)
    for (k=0; k<ROW_LENGTH; k++)
      if (row->col[k] == jcol)
	return(row->entry[k]);

  MSG("matrix entry (%d,%d) not found; size=%d\n",
      irow, jcol, matrix->size);
  MSG("columns found:");
  for (row = matrix->matrix_row[irow]; row; row = row->next)
    for (k=0; k<ROW_LENGTH; k++)
      if (ENTRY_USED(row->col[k]))
	print_msg(" %d", row->col[k]);
  print_msg("\n");

  return(0.0);
}

/****************************************************************************/



static void diff_mat_1way(DOF_MATRIX *m1, DOF_MATRIX *m2)
{
  FUNCNAME("diff_mat_1way");
  MATRIX_ROW *row1, *row2;
  int        irow, jcol, k1,k2;
  REAL       err;

  for (irow = 0; irow < m1->size; irow++)
  {
    if (m1->matrix_row[irow] && !m2->matrix_row[irow])
    {
      ERROR("m2 `%s' row %d missing\n", m2->name, irow);
      WAIT;
      continue;
    }
    for (row1 = m1->matrix_row[irow]; row1; row1 = row1->next)
    {
      for (k1 = 0; k1 < ROW_LENGTH; k1++)
      {
	jcol = row1->col[k1];
	if (ENTRY_USED(jcol))
	{
	  for (row2 = m2->matrix_row[irow]; row2; row2 = row2->next)
	  {
	    for (k2 = 0; k2 < ROW_LENGTH; k2++)
	    {
	      if (row2->col[k2] == jcol)
	      {
		err = row1->entry[k1] - row2->entry[k2];
		err = ABS(err);
		if (err > 1.E-5)
		{
		  ERROR("entries (%d,%d): %.5le, %.5le, err=%.3le\n",
			irow, jcol, row1->entry[k1], row2->entry[k2], err);
		  WAIT;
		}
		break;
	      }
	    }
	    if (k2 < ROW_LENGTH) break;  /* entry found */
	  }
	  if (k2 >= ROW_LENGTH)
	  {
	    ERROR("m2 `%s' entry (%d,%d) missing\n", m2->name, irow, jcol);
	    WAIT;
	  }
	}
      }
    }
  }
  return;
}

static void diff_matrices(DOF_MATRIX *m1, DOF_MATRIX *m2, int size)
{
  FUNCNAME("diff_matrices");
  int     i;
  REAL    *r1, *r2, err;

  TEST_EXIT(m1 && m2)("nil matrix: %8X, %8X\n", m1, m2);

  MSG("compare matrices `%s' and `%s'\n", m1->name, m2->name);

  diff_mat_1way(m1,m2);
  diff_mat_1way(m2,m1);

  return;
}

/****************************************************************************/
/* matrix restriction for one level                                         */
/****************************************************************************/

static void restrict_mg_matrix(MULTI_GRID_INFO *mg_info, int fine_level)
{
  FUNCNAME("restrict_mg_matrix");
  DOF_MATRIX  *fine_mat, *coarse_mat;
  MATRIX_ROW  *row;
  int         nfine, ncoarse, irow, jcol, i0, i1, l1, p1, l2, p2, k;
  int         sirow, sjcol, sp1;
  REAL        entry, sentry;
  static REAL one = 1.0, half = 0.5, quarter = 0.25, eighth=0.125;

  TEST_EXIT(fine_level > 0)("fine_level = %d  <= 0\n", fine_level);

  fine_mat   = mg_info->matrix[fine_level];
  coarse_mat = mg_info->matrix[fine_level - 1];
  TEST_EXIT(fine_mat && coarse_mat)
    ("fine_mat or coarse_mat == nil: %8X, %8X\n", fine_mat, coarse_mat);

  nfine   = mg_info->dofs_per_level[fine_level];
  ncoarse = mg_info->dofs_per_level[fine_level - 1];

  MSG("fine_level %d: nfine=%d, ncoarse=%d\n", fine_level, nfine, ncoarse);

/****************************************************************************/
/* copy coarse/coarse entries to coarse_mat                                 */

  for (irow = 0; irow < ncoarse; irow++)
  {
    for (row = fine_mat->matrix_row[irow]; row; row = row->next)
    {
      for (k=0; k<ROW_LENGTH; k++) {
	jcol = row->col[k];
	if (ENTRY_USED(jcol) && (jcol < ncoarse))
	  add_sparse_dof_entry(coarse_mat, one, irow, jcol, row->entry[k]);
	else if (jcol == NO_MORE_ENTRIES)
	  break;
      }
    }
  }

/****************************************************************************/
/* restrict fine entries to coarse_mat using child/parents relationship     */

  for (irow = ncoarse; irow < nfine; irow++)
  {
    sirow = sort_dof[irow];
    i0 = sort_dof_invers[dof_parent[0][sirow]];
    i1 = sort_dof_invers[dof_parent[1][sirow]];

    for (row = fine_mat->matrix_row[irow]; row; row = row->next)
    {
      for (k=0; k<ROW_LENGTH; k++) {
	jcol = row->col[k];

	if (!(ENTRY_USED(jcol)))
	{
	  if (jcol == NO_MORE_ENTRIES)
	    break;
	  else
	    continue;
	}

	entry = row->entry[k];
	if (irow == jcol)                           /* diagonal element */
	{
	  add_sparse_dof_entry(coarse_mat, quarter, i0, i0, entry);
	  add_sparse_dof_entry(coarse_mat, quarter, i0, i1, entry);
	  add_sparse_dof_entry(coarse_mat, quarter, i1, i0, entry);
	  add_sparse_dof_entry(coarse_mat, quarter, i1, i1, entry);
	  continue;
	}

	if (jcol < ncoarse)
	{
	  /* add rows */
	  add_sparse_dof_entry(coarse_mat, half, i0, jcol, entry);
	  add_sparse_dof_entry(coarse_mat, half, i1, jcol, entry);
	
	  /* add columns with symmetric entry */
	  sentry = matrix_entry(fine_mat, jcol, irow);
	  /*TEST(entry==sentry)("non symmetric? %.5le %.5le\n",entry,sentry);*/

	  add_sparse_dof_entry(coarse_mat, half, jcol, i0, sentry);
	  add_sparse_dof_entry(coarse_mat, half, jcol, i1, sentry);
	}
	else
	{
	  sjcol = sort_dof[jcol];
	  for (l1 = 0; l1 < 2; l1++)            /* both parents of jcol */
	  {
	    p1 = sort_dof_invers[dof_parent[l1][sjcol]];
	    if (p1 < ncoarse)
	    {
	      /* add rows */
	      add_sparse_dof_entry(coarse_mat, quarter, i0, p1, entry);
	      add_sparse_dof_entry(coarse_mat, quarter, i1, p1, entry);
	    }
	    else
	    {
	      ERROR("recursion 2: %d (%d %d), %d >= %d\n",
		    jcol, sort_dof_invers[dof_parent[0][sjcol]],
		    sort_dof_invers[dof_parent[1][sjcol]], p1, ncoarse);
	    }
	  } /* end for (l1) */
	} /* end else (jcol < ncoarse) */

      } /* end for (k) */
    } /* end for (row) */

  } /* end for (irow) */
}

/****************************************************************************/

static void restrict_mg_matrices(MULTI_GRID_INFO *mg_info)
{
  FUNCNAME("restrict_mg_matrices");
  int        i, level;
  clock_t    first, second;

#if TEST_RESTRICT
  DOF_MATRIX *old_mat, *new_mat;

  new_mat             = get_dof_matrix("new_mat", nil);
  new_mat->size       = mg_info->dofs_per_level[mg_info->mg_levels - 1];
  new_mat->matrix_row = MEM_ALLOC(new_mat->size, MATRIX_ROW *);
  for (i = 0; i < new_mat->size; i++)
    new_mat->matrix_row[i] = nil;
#endif


#if SPARSE_MG
  sort_dof_invers = mg_info->sort_dof_invers;
  sort_dof        = mg_info->sort_dof;
#endif
  dof_parent[0]   = mg_info->dof_parent[0]->vec;
  dof_parent[1]   = mg_info->dof_parent[1]->vec;

  first = clock();

  for (level = mg_info->mg_levels - 1; level > 0; level--)
  {
#if TEST_RESTRIC
    /* save old matrix */
    old_mat = mg_info->matrix[level-1];
    mg_info->matrix[level-1] = new_mat;
    clear_dof_matrix_sparse(mg_info, level, mg_info->matrix[level-1]);

    restrict_mg_matrix(mg_info, level);

    /* compare old/new matrices */
    diff_matrices(old_mat, new_mat, mg_info->dofs_per_level[level-1]);

    mg_info->matrix[level-1] = old_mat;
#else

    clear_dof_matrix_sparse(mg_info, level-1, mg_info->matrix[level-1]);
    restrict_mg_matrix(mg_info, level);

#endif
  }

  if (mg_lin_sys_info->info > 0)
  {
    second = clock();
    MSG("needed %.5lf seconds\n",
	(double)(second-first)/(double)CLOCKS_PER_SEC);
  }

#if TEST_RESTRIC
  free_dof_matrix(new_mat);
#endif

  return;
}

#endif /* RESTRICT_MATRICES */


/****************************************************************************/
/*  build matrices for all multigrid levels                                 */
/****************************************************************************/

static int mg_matrix_level;

static DOF_MATRIX *fill_mat;
static const void *fill_info;
static int        n_bas_fcts;
static const DOF_ADMIN  *fill_admin;
static const DOF  *(*get_dof_indices)(const EL *,const DOF_ADMIN *, DOF *);
static const REAL **(*el_stiff_fct)(const EL_INFO *, const void *);


static void setup_mg_matrix_fct(const EL_INFO *el_info)
{
  FUNCNAME("setup_mg_matrix_fct");
  static REAL one = 1.0;
  const REAL  **el_mat;
  const DOF   *el_dof;
  EL          *el = el_info->el;

  if ((el_info->level % DIM) && el->child[0])    /* intermediate element */
  {
    MSG("intermediate element\n");
    return;
  }

#if 0
  MSG("el %5d level %2d mg_level %2d", INDEX(el), el_info->level, 
      mg_matrix_level);
  print_msg("; node->dof.level: %2d %2d %2d\n",
	   multi_grid_info->dof_level->vec[el->dof[0][0]],
	   multi_grid_info->dof_level->vec[el->dof[1][0]],
	   multi_grid_info->dof_level->vec[el->dof[2][0]]);
#endif

  el_dof = (*get_dof_indices)(el, fill_admin, nil);
  el_mat = (*el_stiff_fct)(el_info, fill_info);

#if SPARSE_MG
  update_sparse_dof_matrix(fill_mat, one, n_bas_fcts, el_mat, el_dof);
#else
  update_dof_matrix(fill_mat, one, n_bas_fcts, el_mat, el_dof);
#endif

  return;
}


static void setup_mg_matrices(MESH *mesh, MULTI_GRID_INFO *mg_info,
			      const EL_STIFF_INFO *info)
{
  FUNCNAME("setup_mg_matrices");
  clock_t  first, second;

#if SPARSE_MG
  sort_dof_invers = mg_info->sort_dof_invers;
  sort_dof        = mg_info->sort_dof;
#endif

  TEST_EXIT(mg_info && mg_info->fe_space && mg_info->fe_space->bas_fcts)
    ("no mg_info or fe_space or bas_fcts\n");
  
  if (mg_lin_sys_info->info > 0) first = clock();

  
  fill_admin      = mg_info->fe_space->admin;
  get_dof_indices = mg_info->fe_space->bas_fcts->get_dof_indices;
  n_bas_fcts      = mg_info->fe_space->bas_fcts->n_bas_fcts;
  el_stiff_fct    = info->el_stiff_fct;
  fill_info       = info->fill_info;


#if RESTRICT_MATRICES

  mg_matrix_level = mg_info->mg_levels - 1;
  clear_dof_matrix_sparse(mg_info, mg_matrix_level,
			  mg_info->matrix[mg_matrix_level]);

  fill_mat = mg_info->matrix[mg_matrix_level];
/*  mesh_traverse(mesh, mg_matrix_level, CALL_MG_LEVEL | FILL_COORDS,
		setup_mg_matrix_fct);
 */ /* following should be the same: */
  mesh_traverse(mesh, mg_matrix_level, CALL_LEAF_EL | info->fill_flag,
		setup_mg_matrix_fct);

#if 0
  print_dof_matrix(multi_grid_info->matrix[mg_matrix_level]);
#endif

  restrict_mg_matrices(mg_info);

#else  /* no RESTRICT_MATRICES */

  for (mg_matrix_level = 0; mg_matrix_level < mg_info->mg_levels;
       mg_matrix_level++)
  {
#if SPARSE_MG
    clear_dof_matrix_sparse(mg_info, mg_matrix_level,
			    mg_info->matrix[mg_matrix_level]);
#else
    clear_dof_matrix(mg_info->matrix[mg_matrix_level]);
#endif

    fill_mat = mg_info->matrix[mg_matrix_level];
    mesh_traverse(mesh, mg_matrix_level, CALL_MG_LEVEL | info->fill_flag,
		  setup_mg_matrix_fct);
  }

#endif /* RESTRICT_MATRICES */


  if (mg_lin_sys_info->info > 0)
  {
    second = clock();
    MSG("fill_matrix needed %.5lf seconds\n",
	(double)(second-first)/(double)CLOCKS_PER_SEC);
  }
  return;
}


/****************************************************************************/
/*  restriction of fine grid function to a coarser grid                     */
/****************************************************************************/

static REAL     *fg_fct, *cg_fct;
static int      dim, act_mg_level;
static int      mg_levels;
static U_CHAR   *dof_level;
static S_CHAR   *bound;


static void restrict_value(int dof1, int dof2, int dof3)
{
  FUNCNAME("restrict_value");
  REAL     new_v;
  int      bound1, bound2;

#if SPARSE_MG
  bound1 = bound[sort_dof[dof1]];
  bound2 = bound[sort_dof[dof2]];
#else
  bound1 = bound[dof1];
  bound2 = bound[dof2];
#endif

  /* new_v = (fg_fct[dof3] - (fg_fct[dof1] + fg_fct[dof2])*0.5)*0.5; */
  new_v = fg_fct[dof3] * 0.5;

  if (bound1 <= INTERIOR)
    cg_fct[dof1] += new_v;
  if (bound2 <= INTERIOR)
    cg_fct[dof2] += new_v;
}


#if SPARSE_MG == 0

static void restrict_fct(const EL_INFO *el_info)
{
  FUNCNAME("restrict_fct");
  EL       *el = el_info->el;
  EL_INFO  child_info[1];
  int      mg_level, dof1, dof2, dof3;

  if (!el->child[0])  return;

/****************************************************************************/
/*  get the multi grid level of children                                    */
/****************************************************************************/

  mg_level = ((int)el_info->level + 1) / DIM;

  if (mg_level < act_mg_level + 1  &&  el->child[0]->child[0])
  {
    child_info->el = el->child[0];
    restrict_fct(child_info);
    child_info->el = el->child[1];
    restrict_fct(child_info);
  }

#if DIM == 2
  dof1 = el->dof[1][0];
  dof2 = el->dof[2][0];
  dof3 = el->child[0]->dof[0][0];
#else
  dof1 = el->dof[0][0];
  dof2 = el->dof[1][0];
  dof3 = el->child[0]->dof[3][0];
#endif

  TEST_EXIT(dof1 < dim  &&  dof2 < dim  && dof3 < dim)
    ("dof1 || dof2  ||  dof3 too large");

  if ((int)dof_level[dof3] < mg_levels)
  {
/****************************************************************************/
/*  now restrict the function between one element level, i.e. there is only */
/*  one new node                                                            */
/****************************************************************************/

    restrict_value(dof1, dof2, dof3);


    dof_level[dof3] += mg_levels;
  }
  return;
}
#endif

static void restrict(MESH *mesh, MULTI_GRID_INFO *mg_info, int mg_level)
{
  FUNCNAME("restrict");
  int           i;
  DOF_REAL_VEC  *drv;
  DOF_UCHAR_VEC *duv;
  REAL          *fg_uh;
#if SPARSE_MG
  int           j, first, last;
#endif

  TEST_EXIT(mg_info)("sorry: no mg_info");

  if (mg_level >= mg_info->mg_levels - 1)
  {
    MSG("no restriction possible on finest level\n");
    return;
  }
  mg_levels    = mg_info->mg_levels;
  act_mg_level = mg_level;


  TEST_EXIT(mg_info->lin_sys_info && mg_info->lin_sys_info->bound
	    && (bound = mg_info->lin_sys_info->bound->vec))
    ("sorry: no mg_info->lin_sys_info->bound->vec");
  
  TEST_EXIT(mg_info->r_h  &&  (drv = mg_info->r_h[mg_level+1]))
    ("sorry: no fine grid function");
  TEST_EXIT(fg_fct = drv->vec)
    ("sorry: no vec for fine grid function");
  dim = drv->size;

  TEST_EXIT(mg_info->f_h  &&  (drv = mg_info->f_h[mg_level]))
    ("sorry: no coarse grid function");
  TEST_EXIT(cg_fct = drv->vec)
    ("sorry: no vec for coarse grid function");
  TEST_EXIT(duv = mg_info->dof_level)
    ("sorry: no vec of dof levels");
  TEST_EXIT(dof_level = duv->vec)
    ("sorry: no vec for dof level");

#if !SPARSE_MG
  TEST_EXIT(dim == drv->size)
    ("sorry: different dimensions of vecs, not implemented yet :-(");
  TEST_EXIT(dim == duv->size)
    ("sorry: lenght of dof_level vec doesn't match :-(");
#endif


#if SPARSE_MG

  first = dofs_per_level[mg_level];
  last  = dofs_per_level[mg_level+1];

  for (i = 0; i < first; i++)
    cg_fct[i] = fg_fct[i];

  for (i = first; i < last; i++)
  {
    j = sort_dof[i];
    restrict_value(sort_dof_invers[dof_parent[0][j]],
		   sort_dof_invers[dof_parent[1][j]], i);
  }
#else

  for (i = 0; i < dim; i++)
    if ((int)dof_level[i] <= mg_level)  cg_fct[i] = fg_fct[i];

  mesh_traverse(mesh, mg_level, CALL_MG_LEVEL, restrict_fct);

  for (i = 0; i < dim; i++)
    if ((int)dof_level[i] > mg_levels) dof_level[i] -=  mg_levels;

#endif

  return;
}

/****************************************************************************/
/*  prolongation of a coarse grid function to a finer grid                  */
/****************************************************************************/

static REAL max_prolongated = 0.0;
static int  g_mg_level;

#if SPARSE_MG == 0

static void prolongate_fct(const EL_INFO *el_info)
{
  FUNCNAME("prolongate_fct");
  EL       *el = el_info->el;
  EL_INFO  child_info[1];
  int      mg_level, dof1, dof2, dof3;
  REAL     pro;

  if (!el->child[0])  return;

#if DIM == 2
  dof1 = el->dof[1][0];
  dof2 = el->dof[2][0];
  dof3 = el->child[0]->dof[0][0];
#else
  dof1 = el->dof[0][0];
  dof2 = el->dof[1][0];
  dof3 = el->child[0]->dof[3][0];
#endif
  TEST_EXIT(dof1 < dim  &&  dof2 < dim  && dof3 < dim)
    ("dof1 || dof2  ||  dof3 too large");

  if ((int)dof_level[dof3] < mg_levels)
  {
/****************************************************************************/
/*  prolongate the coarse grid function                                     */
/****************************************************************************/
/*    MSG("setting value at node %d by %d %d\n", dof3, dof1, dof2); */

    pro = 0.5 * (cg_fct[dof1] + cg_fct[dof2]);
    fg_fct[dof3] += pro;

    max_prolongated = MAX(max_prolongated, ABS(pro));

    dof_level[dof3] += mg_levels;
  }

/****************************************************************************/
/*  get the multi grid level of children                                    */
/****************************************************************************/

  mg_level = ((int)el_info->level + 1) / DIM;

  if (mg_level < act_mg_level + 1  &&  el->child[0]->child[0])
  {
    child_info->el = el->child[0];
    prolongate_fct(child_info);
    child_info->el = el->child[1];
    prolongate_fct(child_info);
  }

  return;
}

#endif

static void prolongate(MESH *mesh, MULTI_GRID_INFO *mg_info, int mg_level)
{
  FUNCNAME("prolongate");
  int           i;
  DOF_REAL_VEC  *drv;
  DOF_UCHAR_VEC *duv;
#if SPARSE_MG
  int           j, first, last, dof1,dof2,dof3;
  REAL          pro;
#endif

  TEST_EXIT(mg_info)("sorry: no mg_info");

  if (mg_level >= mg_info->mg_levels - 1)
  {
    MSG("no prolongation possible on finest level\n");
    return;
  }
  mg_levels = mg_info->mg_levels;
  act_mg_level = mg_level;

  TEST_EXIT(mg_info->u_h  &&  (drv = mg_info->u_h[mg_level+1]))
    ("sorry: no fine grid function");
  TEST_EXIT(fg_fct = drv->vec)
    ("sorry: no vector for fine grid function");
  dim = drv->size;

  TEST_EXIT(mg_info->u_h  &&  (drv = mg_info->u_h[mg_level]))
    ("sorry: no coarse grid function");
  TEST_EXIT(cg_fct = drv->vec)
    ("sorry: no vector for coarse grid function");
  TEST_EXIT(duv = mg_info->dof_level)
    ("sorry: no vector of dof levels");
  TEST_EXIT(dof_level = duv->vec)
    ("sorry: no vector for dof level");

#if !SPARSE_MG
  TEST_EXIT(dim == drv->size)
    ("sorry: different dimensions of vecs, not implemented yet :-(");
  TEST_EXIT(dim == duv->size)
    ("sorry: lenght of dof_level vector doesn't match :-(");
#endif

  max_prolongated = 0.0;


#if SPARSE_MG

  first = dofs_per_level[mg_level];
  last  = dofs_per_level[mg_level+1];

  for (i = 0; i < first; i++)
  {
    fg_fct[i] += cg_fct[i];
    max_prolongated = MAX(max_prolongated, ABS(cg_fct[i]));
  }

  for (i = first; i < last; i++)
  {
    j = sort_dof[i];
    dof1 = sort_dof_invers[dof_parent[0][j]];
    dof2 = sort_dof_invers[dof_parent[1][j]];
    dof3 = i;

    pro = 0.5 * (cg_fct[dof1] + cg_fct[dof2]);
    fg_fct[dof3] += pro;

    max_prolongated = MAX(max_prolongated, ABS(pro));
  }

#else  /* SPARSE_MG */

  for (i = 0; i < dim; i++) {
    if ((int)dof_level[i] <= mg_level) 
    {
      fg_fct[i] += cg_fct[i];
      max_prolongated = MAX(max_prolongated, ABS(cg_fct[i]));
    }
  }

  mesh_traverse(mesh, mg_level, CALL_MG_LEVEL, prolongate_fct);

  for (i = 0; i < dim; i++)
    if ((int)dof_level[i] > mg_levels) dof_level[i] -=  mg_levels;

#endif  /* SPARSE_MG */


  if (mg_info->lin_sys_info->info > 4)
    MSG("level %2d: max_prolongated = %12.9lf\n", mg_level, max_prolongated);

  return;
}

/****************************************************************************/

static void test_call_mg_level(EL_INFO *el_info)
{
  FUNCNAME("test_call_mg_level");
  EL      *el = el_info->el;

  MSG("el %5d: level= %2d, childs: %s, dof.levels= %2d %2d %2d\n",
      el->index, el_info->level, el->child[0] ? "yes" : "no ",
      multi_grid_info->dof_level->vec[el->dof[0][0]],
      multi_grid_info->dof_level->vec[el->dof[1][0]],
      multi_grid_info->dof_level->vec[el->dof[2][0]]);

  return;
}

/****************************************************************************/
/*  calculate residuum on level mg_level                                    */
/****************************************************************************/

static void mg_resid(MESH *mesh, MULTI_GRID_INFO *mg_info, int mg_level)
{
  FUNCNAME("mg_resid");
  REAL    *r_h;
  S_CHAR  *bound;
  int     i, last;

  TEST_EXIT(mg_level < mg_info->mg_levels)("mg_level too big");

#if SPARSE_MG
  /*  r_h := f_h  */
  dof_copy_sparse(mg_info, mg_level,
		  mg_info->f_h[mg_level], mg_info->r_h[mg_level]);
  /*  r_h := r_h - A * u_h  */
  dof_gemv_sparse(mg_info, mg_level, NoTranspose, -1.0,
		  mg_info->matrix[mg_level],
		  mg_info->u_h[mg_level], 1.0, mg_info->r_h[mg_level]);
#else
  /*  r_h := f_h  */
  dof_copy(mg_info->f_h[mg_level], mg_info->r_h[mg_level]);
  /*  r_h := r_h - A * u_h  */
  dof_gemv(NoTranspose, -1.0, mg_info->matrix[mg_level],
	   mg_info->u_h[mg_level], 1.0, mg_info->r_h[mg_level]);
#endif

  /*  no residuum at dirichlet boundary nodes  */
  bound    = mg_info->lin_sys_info->bound->vec;
  r_h      = mg_info->r_h[mg_level]->vec;

#if SPARSE_MG
  last = mg_info->dofs_per_level[mg_level];
  for (i = 0; i < last; i++)
  {
    if (bound[sort_dof[i]] >= DIRICHLET) {
      r_h[i] = 0.0;
    }
  }
#else
  FOR_ALL_DOFS(mg_info->fe_space->admin,
	       if (bound[dof] >= DIRICHLET) r_h[dof] = 0.0; );
#endif

  return;
}

/****************************************************************************/
/*  smoothing of the solution on level mg_level                             */
/****************************************************************************/
#if SPARSE_MG
static void sparse_mg_smoother(LIN_SYS_INFO *ls_info,
			       MULTI_GRID_INFO *mg_info, int mg_level)
{
  FUNCNAME("sparse_mg_smoother");
  int        iter, i, j, k, size;
  REAL       sum, omega1, omega = ls_info->omega, unew, max_chg;
  S_CHAR     *bound = ls_info->bound->vec;
  REAL       *fvec = ls_info->f_h->vec;
  REAL       *uvec = ls_info->u_h->vec;
  MATRIX_ROW *row, **matrix_row = ls_info->matrix->matrix_row;

  size = mg_info->dofs_per_level[mg_level];

  omega1 = 1.0 - omega;

  for (iter = 0; iter < ls_info->max_iter; iter++) {
    max_chg = 0.0;
    for (i = 0; i < size; i++) {
      sum = fvec[i];
      if (bound[sort_dof[i]] < DIRICHLET) {
	for (row = matrix_row[i]; row; row = row->next) {
	  for (j = 0; j < ROW_LENGTH; j++) {
	    k = row->col[j];
	    if (ENTRY_USED(k)) {
	      if (k != i) sum = sum - row->entry[j] * uvec[k];
	    }
	    else {
	      if (k == NO_MORE_ENTRIES)  break;
	    }
	  }
	}
	if (row = matrix_row[i]) {
	  sum = sum / row->entry[0];
	  TEST(row->col[0] == i)
	    ("wrong row[%d]->col[0]: %d\n", i, row->col[0]);
	}
	unew = omega * sum + omega1 * uvec[i];
	max_chg = MAX(max_chg, ABS(uvec[i]-unew));
	uvec[i] = unew;
      }
      else {
	uvec[i] = sum;
      }
    }
    /* MSG("iteration %d: max_chg = %12.4le\n", iter, max_chg); */
  }

  if (ls_info->info > 1) {
    MSG("%d SOR iterations with omega=%4.2lf, last max_chg = %.2le\n",
	iter, omega, max_chg);
  }
}
#endif


static REAL max_smooth[3] = {0.0, 0.0, 0.0};

static void mg_smoother(MESH *mesh, MULTI_GRID_INFO *mg_info,
			int mg_level, int nsmooth, int pre_post)
{
  FUNCNAME("mg_smoother");
  LIN_SYS_INFO ls_info[1];

  TEST_EXIT(mg_level < mg_info->mg_levels)("mg_level too big");

  ls_info->matrix   = mg_info->matrix[mg_level];
  ls_info->f_h      = mg_info->f_h[mg_level];
  ls_info->u_h      = mg_info->u_h[mg_level];
  ls_info->bound    = mg_info->lin_sys_info->bound;
  ls_info->max_iter = nsmooth;
  ls_info->tol      = -1.0;
  ls_info->solver   = mg_info->smoother;
  ls_info->precon   = 0;
  ls_info->omega    = mg_info->lin_sys_info->omega;
  ls_info->info     = MAX(mg_info->lin_sys_info->info - 3, 0);

#if SPARSE_MG
  sparse_mg_smoother(ls_info, mg_info, mg_level);
#else
  solve_lin_sys(mesh, ls_info);
#endif

  return;
}

/****************************************************************************/
/*  "exact" solution on the coarsest level                                  */
/****************************************************************************/

static void mg_exact(MESH *mesh, MULTI_GRID_INFO *mg_info, int mg_level)
{
  FUNCNAME("mg_exact");
  LIN_SYS_INFO ls_info[1];

  TEST_EXIT(mg_level < mg_info->mg_levels)("mg_level too big");

  ls_info->matrix   = mg_info->matrix[mg_level];
  ls_info->f_h      = mg_info->f_h[mg_level];
  ls_info->u_h      = mg_info->u_h[mg_level];
  ls_info->bound    = mg_info->lin_sys_info->bound;
  ls_info->max_iter = mg_info->lin_sys_info->max_iter;
  ls_info->tol      = mg_info->tol;
  ls_info->solver   = mg_info->coarse_solver;
  ls_info->precon   = 0;
  ls_info->omega    = mg_info->lin_sys_info->omega;
  ls_info->info     = MAX(mg_info->lin_sys_info->info - 3, 0);

#if SPARSE_MG
  /* ERROR_EXIT("SPARSE_MG exact solver not yet implemented\n"); */
  ls_info->max_iter = 2;
  sparse_mg_smoother(ls_info, mg_info, mg_level);

#else
  solve_lin_sys(mesh, ls_info);
#endif

  return;
}

/****************************************************************************/
/*  recursive MG iteration on level mg_level                                */
/****************************************************************************/

#if SPARSE_MG
#define SORT_DOF_INVERS sort_dof_invers
#else
#define SORT_DOF_INVERS nil
#endif


static void recursive_mg_iteration(MESH *mesh, MULTI_GRID_INFO *mg_info,
				   int mg_level)
{
  FUNCNAME("recursive_mg_iteration");
  int     i, cycle;
  static const REAL zero = 0.0;
  
  if (mg_info->lin_sys_info->info >= 6)
    MSG("mg_level = %d\n", mg_level);


  if (mg_level < mg_info->mg_levels - 1)
  {
#if SPARSE_MG
    dof_set_sparse(mg_info, mg_level, zero, mg_info->u_h[mg_level]);
#else
    dof_set(zero, mg_info->u_h[mg_level]);
#endif
  }


  if (mg_level <= mg_info->exact_level)
  {
    mg_exact(mesh, mg_info, mg_level);
#if MG_GRAPHICS
    graph_values_mg(mg_win, mg_info->u_h[mg_level], GRAPH_REFINE,
		    mg_level, mg_info->fe_space, SORT_DOF_INVERS);
    MSG("<%s> `exact' solution\n", NAME(mg_info->u_h[mg_level]), mg_level);
    WAIT;
#endif
    return;
  }



#if MG_GRAPHICS
  if (mg_level == mg_info->mg_levels - 1) {

#if 1
    graph_values_mg(mg_win, mg_info->u_h[mg_level], GRAPH_REFINE,
		    mg_level, mg_info->fe_space, SORT_DOF_INVERS);
    MSG("<%s>\n", NAME(mg_info->u_h[mg_level]), mg_level); WAIT;
#endif

    graph_values_mg(mg_win, mg_info->f_h[mg_level], GRAPH_REFINE,
		    mg_level, mg_info->fe_space, SORT_DOF_INVERS);
    MSG("<%s>\n", NAME(mg_info->f_h[mg_level]), mg_level); WAIT;
  }
#endif /* MG_GRAPHICS */


  for (cycle = 0; cycle < mg_info->cycle; cycle++)
  {
    if (mg_info->presmooth > 0) {
      mg_smoother(mesh, mg_info, mg_level, mg_info->presmooth, 0);
#if MG_GRAPHICS
      graph_values_mg(mg_win, mg_info->u_h[mg_level], GRAPH_REFINE,
		      mg_level, mg_info->fe_space, SORT_DOF_INVERS);
      MSG("<%s> nach pre_smooth\n", NAME(mg_info->u_h[mg_level]), mg_level);
      WAIT;
#endif
    }
    mg_resid(mesh, mg_info, mg_level);

#if MG_GRAPHICS
    graph_values_mg(mg_win, mg_info->r_h[mg_level], GRAPH_REFINE,
		    mg_level, mg_info->fe_space, SORT_DOF_INVERS);
    MSG("<%s>\n", NAME(mg_info->r_h[mg_level]), mg_level); WAIT;
#endif

#if 0
    interpol(mesh, prob_info->u, mg_info->r_h[mg_level]);
#endif

    restrict(mesh, mg_info, mg_level - 1);

#if MG_GRAPHICS
    graph_values_mg(mg_win, mg_info->f_h[mg_level-1], GRAPH_REFINE,
		    mg_level-1, mg_info->fe_space, SORT_DOF_INVERS);
    MSG("<%s>\n", NAME(mg_info->f_h[mg_level-1]), mg_level-1); WAIT;
#endif

    recursive_mg_iteration(mesh, mg_info, mg_level - 1);

#if MG_GRAPHICS && 0
    graph_values_mg(mg_win, mg_info->u_h[mg_level-1], GRAPH_REFINE,
		    mg_level-1, mg_info->fe_space, SORT_DOF_INVERS);
    MSG("<%s> nach recursive_mg_iteration\n",
	NAME(mg_info->u_h[mg_level-1]), mg_level-1);
    WAIT;
#endif

#if 0
    dof_set(mesh->dof_admin, 0.0, mg_info->u_h[mg_level]);
    interpol(mesh, prob_info->u, mg_info->u_h[mg_level-1]);
#endif

    prolongate(mesh, mg_info, mg_level - 1);

#if MG_GRAPHICS
    graph_values_mg(mg_win, mg_info->u_h[mg_level], GRAPH_REFINE,
		    mg_level, mg_info->fe_space, SORT_DOF_INVERS);
    MSG("<%s> nach prolongate\n", NAME(mg_info->u_h[mg_level]), mg_level);
    WAIT;
#endif
  }

  if (mg_info->postsmooth > 0)
  {
    mg_smoother(mesh, mg_info, mg_level, mg_info->postsmooth, 1);
#if MG_GRAPHICS
    graph_values_mg(mg_win, mg_info->u_h[mg_level], GRAPH_REFINE,
		    mg_level, mg_info->fe_space, SORT_DOF_INVERS);
    MSG("<%s> nach post_smooth\n", NAME(mg_info->u_h[mg_level]), mg_level);
    WAIT;
#endif
  }

}


/****************************************************************************/
/*  one global MG cycle                                                     */
/****************************************************************************/

static REAL multi_grid_iteration(MESH *mesh, MULTI_GRID_INFO *mg_info)
{
  FUNCNAME("multi_grid_iteration");
  int             i;
  REAL            resid = 0.0;

  recursive_mg_iteration(mesh, mg_info, mg_info->mg_levels - 1);

  mg_resid(mesh, mg_info, mg_info->mg_levels - 1);
#if SPARSE_MG
  resid = dof_nrm2_sparse(mg_info, mg_info->mg_levels - 1,
		   mg_info->r_h[mg_info->mg_levels - 1]);
#else
  resid = dof_nrm2(mg_info->r_h[mg_info->mg_levels - 1]);
#endif
  
  return(resid);
}


/****************************************************************************/
/*  MG solver                                                               */
/****************************************************************************/

void multi_grid_solver(MESH *mesh, LIN_SYS_INFO *lin_sys_info,
		       const EL_STIFF_INFO *el_stiff_info)
{
  FUNCNAME("multi_grid_solver");
  REAL    resid, old_resid;
  int     iter;
  clock_t first, second;
  static FILE *MGiter = nil;

#if MG_GRAPHICS
  if (!mg_win) {
    mg_win = graph_open_window("Multigrid Graphics", "400x400+450+0",
			       nil, mesh);
    
  }
  graph_clear_window(mg_win, nil);
#endif

  init_multi_grid(mesh, lin_sys_info);

  if (lin_sys_info->info >= 2)
  {
    MSG("cycle=%d smooth=(%d,%d) smoother=%d coarse=%d levels=%d lev0=%d\n",
	multi_grid_info->cycle,
	multi_grid_info->presmooth, multi_grid_info->postsmooth, 
	multi_grid_info->smoother, multi_grid_info->coarse_solver,
	multi_grid_info->mg_levels-1, multi_grid_info->exact_level);

    if (!MGiter) MGiter = fopen("MGiter.xg", "w");
    fprintf(MGiter, "\n");
    fflush(MGiter);
  }


  setup_mg_matrices(mesh, multi_grid_info, el_stiff_info);

#if SPARSE_MG && 0
  diff_sparse_nonsparse_matrices(multi_grid_info,
				 multi_grid_info->mg_levels-1,
				 lin_sys_info->matrix);
  {
    DOF_REAL_VEC *rh = get_dof_real_vec("temp rh",
					lin_sys_info->matrix->fe_space);
    dof_copy(lin_sys_info->f_h, rh);
    dof_gemv(NoTranspose, -1.0, lin_sys_info->matrix,
	     lin_sys_info->u_h, 1.0, rh);
    FOR_ALL_DOFS(lin_sys_info->matrix->fe_space->admin,
		 if (lin_sys_info->bound->vec[dof] >= DIRICHLET)
	            rh->vec[dof] = 0.0;
      );

    old_resid = dof_nrm2(rh);
    MSG("lin_sys_info old_resid = %le\n", old_resid);
    free_dof_real_vec(rh);
  }
#endif


  mg_resid(mesh, multi_grid_info, multi_grid_info->mg_levels - 1);
#if SPARSE_MG
  old_resid = dof_nrm2_sparse(multi_grid_info, multi_grid_info->mg_levels-1,
		      multi_grid_info->r_h[multi_grid_info->mg_levels-1]);
#else
  old_resid = dof_nrm2(multi_grid_info->r_h[multi_grid_info->mg_levels-1]);
#endif
  if (lin_sys_info->info >= 4)
    MSG("MG          Start Residuum %.4le\n", old_resid);

  WAIT;

  for (iter = 0; iter < lin_sys_info->max_iter; iter++)
  {
    first = clock();

    resid = multi_grid_iteration(mesh, multi_grid_info);

    if (lin_sys_info->info > 0)
    {
      second = clock();
      MSG("multi_grid_iteration needed %.5lf seconds\n",
	  (double)(second-first)/(double)CLOCKS_PER_SEC);
    }

    if (lin_sys_info->info >= 4) {
       MSG("MG iteration %3d: Residuum %.4le  Reduction  %.4lf\n",
	   iter + 1, resid, resid / MAX(old_resid, 1.E-20));
       WAIT;
    }

    if (lin_sys_info->info >= 2)
    {
      fprintf(MGiter, "%3d %.4lf\n", iter+1, resid / MAX(old_resid, 1.E-20));
      fflush(MGiter);
    }

    if (lin_sys_info->info >= 5)
    {
#if SPARSE_MG
      MSG("test_solution() call not implemented yet\n");
#else
      dof_copy(multi_grid_info->lin_sys_info->u_h, lin_sys_info->u_h);
      /* test_solution(mesh); */
#endif
    }

    old_resid = resid;

    if (resid <= lin_sys_info->tol)
      break;
  }


  if (lin_sys_info->info >= 2)
  {
    if (iter < lin_sys_info->max_iter)
      MSG("MG: convergence at iteration %3d: Residuum %.4le\n",
	  iter+1, resid);
    else
    {
      MSG("MG: NO CONVERGENCE after iteration %3d: Residuum %.4le\n",
	  iter, resid);
      WAIT;
    }
  }

#if SPARSE_MG
  dof_copy_from_sparse(multi_grid_info,
		       mg_lin_sys_info->u_h, lin_sys_info->u_h);
#else
  dof_copy(multi_grid_info->lin_sys_info->u_h, lin_sys_info->u_h);
#endif

  return;
}
