#include "albert.h"

/*--------------------------------------------------------------------------*/
/*  matrix-vector multiplication, standard preconditioning and standard     */
/*  solver interface to OEM-library for decoupled vector valued problems:   */
/*                                                                          */
/*  A u_n = f_n, n = 1,...,DIM_OF_WORLD, where                              */
/*  A is stored in a DOF_MATRIX, u and f in DOF_REAL_D_VECs                 */
/*--------------------------------------------------------------------------*/

struct data
{
  const FE_SPACE      *fe_space;
  const DOF_ADMIN     *admin;
  const DOF_MATRIX    *matrix;
  const DOF_SCHAR_VEC *bound;

  struct data         *next;
};


static void diag_precon_d(void *ud, int dim, REAL *r)
{
  FUNCNAME("diag_precon_d");
  struct data   *data = ud;
  MATRIX_ROW    **row = data->matrix->matrix_row;
  const S_CHAR  *b    = data->bound ? data->bound->vec : nil;
  REAL_D        *rd = (REAL_D *) r;
  int           n;

  TEST_EXIT(dim == data->admin->size_used*DIM_OF_WORLD)
    ("argument dim != admin->size_used*DIM_OF_WORLD\n");

  if (b)
  {
    FOR_ALL_DOFS(data->admin, 
		 if (b[dof] <= 0  && row[dof])
		 {
		   for (n = 0; n < DIM_OF_WORLD; n++)
		     rd[dof][n] /= row[dof]->entry[0];
		 });
  }
  else
  {
    FOR_ALL_DOFS(data->admin, 
		 if (row[dof])
		 {
		   for (n = 0; n < DIM_OF_WORLD; n++)
		     rd[dof][n] /= row[dof]->entry[0];
		 });
  }
  return;
}

/*--------------------------------------------------------------------------*/
/*   y <- A x                                                               */
/*--------------------------------------------------------------------------*/

int mat_vec_d(void *ud, int dim, const REAL *x, REAL *y)
{
  FUNCNAME("mat_vec_d");
  DOF_REAL_D_VEC  dof_x = {nil, nil, "x", 0, nil, nil, nil};
  DOF_REAL_D_VEC  dof_y = {nil, nil, "y", 0, nil, nil, nil};
  struct data  *data = ud;
  const S_CHAR *b = data->bound ? data->bound->vec : nil;
  int          n;
  REAL_D       *xd = (REAL_D *) x, *yd = (REAL_D *) y;

  TEST_EXIT(dim == data->admin->size_used*DIM_OF_WORLD)
    ("argument dim != admin->size_used*DIM_OF_WORLD\n");

  dof_x.fe_space = dof_y.fe_space = data->fe_space;
  dof_x.size = dof_y.size = data->admin->size_used;
  dof_x.vec = xd;
  dof_y.vec = yd;

  dof_mv_d(NoTranspose, data->matrix, &dof_x, &dof_y);

  if (b)
  {
    FOR_ALL_DOFS(data->admin, 
		 if (b[dof] > 0)
		 {
		   for (n = 0; n < DIM_OF_WORLD; n++)
		     yd[dof][n] = xd[dof][n];
		 });
  }

  return(0);
}

/*--------------------------------------------------------------------------*/
/*   y <- A^T x                                                             */
/*--------------------------------------------------------------------------*/

static int mat_T_vec_d(void *ud, int dim, const REAL *x, REAL *y)
{
  FUNCNAME("mat_T_vec_d");
  DOF_REAL_D_VEC  dof_x = {nil, nil, "x", 0, nil, nil, nil};
  DOF_REAL_D_VEC  dof_y = {nil, nil, "y", 0, nil, nil, nil};

  struct data  *data = ud;
  const S_CHAR *b = data->bound ? data->bound->vec : nil;
  REAL_D       *xd = (REAL_D *) x, *yd = (REAL_D *) y;

  TEST_EXIT(dim == data->admin->size_used*DIM_OF_WORLD)
    ("argument dim != admin->size_used*DIM_OF_WORLD\n");

  dof_x.fe_space = dof_y.fe_space = data->fe_space;
  dof_x.size = dof_y.size = data->admin->size_used;
  dof_x.vec = xd;
  dof_y.vec = yd;

  dof_mv_d(Transpose, data->matrix, &dof_x, &dof_y);

  if (b)
    ERROR_EXIT("don't know how to handle bound vector???\n");

  return(0);
}

/*--------------------------------------------------------------------------*/
/*---  interface to ofm solver for decoupled vector valued problems  -------*/
/*--------------------------------------------------------------------------*/
#include <oem.h>

int oem_solve_d(const DOF_MATRIX *A, const DOF_REAL_D_VEC *f, 
		const DOF_SCHAR_VEC *bound, DOF_REAL_D_VEC *u, 
		OEM_SOLVER solver, REAL tol, int icon,
		int restart, int max_iter, int info)
{
  FUNCNAME("oem_solve_d");
  int          n, iter = -1, dim, size_used;
  size_t       size;
  OEM_DATA     oem = {nil};
  struct data  data;
  const PRECON *HB = nil;
  WORKSPACE    ws = {0, nil};

  size_used = A->fe_space->admin->size_used;
  FOR_ALL_FREE_DOFS(A->fe_space->admin, 
		    if (dof < size_used)
		    {
		      for (n = 0; n < DIM_OF_WORLD; n++)
			u->vec[dof][n] = f->vec[dof][n] = 0.0;
		    });

  data.fe_space = A->fe_space;
  data.admin    = A->fe_space->admin;
  data.matrix   = A;
  data.bound    = bound;

  oem.mat_vec = mat_vec_d;
  oem.mat_vec_data = &data;

  switch(icon)
  {
  case 0: 
    oem.left_precon = nil;
    break;
  case 2:
    HB = get_HB_precon_d();
    break;
  default:
    oem.left_precon = diag_precon_d;
    oem.left_precon_data = &data;
    break;
  }

  if (HB)
  {
    
    if ((oem.left_precon_data = HB->init_precon(A, bound, info)))
      oem.left_precon = HB->precon;
    else
    {
      oem.left_precon = diag_precon_d;
      oem.left_precon_data = &data;
      HB = nil;
    }
  }

  oem.ws = &ws;

  oem.tolerance = tol;
  oem.restart   = 0;
  oem.max_iter  = max_iter;
  oem.info      = info;

  dim = size_used*DIM_OF_WORLD;
  switch (solver)
  {
  case BiCGStab:
    REALLOC_WORKSPACE(&ws, 5*dim*sizeof(REAL));
    iter = oem_bicgstab(&oem, dim, (REAL *) f->vec, (REAL *) u->vec);
    break;
  case CG:
    REALLOC_WORKSPACE(&ws, 3*dim*sizeof(REAL));
    iter = oem_cg(&oem, dim, (REAL *) f->vec, (REAL *) u->vec);
    break;
  case GMRes:
    oem.restart = restart = MAX(0, MIN(restart, dim));
    size = ((restart+2)*dim + restart*(restart+4))*sizeof(REAL);
    REALLOC_WORKSPACE(&ws, size);
    iter = oem_gmres(&oem, dim, (REAL *) f->vec, (REAL *) u->vec);
    break;
  case ODir:
    REALLOC_WORKSPACE(&ws, 6*dim*sizeof(REAL));
    iter = oem_odir(&oem, dim, (REAL *) f->vec, (REAL *) u->vec);
    break;
  case ORes:
    REALLOC_WORKSPACE(&ws, 7*dim*sizeof(REAL));
    iter = oem_ores(&oem, dim, (REAL *) f->vec, (REAL *) u->vec);
    break;
  default:
    ERROR_EXIT("unknown OEM solver %d\n", (int) solver);
  }

  CLEAR_WORKSPACE(&ws);
  if (HB) HB->exit_precon(oem.left_precon_data);
  return(iter);
}

void *init_mat_vec_d(const DOF_MATRIX *A, const DOF_SCHAR_VEC * bound)
{
  FUNCNAME("init_mat_vec_d");
  static struct data *first = nil;
  struct data        *data;

  for (data = first; data; data = data->next)
    if (data->matrix == A  &&  data->bound == bound)  break;

  if (!data) 
  {
    data = MEM_ALLOC(1, struct data);

    data->fe_space = A->fe_space;
    data->admin    = A->fe_space->admin;
    data->matrix   = A;
    data->bound    = bound;

    data->next     = first;
    first          = data;
  }

  return(data);
}

/*--------------------------------------------------------------------------*/
/*---  initialize diagonal preconditioning  --------------------------------*/
/*--------------------------------------------------------------------------*/

static void *init_diag_precon_d(const DOF_MATRIX *A, 
				const DOF_SCHAR_VEC *bound, int info)
{
  FUNCNAME("init_diag_precon_d");
  static struct data *first = nil;
  struct data        *data;

  for (data = first; data; data = data->next)
    if (data->matrix == A  &&  data->bound == bound)  break;

  if (!data) 
  {
    data = MEM_ALLOC(1, struct data);

    data->fe_space = A->fe_space;
    data->admin    = A->fe_space->admin;
    data->matrix   = A;
    data->bound    = bound;

    data->next     = first;
    first          = data;
  }
  return(data);
}

static void exit_diag_precon_d(void *data){}

const PRECON *get_diag_precon_d(void)
{
  static const PRECON precon = {init_diag_precon_d,
				diag_precon_d,
				exit_diag_precon_d};
  return(&precon);
}
