#include "albert.h"

void daxpy(int n, double alpha, const double *x, int ix, double *y, int iy)
{
  extern void daxpy_(int *, double *, const double *, int *, double *, int *);
  daxpy_(&n, &alpha, x, &iy, y, &iy);
  return;
}

void dcopy(int n, const double *x, int ix, double *y, int iy)
{
  extern void dcopy_(int *, const double *, int *, double *, int *);
  dcopy_(&n, x, &ix, y, &iy);
  return;
}

double ddot(int n, const double *x, int ix, const double *y, int iy)
{
  extern double ddot_(int *, const double *, int *, const double *, int *);
  return(ddot_(&n, x, &ix, y, &iy));
}

double dnrm2(int n, const double *x, int ix)
{
  double dnrm2_(int *, const double *, int *);
  return(dnrm2_(&n, x, &ix));
}

void dscal(int n, double alpha, double *x, int ix)
{
  extern void dscal_(int *, const double *, double *, int *);
  dscal_(&n, &alpha, x, &ix);
}

void dset(int n, double alpha, double *x, int ix)
{
/*   extern void dset_(int *, const double *, double *, int *); */
/*   dset_(&n, &alpha, x, &ix); */
  int i, nix = n*ix;
  for (i = 0; i < nix; i += ix)
    x[i] = alpha;
  return;
}

void dswap(int n, const double *x, int ix, double *y, int iy)
{
  extern void dswap_(int *, const double *, int *, double *, int *);
  dswap_(&n, x, &ix, y, &iy);
}

void xpay(int n, double alpha, const double *x, double *y)
{
  int   i;

  for (i = 0; i < n; i++)
    y[i] = x[i] + alpha*y[i];
  
  return;
}

/****************************************************************************/
/*  orthres for linear, symmetric, non singular problems                    */
/****************************************************************************/

int orthres(int dim, REAL *u, const REAL *b, int max_iter, REAL *tol,
	    int info, void (*mv)(const REAL *, REAL *),
	    void (*precon)(REAL *), WORKSPACE *workspace)
{
  FUNCNAME("orthres");
  int      iter, step, dim2 = 2*dim;
  REAL     *Ad, *Ad_old, *Ah, *d, *d_old, *h, *r;
  REAL     rho, hd_2, alpha, delta, gamma, sigma, tol1, tol2;
  REAL     daad, daad_old, res, res_old, rad;

  tol1 = *tol;
  tol2 = tol1*tol1;

  step = 40;
  step /= MAX(info-2,1);

/*---  check and partion work space                                         */

  if (!workspace)
    ERROR_EXIT("need at least a pointer to a workspace\n");

  if (7*dim*sizeof(REAL) > workspace->size)
  {
    WARNING("need workspace for 7 vectors = %d REALs\n", 7*dim);
    WARNING("reallocating workspace\n");
    workspace = REALLOC_WORKSPACE(workspace, 7*dim*sizeof(REAL));
  }

  d       = (REAL *) workspace->work;
  Ad      = d+dim;
  d_old   = Ad+dim;
  Ad_old  = d_old+dim;
  Ah      = Ad_old+dim;
  r       = Ah+dim;
  h       = r+dim;

/****************************************************************************/
/*  Step 0: Initalization                                                   */
/****************************************************************************/

  if (step == 1)
    INFO(info,2)("iter. |  residuum |  red. |\n");
  else 
    INFO(info,2)("iter. |  residuum |\n");

/*--- initial residual  ----------------------------------------------------*/
  dim2 = 2*dim; /* Important: sequence of vectors in memory for use of dim2!*/

  (*mv)(u, r);
  daxpy(dim, -1.0, b, 1, r, 1);
  dcopy(dim, r, 1, h, 1);

  if (precon) (*precon)(h);

  dcopy(dim, h, 1, d, 1);
  dset(dim2, 0.0, d_old, 1);  /*--- d_old = 0 and Ad_old = 0 ---------------*/
  (*mv)(d, Ad);
  daad_old = 1.0;

  res_old = res = dnrm2(dim, r, 1);

  iter = 0;
  if (step == 1)
    INFO(info,2)("%5d | %8.3le |       |\n", iter, res);
  else
    INFO(info,2)("%5d | %8.3le |\n", iter, res);

  if (res <= tol1)
  {
    *tol = res;
    return(0);
  }

  for (iter = 1; iter <= max_iter; iter++)
  {
/*--- compute r.Ad and d.A.A.d  ---------------------------------------------*/
    rad = ddot(dim, r, 1, Ad, 1);
    daad = ddot(dim, Ad, 1, Ad, 1);

    if (daad <= 1.e-30)
    {
      if (iter > 1)
	INFO(info,2)("%5d | %8.3le | %5.2lf |\n", iter, res);
      INFO(info,2)("(Ad,Ad)_2 = 0\n");
      *tol = sqrt(res);
      return(iter);
    }

/*--- update u and r  ------------------------------------------------------*/
    
    alpha= rad/daad;
    daxpy(dim, -alpha, d, 1, u, 1);
    daxpy(dim, -alpha, Ad, 1, r, 1);

    res = dnrm2(dim, r, 1);

    if (res < tol1)
    {
      INFO(info,2)("%5d | %8.3le |\n", iter, *tol = res);
      return(iter);
    }
    else if (iter == max_iter)
    {
      INFO(info,1)("%5d | %8.3le |\n", iter, *tol = res);
      INFO(info,1)("tolerance not reached after %d iterations\n", iter);
      return(iter);
    }
    if (step == 1)
      INFO(info,2)("%5d | %8.3le | %5.2lf |\n", iter, *tol = res,
		   res/res_old);
    else if (iter%step == 0)
      INFO(info,2)("%5d | %8.3le |\n", iter, res);

    dcopy(dim, Ad, 1, h, 1);
    if (precon) (*precon)(h);
    
    (*mv)(h, Ah);
    gamma = ddot(dim, Ah, 1, Ad, 1)/daad;
    sigma = daad/daad_old;

/*--- compute new d.Ad, save old values!  ----------------------------------*/
    dswap(dim2, d, 1, d_old, 1);         /*--- d_old = d, Ad_old = Ad ------*/
    dscal(dim2, -sigma, d, 1);           /*--- d,Ad *= -sigma --------------*/
    daxpy(dim2, -gamma, d_old, 1, d, 1); /*--- d,Ad -= gamma*d_old,Ad_old --*/
    daxpy(dim, 1.0, h, 1, d, 1);
    daxpy(dim, 1.0, Ah, 1, Ad, 1);
    daad_old = daad;

    res_old = res;
  }

  return(0);
}

/****************************************************************************/
/*--- ALBERT interface -----------------------------------------------------*/
/****************************************************************************/

/****************************************************************************/
/*  diaoganal preconditioning for scalar problems                           */
/****************************************************************************/

static WORKSPACE        workspace = {nil, nil};
static const DOF_MATRIX *matrix;
static const FE_SPACE   *fe_space;

static void diag_precon_s(double *r)
{
  FUNCNAME("diag_precon_s");
  MATRIX_ROW **row = matrix->matrix_row;

  FOR_ALL_DOFS(fe_space->admin, if (row[dof]) r[dof] /= row[dof]->entry[0]);
  return;
}

static void mat_vec_s(const REAL *x, REAL *y)
{
  FUNCNAME("mat_vec");
  DOF_REAL_VEC  dof_x = {nil, nil, "x", 0, nil, nil, nil};
  DOF_REAL_VEC  dof_y = {nil, nil, "y", 0, nil, nil, nil};

  dof_x.fe_space = dof_y.fe_space = fe_space;
  dof_x.size = dof_y.size = fe_space->admin->size_used;
  dof_x.vec = (REAL *) x;
  dof_y.vec = y;

  dof_mv(NoTranspose, matrix, &dof_x, &dof_y);

  return;
}

/****************************************************************************/
/****************************************************************************/
/****************************************************************************/

int orthres_s(const DOF_MATRIX *a, DOF_REAL_VEC *u, const DOF_REAL_VEC *f,
	      REAL tol, int icon, int max_iter, int info)
{
  FUNCNAME("orthres_s");
  int           dim, n_iter;
  void          (*precon)(REAL *);

  matrix    = a;
  fe_space  = matrix->fe_space;
  dim       = fe_space->admin->size_used;

  FOR_ALL_FREE_DOFS(fe_space->admin, if (dof < fe_space->admin->size_used)
                                       u->vec[dof] = f->vec[dof] = 0.0);

  REALLOC_WORKSPACE(&workspace, 7*dim*sizeof(REAL));

  switch(icon)
  {
  case 0: precon = nil; break;
  default:
    precon = diag_precon_s; break;
  }

  n_iter = orthres(dim , (REAL *) u->vec, (const REAL *) f->vec, max_iter,
		   &tol, info, mat_vec_s, precon, &workspace);

  return(n_iter);
}

/****************************************************************************/
/*  diaoganal preconditioning for decoupled vector valued problems          */
/****************************************************************************/

static void diag_precon_d(REAL *r)
{
  FUNCNAME("diag_precon_d");
  MATRIX_ROW    **row = matrix->matrix_row;
  REAL_D        *rd  = (REAL_D *) r;
  int           d;

  FOR_ALL_DOFS(fe_space->admin,
	       if (row[dof]) 
	       {
		 for (d = 0; d < DIM_OF_WORLD; d++)
		   rd[dof][d] /= row[dof]->entry[0];
	       });
  return;
}

/****************************************************************************/
/*   y <- Ax   for decoupled vector valued problems,                        */
/*      i.e.  y_d = A x_d,                                                  */
/****************************************************************************/

static void mat_vec_d(const REAL *x, REAL *y)
{
  FUNCNAME("mat_vec_d");
  DOF_REAL_D_VEC dof_y = {nil, nil, "y", 0, nil, nil, nil};
  DOF_REAL_D_VEC dof_x = {nil, nil, "x", 0, nil, nil, nil};

  dof_x.fe_space = dof_y.fe_space = fe_space;
  dof_x.size = dof_y.size = fe_space->admin->size_used;
  dof_x.vec = (REAL_D *) x;
  dof_y.vec = (REAL_D *) y;

  dof_mv_d(NoTranspose, matrix, &dof_x, &dof_y);
  return;
}

int orthres_d(const DOF_MATRIX *a, DOF_REAL_D_VEC *u, const DOF_REAL_D_VEC *f,
	      REAL tol, int icon, int max_iter, int info)
{
  FUNCNAME("orthres_d");
  int           d, dim, n_iter;
  void          (*precon)(REAL *);

  matrix    = a;
  fe_space  = matrix->fe_space;
  dim       = DIM_OF_WORLD*fe_space->admin->size_used;
  
  FOR_ALL_FREE_DOFS(fe_space->admin, 
		    if (dof < fe_space->admin->size_used)
		       for (d = 0; d < DIM_OF_WORLD; d++)
		          u->vec[dof][d] = f->vec[dof][d] = 0.0);

  REALLOC_WORKSPACE(&workspace, 7*dim*sizeof(REAL));

  switch(icon)
  {
  case 1:
    precon = diag_precon_d;
    break;
  default:
    precon = nil;
    break;
  }
  n_iter = orthres(dim , (REAL *) u->vec, (const REAL *) f->vec, max_iter,
		   &tol, info, mat_vec_d, precon, &workspace);

  return(n_iter);
}
