/****************************************************************************/
/* ALBERT:   an Adaptive multi Level finite element toolbox using           */
/*           Bisectioning refinement and Error control by Residual          */
/*           Techniques                                                     */
/*                                                                          */
/* file:                                                                    */
/*                                                                          */
/*                                                                          */
/* description:                                                             */
/*                                                                          */
/*                                                                          */
/* history:                                                                 */
/*                                                                          */
/*                                                                          */
/****************************************************************************/
/*                                                                          */
/*        authors: Alfred Schmidt and Kunibert G. Siebert                   */
/*                 Institut f"ur Angewandte Mathematik                      */
/*                 Albert-Ludwigs-Universit"at Freiburg                     */
/*                 Hermann-Herder-Str. 10                                   */
/*                 79104 Freiburg                                           */
/*                 Germany                                                  */
/*                                                                          */
/*        email:   alfred,kunibert@mathematik.uni-freiburg.de               */
/*                                                                          */
/*     (c) by A. Schmidt and K.G. Siebert (1996-1999)                       */
/*                                                                          */
/****************************************************************************/
/****************************************************************************/
/****************************************************************************/
/*********       SOR                                                 ********/
/****************************************************************************/
/****************************************************************************/

#include "albert.h"

/* void sor(MESH *mesh, LIN_SYS_INFO *ls_info) */

void sor_s(DOF_MATRIX *a, DOF_REAL_VEC *u, const DOF_REAL_VEC *f, 
	   REAL omega, REAL tol, int max_iter, int info)
{
  FUNCNAME("sor_s");
  const REAL *fvec;
  REAL       *uvec;
  REAL       max, omega1, accu, unew;
  int        j, jcol, iter;
  MATRIX_ROW *row;

  fvec      = f->vec;
  uvec      = u->vec;

  if (omega <= 0  ||  omega > 2)
  {
    ERROR("omega %le not in (0,2], setting omega = 1.0\n", omega);
    omega = 1.0;
  }
  omega1   = 1.0 - omega;

  if (info >= 2)
    MSG("omega = %.3lf, tol = %.3le, max_iter = %d\n", omega, tol, max_iter);

  for (iter = 0; iter < max_iter; iter++)
  {
    max =  0.0;

    FOR_ALL_DOFS(u->fe_space->admin,
	accu = 0.0;
	for (row = a->matrix_row[dof]; row; row = row->next) {
	  for (j=0; j<ROW_LENGTH; j++) {
	    jcol = row->col[j];
	    if (ENTRY_USED(jcol) && (jcol != dof)) {
	      accu += row->entry[j] * uvec[jcol];
	    }
	    else {
	      if (jcol == NO_MORE_ENTRIES)
		break;
	    }
	  }
	}
	if ((row = a->matrix_row[dof]))
	{
	  unew = omega1*uvec[dof] + omega * (fvec[dof]-accu) / row->entry[0];
	}
	else
	{
	  unew = 0.0;
	}
	max = MAX(max, ABS(uvec[dof] - unew));
	uvec[dof] = unew;
    );
    
    if (info >= 4)
      MSG("iter %3d: max = %.3le\n",iter,max);

    if (max < tol) break;
  }
  
  if (info >= 2)
  {
    if (iter < max_iter)
      MSG("convergence after iter %3d: max = %.3le\n", iter, max);
    else
      MSG("NO CONVERGENCE after iter %3d: max = %.3le\n", iter, max);
  }
  return;
}

int sor_d(DOF_MATRIX *a, DOF_REAL_D_VEC *u, const DOF_REAL_D_VEC *f, 
	   REAL omega, REAL tol, int max_iter, int info)
{
  FUNCNAME("sor_d");
  int           n;
  const REAL_D  *fvec;
  REAL_D        *uvec;
  REAL          max, omega1, accu, unew;
  int           j, jcol, iter;
  MATRIX_ROW    *row;

  fvec      = f->vec;
  uvec      = u->vec;

  if (omega <= 0  ||  omega > 2)
  {
    ERROR("omega %le not in (0,2], setting omega = 1.0\n", omega);
    omega = 1.0;
  }
  omega1   = 1.0 - omega;
  ERROR_EXIT("not implemented yet\n");
#if 0
  if (info >= 2)
    MSG("omega = %.3lf, tol = %.3le, max_iter = %d\n", omega, tol, max_iter);

  for (iter = 0; iter < max_iter; iter++)
  {
    max =  0.0;

    FOR_ALL_DOFS(u->fe_space->admin,
	accu = 0.0;
	for (row = a->matrix_row[dof]; row; row = row->next) {
	  for (j=0; j<ROW_LENGTH; j++) {
	    jcol = row->col[j];
	    if (ENTRY_USED(jcol) && (jcol != dof)) {
	      accu += row->entry[j] * uvec[jcol];
	    }
	    else {
	      if (jcol == NO_MORE_ENTRIES)
		break;
	    }
	  }
	}
	if ((row = a->matrix_row[dof]))
	{
	  unew = omega1*uvec[dof] + omega * (fvec[dof]-accu) / row->entry[0];
	}
	else
	{
	  unew = 0.0;
	}
	max = MAX(max, ABS(uvec[dof] - unew));
	uvec[dof] = unew;
    );
    
    if (info >= 4)
      MSG("iter %3d: max = %.3le\n",iter,max);

    if (max < tol) break;
  }
  
  if (info >= 2)
  {
    if (iter < max_iter)
      MSG("convergence after iter %3d: max = %.3le\n", iter, max);
    else
      MSG("NO CONVERGENCE after iter %3d: max = %.3le\n", iter, max);
  }
#endif

  return(iter);
}

/****************************************************************************/

int ssor_s(DOF_MATRIX *a, DOF_REAL_VEC *u, const DOF_REAL_VEC *f, 
	   REAL omega, REAL tol, int max_iter, int info)
{
  FUNCNAME("ssor_s");
  const REAL *fvec;
  REAL       *uvec;
  REAL       max, omega1, accu, unew;
  int        i, j, jcol, iter, dim;
  MATRIX_ROW *row;

  fvec      = f->vec;
  uvec      = u->vec;

  dof_compress(a->fe_space->mesh);

  if (omega <= 0  ||  omega > 2)
  {
    ERROR("omega %le not in (0,2], setting omega = 1.0\n", omega);
    omega = 1.0;
  }
  omega1   = 1.0 - omega;

  if (info >= 2)
  {
    MSG("omega = %.3lf, tol = %.3le, max_iter = %d\n",
	omega, tol, max_iter);
  }

  for (iter = 0; iter < max_iter; iter++)
  {
    max =  0.0;
    dim = u->fe_space->admin->size_used; 


    for (i = 0; i < dim; i++) {
      if (a->matrix_row[i])
      {
	accu = 0.0;
	for (row = a->matrix_row[i]; row; row = row->next) {
	  for (j=0; j<ROW_LENGTH; j++) {
	    jcol = row->col[j];
	    if (ENTRY_USED(jcol) && (jcol != i)) {
	      accu += row->entry[j] * uvec[jcol];
	    }
	    else {
	      if (jcol == NO_MORE_ENTRIES)
		break;
	    }
	  }
	}
	if ((row = a->matrix_row[i]))
	{
	  unew = omega1 * uvec[i]
	    + omega * (fvec[i] - accu) / row->entry[0];
	}
	else
	{
	  unew = 0.0;
	}
	max = MAX(max, ABS(uvec[i] - unew));
	uvec[i] = unew;
      }
    }

    for (i = dim-1; i >= 0; i--) {
      if (a->matrix_row[i])
      {
	accu = 0.0;
	for (row = a->matrix_row[i]; row; row = row->next) {
	  for (j=0; j<ROW_LENGTH; j++) {
	    jcol = row->col[j];
	    if (ENTRY_USED(jcol) && (jcol != i)) {
	      accu += row->entry[j] * uvec[jcol];
	    }
	    else {
	      if (jcol == NO_MORE_ENTRIES)
		break;
	    }
	  }
	}
	if ((row = a->matrix_row[i]))
	{
	  unew = omega1 * uvec[i]
	    + omega * (fvec[i] - accu) / row->entry[0];
	}
	else
	{
	  unew = 0.0;
	}
	max = MAX(max, ABS(uvec[i] - unew));
	uvec[i] = unew;
      }
    }
    
    if (info >= 4)
      MSG("iter %3d: max = %.3le\n",iter,max);

    if (max < tol) break;
  }
  
  if (info >= 2)
  {
    if (iter < max_iter)
      MSG("convergence after iter %3d: max = %.3le\n", iter, max);
    else
      MSG("NO CONVERGENCE after iter %3d: max = %.3le\n", iter, max);
  }
  return(iter);
}

int ssor_d(DOF_MATRIX *a, DOF_REAL_D_VEC *u, const DOF_REAL_D_VEC *f, 
	   REAL omega, REAL tol, int max_iter, int info)
{
  FUNCNAME("ssor_d");
  ERROR_EXIT("not implemented yet\n");
  return(0);
}
/****************************************************************************/
/****************************************************************************/
/******  GMRES                                                        *******/
/****************************************************************************/
/****************************************************************************/

#define SMALL 1.e-15

static REAL householder_vec(int n, double sigma, REAL *u, REAL *x)
{
  FUNCNAME("householder_vec");
  int     i;
  double  beta;
  
  if (n <= 0)
  {
    MSG("ha ha\n");
    return(0.0);
  }

  if (x[0] >= 0)
  {
    beta = sqrt(0.5/(sigma*(sigma + x[0])));
    if (u)
    {
      u[0] = (x[0] + sigma)*beta;
      for (i = 1; i < n; i++)
	u[i] = x[i]*beta;
    }
    return(-sigma);
  }
  else
  {
    beta = sqrt(0.5/(sigma*(sigma - x[0])));
    if (u)
    {
      u[0] = (x[0] - sigma)*beta;
    
      for (i = 1; i < n; i++)
	u[i] = x[i]*beta;
    }
    return(sigma);
  }
}

static void Pm_v(REAL *v, REAL *um, int m, int n)
{
  FUNCNAME("Pm_v");
  int      i;
  double   skp = 0.0;

  for (i = m; i < n; i++)
    skp += um[i]*v[i];
  skp *= 2.0;

  for (i = m; i < n; i++)
    v[i] -= skp*um[i];
}

#if 0
static void cal_new_basisvec1(int m, int n, int k, REAL *U, REAL *LR,
			      REAL *v, REAL *b, REAL *y,
			      void (*mv)(REAL *, const REAL *))
{
  FUNCNAME("cal_new_basisvec1");
  int      i;

  for (i = 0; i < n; i++)
    b[i] = 0.0;
  b[m-1] = 1.0;

  for (i = m-1; i >= 0; i--)
    Pm_v(b, U+(i*n), i, n);

  (*mv)(b, v);

  for (i = 0; i < m; i++)
    Pm_v(v, U+(i*n), i, n);
}
#endif

static void cal_new_basisvec(int m, int n, int k, REAL *U, REAL *LR,
			     REAL *v, REAL *b, REAL *y,
			     void (*mv)(const REAL *, REAL *), 
			     void (*lpc)(REAL *), void (*rpc)(REAL *))
{
  FUNCNAME("cal_new_basisvec");
  int      i, j, l;

/****************************************************************************/
/*  first calculate y = 2 U_m^T e_m                                         */
/****************************************************************************/
  
  for (j = 0; j < m; j++)
    b[j] = 2.0*U[j*n+(m-1)];

/****************************************************************************/
/*  now solve L_m^T y = b in R^m  (upper triagonal system, diagonal == 1.0  */
/****************************************************************************/

  y[m-1] = b[m-1];
  for (j = m-2; j >= 0; j--)
  {
    double yj = b[j];

    for (l = j + 1; l < m; l++)
      yj -= LR[l*k+j]*y[l];

    y[j] = yj;
  }
  
/****************************************************************************/
/*  b = -U_m y + e_m                                                        */
/****************************************************************************/

  for (i = 0; i < n; i++)
  {
    double bi = 0.0;

    for (j = 0; j < MIN(i+1,m); j++)
      bi += U[j*n+i]*y[j];

    b[i] = -bi;
  }
  b[m-1] += 1.0;

/****************************************************************************/
/*  v = Ab                                                                  */
/****************************************************************************/

  if (rpc)
    (*rpc)(b);

  (*mv)(b, v);

  if (lpc)
    (*lpc)(v);

/****************************************************************************/
/*  b = 2 U_m^T v in R^m                                                    */
/****************************************************************************/

  for (j = 0; j < m; j++)
  {
    double bj = 0.0;

    for (i = j; i < n; i++)
      bj += U[j*n+i]*v[i];
    b[j] = 2.0*bj;
  }

/****************************************************************************/
/*  now solve L_m y = b in R^m  (lower triagonal system, diagonal == 1.0    */
/****************************************************************************/

  y[0] = b[0];
  for (j = 1; j < m; j++)
  {
    double yj = b[j];

    for (l = 0; l < j; l++)
      yj -= LR[j*k+l]*y[l];

    y[j] = yj;
  }

/****************************************************************************/
/*  v = v - U_m y                                                           */
/****************************************************************************/

  for (i = 0; i < n; i++)
  {
    double vi = 0.0;

    for (j = 0; j < MIN(i+1,m); j++)
      vi += U[j*n+i]*y[j];

    v[i] -= vi;
  }

  return;
}

double GMRES_k(int n, REAL *x0, REAL *r0, int k, double tol,
	       void (*mv)(const REAL *, REAL *), void (*lpc)(REAL *),
	       void (*rpc)(REAL *), WORKSPACE *workspace)
{
  FUNCNAME("GMRES_k");
  int      i, j, l, m;
  REAL     *v, *rm, *um1;
  REAL     *LR;
  REAL     *U;
  REAL     (*givens)[2];
  REAL     *w, *y;
  double   c, s, wm1, norm;

  for (norm = i = 0; i < n; i++)
    norm += r0[i]*r0[i];
  norm = sqrt(norm);

  if (norm < tol) return(norm);

  if (k <= 0  ||  k > n)
    k = n;

  if (((k+1)*n + k*(k+4))*sizeof(REAL) > workspace->size)
  {
    WARNING("not enough work space; size of workspace must be at least\n");
    WARNING("((k+1)*n + k*(k+4))*sizeof(double)\n");
    WARNING("reallocating workspace\n");
    workspace = REALLOC_WORKSPACE(workspace,((k+1)*n+k*(k+4))*sizeof(double));
  }

  v = workspace->work;
  U = v+n;
  LR = U+(k*n);
  givens = (double (*)[2]) (LR+(k*k));
  w = (REAL *) (givens+k);
  y = w+k;

  wm1 = householder_vec(n, norm, U, r0);
  um1 = U;

/****************************************************************************/
/*  construct k-dimensional Krylov space                                    */
/****************************************************************************/

  for (m = 0; m < k; m++)
  {
    w[m] = wm1;

    rm = r0;
    cal_new_basisvec(m+1, n, k, U, LR, rm, v, y, mv, lpc, rpc);

    if (m+1 < n)
    {
      norm = 0;
      for (i = m+1; i < n; i++)
	norm += rm[i]*rm[i];
      norm = sqrt(norm);

      if (norm > SMALL)
      {
/****************************************************************************/
/*  one of the last components of r is not zero; calculate Householder      */
/*  vector; if m < k-1, we need the Householder vector for the calculation  */
/*  of the next basis vector => store it; if m == k-1 we do not need this   */
/*  vector anymore => do not store it!                                      */
/****************************************************************************/
	if (m < k-1)
	{
	  um1 = um1+n;
	  rm[m+1] = householder_vec(n-(m+1), norm, um1+(m+1), rm+(m+1));
	}
	else
	  rm[m+1] = householder_vec(n-(m+1), norm, nil, rm+(m+1));
      }
    }

    for (j = 0; j < m; j++)
    {
      double   rj = rm[j];

      c = givens[j][0];
      s = givens[j][1];

      rm[j]   =  c*rj + s*rm[j+1];
      rm[j+1] = -s*rj + c*rm[j+1];
    }

    if (m+1 < n  &&  ABS(rm[m+1]) > SMALL)
    {
/****************************************************************************/
/*  Find Givens rotation J_m such that,                                     */
/*       a)   (J_m r)[i] = r[i],  i < m,                                    */
/*       b)   (J_m r)[m+1] = 0.0                                            */
/*         => (J_m r)[m] = +- sqrt(r[m]^2 + r[m+1]^2) =: sigma              */
/*                                                                          */
/*                                                                          */
/*                     |1  0  .  .  .  .  0|                                */
/*                     |0  1  .  .  .  .  .|                                */
/*                     |.     .           .|          c = r[m]/sigma        */
/*            J_m =    |.        .        .|          s = r[m+1]/sigma      */
/*                     |.           .     .|                                */
/*                     |.              c  s| m                              */
/*                     |0  .  .  .  . -s  c| m+1                            */
/*                                     m  m+1                               */
/****************************************************************************/

      double   sigma, maxi;

      maxi = MAX(rm[m], rm[m+1]);

      c = rm[m]/maxi;
      s = rm[m+1]/maxi;
      sigma = maxi*sqrt(c*c + s*s);
      if (rm[m] < 0)
	sigma = -sigma;
      givens[m][0] = c = rm[m]/sigma;
      givens[m][1] = s = rm[m+1]/sigma;

      rm[m] = sigma;  /* rm[m+1] == 0   automatically!                      */
      
      wm1 = -s*w[m];  /*  |wm1| is the new residual!                        */
      w[m] *= c;
    }
    else
      wm1 = 0.0;

/****************************************************************************/
/*  now, store the first m components of the column vector r in the the mth */
/*  column of LR                                                            */
/****************************************************************************/

    for (j = 0; j <= m; j++)
      LR[j*k+m] = rm[j];

/****************************************************************************/
/*  test, whether tolarance is reached or not                               */
/****************************************************************************/

    if (ABS(wm1) < tol)
    {
      m++;
      break;
    }

/****************************************************************************/
/*  tolerance not reached: calculate and store (m+1)th row of matrix L;     */
/*  this row is only needed for the computation of the (m+1)th column of    */
/*  the orthogonal matrix Q_m; this vector is the additional basis vector   */
/*  for the enlargement of the Krylov space; only needed for m < k-1        */
/*  L[m+1][j] = <u_(m+1),u_j>_2                                             */
/*  (m+1)th vector u = umi is allready stored at U+(m+1)                    */
/****************************************************************************/

    if (m < k-1)
    {
      for (j = 0; j < m+1; j++)
      {
	double  *uj = U+(j*n);
	double  val = 0.0;

	for (i = m+1; i < n; i++)
	  val += um1[i]*uj[i];

	LR[(m+1)*k+j] = 2.0*val;
      }
    }
  }

/****************************************************************************/
/*  and now solve the upper triangular system                               */
/****************************************************************************/

  y[m-1] = w[m-1]/LR[(m-1)*(k+1)];  /* = LR[(m-1)*k+(m-1)]!                 */
  for (j = m-2; j >= 0; j--)
  {
    double yj = w[j];

    for (l = j + 1; l < m; l++)
      yj -= LR[j*k+l]*y[l];

    y[j] = yj/LR[j*(k+1)];
  }

/****************************************************************************/
/*  calculate v = 2 U_m^T [e_0,....,e_m-1] y                                */
/****************************************************************************/
  
  for (i = 0; i < m; i++)
  {
    double val = 0.0;

    for (j = i; j < m; j++)
      val += U[i*n+j]*y[j];
    v[i] = 2.0*val;
  }

/****************************************************************************/
/*  now solve L_m^T w = v in R^m  (upper triagonal system, diagonal == 1.0  */
/****************************************************************************/

  w[m-1] = v[m-1];
  for (j = m-2; j >= 0; j--)
  {
    double wj = v[j];

    for (l = j + 1; l < m; l++)
      wj -= LR[l*k+j]*w[l];

    w[j] = wj;
  }

/****************************************************************************/
/*  v = [e_0,....,e_m-1] y - U_m w                                          */
/****************************************************************************/

  for (i = 0; i < n; i++)
  {
    double  vi = 0.0;
    for (j = 0; j < MIN(i+1,m); j++)
      vi += U[j*n+i]*w[j];
    v[i] = -vi;
  }

  for (j = 0; j < m; j++)
    v[j] += y[j];

/****************************************************************************/
/*  and now, make the update of u :-)                                       */
/****************************************************************************/

  if (rpc)
    (*rpc)(v);

  for (i = 0; i < n; i++)
    x0[i] += v[i];

  return(ABS(wm1));
}

/****************************************************************************/
/*                                                                          */
/*  partitioning of the workspace:                                          */
/*                                                                          */
/*  k = MIN(restart,n)                                                      */
/*  workspace:  ((k+2)*n + k*(k+4))*sizeof(double)                          */
/*  we use the following partioning                                         */
/*                                                                          */
/*                                                                          */
/*  |---------------------- n -----------------------------------|          */
/*   ____________________________________________________________           */
/*  |____________________________________________________________| r0, rm   */
/*  |____________________________________________________________| v        */
/*  |____________________________________________________________| u[0]     */
/*  |____________________________________________________________| .        */
/*                                                                 .        */
/*                                                                 .        */
/*   ____________________________________________________________  .        */
/*  |____________________________________________________________| u[k-2]   */
/*  |____________________________________________________________| u[k-1]   */
/*  |_________________________| 0                                           */
/*  |_|_______________________| 1    upper triangular matrix + diogonal:    */
/*  |___|_____________________| 2    transformed Hessenberg R_k matrix;     */
/*  |_____|___________________| .                                           */
/*                              .                                           */
/*                              .    lower triangluar matrix: matrix L_k    */
/*   _________________________       for the calculation of the (m+1)th     */
/*  |_____________________|___| k-2  column of Q_m = P_m ... P_1; diagonal  */
/*  |_______________________|_| k-1  elements L_k[i][i] = 1.0; not stored!  */
/*  |_________________________|      Givens rotations: storing c and s      */
/*  |_________________________|                                             */
/*  |_________________________|      additional workspace!                  */
/*                                                                          */
/*  |--------- k -------------|                                             */
/*                                                                          */
/****************************************************************************/

int GMRES(int n, REAL *x0, REAL *b, int restart, int maxiter,
	  REAL *tol, int info, void (*mv)(const REAL *, REAL *),
          void (*lprecon)(REAL *), void (*rprecon)(REAL *),
	  WORKSPACE *workspace)
{
  FUNCNAME("GMRES");
  int               i, iter, k;
  REAL              *r0;
  double            normb, resid, tolerance = *tol;
  WORKSPACE         gmres_ws[1];
  static const char *s[4] = {"|Ax - b|_2",
			     "|CAx - Cb|_2",
			     "|AC(C^-1)x - b|_2",
			     "|CAC(C^-1)x - Cb|_2"};
  static const char *blank[4] = {"      ","     ","    ","   "};
  const char        *string, *bl;

  if (restart <= 0  ||  restart > n)
    k = n;
  else
    k = restart;

  if (!workspace)
    ERROR_EXIT("need at least a pointer to a workspace\n");

  if (((k+2)*n + k*(k+4))*sizeof(REAL) > workspace->size)
  {
    WARNING("not enough work space; size of workspace must be at least\n");
    WARNING("((k+2)*n + k*(k+4))*sizeof(REAL)\n");
    WARNING("reallocating workspace\n");
    workspace = REALLOC_WORKSPACE(workspace,((k+2)*n+k*(k+4))*sizeof(REAL));
  }

  if (!lprecon  &&  !rprecon)
    string = s[0], bl = blank[0];
  else if (!rprecon)
    string = s[1], bl = blank[1];
  else if (!lprecon)
    string = s[2], bl = blank[2];
  else
    string = s[3], bl = blank[3];

  for (normb = i = 0; i < n; i++)
    normb += b[i]*b[i];
  normb = sqrt(normb);

  r0 = workspace->work;
  gmres_ws->work = (void *) (r0+n);
  gmres_ws->size = workspace->size-n*sizeof(REAL);

  if (normb < SMALL)
  {
    INFO(info,2)("b == 0, x = 0 is the solution of the linear system\n");
    for (i = 0; i < n; i++)
      x0[i] = 0.0;
    *tol = normb;
    return(0);
  }

  INFO(info,2)("GMRES with restart %d\n", k);
  INFO(info,2)("tolerance = %.1le for %s\n", tolerance, string);

  for (iter = 1; iter <= maxiter; iter++)
  {
/****************************************************************************/
/*  initalize: calculate the residual r0                                    */
/****************************************************************************/

    (*mv)(x0, r0);

    for (i = 0; i < n; i++)
      r0[i] = b[i] - r0[i];

    if (lprecon)
      (*lprecon)(r0);

    resid = GMRES_k(n, x0, r0, k, tolerance, mv, lprecon, rprecon, gmres_ws);

    if ((*tol = resid) < tolerance)
    {
      INFO(info,2)("\n");
      INFO(info,2)("finished succesfully with %d iterations\n", iter);
      INFO(info,4)("%s = %.3le\n", string, resid);
      INFO(info,4)("thank you for using GMRES; hope to see you again\n");
      return(iter);
    }
    else
    {
      if (info >= 6)
      {
	if (iter%40 == 1)
	  INFO(info,4)("iteration  | %s%s\n", bl, string);
	INFO(info,6)("%7d    |        %le\n", iter, resid);
      }
      else
      {
	if (iter%400 == 1)
	  INFO(info,4)("iteration  | %s%s\n", bl, string);
	if (iter == 1  ||  iter%10 == 0)
	  INFO(info,4)("%7d    |        %le\n", iter, resid);
      }
    }
  }

  INFO(info,0)("tolerance %le not reached after %d iterations\n",
	       tolerance, iter);
  INFO(info,0)("%s = %le\n", string, resid);
  INFO(info,4)("thank you for using GMRES; hope to see you again\n");

  return(resid);
}

/****************************************************************************/
/*  CG for linear, symmetric, and positive definite problems                */
/****************************************************************************/

int cg(int dim, REAL *u, const REAL *b, int max_iter, REAL *tol,
       int info, void (*mv)(const REAL *, REAL *),
       void (*precon)(REAL *), WORKSPACE *workspace)
{
  FUNCNAME("cg");
  int      iter, step;
  REAL     rho, hd_2, delta, gamma, tolerance;
  REAL     *d, *h, *r, *Cr;

  tolerance = SQR(*tol);
  step = 40;
  step /= MAX(info-2,1);

/****************************************************************************/
/*  get memory out of the work area                                         */
/****************************************************************************/

  if (!workspace)
    ERROR_EXIT("need at least a pointer to a workspace\n");

  if (3*dim*sizeof(REAL) > workspace->size)
  {
    WARNING("need workspace for 3 vectors = %d REALs\n", 3*dim);
    WARNING("reallocating workspace\n");
    workspace = REALLOC_WORKSPACE(workspace, 3*dim*sizeof(REAL));
  }

  r = (REAL *) workspace->work;
  d = r + dim;
  h = d + dim;
  Cr = precon ? h : r;

/****************************************************************************/
/*  Step 0: Initalization                                                   */
/****************************************************************************/

  if (step == 1)
    INFO(info,2)("iter. |  residuum |  red. |\n");
  else 
    INFO(info,2)("iter. |  residuum |\n");

  (*mv)(u, r);
  daxpy(dim, -1.0, b, 1, r, 1);
  if (precon)
  {
    dcopy(dim, r, 1, Cr, 1);
    (*precon)(Cr);
  }

/****************************************************************************/
/*  norm of the residual = (r,r)_C = (r,Cr)_2                               */
/****************************************************************************/
  delta = ddot(dim, r, 1, Cr, 1);

  iter = 0;
  if (step == 1)
    INFO(info,2)("%5d | %8.3le |       |\n", iter, *tol = sqrt(delta));
  else
    INFO(info,2)("%5d | %8.3le |\n", iter, *tol = sqrt(delta));

  if (delta < tolerance)
    return(iter);

/****************************************************************************/
/*  decent direction                                                        */
/****************************************************************************/

  dcopy(dim, Cr, 1, d, 1);

/****************************************************************************/
/*  Step 1: Iteration                                                       */
/****************************************************************************/

  for (iter = 1; iter <= max_iter; iter++)
  {
    (*mv)(d, h);
/****************************************************************************/
/*  decent parameter rho = delta/(h,d)_2                                    */
/****************************************************************************/
    hd_2 = ddot(dim, h, 1, d, 1);

    if (ABS(hd_2) < 1.e-30)
    {
      if (iter > 1)
	INFO(info,2)("%5d | %8.3le | %5.2lf |\n", iter, *tol = sqrt(delta));
      INFO(info,2)("(Ad,d)_2 = 0\n");
      return(iter);
    }

/****************************************************************************/
/*  update                                                                  */
/****************************************************************************/
    rho = delta/hd_2;
    daxpy(dim, -rho, d, 1, u, 1);
    daxpy(dim, -rho, h, 1, r, 1);

/****************************************************************************/
/*  norm of the residual = (r,r)_C = (r,Cr)_2                               */
/****************************************************************************/
    gamma = 1.0/delta;

    if (precon)
    {
      dcopy(dim, r, 1, Cr, 1);
      (*precon)(Cr);
    }

    delta = ddot(dim, r, 1, Cr, 1);

    if (delta < tolerance)
    {
      INFO(info,2)("%5d | %8.3le |\n", iter, *tol = sqrt(delta));
      return(iter);
    }
    else if (iter == max_iter)
    {
      INFO(info,1)("%5d | %8.3le |\n", iter, *tol = sqrt(delta));
      INFO(info,1)("tolerance not reached after %d iterations\n", iter);
      return(iter);
    }
    if (step == 1)
      INFO(info,2)("%5d | %8.3le | %5.2lf |\n", iter, *tol = sqrt(delta),
		   sqrt(delta*gamma));
    else if (iter%step == 0)
      INFO(info,2)("%5d | %8.3le |\n", iter, sqrt(delta));

/****************************************************************************/
/*  update of the decent direction                                          */
/****************************************************************************/
    gamma *= delta;
    xpay(dim, gamma, Cr, d);
  }
  return(0);    /*  statement never reached!!!                              */
}

/****************************************************************************/
/*  memory used by CG                                                       */
/****************************************************************************/

static WORKSPACE         *workspace = nil;
static const FE_SPACE    *fe_space;
static const DOF_MATRIX  *matrix;

/****************************************************************************/
/*  diaoganal preconditioning for scalar problems                           */
/****************************************************************************/

static void diag_precon_s(REAL *r)
{
  FUNCNAME("diag_precon_s");
  MATRIX_ROW **row = matrix->matrix_row;

  FOR_ALL_DOFS(fe_space->admin, if (row[dof]) r[dof] /= row[dof]->entry[0]);

  return;
}


/****************************************************************************/
/*  diaoganal preconditioning for decoupled vector valued problems          */
/****************************************************************************/

static void diag_precon_d(REAL *r)
{
  FUNCNAME("diag_precon_d");
  MATRIX_ROW    **row = matrix->matrix_row;
  REAL_D        *rd  = (REAL_D *) r;
  int           d;

  FOR_ALL_DOFS(fe_space->admin,
	       if (row[dof]) 
	       {
		 for (d = 0; d < DIM_OF_WORLD; d++)
		   rd[dof][d] /= row[dof]->entry[0];
	       });
  return;
}

/****************************************************************************/
/*  hierachical basis preconditioning; only for linear finite elements!!!   */
/****************************************************************************/
static int max_level = 0, max_dof_level = 0, mg_levels = 0;
static int dof_level_size = 0, level_size = 0;
static U_CHAR *dof_level = nil;
static DOF    *dof_parent0 = nil, *dof_parent1 = nil;
static DOF    *sort_dof = nil, *sort_dof_invers = nil;
static int    *dofs_per_level = nil, *tmp_per_level = nil;
static int    n0_vert;

static void max_level_fct(const EL_INFO *el_info)
{
  FUNCNAME("max_level_fct");
  int dof, dof0, dof1, level = (int)(el_info->level);
  EL  *el   = el_info->el;

  max_level = MAX(level, max_level);

  if (!(IS_LEAF_EL(el))) {
    dof   = el->child[0]->dof[N_VERTICES-1][n0_vert];
    dof0  = dof_parent0[dof] = el->dof[0][n0_vert];
    dof1  = dof_parent1[dof] = el->dof[1][n0_vert];
    level = dof_level[dof]   = 1 + MAX(dof_level[dof0], dof_level[dof1]);
    max_dof_level = MAX(level, max_dof_level);
  }
}

int init_HB_precon_s(const FE_SPACE *fe_space, int info)
{
  FUNCNAME("init_HB_precon_s");
  const DOF_ADMIN *admin = fe_space->admin;
  int             i, j, k, size_used = admin->size_used;

  return(false);
  if (fe_space->bas_fcts->degree > 1)  return(false);

  n0_vert = admin->n0_dof[VERTEX];

  if (dof_level_size < size_used) {
    dof_level       = MEM_REALLOC(dof_level, dof_level_size,
				  size_used, U_CHAR);
    dof_parent0     = MEM_REALLOC(dof_parent0, dof_level_size,
				  size_used, DOF);
    dof_parent1     = MEM_REALLOC(dof_parent1, dof_level_size,
				  size_used, DOF);
    sort_dof        = MEM_REALLOC(sort_dof, dof_level_size,
				  size_used, DOF);
    sort_dof_invers = MEM_REALLOC(sort_dof_invers, dof_level_size,
				  size_used, DOF);
    dof_level_size = size_used;
  }

  FOR_ALL_DOFS(admin,
	       dof_level[dof] = 0;
	       dof_parent0[dof] = dof_parent1[dof] = 0; );

  max_level = max_dof_level = 0;
  mesh_traverse(fe_space->mesh, -1, CALL_EVERY_EL_PREORDER, max_level_fct);
  mg_levels = (max_level + DIM-1) / DIM + 1;
  if (mg_levels < 2) return(false);

  TEST_EXIT(mg_levels == (max_dof_level+1))
    ("mg_levels %d != max_dof_level %d + 1\n", mg_levels, max_dof_level);
  
  if (level_size < mg_levels) {
    dofs_per_level = MEM_REALLOC(dofs_per_level, level_size,
				 mg_levels, int);
    tmp_per_level  = MEM_REALLOC(tmp_per_level, level_size,
				 mg_levels, int);
    level_size = mg_levels;
  }
  
  for (i = 0; i < mg_levels; i++) dofs_per_level[i] = 0;
  FOR_ALL_DOFS(admin, dofs_per_level[dof_level[dof]]++);
  if (info > 3) {
    MSG("dofs_per_level:");
    for (i = 0; i < mg_levels; i++) print_msg(" %d", dofs_per_level[i]);
    print_msg("\n");
  }
  for (i = 1; i < mg_levels; i++) {
    tmp_per_level[i]   = dofs_per_level[i-1];
    dofs_per_level[i] += dofs_per_level[i-1];
  }
  tmp_per_level[0] = 0;            /* pointers for filling the sort vectors */

  if (info > 3) {
    MSG("dofs_per_level accumulated:");
    for (i = 0; i < mg_levels; i++) print_msg(" %d", dofs_per_level[i]);
    print_msg("\n");
  }

#if 0
  for (i = 0; i < dofs_per_level[mg_levels-1]; i++)
    MSG("dof_parent[%3d] = (%3d,%3d), lev=%2d (%2d,%2d)\n",
	i, dof_parent0[i], dof_parent1[i],
	dof_level[i], dof_level[dof_parent0[i]],
	dof_level[dof_parent1[i]]);
#endif


/* build sort_dof[] and sort_dof_invers[] vectors                           */

  FOR_ALL_DOFS(admin,
	       j = dof_level[dof];
	       k = tmp_per_level[j]++;
	       sort_dof[k] = dof;
	       sort_dof_invers[dof] = k;
    );

#if 0
  for (i = 0; i < dofs_per_level[mg_levels-1]; i++)
  {
    j = sort_dof[i];
    MSG("sort[%3d]: dof=%3d, lev=%2d; invers[%3d]=%3d\n",
	i, j, dof_level[j], j, sort_dof_invers[j]);
  }
  WAIT;
#endif
  return(true);
}

/****************************************************************************/

void HB_precon_s(REAL *r)
{
  FUNCNAME("HB_precon_s");
  int     i, idof, level, last;
  U_CHAR  pre;

  if (mg_levels < 2)                                       /* nothing to do */
  {
    return;
  }

/* transposed basis transformation (run over refined levels only)          */

  for (level = mg_levels - 1; level > 0; level--)
  {
    last = dofs_per_level[level];

    for (i = dofs_per_level[level-1]; i < last; i++)
    {
      idof = sort_dof[i];
      r[dof_parent0[idof]] += 0.5 * r[idof];
      r[dof_parent1[idof]] += 0.5 * r[idof];
    }
  }

/* basis transformation (run over refined levels only) */

  for (level = 1; level < mg_levels; level++)
  {
    last = dofs_per_level[level];

    for (i = dofs_per_level[level-1]; i < last; i++)
    {
      idof = sort_dof[i];
      r[idof] += 0.5 * (r[dof_parent0[idof]] + r[dof_parent1[idof]]);
    }
  }

  return;
}

/****************************************************************************/
/*   y <- Ax                                                                */
/****************************************************************************/

static void mat_vec_s(const REAL *x, REAL *y)
{
  FUNCNAME("mat_vec_s");
  DOF_REAL_VEC  dof_x = {nil, nil, "x", 0, nil, nil, nil};
  DOF_REAL_VEC  dof_y = {nil, nil, "y", 0, nil, nil, nil};

  dof_x.fe_space = dof_y.fe_space = fe_space;
  dof_x.size = dof_y.size = fe_space->admin->size_used;
  dof_x.vec = (REAL *) x;
  dof_y.vec = y;

  dof_mv(NoTranspose, matrix, &dof_x, &dof_y);
  return;
}

/****************************************************************************/
/*   y <- Ax   for decoupled vector valued problems,                        */
/*      i.e.  y_d = A x_d,                                                  */
/****************************************************************************/

static void mat_vec_d(const REAL *x, REAL *y)
{
  FUNCNAME("mat_vec_d");
  DOF_REAL_D_VEC dof_y = {nil, nil, "y", 0, nil, nil, nil};
  DOF_REAL_D_VEC dof_x = {nil, nil, "x", 0, nil, nil, nil};

  dof_x.fe_space = dof_y.fe_space = fe_space;
  dof_x.size = dof_y.size = fe_space->admin->size_used;
  dof_x.vec = (REAL_D *) x;
  dof_y.vec = (REAL_D *) y;

  dof_mv_d(NoTranspose, matrix, &dof_x, &dof_y);
  return;
}

/****************************************************************************/
/****************************************************************************/
/****************************************************************************/

int cg_s(const DOF_MATRIX *mat, DOF_REAL_VEC *u, const DOF_REAL_VEC *f,
	 REAL tol, int icon, int max_iter, int info)
{
  FUNCNAME("cg_s");
  int           dim, n_iter;
  void          (*precon)(REAL *);

  matrix    = mat;
  fe_space  = matrix->fe_space;
  dim       = fe_space->admin->size_used;

  FOR_ALL_FREE_DOFS(fe_space->admin, if (dof < fe_space->admin->size_used)
                                       u->vec[dof] = f->vec[dof] = 0.0);

  workspace = REALLOC_WORKSPACE(workspace, 3*dim*sizeof(REAL));

  switch(icon)
  {
  case 0: precon = nil; break;
/*   case 2: precon = HB_precon_s; break; */
  default:
    precon = diag_precon_s; break;
  }

  n_iter = cg(dim , (REAL *) u->vec, (const REAL *) f->vec, max_iter,
	      &tol, info, mat_vec_s, precon, workspace);

  return(n_iter);
}

/****************************************************************************/
/****************************************************************************/
/****************************************************************************/

int cg_d(const DOF_MATRIX *mat, DOF_REAL_D_VEC *u, const DOF_REAL_D_VEC *f,
	 REAL tol, int icon, int max_iter, int info)
{
  FUNCNAME("cg_d");
  int           d, dim, n_iter;
  void          (*precon)(REAL *);

  matrix    = mat;
  fe_space  = matrix->fe_space;
  dim       = DIM_OF_WORLD*fe_space->admin->size_used;
  
  FOR_ALL_FREE_DOFS(fe_space->admin, 
		    if (dof < fe_space->admin->size_used)
		       for (d = 0; d < DIM_OF_WORLD; d++)
		          u->vec[dof][d] = f->vec[dof][d] = 0.0);

  workspace = REALLOC_WORKSPACE(workspace, 3*dim*sizeof(REAL));

  switch(icon)
  {
  case 1:
    precon = diag_precon_d;
    break;
  default:
    precon = nil;
    break;
  }
  n_iter = cg(dim , (REAL *) u->vec, (const REAL *) f->vec, max_iter,
	      &tol, info, mat_vec_d, precon, workspace);

  return(n_iter);
}

/****************************************************************************/
/*  gmres:                                                                  */
/****************************************************************************/

int gmres_s(const DOF_MATRIX *A, DOF_REAL_VEC *u, const DOF_REAL_VEC *f,
	    REAL tol, int icon, int max_iter, int restart, int info)
{
  FUNCNAME("gmres_s");
  int            n, iter;
  static int     size_w = 0;
  static WORKSPACE *workspace = nil;

  static         void (*lprecon)(REAL *) = nil;
  static         void (*rprecon)(REAL *) = nil;
  REAL           *x0, *b;

  matrix    = A;
  fe_space  = u->fe_space;

  n         = fe_space->admin->size_used;
  x0        = u->vec;
  b         = f->vec;

  FOR_ALL_FREE_DOFS(fe_space->admin, if (dof < fe_space->admin->size_used)
                                       x0[dof] = b[dof] = 0.0);

  if (icon == 1)
    lprecon = diag_precon_s;
  else if (icon == 2)
    rprecon = diag_precon_s;

  restart = MAX(1,restart);
  restart = MIN(restart,n);

  size_w = (restart+2)*n + restart*(restart+4);
  workspace = REALLOC_WORKSPACE(workspace, size_w*sizeof(REAL));

  iter = GMRES(n, x0, b, restart, max_iter, &tol, info, mat_vec_s,
	       lprecon, rprecon, workspace);

  return(iter);
}

int gmres_d(const DOF_MATRIX *A, DOF_REAL_D_VEC *u, const DOF_REAL_D_VEC *f,
	    REAL tol, int icon, int max_iter, int restart, int info)
{
  FUNCNAME("gmres_d");
  
  ERROR_EXIT("not implemented yet\n");
  return(0);
}
