#include "albert.h"

/****************************************************************************/
/* ALBERT:   an Adaptive multi Level finite element toolbox using           */
/*           Bisectioning refinement and Error control by Residual          */
/*           Techniques                                                     */
/*                                                                          */
/* file:     Common/sp_cg.c                                                 */
/*                                                                          */
/*                                                                          */
/* description:  cg solver for saddle point problems                        */
/****************************************************************************/
/*                                                                          */
/*        authors: Alfred Schmidt and Kunibert G. Siebert                   */
/*                 Institut f"ur Angewandte Mathematik                      */
/*                 Albert-Ludwigs-Universit"at Freiburg                     */
/*                 Hermann-Herder-Str. 10                                   */
/*                 79104 Freiburg                                           */
/*                 Germany                                                  */
/*                                                                          */
/*        email:   alfred,kunibert@mathematik.uni-freiburg.de               */
/*                                                                          */
/*     (c) by A. Schmidt and K.G. Siebert (1996-1999)                       */
/*                                                                          */
/****************************************************************************/

static void set(REAL alpha, int n, REAL *u)
{
  int  i;
  for (i = 0; i < n; i++)
    u[i] = alpha;
  return;
}

/****************************************************************************/
/*  two routines for screen output                                          */
/****************************************************************************/

static int info, msg_step, msg_precon;

static void head_info(const char *funcName, int inf, int step, int precon)
{
  info = inf; msg_step = step; msg_precon = precon;
  if (step == 1  &&  precon)
    INFO(inf,2)("iter. |  residuum |   red. |elliptic|project|precon\n");
  else if (step == 1  &&  !precon)
    INFO(inf,2)("iter. |  residuum |   red. |elliptic |project\n");
  else if (precon)
    INFO(inf,2)("iter. |  residuum |elliptic|project|precon\n");
  else
    INFO(inf,2)("iter. |  residuum |elliptic|project\n");
}

static void print_info(int iter, REAL delta, REAL red, int n0, int n1, int n2)
{
  if (msg_step == 1  &&  msg_precon)
  {
    if (red >= 0)
      INFO(info,2)("%5d | %8.3le | %6.3lf | %6d | %5d | %5d\n",
		   iter, delta, red, n0, n1, n2);
    else
      INFO(info,2)("%5d | %8.3le |        | %6d | %5d | %5d\n",
		  iter, delta, n0, n1, n2);
  }      
  else if (msg_step == 1  &&  !msg_precon)
  {
    if (red >= 0)
      INFO(info,2)("%5d | %8.3le | %6.3lf | %6d | %5d\n",
		   iter, delta, red, n0, n1);
    else
      INFO(info,2)("%5d | %8.3le |        | %6d | %5d\n", iter, delta,n0,n1);
  }
  else if (msg_precon)
      INFO(info,2)("%5d | %8.3le | %6d | %5d | %5d\n", iter, delta,n0,n1,n2);
  else
    INFO(info,2)("%5d | %8.3le | %6d | %5d\n", iter, delta, n0, n1);
}

/****************************************************************************/
/*  CG for saddle point problems                                            */
/****************************************************************************/

REAL sp_cg(SP_CG_INFO *cg_info)
{
  FUNCNAME("sp_cg");
  int      iter, size_used, step, info, precon = false;
  REAL     rho, bdchi, delta, gamma;

  REAL     tolerance;
  REAL     *u, *p, *chi, *b_p_psi, *b_phi_u, *r, *Cr, *d;
  int      n_u = cg_info->n_u, n_p = cg_info->n_p;
  int      nse = 0, npr1 = 0, npr2 = 0;

  tolerance = SQR(cg_info->tolerance);
  info = cg_info->info;
  step = 10;
  step /= MAX(info-2,1);

  u = cg_info->u;
  p = cg_info->p;

/****************************************************************************/
/*  get memory out of the work area                                         */
/****************************************************************************/

  chi = cg_info->work;
  size_used = n_u;
  r = cg_info->work + size_used;
  size_used += n_p;
  d = cg_info->work + size_used;
  size_used += n_p;
  if (cg_info->precon)
  {
    Cr = cg_info->work + size_used;
    size_used += n_p;
    precon = true;
  }
  else
    Cr = r;

/****************************************************************************/
/*    b_p_psi and  b_phi_u use the same memory area!!!                      */
/****************************************************************************/

  b_p_psi = b_phi_u = cg_info->work+size_used;
  size_used += MAX(n_u, n_p);
  
  if (size_used > cg_info->size)
  {
    if (cg_info->precon)
    {
      ERROR("need workspace for 1 u, 3 p and 1 u/p vector = %d REALs\n", 
	    1*n_u + 3*n_p + MAX(n_u, n_p));
      ERROR_EXIT("but only memory for %d REALs available\n", cg_info->size);
    }
    else
    {
      ERROR("need workspace for 1 u, 2 p and 1 u/p vector = %d REALs\n", 
	    1*n_u + 2*n_p + MAX(n_u, n_p));
      ERROR_EXIT("but only memory for %d REALs available\n", cg_info->size);
    }
  }

/****************************************************************************/
/*  start of CG for Quasi--Stokes                                           */
/****************************************************************************/
/*  Step 0: Initalization                                                   */
/****************************************************************************/
/****************************************************************************/
/*  compute correponding initial velocity                                   */
/****************************************************************************/

  dcopy(n_u, cg_info->f, 1, b_p_psi, 1);
  cg_info->add_b_p_psi(p, b_p_psi, -1.0, cg_info->user_data);

  nse = cg_info->solve_elliptic(u, b_p_psi, cg_info->user_data);

/****************************************************************************/
/*  preconditioning of the pressure; calculation of the residual            */
/****************************************************************************/

  set(0.0, n_p, b_phi_u);
  cg_info->add_b_phi_u_add_g(u, b_phi_u, -1.0, cg_info->user_data);

  set(0.0, n_p, r);
  npr1 = cg_info->project(r, b_phi_u, cg_info->user_data);

  if (cg_info->precon)
  {
    set(0.0, n_p, Cr);
    npr2 = cg_info->precon(Cr, r, b_phi_u, cg_info->user_data);
  }


/****************************************************************************/
/*  norm of the residual = (r,r) = b(r,u)                                   */
/****************************************************************************/
  if (cg_info->cal_Y_scp)
    delta = cg_info->cal_Y_scp(n_p, r, Cr, cg_info->user_data);
  else
    delta = ddot(n_p, Cr, 1, b_phi_u, 1);


  iter = 0;
  head_info(funcName, info, step, precon);
  print_info(iter, sqrt(delta), -1.0, nse, npr1, npr2);

  if (delta < tolerance)
  {
    return(sqrt(delta));
  }

/****************************************************************************/
/*  decent direction                                                        */
/****************************************************************************/

  dcopy(n_p, Cr, 1, d, 1);

/****************************************************************************/
/*  Step 1: Iteration                                                       */
/****************************************************************************/

  for (iter = 1; iter <= cg_info->max_iter; iter++)
  {
    set(0.0, n_u, b_p_psi);
    cg_info->add_b_p_psi(d, b_p_psi, +1.0, cg_info->user_data);
    set(0.0, n_u, chi);
    nse = cg_info->solve_elliptic(chi, b_p_psi, cg_info->user_data);

/****************************************************************************/
/*  decent parameter rho = delta/b(d,chi)                                   */
/****************************************************************************/

    bdchi = ddot(n_u, chi, 1, b_p_psi, 1);

    if (ABS(bdchi) < 1.e-30)
    {
      if (iter > 1 &&  (iter-1)%step != 0)
	print_info(iter, sqrt(delta), sqrt(delta*gamma), nse, npr1, npr2);
      INFO(info,2)("b(d,chi) = 0\n");
      return(sqrt(delta));
    }

    rho = delta/bdchi;

/****************************************************************************/
/*  update of the pressure and velocity                                     */
/****************************************************************************/
    daxpy(n_p, -rho, d, 1, p, 1);
    daxpy(n_u, rho, chi, 1, u, 1);

/****************************************************************************/
/*  calculation of the residual                                             */
/****************************************************************************/

    set(0.0, n_p, b_phi_u);
    cg_info->add_b_phi_u_add_g(u, b_phi_u, -1.0, cg_info->user_data);

    set(0.0, n_p, r);
    npr1 = cg_info->project(r, b_phi_u, cg_info->user_data);
    if (cg_info->precon)
    {
      set(0.0, n_p, Cr);
      npr2 = cg_info->precon(Cr, r, b_phi_u, cg_info->user_data);
    }

/****************************************************************************/
/*  norm of the residual = (r,r) = b(r,u)                                   */
/****************************************************************************/
    gamma = 1.0/delta;

    if (cg_info->cal_Y_scp)
      delta = cg_info->cal_Y_scp(n_p, r, Cr, cg_info->user_data);
    else
      delta = ddot(n_p, Cr, 1, b_phi_u, 1);

    if (delta < tolerance)
    {
      print_info(iter, sqrt(delta), sqrt(delta*gamma), nse, npr1, npr2);
      return(sqrt(delta));
    }
    else if (iter == cg_info->max_iter)
    {
      print_info(iter, sqrt(delta), sqrt(delta*gamma), nse, npr1, npr2);
      INFO(info,1)("tolerance not reached after %d iterations\n", iter);
      return(sqrt(delta));
    }

    if (iter%step == 0)
      print_info(iter, sqrt(delta), sqrt(delta*gamma), nse, npr1, npr2);

/****************************************************************************/
/*  update of the decent direction                                          */
/****************************************************************************/
    gamma *= delta;
    xpay(n_p, gamma, Cr, d);
  }
  return(0.0);  /*  statement never reached!!!                              */
}
