#include "albert.h"

/****************************************************************************/
/* ALBERT:   an Adaptive multi Level finite element toolbox using           */
/*           Bisectioning refinement and Error control by Residual          */
/*           Techniques                                                     */
/*                                                                          */
/* file:     Common/stiff.c                                                 */
/*                                                                          */
/*                                                                          */
/* description:  routines for general assembling of matrices                */
/*                                                                          */
/****************************************************************************/
/*                                                                          */
/*        authors: Alfred Schmidt and Kunibert G. Siebert                   */
/*                 Institut f"ur Angewandte Mathematik                      */
/*                 Albert-Ludwigs-Universit"at Freiburg                     */
/*                 Hermann-Herder-Str. 10                                   */
/*                 79104 Freiburg                                           */
/*                 Germany                                                  */
/*                                                                          */
/*        email:   alfred,kunibert@mathematik.uni-freiburg.de               */
/*                                                                          */
/*     (c) by A. Schmidt and K.G. Siebert (1996-1999)                       */
/*                                                                          */
/****************************************************************************/

/****************************************************************************/
/*  compute the number of non zero entries for all basis functions in the   */
/*  matrices for the second order term                                      */
/****************************************************************************/

static int get_entries2(const BAS_FCTS *psi, const BAS_FCTS *phi, 
			const QUAD *quad)
{
  FUNCNAME("get_entries2");
  const QUAD_FAST *q_phi, *q_psi;
  int             i, j, k, l, iq, n_entries, n_psi, n_phi;
  REAL            val, grdi, grdj;

  n_psi = psi->n_bas_fcts;
  q_psi = get_quad_fast(psi, quad, INIT_GRD_PHI);
  n_phi = phi->n_bas_fcts;
  q_phi = get_quad_fast(phi, quad, INIT_GRD_PHI);

  for (n_entries = i = 0; i < n_psi; i++)
  {
    for (j = 0; j < n_phi; j++)
    {
      for (k = 0; k < DIM+1; k++)
      {
	for (l =  0; l < DIM+1; l++)
	{
	  for (val = iq = 0; iq < quad->n_points; iq++)
	  {
	    grdi = q_psi->grd_phi[iq][i][k];
	    grdj = q_phi->grd_phi[iq][j][l];

	    val += quad->w[iq]*grdi*grdj;
	  }
	  if (ABS(val) > TOO_SMALL)
	    n_entries++;
	}
      }
    }
  }

  return(n_entries);
}

static void init2(const BAS_FCTS *psi, const BAS_FCTS *phi, const QUAD *quad,
		  PSI_K_PHI_L **psi_k_phi_l, int n_entries, REAL *values,
		  int *k_vec, int *l_vec)
{
  FUNCNAME("init2");
  const QUAD_FAST *q_psi, *q_phi;
  int              i, j, k, l, iq, n_psi, n_phi , entries;
  REAL             grdi, grdj, val;

  TEST_EXIT(psi  &&  phi)("no basis functions passed over");

  n_psi = psi->n_bas_fcts;
  q_psi = get_quad_fast(psi, quad, INIT_GRD_PHI);
  n_phi = phi->n_bas_fcts;
  q_phi = get_quad_fast(phi, quad, INIT_GRD_PHI);

  for (i = 0; i < n_psi; i++)
  {
    for (j = 0; j < n_phi; j++)
    {
      psi_k_phi_l[i][j].values = values; 
      psi_k_phi_l[i][j].k = k_vec;
      psi_k_phi_l[i][j].l = l_vec;
      for (entries = k = 0; k < DIM+1; k++)
      {
	for (l =  0; l < DIM+1; l++)
	{
	  for (val = iq = 0; iq < quad->n_points; iq++)
	  {
	    grdi = q_psi->grd_phi[iq][i][k];
	    grdj = q_phi->grd_phi[iq][j][l];

	    val += quad->w[iq]*grdi*grdj;
	  }
	  if (ABS(val) > TOO_SMALL)
	  {
	    TEST_EXIT(n_entries-- > 0)
	      ("now more entries found than counted by `get_entries2'\n");
	    entries++;
	    *values++ = val;
	    *k_vec++ = k;
	    *l_vec++ = l;
	  }
	}
      }
      psi_k_phi_l[i][j].n_entries = entries;
    }
  }

  return;
}

/****************************************************************************/
/*  compute the number of non zero entries for all basis functions in the   */
/*  matrices for the first order term                                       */
/****************************************************************************/

static int get_entries1(const BAS_FCTS *psi, const BAS_FCTS *phi, 
			const QUAD *quad)
{
  FUNCNAME("get_entries1");
  const QUAD_FAST *q_psi, *q_phi;
  int              i, j, l, iq, n_psi, n_phi, n_entries;
  REAL             val, phii, grdj;

  n_psi = psi->n_bas_fcts;
  q_psi = get_quad_fast(psi, quad, INIT_PHI);
  n_phi = phi->n_bas_fcts;
  q_phi = get_quad_fast(phi, quad, INIT_GRD_PHI);

  for (n_entries = i = 0; i < n_psi; i++)
  {
    for (j = 0; j < n_phi; j++)
    {
      for (l = 0; l < DIM+1; l++)
      {
	for (val = iq = 0; iq < quad->n_points; iq++)
	{
	  phii = q_psi->phi[iq][i];
	  grdj = q_phi->grd_phi[iq][j][l];

	  val += quad->w[iq]*phii*grdj;
	}
	if (ABS(val) > TOO_SMALL)
	  n_entries++;
      }
    }
  }

  return(n_entries);
}

static void init1(const BAS_FCTS *psi, const BAS_FCTS *phi, const QUAD *quad,
		  PSI_PHI_L **psi_phi_l, int n_entries, REAL *values,
		  int *l_vec)
{
  FUNCNAME("init");
  const QUAD_FAST *q_psi, *q_phi;
  int              i, j, l, iq, n_psi, n_phi, entries;
  REAL             phii, grdj, val;
  
  TEST_EXIT(psi  &&  phi)("no basis functions passed over");
  MSG("%d entries for %s and %s\n", n_entries, NAME(psi), NAME(phi));

  n_psi = psi->n_bas_fcts;
  q_psi = get_quad_fast(psi, quad, INIT_PHI);
  n_phi = phi->n_bas_fcts;
  q_phi = get_quad_fast(phi, quad, INIT_GRD_PHI);

  for (i = 0; i < n_psi; i++)
  {
    for (j = 0; j < n_phi; j++)
    {
      psi_phi_l[i][j].values = values;
      psi_phi_l[i][j].l = l_vec;

      for (entries = l = 0; l < DIM+1; l++)
      {
	for (val = iq = 0; iq < quad->n_points; iq++)
	{
	  phii = q_psi->phi[iq][i];
	  grdj = q_phi->grd_phi[iq][j][l];

	  val += quad->w[iq]*phii*grdj;
	}
	if (ABS(val) > TOO_SMALL)
	{
	  TEST_EXIT(n_entries-- > 0)
	    ("now more entries found than counted by `get_entries1'\n");
	  entries++;
	  *values++ = val;
	  *l_vec++ = l;
	}
      }
      psi_phi_l[i][j].n_entries = entries;
    }
  }
  return;
}

static void init0(const BAS_FCTS *psi, const BAS_FCTS *phi, const QUAD *quad,
		  PSI_PHI **psi_phi)
{
  FUNCNAME("init0");
  const QUAD_FAST *q_psi, *q_phi;
  int              i, j, iq, n_psi, n_phi;
  REAL             val;

  TEST_EXIT(psi  &&  phi)("no basis functions passed over");

  n_psi = psi->n_bas_fcts;
  q_psi = get_quad_fast(psi, quad, INIT_PHI);
  n_phi = phi->n_bas_fcts;
  q_phi = get_quad_fast(phi, quad, INIT_PHI);

  for (i = 0; i < n_psi; i++)
  {
    for (j = 0; j < n_phi; j++)
    {
      for (val = iq = 0; iq < quad->n_points; iq++)
	val += quad->w[iq]*q_psi->phi[iq][i]*q_phi->phi[iq][j];

      if (ABS(val) < TOO_SMALL)
	psi_phi[i][j] = 0.0;
      else
	psi_phi[i][j] = val;
    }
  }

  return;
}

/****************************************************************************/
/*  initialize a new assemble info; look in the list of all infos wheter    */
/*  there are precalculated values matching the new demands                 */
/****************************************************************************/

const ASS_INFO *get_ass_info(const BAS_FCTS *psi, const BAS_FCTS *phi, 
			     FLAGS init_flag, const QUAD *quad[3])
{
  FUNCNAME("get_ass_info");
  static ASS_INFO  *first_ass_info = nil;
  ASS_INFO         *ass_info, *info;
  const QUAD       **ass_quad;
  size_t           size2, size1, size_p, size_r, size_i;
  void             **mem_p = nil;
  PSI_K_PHI_L      *mem2 = nil;
  PSI_PHI_L        *mem1 = nil;
  REAL             *mem_r = nil;
  int              *mem_i = nil;
  int              i, n_psi, n_phi, entries2 = 0, entries1 = 0;

  ass_info = MEM_CALLOC(1, ASS_INFO);

  ass_info->psi = psi;
  ass_info->phi = phi;
  ass_quad = ass_info->quad;

  n_psi = psi->n_bas_fcts;
  n_phi = phi->n_bas_fcts;

  size2 = size1 = size_p = size_r = size_i = 0;
      
  if (init_flag & INIT_PSI_K_PHI_L)
  {
    if (!quad[2])
      ass_quad[2] = get_quadrature(DIM, psi->degree+phi->degree-2);
    else
      ass_quad[2] = quad[2];

    for (info = first_ass_info; info; info = info->next)
    {
      if (info->psi == psi && info->phi == phi && info->quad[2] == ass_quad[2])
      {
	ass_info->psi_k_phi_l = info->psi_k_phi_l;
	ass_info->init_flag |= INIT_PSI_K_PHI_L;
	break;
      }
    }
    if (!info)
    {
      size_p += n_psi;
      size2 = n_psi*n_phi;
      entries2 = get_entries2(psi, phi, quad[2]);
      size_r += entries2;
      size_i += 2*entries2;
    }
  }

  if (init_flag & INIT_PSI_PHI_L)
  {
    if (!quad[1])
      ass_quad[1] = get_quadrature(DIM, psi->degree+phi->degree-1);
    else
      ass_quad[1] = quad[1];

    for (info = first_ass_info; info; info = info->next)
    {
      if (info->psi == psi && info->phi == phi && info->quad[1] == ass_quad[1])
      {
	ass_info->psi_phi_l = info->psi_phi_l;
	ass_info->init_flag |= INIT_PSI_PHI_L;
	break;
      }
    }
    if (!info)
    {
      size_p += n_psi;
      size1 = n_psi*n_phi;
      entries1 = get_entries1(psi, phi, quad[1]);
      size_r += entries1;
      size_i += entries1;
    }
  }

  if (init_flag & INIT_PSI_PHI)
  {
    if (!quad[0])
      ass_quad[0] = get_quadrature(DIM, psi->degree+phi->degree);
    else
      ass_quad[0] = quad[0];

    for (info = first_ass_info; info; info = info->next)
    {
      if (info->psi == psi && info->phi == phi && info->quad[0] == ass_quad[0])
      {
	ass_info->psi_phi = info->psi_phi;
	ass_info->init_flag |= INIT_PSI_PHI;
	break;
      }
    }
    if (!info)
    {
      size_p += n_psi;
      size_r += n_psi*n_phi;
    }
  }

/****************************************************************************/
/*  add ass_info to the linked list                                         */
/****************************************************************************/

  ass_info->next = first_ass_info;
  first_ass_info = ass_info;
  if (size_p == 0)
  {
    TEST_EXIT(ass_info->init_flag == init_flag)
      ("could not fill ass_info correctly\n");
    return((const ASS_INFO *) ass_info);
  }

  mem_p = MEM_ALLOC(size_p, void *);
  if (size2) 
    mem2 = MEM_ALLOC(size2, PSI_K_PHI_L);
  if (size1)
    mem1 = MEM_ALLOC(size1, PSI_PHI_L);
  if (size_r)
    mem_r = MEM_ALLOC(size_r, REAL); 
  if (size_i)
    mem_i = MEM_ALLOC(size_i, int);

  if (!ass_info->psi_k_phi_l  &&  (init_flag & INIT_PSI_K_PHI_L))
  {
    PSI_K_PHI_L  **psi_k_phi_l;
    REAL         *values;
    int          *k_vec, *l_vec;

    psi_k_phi_l = ass_info->psi_k_phi_l = (PSI_K_PHI_L **) mem_p;
    mem_p = (void **) (psi_k_phi_l+n_psi);

    for (i = 0; i < n_psi; i++)
    {
      psi_k_phi_l[i] = mem2+(i*n_phi);
    }
  
    values = (REAL *) mem_r;
    mem_r += entries2;
    
    k_vec = mem_i;
    l_vec = mem_i+entries2;
    mem_i += 2*entries2;

    init2(psi, phi, ass_quad[2], psi_k_phi_l, entries2, values, k_vec, l_vec);

    ass_info->init_flag |= INIT_PSI_K_PHI_L;
  }

  if (!ass_info->psi_phi_l  &&  (init_flag & INIT_PSI_PHI_L))
  {
    PSI_PHI_L   **psi_phi_l;
    REAL        *values;
    int         *l_vec;

    psi_phi_l = ass_info->psi_phi_l = (PSI_PHI_L **) mem_p;
    mem_p = (void **) (psi_phi_l + n_psi);

    for (i = 0; i < n_psi; i++)
      psi_phi_l[i] = mem1+(i*n_phi);

    values = mem_r;
    mem_r += entries1;
    l_vec = mem_i;
    mem_i += entries1;

    init1(psi, phi, ass_quad[1], psi_phi_l, entries1, values, l_vec);

    ass_info->init_flag |= INIT_PSI_PHI_L;
  }

  if (!ass_info->psi_phi  &&  (init_flag & INIT_PSI_PHI))
  {
    PSI_PHI   **psi_phi;

    psi_phi = ass_info->psi_phi = (PSI_PHI **) mem_p;

    for (i = 0; i < n_psi; i++)
      psi_phi[i] = mem_r+(i*n_phi);

    init0(psi, phi, ass_quad[0], psi_phi);

    ass_info->init_flag |= INIT_PSI_PHI;
  }

  TEST_EXIT((ass_info->init_flag & init_flag) == init_flag)
    ("could not initialize ass_info!\n");

  return((const ASS_INFO *) ass_info);
}


static void init_quad_fast(const BAS_FCTS *bas_fcts, FLAGS init_flag, 
			   const QUAD *quad[3], const QUAD_FAST *qf[3])
{
  FUNCNAME("init_quad");

  qf[0] = qf[1] = qf[2] = nil;

  if (init_flag & INIT_PSI_PHI_L)
    qf[1] = get_quad_fast(bas_fcts, quad[1], INIT_GRD_PHI|INIT_PHI);

  if (init_flag & INIT_PSI_K_PHI_L)
  {
    if (init_flag & INIT_PSI_PHI  &&  quad[2] == quad[0])
      qf[2] = qf[0] = get_quad_fast(bas_fcts, quad[2], INIT_GRD_PHI|INIT_PHI);
    else
      qf[2] = get_quad_fast(bas_fcts, quad[2], INIT_GRD_PHI);
  }

  if (init_flag & INIT_PSI_PHI  &&  !qf[0])
    qf[0] = get_quad_fast(bas_fcts, quad[0], INIT_PHI);

  return;
}

typedef struct fill_info  FILL_INFO;

struct fill_info
{
  const ASS_INFO   *ass_info;
  const QUAD_FAST  *quad_fast[3];

  int        n_bas_fcts;
  REAL       **element_stiff;

  void       (*init_element_pre)(const EL_INFO *);
  const REAL (*(*LALt_on_T)(void))[DIM+1];
  const REAL *(*Lb_on_T)(void);
  REAL       (*c_on_T)(void);

  void       (*init_element_quad)(const EL_INFO *, const QUAD_FAST *[3]);
  const REAL (*(*LALt_quad)(const QUAD_FAST *, int))[DIM+1];
  const REAL *(*Lb_quad)(const QUAD_FAST *, int);
  REAL       (*c_quad)(const QUAD_FAST *, int);
};

/****************************************************************************/
/*  functions for calculating element stiffness matrices using              */
/*  precalculated integrals of the basis functions on the reference element */
/****************************************************************************/

static const REAL **sym2_1_0_pre(const EL_INFO *el_info, const void *info)
{
  FUNCNAME("sym2_1_0_pre");
  int         i, j, m;
  FILL_INFO   *fill_info;
  REAL        **mat;

  fill_info = (FILL_INFO *) info;
  mat = fill_info->element_stiff;

  for (i = 0; i < fill_info->n_bas_fcts; i++)
    for (j = 0; j < fill_info->n_bas_fcts; j++)
      mat[i][j] = 0;

  if (fill_info->init_element_pre)  fill_info->init_element_pre(el_info);

  if (fill_info->LALt_on_T)
  {
    const REAL   (*LALt)[DIM+1] = fill_info->LALt_on_T();
    PSI_K_PHI_L  **psi_k_phi_l = fill_info->ass_info->psi_k_phi_l;
    REAL         val;
    int          *k, *l;

    for (i = 0; i < fill_info->n_bas_fcts; i++)
    {
      k = psi_k_phi_l[i][i].k;
      l = psi_k_phi_l[i][i].l;
      for (val = m = 0; m < psi_k_phi_l[i][i].n_entries; m++)
	val += psi_k_phi_l[i][i].values[m]*LALt[k[m]][l[m]];
      mat[i][i] += val;
      for (j = i+1; j < fill_info->n_bas_fcts; j++)
      {
	k = psi_k_phi_l[i][j].k;
	l = psi_k_phi_l[i][j].l;
	for (val = m = 0; m < psi_k_phi_l[i][j].n_entries; m++)
	  val += psi_k_phi_l[i][j].values[m]*LALt[k[m]][l[m]];
	mat[i][j] += val;
	mat[j][i] += val;
      }
    }
  }

  if (fill_info->Lb_on_T)
  {
    const REAL  *Lb = fill_info->Lb_on_T();
    PSI_PHI_L   **psi_phi_l = fill_info->ass_info->psi_phi_l;
    REAL        val;
    int         *l;

    for (i = 0; i < fill_info->n_bas_fcts; i++)
    {
      for (j = 0; j < fill_info->n_bas_fcts; j++)
      {
	l = psi_phi_l[i][j].l;
	for (val = m = 0; m < psi_phi_l[i][j].n_entries; m++)
	  val += psi_phi_l[i][j].values[m]*Lb[l[m]];
	mat[i][j] += val;
      }
    }
  }

  if (fill_info->c_on_T)
  {
    REAL     val, c = fill_info->c_on_T();
    PSI_PHI  **psi_phi = fill_info->ass_info->psi_phi;

    for (i = 0; i < fill_info->n_bas_fcts; i++)
    {
      mat[i][i] += c*psi_phi[i][i];
      for (j = i+1; j < fill_info->n_bas_fcts; j++)
      {
	val = c*psi_phi[i][j];
	mat[i][j] += val;
	mat[j][i] += val;
      }
    }
  }

  return((const REAL **) mat);
}

static const REAL **nonsym2_1_0_pre(const EL_INFO *el_info, const void *info)
{
  FUNCNAME("nonsym2_1_0_pre");
  int         i, j, m;
  FILL_INFO   *fill_info;
  REAL        **mat;

  fill_info = (FILL_INFO *) info;
  mat = fill_info->element_stiff;

  for (i = 0; i < fill_info->n_bas_fcts; i++)
    for (j = 0; j < fill_info->n_bas_fcts; j++)
      mat[i][j] = 0;

  if (fill_info->init_element_pre)  fill_info->init_element_pre(el_info);

  if (fill_info->LALt_on_T)
  {
    const REAL  (*LALt)[DIM+1] = fill_info->LALt_on_T();
    REAL        val;
    PSI_K_PHI_L     **psi_k_phi_l = fill_info->ass_info->psi_k_phi_l;
    int         *k, *l;

    for (i = 0; i < fill_info->n_bas_fcts; i++)
    {
      for (j = 0; j < fill_info->n_bas_fcts; j++)
      {
	k = psi_k_phi_l[i][j].k;
	l = psi_k_phi_l[i][j].l;
	for (val = m = 0; m < psi_k_phi_l[i][j].n_entries; m++)
	  val += psi_k_phi_l[i][j].values[m]*LALt[k[m]][l[m]];
	mat[i][j] += val;
      }
    }
  }

  if (fill_info->Lb_on_T)
  {
    const REAL *Lb = fill_info->Lb_on_T();
    PSI_PHI_L    **psi_phi_l = fill_info->ass_info->psi_phi_l;
    REAL       val;
    int        *l;

    for (i = 0; i < fill_info->n_bas_fcts; i++)
    {
      for (j = 0; j < fill_info->n_bas_fcts; j++)
      {
	l = psi_phi_l[i][j].l;
	for (val = m = 0; m < psi_phi_l[i][j].n_entries; m++)
	  val += psi_phi_l[i][j].values[m]*Lb[l[m]];
	mat[i][j] += val;
      }
    }
  }

  if (fill_info->c_on_T)
  {
    REAL     val, c = fill_info->c_on_T();
    PSI_PHI  **psi_phi = fill_info->ass_info->psi_phi;

    for (i = 0; i < fill_info->n_bas_fcts; i++)
    {
      mat[i][i] += c*psi_phi[i][i];
      for (j = i+1; j < fill_info->n_bas_fcts; j++)
      {
	val = c*psi_phi[i][j];
	mat[i][j] += val;
	mat[j][i] += val;
      }
    }
  }

  return((const REAL **) mat);
}

/****************************************************************************/
/*  functions for calculating element stiffness matrices by quadrature      */
/****************************************************************************/

static const REAL **quad_210(const EL_INFO *el_info, const void *info)
{
  FUNCNAME("quad_210");
  const REAL       (*LALt)[DIM+1], *Lb;
  REAL             c, val, (*grd_phi)[DIM+1], *phi;
  int              iq, i, j, k, l, n_bas_fcts;
  FILL_INFO        *fill_info;
  const QUAD_FAST  *q_fast;
  REAL             **mat;

  const REAL (*(*LALt_quad)(const QUAD_FAST *, int))[DIM+1];
  const REAL *(*Lb_quad)(const QUAD_FAST *, int);
  REAL       (*c_quad)(const QUAD_FAST *, int);

  fill_info = (FILL_INFO *) info;
  mat = fill_info->element_stiff;
  LALt_quad = fill_info->LALt_quad;
  Lb_quad   = fill_info->Lb_quad;
  c_quad    = fill_info->c_quad;

  q_fast = fill_info->quad_fast[2];
  n_bas_fcts = q_fast->n_bas_fcts;

  for (i = 0; i < n_bas_fcts; i++)
    for (j = 0; j < n_bas_fcts; j++)
      mat[i][j] = 0.0;

  if (fill_info->init_element_quad)
    fill_info->init_element_quad(el_info, fill_info->quad_fast);

  for (iq = 0; iq < q_fast->n_points; iq++)
  {
    LALt = LALt_quad(q_fast, iq);
    Lb   = Lb_quad(q_fast, iq);
    c    = c_quad(q_fast, iq);

    grd_phi = q_fast->grd_phi[iq];
    phi     = q_fast->phi[iq];

    for (i = 0; i < n_bas_fcts; i++)
    {
      for (j = 0; j < n_bas_fcts; j++)
      {
	val = 0.0;
	for (k = 0; k <= DIM; k++)
	{
	  for (l = 0; l <= DIM; l++)
	    val += LALt[k][l]*grd_phi[i][k]*grd_phi[j][l];
	  val += Lb[k]*phi[i]*grd_phi[j][k];
	}
	mat[i][j] += q_fast->w[iq]*(val + c*phi[i]*phi[j]);
      }
    }
  }

  return((const REAL **) mat);
}

static const REAL **quad_21(const EL_INFO *el_info, const void *info)
{
  FUNCNAME("quad_21");
  const REAL       (*LALt)[DIM+1], *Lb;
  REAL             val, (*grd_phi)[DIM+1], *phi;
  int              iq, i, j, k, l, n_bas_fcts;
  FILL_INFO        *fill_info;
  const QUAD_FAST  *q_fast;
  REAL             **mat;

  const REAL (*(*LALt_quad)(const QUAD_FAST *, int))[DIM+1];
  const REAL *(*Lb_quad)(const QUAD_FAST *, int);

  fill_info = (FILL_INFO *) info;
  mat = fill_info->element_stiff;
  LALt_quad = fill_info->LALt_quad;
  Lb_quad   = fill_info->Lb_quad;

  q_fast = fill_info->quad_fast[2];
  n_bas_fcts = q_fast->n_bas_fcts;

  for (i = 0; i < n_bas_fcts; i++)
    for (j = 0; j < n_bas_fcts; j++)
      mat[i][j] = 0.0;

  if (fill_info->init_element_quad)
    fill_info->init_element_quad(el_info, fill_info->quad_fast);

  for (iq = 0; iq < q_fast->n_points; iq++)
  {
    LALt = LALt_quad(q_fast, iq);
    Lb   = Lb_quad(q_fast, iq);

    grd_phi = q_fast->grd_phi[iq];
    phi     = q_fast->phi[iq];

    for (i = 0; i < n_bas_fcts; i++)
    {
      for (j = 0; j < n_bas_fcts; j++)
      {
	val = 0.0;
	for (k = 0; k <= DIM; k++)
	{
	  for (l = 0; l <= DIM; l++)
	    val += LALt[k][l]*grd_phi[i][k]*grd_phi[j][l];
	  val += Lb[k]*phi[i]*grd_phi[j][k];
	}
	mat[i][j] += q_fast->w[iq]*val;
      }
    }
  }

  return((const REAL **) mat);
}

static const REAL **quad_20(const EL_INFO *el_info, const void *info)
{
  FUNCNAME("quad_20");
  const REAL       (*LALt)[DIM+1];
  REAL             c, val, (*grd_phi)[DIM+1], *phi;
  int              iq, i, j, k, l, n_bas_fcts;
  FILL_INFO        *fill_info;
  const QUAD_FAST  *q_fast;
  REAL             **mat;

  const REAL (*(*LALt_quad)(const QUAD_FAST *, int))[DIM+1];
  REAL       (*c_quad)(const QUAD_FAST *, int);

  fill_info = (FILL_INFO *) info;
  mat = fill_info->element_stiff;
  LALt_quad = fill_info->LALt_quad;
  c_quad    = fill_info->c_quad;

  q_fast = fill_info->quad_fast[2];
  n_bas_fcts = q_fast->n_bas_fcts;

  for (i = 0; i < n_bas_fcts; i++)
    for (j = 0; j < n_bas_fcts; j++)
      mat[i][j] = 0.0;

  if (fill_info->init_element_quad)
    fill_info->init_element_quad(el_info, fill_info->quad_fast);

  for (iq = 0; iq < q_fast->n_points; iq++)
  {
    LALt = LALt_quad(q_fast, iq);
    c    = c_quad(q_fast, iq);

    grd_phi = q_fast->grd_phi[iq];
    phi     = q_fast->phi[iq];

    for (i = 0; i < n_bas_fcts; i++)
    {
      for (j = 0; j < n_bas_fcts; j++)
      {
	for (val = k = 0; k <= DIM; k++)
	  for (l = 0; l <= DIM; l++)
	    val += LALt[k][l]*grd_phi[i][k]*grd_phi[j][l];
	mat[i][j] += q_fast->w[iq]*(val + c*phi[i]*phi[j]);
      }
    }
  }

  return((const REAL **) mat);
}

static const REAL **quad_sym20(const EL_INFO *el_info, const void *info)
{
  FUNCNAME("quad_sym20");
  const REAL       (*LALt)[DIM+1];
  REAL             c, val, (*grd_phi)[DIM+1], *phi;
  int              iq, i, j, k, l, n_bas_fcts;
  FILL_INFO        *fill_info;
  const QUAD_FAST  *q_fast;
  REAL             **mat;

  const REAL (*(*LALt_quad)(const QUAD_FAST *, int))[DIM+1];
  REAL       (*c_quad)(const QUAD_FAST *, int);

  fill_info = (FILL_INFO *) info;
  mat = fill_info->element_stiff;
  LALt_quad = fill_info->LALt_quad;
  c_quad    = fill_info->c_quad;

  q_fast = fill_info->quad_fast[2];
  n_bas_fcts = q_fast->n_bas_fcts;

  for (i = 0; i < n_bas_fcts; i++)
    for (j = i; j < n_bas_fcts; j++)
      mat[i][j] = 0.0;

  if (fill_info->init_element_quad)
    fill_info->init_element_quad(el_info, fill_info->quad_fast);

  for (iq = 0; iq < q_fast->n_points; iq++)
  {
    LALt = LALt_quad(q_fast, iq);
    c    = c_quad(q_fast, iq);

    grd_phi = q_fast->grd_phi[iq];
    phi     = q_fast->phi[iq];

    for (i = 0; i < n_bas_fcts; i++)
    {
      for (j = i; j < n_bas_fcts; j++)
      {
	for (val = k = 0; k <= DIM; k++)
	  for (l = 0; l <= DIM; l++)
	    val += LALt[k][l]*grd_phi[i][k]*grd_phi[j][l];
	mat[i][j] += q_fast->w[iq]*(val + c*phi[i]*phi[j]);
      }
    }
  }

  for (i = 0; i < n_bas_fcts; i++)
    for (j = 0; j < i; j++)
      mat[i][j] = mat[j][i];

  return((const REAL **) mat);
}

static const REAL **quad_10(const EL_INFO *el_info, const void *info)
{
  FUNCNAME("quad_10");
  const REAL       *Lb;
  REAL             c, val, (*grd_phi)[DIM+1], *phi;
  int              iq, i, j, l, n_bas_fcts;
  FILL_INFO        *fill_info;
  const QUAD_FAST  *q_fast;
  REAL             **mat;

  const REAL *(*Lb_quad)(const QUAD_FAST *, int);
  REAL       (*c_quad)(const QUAD_FAST *, int);

  fill_info = (FILL_INFO *) info;
  mat = fill_info->element_stiff;
  Lb_quad   = fill_info->Lb_quad;
  c_quad    = fill_info->c_quad;

  q_fast = fill_info->quad_fast[1];
  n_bas_fcts = q_fast->n_bas_fcts;

  for (i = 0; i < n_bas_fcts; i++)
    for (j = 0; j < n_bas_fcts; j++)
      mat[i][j] = 0.0;

  if (fill_info->init_element_quad)
    fill_info->init_element_quad(el_info, fill_info->quad_fast);

  for (iq = 0; iq < q_fast->n_points; iq++)
  {
    Lb   = Lb_quad(q_fast, iq);
    c    = c_quad(q_fast, iq);

    grd_phi = q_fast->grd_phi[iq];
    phi     = q_fast->phi[iq];

    for (i = 0; i < n_bas_fcts; i++)
    {
      for (j = 0; j < n_bas_fcts; j++)
      {
	for (val = l = 0; l <= DIM; l++)
	  val += Lb[l]*phi[i]*grd_phi[j][l];
	mat[i][j] += q_fast->w[iq]*(val + c*phi[i]*phi[j]);
      }
    }
  }

  return((const REAL **) mat);
}

static const REAL **quad_2(const EL_INFO *el_info, const void *info)
{
  FUNCNAME("quad_2");
  const REAL       (*LALt)[DIM+1];
  REAL             val, (*grd_phi)[DIM+1];
  int              iq, i, j, k, l, n_bas_fcts;
  FILL_INFO        *fill_info;
  const QUAD_FAST  *q_fast;
  REAL             **mat;

  const REAL (*(*LALt_quad)(const QUAD_FAST *, int))[DIM+1];

  fill_info = (FILL_INFO *) info;

  n_bas_fcts = fill_info->n_bas_fcts;
  mat        = fill_info->element_stiff;
  LALt_quad  = fill_info->LALt_quad;
  q_fast     = fill_info->quad_fast[2];

  for (i = 0; i < n_bas_fcts; i++)
    for (j = 0; j < n_bas_fcts; j++)
      mat[i][j] = 0.0;

  if (fill_info->init_element_quad)
    fill_info->init_element_quad(el_info, fill_info->quad_fast);


  for (iq = 0; iq < q_fast->n_points; iq++)
  {
    LALt = LALt_quad(q_fast, iq);

    grd_phi = q_fast->grd_phi[iq];
    for (i = 0; i < n_bas_fcts; i++)
    {
      for (j = 0; j < n_bas_fcts; j++)
      {
	for (val = k = 0; k <= DIM; k++)
	  for (l = 0; l <= DIM; l++)
	    val += LALt[k][l]*grd_phi[i][k]*grd_phi[j][l];
	mat[i][j] += q_fast->w[iq]*val;
      }
    }
  }

  return((const REAL **) mat);
}

static const REAL **quad_sym2(const EL_INFO *el_info, const void *info)
{
  FUNCNAME("quad_sym2");
  const REAL       (*LALt)[DIM+1];
  REAL             val, (*grd_phi)[DIM+1];
  int              iq, i, j, k, l, n_bas_fcts;
  FILL_INFO        *fill_info;
  const QUAD_FAST  *q_fast;
  REAL             **mat;

  const REAL (*(*LALt_quad)(const QUAD_FAST *, int))[DIM+1];

  fill_info = (FILL_INFO *) info;
  n_bas_fcts = fill_info->n_bas_fcts;
  mat = fill_info->element_stiff;
  LALt_quad = fill_info->LALt_quad;
  q_fast = fill_info->quad_fast[2];

  for (i = 0; i < n_bas_fcts; i++)
    for (j = i; j < n_bas_fcts; j++)
      mat[i][j] = 0.0;

  if (fill_info->init_element_quad)
    fill_info->init_element_quad(el_info, fill_info->quad_fast);

  for (iq = 0; iq < q_fast->n_points; iq++)
  {
    LALt = LALt_quad(q_fast, iq);

    grd_phi = q_fast->grd_phi[iq];
    for (i = 0; i < n_bas_fcts; i++)
    {
      for (j = i; j < n_bas_fcts; j++)
      {
	for (val = k = 0; k <= DIM; k++)
	  for (l = 0; l <= DIM; l++)
	    val += LALt[k][l]*grd_phi[i][k]*grd_phi[j][l];
	mat[i][j] += q_fast->w[iq]*val;
      }
    }
  }

  for (i = 0; i < n_bas_fcts; i++)
    for (j = 0; j < i; j++)
      mat[i][j] = mat[j][i];

  return((const REAL **) mat);
}

static const REAL **quad_1(const EL_INFO *el_info, const void *info)
{
  FUNCNAME("quad_1");
  const REAL       *Lb;
  REAL             val, (*grd_phi)[DIM+1], *phi;
  int              iq, i, j, l, n_bas_fcts;
  FILL_INFO        *fill_info;
  const QUAD_FAST  *q_fast;
  REAL             **mat;

  const REAL *(*Lb_quad)(const QUAD_FAST *, int);

  fill_info = (FILL_INFO *) info;
  mat = fill_info->element_stiff;
  Lb_quad   = fill_info->Lb_quad;

  q_fast = fill_info->quad_fast[1];
  n_bas_fcts = q_fast->n_bas_fcts;

  for (i = 0; i < n_bas_fcts; i++)
    for (j = 0; j < n_bas_fcts; j++)
      mat[i][j] = 0.0;

  if (fill_info->init_element_quad)
    fill_info->init_element_quad(el_info, fill_info->quad_fast);

  for (iq = 0; iq < q_fast->n_points; iq++)
  {
    Lb   = Lb_quad(q_fast, iq);

    grd_phi = q_fast->grd_phi[iq];
    phi     = q_fast->phi[iq];

    for (i = 0; i < n_bas_fcts; i++)
    {
      for (j = 0; j < n_bas_fcts; j++)
      {
	for (val = l = 0; l <= DIM; l++)
	  val += Lb[l]*phi[i]*grd_phi[j][l];
	mat[i][j] += q_fast->w[iq]*val;
      }
    }
  }

  return((const REAL **) mat);
}

static const REAL **quad_0(const EL_INFO *el_info, const void *info)
{
  FUNCNAME("quad_0");
  REAL             c, val, *phi;
  int              iq, i, j, n_bas_fcts;
  FILL_INFO        *fill_info;
  const QUAD_FAST  *q_fast;
  REAL             **mat;

  REAL       (*c_quad)(const QUAD_FAST *, int);

  fill_info = (FILL_INFO *) info;
  n_bas_fcts = fill_info->n_bas_fcts;
  mat = fill_info->element_stiff;

  for (i = 0; i < n_bas_fcts; i++)
    for (j = 0; j < n_bas_fcts; j++)
      mat[i][j] = 0.0;

  if (fill_info->init_element_quad)
    fill_info->init_element_quad(el_info, fill_info->quad_fast);

  c_quad = fill_info->c_quad;
  q_fast = fill_info->quad_fast[0];

  for (iq = 0; iq < q_fast->n_points; iq++)
  {
    c   = c_quad(q_fast, iq);
    phi = q_fast->phi[iq];

    for (i = 0; i < n_bas_fcts; i++)
    {
      mat[i][i] += q_fast->w[iq]*c*phi[i]*phi[i];
      for (j = i+1; j < n_bas_fcts; j++)
      {
	val = q_fast->w[iq]*c*phi[i]*phi[j];
	mat[i][j] += val;
	mat[j][i] += val;
      }
    }
  }
  return((const REAL **) mat);
}

static const REAL **quad_2_1_0(const EL_INFO *el_info, const void *info)
{
  FUNCNAME("quad_2_1_0");
  const REAL       (*LALt)[DIM+1], *Lb;
  REAL             c, val, (*grd_phi)[DIM+1], *phi;
  int              iq, i, j, k, l, n_bas_fcts;
  FILL_INFO        *fill_info;
  const QUAD_FAST  *q_fast;
  REAL             **mat;

  const REAL (*(*LALt_quad)(const QUAD_FAST *, int))[DIM+1];
  const REAL *(*Lb_quad)(const QUAD_FAST *, int);
  REAL       (*c_quad)(const QUAD_FAST *, int);

  fill_info = (FILL_INFO *) info;
  n_bas_fcts = fill_info->n_bas_fcts;
  mat = fill_info->element_stiff;

  for (i = 0; i < n_bas_fcts; i++)
    for (j = 0; j < n_bas_fcts; j++)
      mat[i][j] = 0.0;

  if (fill_info->init_element_quad)
    fill_info->init_element_quad(el_info, fill_info->quad_fast);

  if ((LALt_quad = fill_info->LALt_quad))
  {
    q_fast = fill_info->quad_fast[2];

    for (iq = 0; iq < q_fast->n_points; iq++)
    {
      LALt = LALt_quad(q_fast, iq);

      grd_phi = q_fast->grd_phi[iq];
      for (i = 0; i < n_bas_fcts; i++)
      {
	for (j = 0; j < n_bas_fcts; j++)
	{
	  for (val = k = 0; k <= DIM; k++)
	    for (l = 0; l <= DIM; l++)
	      val += LALt[k][l]*grd_phi[i][k]*grd_phi[j][l];
	  mat[i][j] += q_fast->w[iq]*val;
	}
      }
    }
  }

  if ((Lb_quad = fill_info->Lb_quad))
  {
    q_fast = fill_info->quad_fast[1];

    for (iq = 0; iq < q_fast->n_points; iq++)
    {
      Lb = Lb_quad(q_fast, iq);

      grd_phi = q_fast->grd_phi[iq];
      phi     = q_fast->phi[iq];

      for (i = 0; i < n_bas_fcts; i++)
      {
	for (j = 0; j < n_bas_fcts; j++)
	{
	  for (val = l = 0; l <= DIM; l++)
	    val += Lb[l]*phi[i]*grd_phi[j][l];
	  mat[i][j] += q_fast->w[iq]*val;
	}
      }
    }
  }

  if ((c_quad = fill_info->c_quad))
  {
    q_fast = fill_info->quad_fast[0];

    for (iq = 0; iq < q_fast->n_points; iq++)
    {
      c   = c_quad(q_fast, iq);
      phi = q_fast->phi[iq];

      for (i = 0; i < n_bas_fcts; i++)
      {
	mat[i][i] += q_fast->w[iq]*c*phi[i]*phi[i];
	for (j = i+1; j < n_bas_fcts; j++)
	{
	  val = q_fast->w[iq]*c*phi[i]*phi[j];
	  mat[i][j] += val;
	  mat[j][i] += val;
	}
      }
    }
  }

  return((const REAL **) mat);
}

static const REAL **quad_sym2_1_0(const EL_INFO *el_info, const void *info)
{
  FUNCNAME("quad_sym2_1_0");
  const REAL       (*LALt)[DIM+1], *Lb;
  REAL             c, val, (*grd_phi)[DIM+1], *phi;
  int              iq, i, j, k, l, n_bas_fcts;
  FILL_INFO        *fill_info;
  const QUAD_FAST  *q_fast;
  REAL             **mat;

  const REAL (*(*LALt_quad)(const QUAD_FAST *, int))[DIM+1];
  const REAL *(*Lb_quad)(const QUAD_FAST *, int);
  REAL       (*c_quad)(const QUAD_FAST *, int);

  fill_info = (FILL_INFO *) info;
  n_bas_fcts = fill_info->n_bas_fcts;
  mat = fill_info->element_stiff;

  for (i = 0; i < n_bas_fcts; i++)
    for (j = i; j < n_bas_fcts; j++)
      mat[i][j] = 0.0;

  if (fill_info->init_element_quad)
    fill_info->init_element_quad(el_info, fill_info->quad_fast);

  if ((LALt_quad = fill_info->LALt_quad))
  {
    q_fast = fill_info->quad_fast[2];

    for (iq = 0; iq < q_fast->n_points; iq++)
    {
      LALt = LALt_quad(q_fast, iq);

      grd_phi = q_fast->grd_phi[iq];
      for (i = 0; i < n_bas_fcts; i++)
      {
	for (j = i; j < n_bas_fcts; j++)
	{
	  for (val = k = 0; k <= DIM; k++)
	    for (l = 0; l <= DIM; l++)
	      val += LALt[k][l]*grd_phi[i][k]*grd_phi[j][l];
	  mat[i][j] += q_fast->w[iq]*val;
	}
      }
    }
  }

  for (i = 0; i < n_bas_fcts; i++)
    for (j = 0; j < i; j++)
      mat[i][j] = mat[j][i];

  if ((Lb_quad = fill_info->Lb_quad))
  {
    q_fast = fill_info->quad_fast[1];

    for (iq = 0; iq < q_fast->n_points; iq++)
    {
      Lb = Lb_quad(q_fast, iq);

      grd_phi = q_fast->grd_phi[iq];
      phi     = q_fast->phi[iq];

      for (i = 0; i < n_bas_fcts; i++)
      {
	for (j = 0; j < n_bas_fcts; j++)
	{
	  for (val = l = 0; l <= DIM; l++)
	    val += Lb[l]*phi[i]*grd_phi[j][l];
	  mat[i][j] += q_fast->w[iq]*val;
	}
      }
    }
  }

  if ((c_quad = fill_info->c_quad))
  {
    q_fast = fill_info->quad_fast[0];

    for (iq = 0; iq < q_fast->n_points; iq++)
    {
      c   = c_quad(q_fast, iq);
      phi = q_fast->phi[iq];

      for (i = 0; i < n_bas_fcts; i++)
      {
	mat[i][i] += q_fast->w[iq]*c*phi[i]*phi[i];
	for (j = i+1; j < n_bas_fcts; j++)
	{
	  val = q_fast->w[iq]*c*phi[i]*phi[j];
	  mat[i][j] += val;
	  mat[j][i] += val;
	}
      }
    }
  }

  return((const REAL **) mat);
}


const EL_STIFF_INFO *el_stiff_pre(const BAS_FCTS *bas_fcts, 
				  const QUAD *_quad[3], 
				  FLAGS fill_flag, int sym_2,
				  void (*init_element)(const EL_INFO *),
				  REAL (*c_on_T)(void), 
				  const REAL  *(*Lb_on_T)(void),
				  const REAL(*(*LALt_on_T)(void))[DIM+1])
{
  FUNCNAME("el_stiff_pre");
  FLAGS            flag = 0;
  FILL_INFO        *fill_info;
  const ASS_INFO   *ass_info;
  EL_STIFF_INFO    *el_stiff_info;
  const QUAD       *_quad_[3] = {nil, nil, nil}, **quad;
  int              n_bas_fcts;

  quad = _quad ? _quad : _quad_;

  if (!bas_fcts)
  {
    ERROR("`bas_fcts' nil; can not initialize el_stiff_fct;\n");
    ERROR("returning nil\n");
    return(nil);
  }
  if (!c_on_T   &&  !Lb_on_T  &&  !LALt_on_T)
  {
    ERROR("no function for 2., 1., or 0. order term;\n");
    ERROR("can not initialize el_stiff_fct; returning nil\n");
    return(nil);
  }

  fill_info = MEM_CALLOC(1, FILL_INFO);
  n_bas_fcts = fill_info->n_bas_fcts = bas_fcts->n_bas_fcts;
  fill_info->element_stiff = MAT_ALLOC(n_bas_fcts, n_bas_fcts, REAL);

  flag = INIT_NONE;
  if (c_on_T)    flag |= INIT_PSI_PHI;
  if (Lb_on_T)   flag |= INIT_PSI_PHI_L;
  if (LALt_on_T) flag |= INIT_PSI_K_PHI_L;

  if (c_on_T     &&  !quad[0])
  {
    quad[0] = get_quadrature(DIM, 2*bas_fcts->degree);
    INFO(2,2)("using quadrature of degree %d for 0. order term\n",
	      2*bas_fcts->degree);
  }
  if (Lb_on_T    &&  !quad[1])
  {
    quad[1] = get_quadrature(DIM, 2*bas_fcts->degree-1);
    INFO(2,2)("using quadrature of degree %d for 1. order term\n",
	      2*bas_fcts->degree-1);
  }

  if (LALt_on_T  &&  !quad[2]) 
  {
    quad[2] = get_quadrature(DIM, 2*bas_fcts->degree-2);
    INFO(2,2)("using quadrature of degree %d for 2. order term\n",
	      2*bas_fcts->degree-2);
  }

  ass_info = get_ass_info(bas_fcts, bas_fcts, flag, quad);
  fill_info->ass_info = ass_info;

  fill_info->init_element_pre = init_element;
  if (c_on_T)    fill_info->c_on_T    = c_on_T;
  if (Lb_on_T)   fill_info->Lb_on_T   = Lb_on_T;
  if (LALt_on_T) fill_info->LALt_on_T = LALt_on_T;
  
  el_stiff_info = MEM_CALLOC(1, EL_STIFF_INFO);

  if (sym_2)
    el_stiff_info->el_stiff_fct = sym2_1_0_pre;
  else
    el_stiff_info->el_stiff_fct = nonsym2_1_0_pre;

  el_stiff_info->fill_info = fill_info;
  el_stiff_info->fill_flag = FILL_COORDS | fill_flag;

  return((const EL_STIFF_INFO *) el_stiff_info);
}


const EL_STIFF_INFO *el_stiff_quad(const BAS_FCTS *bas_fcts, 
				   const QUAD *_quad[3], 
				   FLAGS fill_flag, int sym_2,
				   void (*init_element)(const EL_INFO *,
							const QUAD_FAST *[3]),
				   REAL (*c)(const QUAD_FAST *, int),
				   const REAL *(*Lb)(const QUAD_FAST *, int),
				   const REAL (*(*LALt)(const QUAD_FAST *, 
							int))[DIM+1])
{
  FUNCNAME("el_stiff_quad");
  FLAGS           flag = 0UL;
  FILL_INFO       *fill_info;
  int             same, n_bas_fcts;
  EL_STIFF_INFO   *el_stiff_info;
  const QUAD       *_quad_[3] = {nil, nil, nil}, **quad;
  
  quad = _quad ? _quad : _quad_;


  if (!bas_fcts)
  {
    ERROR("`bas_fcts' nil; can not initialize el_stiff_fct;\n");
    ERROR("returning nil\n");
    return(nil);
  }
  if (!c  &&  !Lb  &&  !LALt)
  {
    ERROR("no function for 2., 1., or 0. order term;\n");
    ERROR("can not initialize el_stiff_fct; returning nil\n");
    return(nil);
  }

  fill_info = MEM_CALLOC(1, FILL_INFO);
  n_bas_fcts = fill_info->n_bas_fcts = bas_fcts->n_bas_fcts;
  fill_info->element_stiff = MAT_ALLOC(n_bas_fcts, n_bas_fcts, REAL);

  flag = INIT_NONE;
  if (c)    flag |= INIT_PSI_PHI;
  if (Lb)   flag |= INIT_PSI_PHI_L;
  if (LALt) flag |= INIT_PSI_K_PHI_L;


  if (LALt  &&  !quad[2]) 
  {
    quad[2] = get_quadrature(DIM, 2*bas_fcts->degree-2);
    INFO(2,2)("using quadrature of degree %d for 2. order term\n",
	      2*bas_fcts->degree-2);
  }

  if (Lb    &&  !quad[1])
  {
    if (quad[2]) 
    {
      quad[1] = quad[2];
      INFO(2,2)("using for 1. order term quadrature of 2. order term\n");
    }
    else
    {
      quad[1] = get_quadrature(DIM, 2*bas_fcts->degree-1);
      INFO(2,2)("using quadrature of degree %d for 1. order term\n",
		2*bas_fcts->degree-1);
    }
  }

  if (c     &&  !quad[0])
  {
    if (quad[2]) 
    {
      quad[0] = quad[2];
      INFO(2,2)("using for 0. order term quadrature of 2. order term\n");
    }
    else if (quad[1])
    {
      quad[0] = quad[1];
      INFO(2,2)("using for 0. order term quadrature of 1. order term\n");
    }
    else
    {
      quad[0] = get_quadrature(DIM, 2*bas_fcts->degree);
      INFO(2,2)("using quadrature of degree %d for 0. order term\n",
		2*bas_fcts->degree);
    }
  }

  fill_info->n_bas_fcts = bas_fcts->n_bas_fcts;
  init_quad_fast(bas_fcts, flag, quad, fill_info->quad_fast); 

  fill_info->init_element_quad = init_element;
  fill_info->c_quad    = c;
  fill_info->Lb_quad   = Lb;
  fill_info->LALt_quad = LALt;

  el_stiff_info = MEM_CALLOC(1, EL_STIFF_INFO);

  same = true;
  if ((flag & INIT_PSI_K_PHI_L)  &&  (flag & INIT_PSI_PHI_L))
    if (quad[2] != quad[1])  same = false;
  if ((flag & INIT_PSI_K_PHI_L)  &&  (flag & INIT_PSI_PHI))
    if (quad[2] != quad[0])  same = false;
  if ((flag & INIT_PSI_PHI_L)  &&  (flag & INIT_PSI_PHI))
    if (quad[1] != quad[0])  same = false;

  if (same)
  {
    switch(flag)
    {
    case (INIT_PSI_K_PHI_L|INIT_PSI_PHI_L|INIT_PSI_PHI):
      el_stiff_info->el_stiff_fct = quad_210;
      break;
    case (INIT_PSI_K_PHI_L|INIT_PSI_PHI_L):
      el_stiff_info->el_stiff_fct = quad_21;
      break;
    case (INIT_PSI_K_PHI_L|INIT_PSI_PHI):
      if (sym_2)
	el_stiff_info->el_stiff_fct = quad_sym20;
      else
	el_stiff_info->el_stiff_fct = quad_20;
      break;
    case (INIT_PSI_PHI_L|INIT_PSI_PHI):
      el_stiff_info->el_stiff_fct = quad_10;
      break;
    case (INIT_PSI_K_PHI_L):
      if (sym_2)
	el_stiff_info->el_stiff_fct = quad_sym2;
      else
	el_stiff_info->el_stiff_fct = quad_2;
      break;
    case (INIT_PSI_PHI_L):
      el_stiff_info->el_stiff_fct = quad_1;
      break;
    case (INIT_PSI_PHI):
      el_stiff_info->el_stiff_fct = quad_0;
      break;
    default:
      ERROR_EXIT("no el_stiff_fct available for flag %d\n", flag);
    }
  }
  else
  {
    if (sym_2)
      el_stiff_info->el_stiff_fct = quad_sym2_1_0;
    else
      el_stiff_info->el_stiff_fct = quad_2_1_0;
  }

  el_stiff_info->fill_info = fill_info;
  el_stiff_info->fill_flag = FILL_COORDS | fill_flag;

  return((const EL_STIFF_INFO *) el_stiff_info);
}

/****************************************************************************/
/****************************************************************************/
/****************************************************************************/
#if DIM == 2
#define N_FACES N_EDGES
#endif

const FACE_QUAD_FAST *get_face_quad_fast(const BAS_FCTS *bas_fcts, 
					 const QUAD *quad, U_CHAR init_flag)
{
  FUNCNAME("get_face_quad_fast");
  static FACE_QUAD_FAST  *first = nil;
  FACE_QUAD_FAST   *quad_fast;
  BAS_FCTS         *bas_fct;
  int              d, i, j, k, l, n_points, n_bas_fcts;
  size_t           size, size_p;
  REAL             **g_phi;
  double           *lambda;
  void             *mem = nil, *mem_p = nil, *mem_last = nil;

  if (quad->dim != DIM-1)
  {
    ERROR("FACE_QUAD_FAST not for %d != DIM-1\n");
    return(nil);
  }

  if (init_flag != INIT_PHI)
  {
    ERROR("FACE_QUAD_FAST not for init_flag != INIT_PHI\n");
    return(nil);
  }

  for (quad_fast = first; quad_fast; quad_fast = quad_fast->next)
  {
    if (quad_fast->quad == quad  &&  quad_fast->bas_fcts == bas_fcts)
      break;
  }

  if (quad_fast  &&  ((quad_fast->init_flag & init_flag) == init_flag))
    return((const FACE_QUAD_FAST *) quad_fast);

  if (!quad_fast)
  {
    quad_fast = MEM_ALLOC(1, FACE_QUAD_FAST);
    quad_fast->quad = quad;
    quad_fast->n_points = quad->n_points;
    quad_fast->n_bas_fcts = bas_fcts->n_bas_fcts;
    quad_fast->w = quad->w;
    quad_fast->init_flag = 0;
    for (k = 0; k < N_FACES; k++)
    {
      quad_fast->lambda[k] = nil;
      quad_fast->phi[k] = nil;
    }
    quad_fast->next = first;
    quad_fast->bas_fcts = (BAS_FCTS *) bas_fcts;
    first = quad_fast;
  }

  n_points = quad->n_points;
  n_bas_fcts = bas_fcts->n_bas_fcts;

  size = size_p = 0;
  if (!quad_fast->lambda[0])
    size += N_FACES*quad_fast->n_points*sizeof(double [DIM+1]);
  /*  for barycentric coordinates  */

  if (!quad_fast->phi[0]  &&  (init_flag & INIT_PHI))
  {
    size +=  N_FACES*n_points*n_bas_fcts*sizeof(REAL);
    size_p += N_FACES*n_points*sizeof(REAL *);
  }

  if (size)
  {
    mem = albert_alloc(size, funcName, __FILE__,  __LINE__);
    mem_last = (void *) (((char *) mem)+size);
    mem_p = albert_alloc(size_p, funcName, __FILE__,  __LINE__);

    if (!quad_fast->lambda[0])
    {
      for (k = 0; k < N_FACES; k++)
      {
	quad_fast->lambda[k] = mem;
	mem = (void *) (quad_fast->lambda[k]+n_points);
      }
    }

    if (!quad_fast->phi[0]  &&  (init_flag & INIT_PHI))
    {
      BAS_FCT    **phi = bas_fcts->phi;
      
      for (k = 0; k < N_FACES; k++)
      {
	g_phi = quad_fast->phi[k] = (REAL **) mem_p;
	mem_p = (void *) (g_phi+n_points);
	for (i = 0; i < n_points; i++)
	{
	  lambda = (double *) quad_fast->lambda[k][i];
	  for (d = 0; d < k; d++)
	    lambda[d] = quad->lambda[i][d];
	  lambda[k] = 0.0;
	  for (d = k+1; d <= DIM; d++)
	    lambda[d] = quad->lambda[i][d-1];
	  
	  g_phi[i] = (REAL *) mem;
	  mem = (void *) (g_phi[i]+n_bas_fcts);
	  for (j = 0; j < n_bas_fcts; j++)
	  {
	    g_phi[i][j] = phi[j](lambda);
	  }
	}
      }
      quad_fast->init_flag |= INIT_PHI;
    }

#if 0
    if (!quad_fast->grd_phi  && (init_flag & INIT_GRD_PHI))
    {
      GRD_BAS_FCT  **grd_phi = bas_fcts->grd_phi;
      const REAL   *grd;

      g_grd_phi = quad_fast->grd_phi  = (REAL (**)[DIM+1]) mem_p;
      mem_p = (void *) (quad_fast->grd_phi+n_points);

      for (i = 0; i < n_points; i++)
      { 
	lambda = quad->lambda[i];
	g_grd_phi[i] = (REAL (*)[DIM+1]) mem;
	mem = (void *) (g_grd_phi[i]+n_bas_fcts);

	for (j = 0; j < n_bas_fcts; j++)
	{
	  grd = grd_phi[j](lambda);
	  for (k = 0; k < DIM+1; k++)
	  {
	    g_grd_phi[i][j][k] = grd[k];
	  }	    
	}
      }
      quad_fast->init_flag |= INIT_GRD_PHI;
    }

    if (!quad_fast->D2_phi  && (init_flag & INIT_D2_PHI))
    {
      D2_BAS_FCT   **D2_phi = bas_fcts->D2_phi;
      const REAL   (*D2)[DIM+1];

      g_D2_phi = quad_fast->D2_phi  = (REAL (**)[DIM+1][DIM+1]) mem_p;
      mem_p = (void *) (quad_fast->D2_phi+n_points);

      for (i = 0; i < n_points; i++)
      {
	lambda = quad->lambda[i];
	g_D2_phi[i] = (REAL (*)[DIM+1][DIM+1]) mem;
	mem = (void *) (g_D2_phi[i]+n_bas_fcts);
	for (j = 0; j < n_bas_fcts; j++)
	{
	  D2 = D2_phi[j](lambda);
	  for (k = 0; k < DIM+1; k++)
	    for (l = 0; l < DIM+1; l++)
	      g_D2_phi[i][j][k][l] = D2[k][l];
	}
      }
      quad_fast->init_flag |= INIT_D2_PHI;
    }
#endif
  }

  if (mem > mem_last)
    ERROR_EXIT("allocated size too small\n");
  if (mem < mem_last)
    ERROR_EXIT("allocated size too big\n");

  if ((quad_fast->init_flag & init_flag) != init_flag)
  {
    ERROR("could not initialize quad_fast, returning pointer to nil\n");
    return(nil);
  }

  return((const FACE_QUAD_FAST *) quad_fast);
}


static void face_init0(const BAS_FCTS *psi, const BAS_FCTS *phi,
		       const QUAD *quad, PSI_PHI **psi_phi[N_FACES])
{
  FUNCNAME("face_init0");
  const FACE_QUAD_FAST *q_psi, *q_phi;
  int              i, j, k, iq, n_psi, n_phi;
  REAL             val;

  TEST_EXIT(psi  &&  phi)("no basis functions passed over");

  n_psi = psi->n_bas_fcts;
  q_psi = get_face_quad_fast(psi, quad, INIT_PHI);
  n_phi = phi->n_bas_fcts;
  q_phi = get_face_quad_fast(phi, quad, INIT_PHI);

  for (k = 0; k < N_FACES; k++)
  {
    for (i = 0; i < n_psi; i++)
    {
      for (j = 0; j < n_phi; j++)
      {
	for (val = iq = 0; iq < quad->n_points; iq++)
	  val += quad->w[iq]*q_psi->phi[k][iq][i]*q_phi->phi[k][iq][j];

      if (ABS(val) < TOO_SMALL)
	psi_phi[k][i][j] = 0.0;
      else
	psi_phi[k][i][j] = val;
      }
    }
  }

  return;
}

const FACE_ASS_INFO *get_face_ass_info(const BAS_FCTS *psi, 
				       const BAS_FCTS *phi, FLAGS init_flag,
				       const QUAD *quad[3])
{
  FUNCNAME("get_face_ass_info");
  static FACE_ASS_INFO  *first_ass_info = nil;
  FACE_ASS_INFO   *ass_info, *info;
  const QUAD       **ass_quad;
  size_t           size2, size1, size_p, size_r, size_i;
  void             **mem_p = nil;
  REAL             *mem_r = nil;
  int              *mem_i = nil;
  int              i, k, n_psi, n_phi;

  ass_info = MEM_CALLOC(1, FACE_ASS_INFO);

  ass_info->psi = psi;
  ass_info->phi = phi;
  ass_quad = ass_info->quad;

  n_psi = psi->n_bas_fcts;
  n_phi = phi->n_bas_fcts;

  size_p = size_r = 0;

  if (init_flag & INIT_PSI_PHI)
  {
    if (!quad[0])
      ass_quad[0] = get_quadrature(DIM-1, psi->degree+phi->degree-2);
    else
      ass_quad[0] = quad[0];

    for (info = first_ass_info; info; info = info->next)
    {
      if (info->psi == psi && info->phi == phi && info->quad[0] == ass_quad[0])
      {
	for (k = 0; k < N_FACES; k++)
	  ass_info->psi_phi[k] = info->psi_phi[k];
	ass_info->init_flag |= INIT_PSI_PHI;
	break;
      }
    }
    if (!info)
    {
      size_p += n_psi*N_FACES;
      size_r += n_psi*n_phi*N_FACES;
    }
  }

/****************************************************************************/
/*  add ass_info to the linked list                                         */
/****************************************************************************/

  ass_info->next = first_ass_info;
  first_ass_info = ass_info;

  if (size_p == 0)
  {
    TEST_EXIT(ass_info->init_flag == init_flag)
      ("could not fill ass_info correctly\n");
    return((const FACE_ASS_INFO *) ass_info);
  }

  mem_p = MEM_ALLOC(size_p, void *);
  if (size_r)
    mem_r = MEM_ALLOC(size_r, REAL); 

  if (!ass_info->psi_phi[0]  &&  (init_flag & INIT_PSI_PHI))
  {
    PSI_PHI   **psi_phi;

    for (k = 0; k < N_FACES; k++)
    {
      psi_phi = ass_info->psi_phi[k] = (PSI_PHI **) mem_p;
      mem_p = (void *) (psi_phi+n_psi);

      for (i = 0; i < n_psi; i++)
      {
	psi_phi[i] = mem_r;
	mem_r = psi_phi[i]+n_phi;
      }
    }
    face_init0(psi, phi, ass_quad[0], ass_info->psi_phi);

    ass_info->init_flag |= INIT_PSI_PHI;
  }

  TEST_EXIT((ass_info->init_flag & init_flag) == init_flag)
    ("could not initialize ass_info!\n");

  return((const FACE_ASS_INFO *) ass_info);
}

#if DIM == 2
#undef N_FACES
#endif

#if 0
void print_int_psi_phi(const BAS_FCTS *psi, const BAS_FCTS *phi,
		       PSI_PHI **psi_phi)
{
  FUNCNAME("print_int_psi_phi");
  int      i, j;
  REAL     sum;

  for (sum = i = 0; i < psi->n_bas_fcts; i++)
  {
    MSG("%2d: ", i);
    for (j = 0; j < phi->n_bas_fcts; j++)
    {
      sum +=  psi_phi[i][j];
      print_msg("%12.5le%s", psi_phi[i][j],
		j < phi->n_bas_fcts-1 ? ", " : "\n");
    }
  }
  if (ABS(sum - 0.5) > 1.e-15)
  {
    MSG("sum != 0.5, sum = %e\n", sum);
    WAIT;
  }
  MSG("sum = %le\n", sum);
  return;
}

void print_face_ass_info(const BAS_FCTS *psi, const BAS_FCTS *phi)
{
  FUNCNAME("print_face_ass_info");
  const QUAD          *quad[3] = {nil, nil, nil};
  const FACE_ASS_INFO *info;
  int                 k;

  quad[0] = get_quadrature(DIM-1, psi->degree+phi->degree);
  info = get_face_ass_info(psi, phi, INIT_PSI_PHI, quad);

  MSG("face_ass_info for %s and %s\n", NAME(psi), NAME(phi));

  for (k = 0; k < N_FACES; k++)
  {
    MSG("face %d\n", k);
    print_int_psi_phi(psi, phi, info->psi_phi[k]);
  }

  return;
}

void print_ass_info(const BAS_FCTS *psi, const BAS_FCTS *phi)
{
  FUNCNAME("print_ass_info");
  const QUAD          *quad[3] = {nil, nil, nil};
  const ASS_INFO *info;

  quad[0] = get_quadrature(DIM, psi->degree+phi->degree);
  info = get_ass_info(psi, phi, INIT_PSI_PHI, quad);

  MSG("ass_info for %s and %s\n", NAME(psi), NAME(phi));
  print_int_psi_phi(psi, phi, info->psi_phi);

  return;
}
#endif
  
