/****************************************************************************/
/* ALBERT:   an Adaptive multi Level finite element toolbox using           */
/*           Bisectioning refinement and Error control by Residual          */
/*           Techniques                                                     */
/*                                                                          */
/* file:     trav_xy.c                                                      */
/*                                                                          */
/*                                                                          */
/* description: Find element at position (xy) and local coordinates         */
/*              return 0 if xy \not\in domain.                              */
/*                                                                          */
/* history:                                                                 */
/*                                                                          */
/*                                                                          */
/****************************************************************************/
/*                                                                          */
/*        authors: Alfred Schmidt and Kunibert G. Siebert                   */
/*                 Institut f"ur Angewandte Mathematik                      */
/*                 Albert-Ludwigs-Universit"at Freiburg                     */
/*                 Hermann-Herder-Str. 10                                   */
/*                 79104 Freiburg                                           */
/*                 Germany                                                  */
/*                                                                          */
/*        email:   alfred,kunibert@mathematik.uni-freiburg.de               */
/*                                                                          */
/*     (c) by A. Schmidt and K.G. Siebert (1996,97,98)                      */
/*                                                                          */
/****************************************************************************/
#include "albert.h"

/****************************************************************************/
/*  world_to_coord():  return -1 if inside, otherwise index of lambda<0     */
/****************************************************************************/

int world_to_coord(const EL_INFO *el_info, const REAL *xy, REAL lambda[DIM+1])
{
  FUNCNAME("world_to_coord");
  REAL  edge[DIM][DIM_OF_WORLD], x[DIM_OF_WORLD];
  REAL  x0, det, det0, det1, lmin;
  int   i, j, k;
#if DIM == 3
  REAL  det2;
#endif

#if (DIM != DIM_OF_WORLD)
  ERROR_EXIT("not yet for DIM != DIM_OF_WORLD\n");
#endif


#if DIM==2
/*  wir haben das gleichungssystem zu loesen: */
/*       ( q1x q2x )  (lambda1)     (qx)      */
/*       ( q1y q2y )  (lambda2)  =  (qy)      */
/*      mit qi=pi-p3, q=xy-p3                 */

  for (j=0; j<DIM_OF_WORLD; j++) {
    x0 = el_info->coord[DIM][j];
    x[j] = xy[j] - x0;
    for (i=0; i<DIM; i++)
      edge[i][j] = el_info->coord[i][j] - x0;
  }

  det  = edge[0][0] * edge[1][1] - edge[0][1] * edge[1][0]; 
  det0 =       x[0] * edge[1][1] -       x[1] * edge[1][0]; 
  det1 = edge[0][0] * x[1]       - edge[0][1] * x[0]; 

  if (ABS(det) < 1.E-20) {
    ERROR("det = %le; abort\n", det);
    for (i=0; i<=DIM; i++) lambda[i] = 1.0/DIM;
    return(-1);
  }

  lambda[0] = det0 / det;
  lambda[1] = det1 / det;
  lambda[2] = 1.0 - lambda[0] - lambda[1];
#else
/*  wir haben das gleichungssystem zu loesen: */
/*       ( q1x q2x q3x)  (lambda1)     (qx)      */
/*       ( q1y q2y q3y)  (lambda2)  =  (qy)      */
/*       ( q1z q2z q3z)  (lambda3)     (qz)      */
/*      mit qi=pi-p3, q=xy-p3                 */

  for (j=0; j<DIM_OF_WORLD; j++) {
    x0 = el_info->coord[DIM][j];
    x[j] = xy[j] - x0;
    for (i=0; i<DIM; i++)
      edge[i][j] = el_info->coord[i][j] - x0;
  }

  det =  edge[0][0] * edge[1][1] * edge[2][2]
       + edge[0][1] * edge[1][2] * edge[2][0]
       + edge[0][2] * edge[1][0] * edge[2][1]
       - edge[0][2] * edge[1][1] * edge[2][0]
       - edge[0][0] * edge[1][2] * edge[2][1]
       - edge[0][1] * edge[1][0] * edge[2][2];
  det0 =       x[0] * edge[1][1] * edge[2][2]
       +       x[1] * edge[1][2] * edge[2][0]
       +       x[2] * edge[1][0] * edge[2][1]
       -       x[2] * edge[1][1] * edge[2][0]
       -       x[0] * edge[1][2] * edge[2][1]
       -       x[1] * edge[1][0] * edge[2][2];
  det1 = edge[0][0] *       x[1] * edge[2][2]
       + edge[0][1] *       x[2] * edge[2][0]
       + edge[0][2] *       x[0] * edge[2][1]
       - edge[0][2] *       x[1] * edge[2][0]
       - edge[0][0] *       x[2] * edge[2][1]
       - edge[0][1] *       x[0] * edge[2][2];
  det2 = edge[0][0] * edge[1][1] *       x[2]
       + edge[0][1] * edge[1][2] *       x[0]
       + edge[0][2] * edge[1][0] *       x[1]
       - edge[0][2] * edge[1][1] *       x[0]
       - edge[0][0] * edge[1][2] *       x[1]
       - edge[0][1] * edge[1][0] *       x[2];

  if (ABS(det) < 1.E-20) {
    ERROR("det = %le; abort\n", det);
    for (i=0; i<=DIM; i++) lambda[i] = 1.0/DIM;
    return(-1);
  }

  lambda[0] = det0 / det;
  lambda[1] = det1 / det;
  lambda[2] = det2 / det;
  lambda[3] = 1.0 - lambda[0] - lambda[1] - lambda[2];
#endif

  k = -1;
  lmin = 0.0;
  j = 0;
  for (i = 0; i <= DIM; i++) {
    if (lambda[i] < -1.E-5) {
      if (lambda[i] < lmin) {
	k = i;
	lmin = lambda[i];
      }
      j++;
    }
  }

  return(k);
}

/****************************************************************************/


void fill_elinfo(int ichild, const EL_INFO *elinfo_old, EL_INFO *elinfo);

static const REAL *g_xy0, *g_xy;
static REAL       *g_sp;
static EL     *final_el;
static REAL   final_lambda[DIM+1] = {0.0};

static int find_el_at_pt_recursive(EL_INFO *el_info, REAL lambda[DIM+1],
				   int outside)
{
  FUNCNAME("find_el_at_pt_recursive");
  EL      *el = el_info->el;
  EL_INFO c_el_info[1];
  REAL    c_lambda[DIM+1];
  int     i, inside;
#if DIM == 3
  int     ichild, c_outside;
#endif

#if (DIM==3) && 0
  if (outside >= 0) {
    MSG("el %2d: leaf=%d, outside=%d, lambda=(%.2lf %.2lf %.2lf %.2lf)\n",
	INDEX(el), IS_LEAF_EL(el), outside,
	lambda[0],lambda[1],lambda[2],lambda[3]);
  }
  else {
    MSG("el %2d: leaf=%d, inside, lambda=(%.2lf %.2lf %.2lf %.2lf)\n",
	INDEX(el), IS_LEAF_EL(el), lambda[0],lambda[1],lambda[2],lambda[3]);
  }
#endif

  if (IS_LEAF_EL(el)) {
    final_el = el;
    if (outside < 0) {
      for (i=0; i<=DIM; i++)  final_lambda[i] = lambda[i];

#if 0
      for (i=0; i<=DIM; i++)  {
	if (final_lambda[i] < -1.0E-5)
#if DIM==2
	  ERROR("outside element: %.5lf %.5lf %.5lf\n",
		lambda[0],lambda[1],lambda[2]);
#else
	  ERROR("outside element: %.5lf %.5lf %.5lf %.5lf\n",
		lambda[0],lambda[1],lambda[2],lambda[3]);
#endif
      }
#endif

#if 0
      {
	REAL_D world;
	REAL   s;

	coord_to_world(el_info, lambda, world);
	s = SQR(world[0]-g_xy[0]) + SQR(world[1]-g_xy[1]);
#if DIM==3
	s += SQR(world[2]-g_xy[2]);
#endif
	if (s > 1.E-10) {
	  MSG("el %d:\n", INDEX(el));
#if DIM==3
	  MSG("? xy=(%.2le, %.2le, %.2le), world=(%.2le, %.2le, %.2le)\n",
	      g_xy[0], g_xy[1], g_xy[2], world[0], world[1], world[2]);
#else
	  MSG("? xy=(%.2le, %.2le), world=(%.2le, %.2le)\n",
	      g_xy[0], g_xy[1], world[0], world[1]);
#endif
	  WAIT_REALLY;
	}
      }
#endif

      return(true);
    }
    else {  /* outside */
      if (g_xy0) {             /* find boundary point of [xy0, xy] */
	REAL s;
	world_to_coord(el_info, g_xy0, c_lambda);
	s = lambda[outside] / (lambda[outside] - c_lambda[outside]);
	for (i=0; i<=DIM; i++) {
	  final_lambda[i] = s * c_lambda[i] + (1.0-s) * lambda[i];
	}
	if (g_sp) *g_sp = s;
#if DIM==3 && 1
	MSG("outside finest level on el %d: s=%.3le\n", INDEX(el), s);
#endif
	return(false);  /* ??? */
      }
      else {
	return(false);
      }
    }
  }

#if DIM==2
  if (lambda[0] >= lambda[1]) {
    fill_elinfo(0, el_info, c_el_info);
    if (el->new_coord) {
      outside = world_to_coord(c_el_info, g_xy, c_lambda);
      if (outside >= 0) {
	ERROR("outside curved boundary child 0\n");
      }
    }
    else {
      c_lambda[0] = lambda[2];
      c_lambda[1] = lambda[0] - lambda[1];
      c_lambda[2] = 2.0 * lambda[1];
    }
  }
  else {
    fill_elinfo(1, el_info, c_el_info);
    if (el->new_coord) {
      outside = world_to_coord(c_el_info, g_xy, c_lambda);
      if (outside >= 0) {
	ERROR("outside curved boundary child 1\n");
      }
    }
    else {
      c_lambda[0] = lambda[1] - lambda[0];
      c_lambda[1] = lambda[2];
      c_lambda[2] = 2.0 * lambda[0];
    }
  }

#else   /* DIM==3 */

  if (el->new_coord) {
    if (lambda[0] >= lambda[1])
      ichild = 0;
    else
      ichild = 1;
    fill_elinfo(ichild, el_info, c_el_info);
    c_outside = world_to_coord(c_el_info, g_xy, c_lambda);

    if (c_outside>=0) {  /* test is other child is better... */
      REAL c_lambda2[DIM+1];
      int  c_outside2;
      EL_INFO c_el_info2[1];

      fill_elinfo(1-ichild, el_info, c_el_info2);
      c_outside2 = world_to_coord(c_el_info2, g_xy, c_lambda2);

      MSG("new_coord CHILD %d: outside=%d, lambda=(%.2lf %.2lf %.2lf %.2lf)\n",
	  ichild, c_outside, c_lambda[0],c_lambda[1],c_lambda[2],c_lambda[3]);
      MSG("new_coord CHILD %d: outside=%d, lambda=(%.2lf %.2lf %.2lf %.2lf)\n",
    1-ichild, c_outside2, c_lambda2[0],c_lambda2[1],c_lambda2[2],c_lambda2[3]);

/*       if (c_lambda[0] < c_lambda2[0]) { */
      if ((c_outside2 < 0) || (c_lambda2[c_outside2] > c_lambda[c_outside])) {
	for (i=0; i<=DIM; i++) c_lambda[i] = c_lambda2[i];
	c_outside = c_outside2;
	*c_el_info = *c_el_info2;
	ichild = 1 - ichild;
      }
    }
    outside = c_outside;

#if 0
    MSG("new_coord child %d: el %d: outside=%2d, lambda=(%.2lf %.2lf %.2lf %.2lf)\n",
	ichild, INDEX(el->child[0]), outside, c_lambda[0],c_lambda[1],c_lambda[2],c_lambda[3]);
    {
      REAL_D world;
      coord_to_world(c_el_info, c_lambda, world);
      MSG("xy=(%.2le, %.2le, %.2le), world=(%.2le, %.2le, %.2le)\n",
	  g_xy[0], g_xy[1], g_xy[2], world[0], world[1], world[2]);
    }
#endif

    if (outside >= 0) {
/* 	ERROR("outside curved boundary child 0\n"); */
    }
  }
  else {  /* no new_coord */
    if (lambda[0] >= lambda[1]) {
      fill_elinfo(0, el_info, c_el_info);
      MSG("transform lambda child 0\n");
      c_lambda[0] = lambda[0] - lambda[1];
      c_lambda[1] = lambda[child_vertex[el_info->el_type][0][1]];
      c_lambda[2] = lambda[child_vertex[el_info->el_type][0][2]];
      c_lambda[3] = 2.0 * lambda[1];
    }
    else {
      fill_elinfo(1, el_info, c_el_info);
      MSG("transform lambda child 1\n");
      c_lambda[0] = lambda[1] - lambda[0];
      c_lambda[1] = lambda[child_vertex[el_info->el_type][1][1]];
      c_lambda[2] = lambda[child_vertex[el_info->el_type][1][2]];
      c_lambda[3] = 2.0 * lambda[0];
    }
  }

/****************************************************************************/
#if 0
  if (lambda[0] >= lambda[1]) {
    fill_elinfo(0, el_info, c_el_info);
    if (el->new_coord) {
      outside = world_to_coord(c_el_info, g_xy, c_lambda);
#if 1
      MSG("new_coord child 0: el %d: outside=%2d, lambda=(%.2lf %.2lf %.2lf %.2lf)\n",
	  INDEX(el->child[0]), outside, c_lambda[0],c_lambda[1],c_lambda[2],c_lambda[3]);
#endif
      if (outside >= 0) {
/* 	ERROR("outside curved boundary child 0\n"); */
      }
    }
    else {
      MSG("transform lambda child 0\n");
      c_lambda[0] = lambda[0] - lambda[1];
      c_lambda[1] = lambda[child_vertex[el_info->el_type][0][1]];
      c_lambda[2] = lambda[child_vertex[el_info->el_type][0][2]];
      c_lambda[3] = 2.0 * lambda[1];
    }
  }
  else {
    fill_elinfo(1, el_info, c_el_info);
    if (el->new_coord) {
      outside = world_to_coord(c_el_info, g_xy, c_lambda);
#if 1
      MSG("new_coord child 1: el %d: outside=%2d, lambda=(%.2lf %.2lf %.2lf %.2lf)\n",
	  INDEX(el->child[1]), outside, c_lambda[0],c_lambda[1],c_lambda[2],c_lambda[3]);
      {
	REAL_D world;
	coord_to_world(c_el_info, c_lambda, world);
	MSG("xy=(%.2le, %.2le, %.2le), world=(%.2le, %.2le, %.2le)\n",
	    g_xy[0], g_xy[1], g_xy[2], world[0], world[1], world[2]);
      }
#endif
      if (outside >= 0) {
/* 	ERROR("outside curved boundary child 1\n"); */
      }
    }
    else {
      MSG("transform lambda child 1\n");
      c_lambda[0] = lambda[1] - lambda[0];
      c_lambda[1] = lambda[child_vertex[el_info->el_type][1][1]];
      c_lambda[2] = lambda[child_vertex[el_info->el_type][1][2]];
      c_lambda[3] = 2.0 * lambda[0];
    }
  }

  {
    REAL_D world;
    coord_to_world(c_el_info, c_lambda, world);
    MSG("xy=(%.2le, %.2le, %.2le), world=(%.2le, %.2le, %.2le)\n",
	g_xy[0], g_xy[1], g_xy[2], world[0], world[1], world[2]);
  }

#endif
/****************************************************************************/


#endif  /* DIM */



  inside = find_el_at_pt_recursive(c_el_info, c_lambda, outside);
  return(inside);
}
  
/****************************************************************************/

int find_el_at_pt(const MESH *mesh, const REAL_D xy,
		  EL **elp, REAL bary[DIM+1],
		  const MACRO_EL *start_mel, const REAL_D xy0, REAL *sp)
{
  const MACRO_EL  *mel;
  REAL      lambda[DIM+1];
  EL_INFO   mel_info[1];
  int       i, k, inside;

  if (start_mel)
    mel = start_mel;
  else
    mel = mesh->first_macro_el;

  mel_info->fill_flag = FILL_COORDS;
  g_xy  = xy;
  g_xy0 = xy0;
  g_sp  = sp;

  fill_macro_info(mel, mel_info);


  while ((k = world_to_coord(mel_info, xy, lambda)) >= 0) {
    if (mel->neigh[k]) {
      mel = mel->neigh[k];
      fill_macro_info(mel, mel_info);
      continue;
    }

#if 0
    if (xy0) {             /* find boundary point of [xy0, xy] */
      if (mel_info->el->new_coord) {
	MSG("may produce incorrect results near curved boundary\n");
      }

      world_to_coord(mel_info, xy0, bary);
      s = lambda[k] / (lambda[k] - bary[k]);
      for (i=0; i<=DIM; i++) {
	lambda[i] = s * bary[i] + (1.0-s) * lambda[i];
      }
      if (sp) *sp = s;
    }
    else {
      return(false);
    }
#endif

    break;
  }

#if 0
  if (k >= 0) {
    MSG("outside=%d on macro level.\n", k);
  }
  else {
    MSG("inside on macro level.\n", k);
  }
#endif
  /* now, descend in tree to find leaf element at point */

  inside = find_el_at_pt_recursive(mel_info, lambda, k);
  for (i=0; i<=DIM; i++) bary[i] = final_lambda[i];
  *elp = final_el;
  
/*   if ((k>=0) || (!inside)) WAIT; */

  return(inside);
}

/****************************************************************************/
