/****************************************************************************/
/* ALBERT:   an Adaptive multi Level finite element toolbox using           */
/*           Bisectioning refinement and Error control by Residual          */
/*           Techniques                                                     */
/*                                                                          */
/* file:                                                                    */
/*                                                                          */
/*                                                                          */
/* description:                                                             */
/*                                                                          */
/*                                                                          */
/* history:                                                                 */
/*                                                                          */
/*                                                                          */
/****************************************************************************/
/*                                                                          */
/*        authors: Alfred Schmidt and Kunibert G. Siebert                   */
/*                 Institut f"ur Angewandte Mathematik                      */
/*                 Albert-Ludwigs-Universit"at Freiburg                     */
/*                 Hermann-Herder-Str. 10                                   */
/*                 79104 Freiburg                                           */
/*                 Germany                                                  */
/*                                                                          */
/*        email:   alfred,kunibert@mathematik.uni-freiburg.de               */
/*                                                                          */
/*     (c) by A. Schmidt and K.G. Siebert (1996,97)                         */
/*                                                                          */
/****************************************************************************/
/****************************************************************************/
/* common (static) traversal routines for 2d and 3d                         */
/* to be #included in traverse.c                                            */
/****************************************************************************/
/* to_do:                                                                   */
/*   traverse_every_el_inorder()                              		    */
/*                                                      		    */
/****************************************************************************/

/* traverse hierarchical mesh,  call el_fct() for each/some element */

typedef struct traverse_info TRAVERSE_INFO;

struct traverse_info {
  MESH     *mesh;
  FLAGS    flag; 
  U_CHAR   level;
  void     (*el_fct)(EL_INFO *);
};


static void recursive_traverse(EL_INFO *, TRAVERSE_INFO *);
static EL_INFO   *traverse_every_el_preorder(void);
static EL_INFO   *traverse_every_el_inorder(void);
static EL_INFO   *traverse_every_el_postorder(void);
static EL_INFO   *traverse_leaf_el(void);
static EL_INFO   *traverse_leaf_el_level(void);
static EL_INFO   *traverse_el_level(void);
static EL_INFO   *traverse_mg_level(void);

/****************************************************************************/
/*   mesh_traverse:                                            		    */
/*   --------------                                     		    */
/*   run through all macro elements and start mesh traversal      	    */
/*   call recursive_traverse() for each macro element's element		    */
/****************************************************************************/

void mesh_traverse(MESH *mesh, int level,
		   FLAGS flag, void (*el_fct)())
{
  FUNCNAME("mesh_traverse");
  MACRO_EL *mel;
  EL_INFO elinfo;
  TRAVERSE_INFO traverse_info;

  if (mesh==nil) return;

  elinfo.fill_flag = (flag & FILL_ANY);

  traverse_info.mesh   = mesh;
  traverse_info.level  = level;
  traverse_info.el_fct = el_fct;
  traverse_info.flag = flag;
  elinfo.mesh = mesh;

  if (flag & (CALL_LEAF_EL_LEVEL | CALL_EL_LEVEL | CALL_MG_LEVEL)) {
    TEST(level >= 0)("invalid level: %d\n",level);
  }

  for (mel = mesh->first_macro_el; mel; mel = mel->next) {
    fill_macro_info(mel, &elinfo);
    recursive_traverse(&elinfo, &traverse_info);
  }

}

void par_mesh_traverse(MESH *mesh, int level, FLAGS flag, void (*el_fct)())
{
  FUNCNAME("par_mesh_traverse");
  MACRO_EL        *mel;
  static MACRO_EL **mel_vec = nil;
  static int      size_mel_vec = 0;
  int             i;
  EL_INFO         elinfo;
  TRAVERSE_INFO   traverse_info;

  if (mesh==nil) return;

  elinfo.fill_flag = (flag & FILL_ANY);

  traverse_info.mesh   = mesh;
  traverse_info.level  = level;
  traverse_info.el_fct = el_fct;
  traverse_info.flag = flag;
  elinfo.mesh = mesh;

  if (flag & (CALL_LEAF_EL_LEVEL | CALL_EL_LEVEL | CALL_MG_LEVEL)) {
    TEST(level >= 0)("invalid level: %d\n",level);
  }

  if (size_mel_vec < mesh->n_macro_el)
  {
    mel_vec = MEM_REALLOC(mel_vec, size_mel_vec, mesh->n_macro_el, MACRO_EL *);
    size_mel_vec = mesh->n_macro_el;
  }

  i = 0;
  for (mel = mesh->first_macro_el; mel; mel = mel->next)
    mel_vec[i++] = mel;
  TEST_EXIT(i == mesh->n_macro_el)("i = %d, but mesh->n_macro_el = %d\n", 
				   i, mesh->n_macro_el);

  for (i = 0; i < mesh->n_macro_el; i++)
  {
    fill_macro_info(mel_vec[i], &elinfo);
    recursive_traverse(&elinfo, &traverse_info);
  }
}

/****************************************************************************/
/*   recursive_traverse:                                		    */
/*   -------------------                                		    */
/*   recursively traverse the mesh hierarchy tree       		    */
/*   call the routine traverse_info->el_fct(el_info) with    		    */
/*    - all tree leaves                                 		    */
/*    - all tree leaves at a special level              		    */
/*    - all tree elements in pre-/in-/post-order        		    */
/*   depending on the traverse_info->level variable     		    */
/****************************************************************************/

static void recursive_traverse(EL_INFO *elinfo, TRAVERSE_INFO *trinfo)
{
  FUNCNAME("recursive_traverse");
  EL      *el = elinfo->el;
  EL_INFO elinfo_new;
  int     mg_level;


  if (trinfo->flag & CALL_LEAF_EL) {
    if (el->child[0]) {
      fill_elinfo(0, elinfo, &elinfo_new);
      recursive_traverse(&elinfo_new, trinfo);
      fill_elinfo(1, elinfo, &elinfo_new);
      recursive_traverse(&elinfo_new, trinfo);
    }
    else {
      (trinfo->el_fct)(elinfo);
    }
    return;
  }

  if (trinfo->flag & CALL_LEAF_EL_LEVEL) {
    if (el->child[0]) {
      if ((elinfo->level >= trinfo->level))  return;
      fill_elinfo(0, elinfo, &elinfo_new);
      recursive_traverse(&elinfo_new, trinfo);
      fill_elinfo(1, elinfo, &elinfo_new);
      recursive_traverse(&elinfo_new, trinfo);
    }
    else {
      if (elinfo->level == trinfo->level) (trinfo->el_fct)(elinfo);
    }
    return;
  }

  if (trinfo->flag & CALL_EL_LEVEL) {
    if (elinfo->level == trinfo->level)
      (trinfo->el_fct)(elinfo);
    else {
      if (elinfo->level > trinfo->level)  return;
      else
	if (el->child[0]) {
	  fill_elinfo(0, elinfo, &elinfo_new);
	  recursive_traverse(&elinfo_new, trinfo);
	  fill_elinfo(1, elinfo, &elinfo_new);
	  recursive_traverse(&elinfo_new, trinfo);
	}
    }
    return;
  }

  if (trinfo->flag & CALL_MG_LEVEL) {
    mg_level = ((int)elinfo->level + DIM-1) / DIM;

#if 0
    MSG("el %5d: level=%2d, mg_level=%2d, tr_level=%2d, childs=%s\n",
	INDEX(el), elinfo->level, mg_level, trinfo->level,
	el->child[0] ? "yes" : "no " );
#endif

    if (mg_level > (int)trinfo->level) return;

    if (!(el->child[0])) {
      (trinfo->el_fct)(elinfo);
      return;
    }
    
    if ((mg_level == trinfo->level) && ((elinfo->level % DIM) == 0))
    {
      (trinfo->el_fct)(elinfo);
      return;
    }

    fill_elinfo(0, elinfo, &elinfo_new);
    recursive_traverse(&elinfo_new, trinfo);

    fill_elinfo(1, elinfo, &elinfo_new);
    recursive_traverse(&elinfo_new, trinfo);

    return;
  }

  if (trinfo->flag & CALL_EVERY_EL_PREORDER)  (trinfo->el_fct)(elinfo);

  if (el->child[0]) {
    fill_elinfo(0, elinfo, &elinfo_new);
    recursive_traverse(&elinfo_new, trinfo);

    if (trinfo->flag & CALL_EVERY_EL_INORDER) (trinfo->el_fct)(elinfo);

    fill_elinfo(1, elinfo, &elinfo_new);
    recursive_traverse(&elinfo_new, trinfo);
  }
  else {
    if (trinfo->flag & CALL_EVERY_EL_INORDER) (trinfo->el_fct)(elinfo);
  }

  if (trinfo->flag & CALL_EVERY_EL_POSTORDER)  (trinfo->el_fct)(elinfo);

  return;

}



/****************************************************************************/
/*   test_traverse_fct:                              		            */
/*   ------------------                                 		    */
/*   display information about an element from EL and EL_INFO structures    */
/****************************************************************************/

static void test_traverse_fct(const EL_INFO *elinfo)
{
  FUNCNAME("test_traverse_fct");
  EL *el = elinfo->el;
  int i,j;

  MSG("\n");
  MSG("traversing element %d: at %8X -----------------------\n",
      INDEX(el), el);

  print_msg("level:        %3d\n",elinfo->level);

  print_msg("macro_el:     %8X\n", elinfo->macro_el);
  
  if (el->child[0]) {
    print_msg("childs:        ");
    for (i=0;i<2;i++) 
      if (el->child[i])
	         print_msg(" %3d at %8X",INDEX(el->child[i]), el->child[i]);
      else       print_msg(" ---");
    print_msg("\n");
  }

  if (elinfo->fill_flag & FILL_COORDS) {
    print_msg("coords:      ");
    for (i=0; i<N_VERTICES; i++) 
    {
      print_msg("%1d: (", i);
      for (j=0; j<DIM_OF_WORLD; j++)
	print_msg("%10.6lf%s", elinfo->coord[i][j],
		 j < (DIM_OF_WORLD-1) ? ", " : ")\n");
      if (i < (N_VERTICES-1)) print_msg("             ");
    }
  }

  if (elinfo->fill_flag & FILL_NEIGH) {
#if EL_INDEX
    print_msg("neigh index :");
    for (i=0; i<N_NEIGH; i++) 
      if (NEIGH(el,elinfo)[i]) print_msg(" %3d",INDEX(NEIGH(el,elinfo)[i]));
      else                  print_msg(" ---");
    print_msg("\n");
#endif
    print_msg("opp_vertex:  ");
    for (i=0; i<N_NEIGH; i++) 
      if (NEIGH(el,elinfo)[i]) print_msg(" %3d",OPP_VERTEX(el,elinfo)[i]);
      else                  print_msg(" ---");
    print_msg("\n");

    print_msg("neigh:      ");
    for (i=0; i<N_NEIGH; i++) print_msg(" %8X",NEIGH(el,elinfo)[i]);
    print_msg("\n");
  }
   
  if (elinfo->fill_flag & FILL_OPP_COORDS) {
    print_msg("opp_coords:  ");
    for (i=0; i<N_NEIGH; i++) {
      if (NEIGH(el,elinfo)[i]) {
	print_msg("%1d (ov=%1d): (", i, OPP_VERTEX(el,elinfo)[i]);
	for (j=0; j<DIM_OF_WORLD; j++)
	  print_msg("%10.6lf%s", elinfo->opp_coord[i][j],
		   j < (DIM_OF_WORLD-1) ? ", " : ")\n");
      }
      else {
	print_msg("%1d             : ---\n",i);
      }
      if (i < (N_VERTICES-1)) print_msg("             ");
    }
  }
}

/****************************************************************************/
/*   test_traverse:                                  		            */
/*   --------------                                     		    */
/*   display information about all elements from EL and EL_INFO structures  */
/*   by calling traverse with test_traverse_fct().      		    */
/****************************************************************************/

void test_traverse(MESH *mesh, int level, FLAGS fill_flag)
{
  FUNCNAME("test_traverse");

  MSG("with level    : %3d\n", level);
  MSG("with fill_flag:");
  if (fill_flag & FILL_ANY) {
    if (fill_flag & FILL_COORDS)   print_msg(" FILL_COORDS");
    if (fill_flag & FILL_BOUND)    print_msg(" FILL_BOUND");
    if (fill_flag & FILL_NEIGH)    print_msg(" FILL_NEIGH");
    if (fill_flag & FILL_OPP_COORDS)  print_msg(" FILL_OPP_COORDS");
  }
  else print_msg(" none");
  print_msg("\n");

  mesh_traverse(mesh, level, fill_flag, test_traverse_fct);
  MSG("done.\n");
 }




/****************************************************************************/
/*   List of static variables for stack-based traversal                     */
/****************************************************************************/
/*   traverse_mesh:       save pointer to current mesh		            */
/*   traverse_level:      save level from traverse_first() call		    */
/*   traverse_fill_flag:  save fill flag for traversal   		    */
/*   traverse_mel:        current macro element during traversal	    */
/*   stack_size:          current stack size                     	    */
/*   stack_used:          current level of hierarchy recursion   	    */
/*   elinfo_stack:        el_info hierarchy                      	    */
/*   info_stack:          describe path through hierarchy tree              */
/*                        =0: leaf of current hierarchy          	    */
/*                        =1: next element is child[0] of this one 	    */
/*                        =2: next element is child[1] of this one 	    */
/*   save_elinfo_stack:   backup for elinfo_stack during traverse_neighbour */
/*   save_info_stack:     backup for info_stack during traverse_neighbour   */
/****************************************************************************/

static MESH      *traverse_mesh;
static int       traverse_level;
static FLAGS     traverse_fill_flag;
		 
static MACRO_EL  *traverse_mel;
static int       stack_size = 0;
static int       stack_used = 0;
static EL_INFO   *elinfo_stack = nil;
static U_CHAR    *info_stack = nil;

static MACRO_EL  *save_traverse_mel;
static EL_INFO   *save_elinfo_stack = nil;
static U_CHAR    *save_info_stack = nil;
static int       save_stack_used = 0;

static int       el_count = 0;

/****************************************************************************/

static void enlarge_traverse_stack(void)
{
  FUNCNAME("enlarge_traverse_stack");
  int     i;
  int     new_stack_size = stack_size + 10;

  elinfo_stack = MEM_REALLOC(elinfo_stack, stack_size, new_stack_size,
			     EL_INFO);
  if (stack_size > 0)
    for (i=stack_size; i<new_stack_size; i++) 
      elinfo_stack[i].fill_flag = elinfo_stack[0].fill_flag;

  info_stack = MEM_REALLOC(info_stack, stack_size, new_stack_size, U_CHAR);
  save_elinfo_stack = MEM_REALLOC(save_elinfo_stack, stack_size, 
				  new_stack_size, EL_INFO);
  save_info_stack   = MEM_REALLOC(save_info_stack, stack_size,
				  new_stack_size, U_CHAR);

#if 0
  MSG("increase stack_size from %d to %d\n", stack_size, new_stack_size); 
  MSG("elinfo_stack      = %8X\n",elinfo_stack);
  MSG("info_stack        = %8X\n",info_stack);
  MSG("save_elinfo_stack = %8X\n",save_elinfo_stack);
  MSG("save_info_stack   = %8X\n",save_info_stack);
#endif

  stack_size = new_stack_size;    
}


/****************************************************************************/
/*                                                                          */
/*   return next element in tree traversal, starting from current one       */
/*                                                                          */
/****************************************************************************/


#define CALL_EVERY_EL \
  (CALL_EVERY_EL_PREORDER | CALL_EVERY_EL_INORDER | CALL_EVERY_EL_POSTORDER)


EL_INFO *traverse_next(EL_INFO *elinfo_old)
{
  FUNCNAME("traverse_next");
  EL_INFO       *elinfo=nil;

  if (stack_used) {
    TEST_EXIT(elinfo_old == elinfo_stack+stack_used)("invalid old elinfo\n");
  }
  else {
    TEST_EXIT(elinfo_old == nil)("invalid old elinfo != nil\n");
  }

  if (traverse_fill_flag & CALL_LEAF_EL) elinfo = traverse_leaf_el();
  else
    if (traverse_fill_flag & CALL_LEAF_EL_LEVEL)
      elinfo = traverse_leaf_el_level();
    else if (traverse_fill_flag & CALL_EL_LEVEL)
      elinfo = traverse_el_level();
    else if (traverse_fill_flag & CALL_MG_LEVEL)
      elinfo = traverse_mg_level();
    else
      if (traverse_fill_flag & CALL_EVERY_EL_PREORDER)
	elinfo = traverse_every_el_preorder();
      else if (traverse_fill_flag & CALL_EVERY_EL_INORDER)
	elinfo = traverse_every_el_inorder();
      else if (traverse_fill_flag & CALL_EVERY_EL_POSTORDER)
	elinfo = traverse_every_el_postorder();
      else
	ERROR_EXIT("invalid traverse_flag: %8x\n", traverse_fill_flag);

  if (elinfo)
    el_count++;
  else {
    /* MSG("total element count:%d\n",el_count); */
  }

  return(elinfo);
}

#undef CALL_EVERY_EL


#if 0   /*  Kuni 12.03.97: gcc not used    */
static EL_INFO *traverse_every_el(void)
{
  FUNCNAME("traverse_every_el");
  EL *el;
  int i;
  
  if (stack_used == 0)   /* first call */
  {
    traverse_mel = traverse_mesh->first_macro_el;
    if (traverse_mel == nil)  return(nil);

    stack_used = 1;
    fill_macro_info(traverse_mel, elinfo_stack+stack_used);
    info_stack[stack_used] = 0;

    return(elinfo_stack+stack_used);
  }
  
  el = elinfo_stack[stack_used].el;

  /* go up in tree until we can go down again */
  while((stack_used > 0) && 
	((info_stack[stack_used] >= 2) || (el->child[0]==nil)))
  {
    stack_used--;
    el = elinfo_stack[stack_used].el;
  }


  /* goto next macro element */
  if (stack_used < 1) {
    traverse_mel = traverse_mel->next;
    if (traverse_mel == nil)  return(nil);

    stack_used = 1;
    fill_macro_info(traverse_mel, elinfo_stack+stack_used);
    info_stack[stack_used] = 0;

    return(elinfo_stack+stack_used);
  }


  /* go down tree */

  if (stack_used >= stack_size-1) enlarge_traverse_stack();
  i = info_stack[stack_used];
  info_stack[stack_used]++;
  fill_elinfo(i, elinfo_stack+stack_used, elinfo_stack+stack_used+1);
  stack_used++;

  TEST_EXIT(stack_used < stack_size)
    ("stack_size=%d too small, level=%d\n",
     stack_size, elinfo_stack[stack_used].level);

  info_stack[stack_used] = 0;
  
  return(elinfo_stack+stack_used); 
}
#endif

static EL_INFO *traverse_every_el_preorder(void)
{
  FUNCNAME("traverse_every_el_preorder");
  EL *el;
  int i;
  
  if (stack_used == 0)   /* first call */
  {
    traverse_mel = traverse_mesh->first_macro_el;
    if (traverse_mel == nil)  return(nil);

    stack_used = 1;
    fill_macro_info(traverse_mel, elinfo_stack+stack_used);
    info_stack[stack_used] = 0;

    return(elinfo_stack+stack_used);
  }
  
  el = elinfo_stack[stack_used].el;

  /* go up in tree until we can go down again */
  while((stack_used > 0) && 
	((info_stack[stack_used] >= 2) || (el->child[0]==nil)))
  {
    stack_used--;
    el = elinfo_stack[stack_used].el;
  }


  /* goto next macro element */
  if (stack_used < 1) {
    traverse_mel = traverse_mel->next;
    if (traverse_mel == nil)  return(nil);

    stack_used = 1;
    fill_macro_info(traverse_mel, elinfo_stack+stack_used);
    info_stack[stack_used] = 0;

    return(elinfo_stack+stack_used);
  }


  /* go down tree */

  if (stack_used >= stack_size-1) enlarge_traverse_stack();
  i = info_stack[stack_used];
  info_stack[stack_used]++;
  fill_elinfo(i, elinfo_stack+stack_used, elinfo_stack+stack_used+1);
  stack_used++;

  TEST_EXIT(stack_used < stack_size)
    ("stack_size=%d too small, level=%d\n",
     stack_size, elinfo_stack[stack_used].level);

  info_stack[stack_used] = 0;
  
  return(elinfo_stack+stack_used); 
}


static EL_INFO *traverse_every_el_inorder(void)
{
  FUNCNAME("traverse_every_el_inorder");
  EL *el;
  int i;

  ERROR_EXIT("not yet implemented\n");
  
  if (stack_used == 0)   /* first call */
  {
    traverse_mel = traverse_mesh->first_macro_el;
    if (traverse_mel == nil)  return(nil);

    stack_used = 1;
    fill_macro_info(traverse_mel, elinfo_stack+stack_used);
    info_stack[stack_used] = 0;

    return(elinfo_stack+stack_used);
  }
  
  el = elinfo_stack[stack_used].el;

  /* go up in tree until we can go down again */    /* inorder !!! */
  while((stack_used > 0) && 
	((info_stack[stack_used] >= 2) || (el->child[0]==nil)))
  {
    stack_used--;
    el = elinfo_stack[stack_used].el;
  }


  /* goto next macro element */
  if (stack_used < 1) {
    traverse_mel = traverse_mel->next;
    if (traverse_mel == nil)  return(nil);

    stack_used = 1;
    fill_macro_info(traverse_mel, elinfo_stack+stack_used);
    info_stack[stack_used] = 0;

    return(elinfo_stack+stack_used);
  }


  /* go down tree */

  if (stack_used >= stack_size-1) enlarge_traverse_stack();
  i = info_stack[stack_used];
  info_stack[stack_used]++;
  fill_elinfo(i, elinfo_stack+stack_used, elinfo_stack+stack_used+1);
  stack_used++;

  TEST_EXIT(stack_used < stack_size)
    ("stack_size=%d too small, level=%d\n",
     stack_size, elinfo_stack[stack_used].level);

  info_stack[stack_used] = 0;
  
  return(elinfo_stack+stack_used); 
}


static EL_INFO *traverse_every_el_postorder(void)
{
  FUNCNAME("traverse_every_el_postorder");
  EL *el;
  int i;
  
  if (stack_used == 0)   /* first call */
  {
    traverse_mel = traverse_mesh->first_macro_el;
    if (traverse_mel == nil)  return(nil);

    stack_used = 1;
    fill_macro_info(traverse_mel, elinfo_stack+stack_used);
    info_stack[stack_used] = 0;

    return(elinfo_stack+stack_used);
  }
  
  el = elinfo_stack[stack_used].el;

  /* go up in tree until we can go down again */          /* postorder!!! */
  while((stack_used > 0) && 
	((info_stack[stack_used] >= 3) || (el->child[0]==nil)))
  {
    stack_used--;
    el = elinfo_stack[stack_used].el;
  }


  /* goto next macro element */
  if (stack_used < 1) {
    traverse_mel = traverse_mel->next;
    if (traverse_mel == nil)  return(nil);

    stack_used = 1;
    fill_macro_info(traverse_mel, elinfo_stack+stack_used);
    info_stack[stack_used] = 0;

/*    return(elinfo_stack+stack_used); */
  }


  /* go down tree */

  while((elinfo_stack+stack_used)->el->child[0]
	&& (info_stack[stack_used] < 2)) {
    if (stack_used >= stack_size-1) enlarge_traverse_stack();
    i = info_stack[stack_used];
    info_stack[stack_used]++;
    fill_elinfo(i, elinfo_stack+stack_used, elinfo_stack+stack_used+1);
    stack_used++;
    info_stack[stack_used] = 0;
  }
  
  info_stack[stack_used]++;      /* postorder!!! */
  return(elinfo_stack+stack_used); 
}


static EL_INFO *traverse_leaf_el(void)
{
  FUNCNAME("traverse_leaf_el");
  EL *el;
  int i;
  
  if (stack_used == 0)   /* first call */
  {
    traverse_mel = traverse_mesh->first_macro_el;
    if (traverse_mel == nil)  return(nil);

    stack_used = 1;
    fill_macro_info(traverse_mel, elinfo_stack+stack_used);
    info_stack[stack_used] = 0;

    el = elinfo_stack[stack_used].el;
    if ((el == nil) || (el->child[0] == nil)) {
      return(elinfo_stack+stack_used);
    }
  }
  else
  {
    el = elinfo_stack[stack_used].el;

    /* go up in tree until we can go down again */
    while((stack_used > 0) && 
	  ((info_stack[stack_used] >= 2) || (el->child[0]==nil)))
    {
      stack_used--;
      el = elinfo_stack[stack_used].el;
    }

    /* goto next macro element */
    if (stack_used < 1) {
      traverse_mel = traverse_mel->next;
      if (traverse_mel == nil)  return(nil);

      stack_used = 1;
      fill_macro_info(traverse_mel, elinfo_stack+stack_used);
      info_stack[stack_used] = 0;

      el = elinfo_stack[stack_used].el;
      if ((el == nil) || (el->child[0] == nil)) {
	return(elinfo_stack+stack_used);
      }
    }
  }

  /* go down tree until leaf */
  while (el->child[0])
  {
    if (stack_used >= stack_size-1) enlarge_traverse_stack();
    i = info_stack[stack_used];
    el = el->child[i];
    info_stack[stack_used]++;
    fill_elinfo(i, elinfo_stack+stack_used,
		elinfo_stack+stack_used+1);
    stack_used++;
    

    TEST_EXIT(stack_used < stack_size)
      ("stack_size=%d too small, level=(%d,%d)\n",
       stack_size, elinfo_stack[stack_used].level);

    info_stack[stack_used] = 0;
  }

  return(elinfo_stack+stack_used); 

}

static EL_INFO *traverse_leaf_el_level(void)
{
  FUNCNAME("traverse_leaf_el_level");

  ERROR_EXIT("not yet");
  return(nil);
}

static EL_INFO *traverse_el_level(void)
{
  FUNCNAME("traverse_el_level");
  ERROR_EXIT("not yet");
  return(nil);
}

static EL_INFO *traverse_mg_level(void)
{
  FUNCNAME("traverse_mg_level");
  ERROR_EXIT("not yet");
  return(nil);
}



/****************************************************************************/
/*   traverse_first:                                 		            */
/*   ---------------                                    		    */
/*   initiate traversal and return first el_info                            */
/****************************************************************************/

EL_INFO *traverse_first(MESH *mesh, int level, FLAGS fill_flag)
{
  FUNCNAME("traverse_first");
  int i;

  traverse_mesh      = mesh;
  traverse_level     = level; 
  traverse_fill_flag = fill_flag; 

  if (stack_size < 1)
    enlarge_traverse_stack();

  for (i=0; i<stack_size; i++)
    elinfo_stack[i].fill_flag = fill_flag & FILL_ANY;

  elinfo_stack[0].mesh = elinfo_stack[1].mesh = mesh;

  if (fill_flag & CALL_LEAF_EL_LEVEL) {
    TEST_EXIT(level >= 0)("invalid level: %d\n",level);
  }

  traverse_mel = nil;
  stack_used   = 0;
  el_count     = 0;

  return(traverse_next(nil));
}




/****************************************************************************/
/*   test_traverse_fct:                              		            */
/*   ------------------                                 		    */
/*   display information about an element from EL and EL_INFO structures    */
/****************************************************************************/

static void test_traverse2_fct(EL_INFO *elinfo)
{
  FUNCNAME("test_traverse2_fct");
  EL      *el = elinfo->el;

  MSG("\n");
  MSG("traversing element %d: at %8X -----------------------\n",
      INDEX(el), el);

  print_msg("level:        %3d\n",elinfo->level);

}

/****************************************************************************/
/*   test_traverse2:                                  		            */
/*   ---------------                                     		    */
/*   display information about all elements from EL and EL_INFO structures  */
/*   by calling traverse with test_traverse_fct().      		    */
/****************************************************************************/

void test_traverse2(MESH *mesh, int level, FLAGS fill_flag)
{
  FUNCNAME("test_traverse2");
  EL_INFO *elinfo;

  MSG("with level    : %3d\n", level);
  MSG("with fill_flag:");
  if (fill_flag & FILL_ANY) {
    if (fill_flag & FILL_COORDS)      print_msg(" FILL_COORDS");
    if (fill_flag & FILL_NEIGH)       print_msg(" FILL_NEIGH");
    if (fill_flag & FILL_BOUND)       print_msg(" FILL_BOUND");
    if (fill_flag & FILL_OPP_COORDS)  print_msg(" FILL_OPP_COORDS");
  }
  else print_msg(" none");
  print_msg("\n");


  elinfo = traverse_first(mesh, level, fill_flag);
  while (elinfo)
  {
    test_traverse2_fct(elinfo);

    elinfo = traverse_next(elinfo);
  }

  MSG("done.\n");
 }
