/****************************************************************************/
/* ALBERT:   an Adaptive multi Level finite element toolbox using           */
/*           Bisectioning refinement and Error control by Residual          */
/*           Techniques                                                     */
/*                                                                          */
/* file:     Common/util.c                                                  */
/*                                                                          */
/*                                                                          */
/* description:  utilities for messages, memory allocation, parameters      */
/*                                                                          */
/*                                                                          */
/* history:                                                                 */
/*                                                                          */
/*                                                                          */
/****************************************************************************/
/*                                                                          */
/*        authors: Alfred Schmidt and Kunibert G. Siebert                   */
/*                 Institut f"ur Angewandte Mathematik                      */
/*                 Albert-Ludwigs-Universit"at Freiburg                     */
/*                 Hermann-Herder-Str. 10                                   */
/*                 79104 Freiburg                                           */
/*                 Germany                                                  */
/*                                                                          */
/*        email:   alfred,kunibert@mathematik.uni-freiburg.de               */
/*                                                                          */
/*     (c) by A. Schmidt and K.G. Siebert (1996-1999)                       */
/*                                                                          */
/****************************************************************************/
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>

#include "albert_util.h"

/****************************************************************************/
/*  global variable funcName; used if no local funcName declared            */
/****************************************************************************/

const char   *funcName = nil;

/****************************************************************************/
/*  global variables for util.c                                             */
/****************************************************************************/

/*  static FILE  *out_file = stdout, *error_file = stderr; */
static FILE  *out_file = nil, *error_file = nil;
static const char *old_func_name = nil;
static char  func_name[256];

void albert_wait(const char *funcName, int wait)
{
  if (wait)
  {
    char   line[10];
    MSG("wait for <enter> ...");
    fgets(line, 9, stdin);
  }
  return;
}

void change_msg_out(FILE *fp)
{
  FUNCNAME("change_msg_out");
  
  if (fp)
  {
    if (out_file  &&  out_file != stdout  &&  out_file != stderr)
      fclose(out_file);

    out_file = fp;
  }
  else
  {
    ERROR("file pointer is pointer to nil;\n");
    ERROR("use previous stream for messages furthermore\n");
  }
  return;
}

void open_msg_file(const char *filename, const char *type)
{
  FUNCNAME("open_msg_file");
  FILE  *fp;

  if (filename  &&  (fp = fopen(filename, type)))
  {
    if (out_file  &&  out_file != stdout  &&  out_file != stderr)
      fclose(out_file);

    out_file = fp;
  }
  else
  {
    if (filename)
      ERROR("can not open %s;\n", filename);
    else
      ERROR("no filename specified;\n");
    ERROR("use previous stream for messages furthermore\n");
  }
  return;
}

void change_error_out(FILE *fp)
{
  FUNCNAME("change_error_out");
  if (fp)
  {
    if (error_file  &&  error_file != stdout  &&  error_file != stderr)
      fclose(error_file);

    error_file = fp;
  }
  else
  {
    ERROR("file pointer is pointer to nil;\n");
    ERROR("use previous stream for errors furthermore\n");
  }

  return;
}

void open_error_file(const char *filename, const char *type)
{
  FUNCNAME("open_error_file");
  FILE  *fp;

  if (filename  &&  (fp = fopen(filename, type)))
  {
    if (error_file  &&  error_file != stdout  &&  error_file != stderr)
      fclose(error_file);

    error_file = fp;
  }
  else
  {
    if (filename)
      ERROR("can not open %s;\n", filename);
    else
      ERROR("no filename specified;\n");
    ERROR("use previous stream for errors furthermore\n");
  }
  return;
}

void print_funcname(const char *funcname)
{
  if (!out_file) out_file = stdout;

  if (funcname  &&  old_func_name != funcname)
  {
    strcpy(func_name, funcname);
    strcat(func_name, ": ");
  }
  else if (!funcname)
  {
    strcpy(func_name, "*unknown function*");
  }
  else
  {
    strcpy(func_name, "");
  }
  fprintf(out_file, "%-20s", func_name);
  old_func_name = funcname;
  return;
}

void print_error_funcname(const char *funcname, const char *file, int line)
{
  static int        old_line = -1;

  if (!error_file) error_file = stderr;

  if (funcname  &&  old_func_name != funcname)
  {
    strcpy(func_name, funcname);
    strcat(func_name, ": ");
  }
  else if (!funcname)
  {
    strcpy(func_name, "*unknown function*");
  }
  else
  {
    strcpy(func_name, "");
  }
  if (old_func_name != funcname)
    fprintf(error_file, "%-20sERROR in %s, line %d\n", func_name, file, line);
  else if (line - old_line > 5)
    fprintf(error_file, "%-20sERROR in %s, line %d\n", func_name, file, line);

  fprintf(error_file, "%-20s", "");
  old_func_name = funcname;
  old_line = line;

  return;
}

void print_error_msg_exit(const char *format, ...)
{
  va_list   arg;

  if (!error_file) error_file = stderr;

  va_start(arg, format);
  vfprintf(error_file, format, arg);
  va_end(arg);

  exit(1);
}

void print_error_msg(const char *format, ...)
{
  va_list   arg;

  if (!error_file) error_file = stderr;

  va_start(arg, format);
  vfprintf(error_file, format, arg);
  va_end(arg);

  return;
}

void print_warn_funcname(const char *funcname, const char *file, int line)
{
  static int        old_line = -1;

  if (!out_file) out_file = stdout;

  if (funcname  &&  old_func_name != funcname)
  {
    strcpy(func_name, funcname);
    strcat(func_name, ": ");
  }
  else if (!funcname)
  {
    strcpy(func_name, "*unknown function*");
  }
  else
  {
    strcpy(func_name, "");
  }
  if (old_func_name != funcname)
    fprintf(out_file, "%-20sWARNING in %s, line %d\n", func_name,file,line);
  else if (line - old_line > 5)
    fprintf(out_file, "%-20sWARNING in %s, line %d\n", func_name,file,line);

  fprintf(out_file, "%-20s", "");
  old_func_name = funcname;
  old_line = line;

  return;
}

void print_warn_msg(const char *format, ...)
{
  va_list   arg;

  if (!out_file) out_file = stdout;

  va_start(arg, format);
  vfprintf(out_file, format, arg);
  va_end(arg);

  return;
}


void print_msg(const char *format, ...)
{
  va_list   arg;

  if (!out_file) out_file = stdout;

  va_start(arg, format);
  vfprintf(out_file, format, arg);
  va_end(arg);
}

/****************************************************************************/
/*  routines for writing int and REAl vectors to out_file                   */
/****************************************************************************/

void print_int_vec(const char *s, const int *vec, int no)
{
  int   i;

  print_msg("%s = (", s);
  for (i = 0; i < no; i++)
    print_msg("%d%s", vec[i], i < no-1 ? ", " : ")\n");
}

void print_real_vec(const char *s, const REAL *vec, int no)
{
  int   i;

  print_msg("%s = (", s);
  for (i = 0; i < no; i++)
    print_msg("%.5lf%s", vec[i], i < no-1 ? ", " : ")\n");
}

/****************************************************************************/
/*  routines for allocation/deallocation of memory. every (de)allocation is */
/*  done via albert_alloc(), albert_realloc(), albert_free() resp.          */
/****************************************************************************/

static const char *size_as_string(double size)
{
  static char  sas[128];

  if (size < 1024)
    sprintf(sas, "%d %s", (int) size, "B");
  else if (size < 1048576)
    sprintf(sas, "%.2lf %s", size/1024.0, "KB");
  else
    sprintf(sas, "%.2lf %s", size/1048576.0, "MB");
  return(sas);
}

#define MALLOC(size)        malloc(size)
#define REALLOC(ptr, size)  (ptr) ? realloc(ptr,size) : malloc(size)
#define FREE(ptr,size)      free(ptr)
#define CALLOC(size,elsize) calloc(size,elsize)

#define ALLOC_ERR(fct,file,line)\
if (fct && file) ERROR_EXIT("called by %s in %s, line %d\n",fct,file,line);\
else if (fct)    ERROR_EXIT("called by %s, (unknown filename)\n", fct);\
else if (file)   ERROR_EXIT("called in %s, line %d\n", file, line);\
else             ERROR_EXIT("location unknow\n")

static double  size_used = 0;

void *albert_alloc(size_t size, const char *fct, const char *file, int line)
{
  FUNCNAME("albert_alloc");
  void  *mem;

  if (size <= 0)
  {
    ERROR("size == 0\n");
    ALLOC_ERR(fct, file, line);
  }

  if (!(mem = MALLOC(size)))
  {
    ERROR("can not allocate %s\n", size_as_string(size));
    ALLOC_ERR(fct, file, line);
  }

  size_used += size;
  return(mem);
}


void *albert_realloc(void *ptr, size_t old_size, size_t new_size,
		     const char *fct, const char *file, int line)
{
  FUNCNAME("albert_realloc");
  void  *mem;

  if (old_size >= new_size)
  {
    if (!ptr)
    {
      ERROR("ptr == nil and old_size > new_size!\n");
      ALLOC_ERR(fct, file, line);
    }
    return(ptr);
  }

  if (new_size <= 0)
  {
    ERROR("size == 0\n");
    ALLOC_ERR(fct, file, line);
  }
  if (!(mem = REALLOC(ptr,new_size)))
  {
    ERROR("can not allocate %s\n", size_as_string(new_size));
    ALLOC_ERR(fct, file, line);
  }

  size_used += new_size - old_size;

  return(mem);
}

void *albert_calloc(size_t size, size_t elsize, const char *fct, 
		    const char *file, int line)
{
  FUNCNAME("albert_calloc");
  void  *mem;

  if (size <= 0 || elsize <= 0)
  {
    ERROR("one of size = %d or data size = %d is zero\n", size, elsize);
    ALLOC_ERR(fct, file, line);
  }

  if (!(mem = CALLOC(size,elsize)))
  {
    ERROR("can not allocate %s\n", size_as_string(size*elsize));
    ALLOC_ERR(fct, file, line);
  }

  size_used += size*elsize;
  return(mem);
}

void albert_free(void *ptr, size_t size)
{
  if (ptr)
  {
    FREE(ptr, size);
    size_used -= size;
  }
  return;
}

void print_mem_use(void)
{
  FUNCNAME("print_mem_use");

  MSG("%s memory allocated\n", size_as_string(size_used));

  return;
}

WORKSPACE *get_workspace(size_t size, const char *fct, const char *f, int l)
{
  WORKSPACE *ws = albert_alloc(sizeof(WORKSPACE), fct, f, l);

  ws->work = albert_alloc(size, fct, f, l);
  ws->size = size;

  return(ws);
}

WORKSPACE  *realloc_workspace(WORKSPACE *ws, size_t new_size, 
			      const char *fct, const char *file, int line)
{
  WORKSPACE  *workspace = ws;

  if (!workspace)
  {
    workspace = albert_alloc(sizeof(WORKSPACE), fct, file, line);
    workspace->work = nil;
  }

  if (!workspace->work)
  {
    workspace->work = albert_alloc(new_size, fct, file, line);
    workspace->size = new_size;
  }
  else if (workspace->size < new_size)
  {
    workspace->work = albert_realloc(workspace->work, workspace->size,
				     new_size, fct, file, line);
    workspace->size = new_size;
  }
  return(workspace);
}

void clear_workspace(WORKSPACE *ws)
{
  if (!ws) return;
  albert_free(ws->work, ws->size);
  ws->work = nil;
  ws->size = 0;
  return;
}

void free_workspace(WORKSPACE *ws)
{
  if (!ws) return;
  albert_free(ws->work, ws->size);
  MEM_FREE(ws, 1, WORKSPACE);

  return;
}

/****************************************************************************/
/*  matrix (de-)allocation routines: 					    */
/*  free_matrix deallocates such a matrix				    */
/****************************************************************************/


void **albert_matrix(size_t nr, size_t nc, size_t size,
		     const char *fct, const char *file, int line)
{
  int       i;
  size_t    row_length = nc*size;
  char      **mat, *mrows;

  mat = (char **) albert_alloc(nr*sizeof(char *), fct, file, line);
  mrows = (char *) albert_alloc(nr*nc*size, fct, file, line);

  for(i = 0; i < nr; i++)
    mat[i] = mrows + i*row_length;

  return((void **)mat);
}

void  free_albert_matrix(void **ptr, size_t nr, size_t nc, size_t size)
{
  albert_free(ptr[0],nr*nc*size);
  albert_free(ptr,nr*sizeof(char *));

  return;
}

/****************************************************************************/
/*  utilities for parameters and parameter initialization                   */
/****************************************************************************/

#define COMMENT   '%'
#define N_KEY     100
#define N_CHAR    1000
#define LENGTH    256

static void read_parameters(const char *filename);

struct param
{
  char       *key;
  char       *parameters;
  const char *filename;
  const char *funcname;
  int        line_no;
};

static int cpp_read = 0;

#define IS_BLANK_CHAR(s)   (s==' '||s=='\t'||s =='\f'||s=='\r')


static char *get_key(char *s, int n_line, const char *filename)
{
  FUNCNAME("get_key");
  static  char key[LENGTH];
  char    fn[LENGTH], c;
  int     i;

  if (cpp_read)
  {
    if (*s == '#')  /*  Praeprozessor line   */
      return(s);
  }

  while (*s  &&  IS_BLANK_CHAR(*s)) s++;
  
  if (*s == COMMENT  ||  *s == '\0'  ||  *s == '\n')
    return(nil);

  if (*s == '#')
  {
    if (strstr(s, "#include") == s)
/****************************************************************************/
/*  another init file has to be included                                    */
/****************************************************************************/
    {
      s += strlen("#include");
      while (*s  &&  IS_BLANK_CHAR(*s))
	s++;	

      i = 0;
      switch (c = *s++)
      {
      case '<':
	c = '>';
      case '\"':
	while(*s  &&  *s != c  &&  !IS_BLANK_CHAR(*s))
	{
	  if (i == LENGTH-1)
	  {
	    ERROR("filename more than %d characters.\n", i);
	    ERROR("Skipping line %d of file %s\n", n_line, filename);
	    return(nil);
	  }
	  fn[i++] = *s++;
	}
	if (*s != c)
	{
	  ERROR("filename of include not terminated by %c\n", c);
	  ERROR("skipping line %d of file %s\n", n_line, filename);
	  return(nil);
	}
	fn[i] = '\0';
	break;
      default:
	ERROR("no filename of include file found\n");
	ERROR("skipping line %d of file %s\n", n_line, filename);
	return(nil);
      }

      read_parameters(fn);
      return(nil);
    }
    else
    {
      ERROR("# must not be the first character on a line; except #include\n");
      return(nil);
    }
  }

/****************************************************************************/
/*  now get the key                                                         */
/****************************************************************************/
  i = 0;
  while (*s  &&  *s != ':')
  {
    if (*s == COMMENT)
    {
      ERROR("key must not contain '%c'.\n", COMMENT);
      ERROR("Skipping line %d of file %s\n", n_line, filename);
      return(nil);
    }

    if (i == LENGTH-1)
    {
      ERROR("key more than %d characters.\n", i);
      ERROR("Skipping line %d of file %s\n", n_line, filename);
      return(nil);
    }
    key[i++] = *s++;
  }
  key[i] = '\0';

  if (*s != ':')
  {
    ERROR("key was not terminated by ':'.\n");
    ERROR("Skipping line %d of file %s\n", n_line, filename);
    return(nil);
  }

  if (i == 0)
  {
    ERROR("use of ':' without key.\n");
    ERROR("Skipping line %d of file %s\n", n_line, filename);
    return(nil);
  }
  return(key);
}

static char *get_par(char *key, char *s, int *nl, const char *fn, FILE *fp)
{
  FUNCNAME("get_par");
  static char    parameter[2*LENGTH];
  int     i, ol = *nl;

  while (*s  &&  *s++ != ':');
  while (*s  &&  IS_BLANK_CHAR(*s))  s++;
  
  i = 0;
  while (*s  &&  *s != COMMENT  && *s != '\n')
  {
    if (*s == '\\'  &&  *(s+1) == '\n')
    {
      (*nl)++;
      if (fgets(s, LENGTH, fp) == nil)
      {
	ERROR("EOF reached while reading parameters of key %s\n", key);
	if (ol == *nl-1) ERROR("Skipping line %d of file %s\n", *nl, fn);
	else     ERROR("Skipping lines %d-%d of file %s\n", ol, *nl, fn);
	
	return(nil);
      }
    }
    else
    {
      if (i == 2*LENGTH-1)
      {
	ERROR("parameter has more than %d characters.\n", i);
	if (ol == *nl)  ERROR("Skipping line %d of file %s\n", *nl, fn);
	else    ERROR("Skipping lines %d-%d of file %s\n", ol, *nl, fn);

	return(nil);
      }
      if (IS_BLANK_CHAR(*s))
      {
	parameter[i++] = ' ';
	s++;
	while (*s  &&  IS_BLANK_CHAR(*s)) s++;
      }
      else
      {
	parameter[i++] = *s++;
      }
    }
  }

  do
  {
    parameter[i--] = '\0';
  } while (i  &&  IS_BLANK_CHAR(parameter[i]));

  if (i == 0  &&  IS_BLANK_CHAR(parameter[0]))
  {
    ERROR("no parameter of key %s.\n", key);
    if (ol == *nl) ERROR("Skipping line %d of file %s\n", *nl, fn);
    else   ERROR("Skipping lines %d-%d of file %s\n", ol, *nl, fn);

    return(nil);
  }

  return(parameter);
}


static int bin_search(struct param *param, const char *key, int n_keys)
{
  int      cond, left, right, mid;

  left = 0;
  right = n_keys-1;
  
  while (left <= right)
  {
    mid = (left+right)/2;
    if ((cond = strcmp(param[mid].key, key)) < 0)
    {
      left = mid + 1;
    }
    else if (cond > 0)
    {
      right = mid - 1;
    }
    else
      return(mid);
  }

  return(-1);
}

static struct param  *all_param = nil;
static int  n_param = 0, max_param = 0;

static char     *buffer = nil;
static unsigned buffer_size = 0, buffer_used = 0;
static unsigned total_buffer_size = 0;

static int  param_info = 1;
int         msg_info = 10;
int         msg_wait = 1;


static const char *get_actfile(const char *filename)
{
  FUNCNAME("get_actfilename");
  static char   **filenames = nil;
  static size_t n_files = 0, max_files = 0;
  int           i, size_f;
  char          *actfile;


  for (i = 0; i < n_files; i++)
    if (!strcmp(filename,filenames[i])) break;
  
  if (i < n_files)
  {
    return((const char *) filenames[i]);
  }
  else
  {
    if (n_files == max_files)
    {
      filenames = MEM_REALLOC(filenames, max_files, max_files+10, char *);
      max_files += 10;
    }
    size_f = strlen(filename)+1;
    if (buffer_used + size_f >= buffer_size)
    {
      buffer_used = 0;
      buffer_size = N_CHAR;
      buffer = MEM_ALLOC(N_CHAR, char);
    }
    actfile = buffer+buffer_used;
    buffer_used += size_f;
    strncpy(actfile,filename,size_f);
    return((const char *) actfile);
  }
}

static void add_param(const char *key, const char *parameter,
		      const char *actfile, int  n_line, const char *fname)
{
  FUNCNAME("add_param");
  unsigned size_k, size_p;
  int      i, j, scmp;

  size_k = strlen(key)+1;
  size_p = strlen(parameter)+1;

  for (i = 0; i < n_param; i++)
  {
    if ((scmp = strcmp(all_param[i].key, key)) >= 0)  break;
  }


  if (i < n_param  &&  scmp == 0)
  {
/****************************************************************************/
/*  key does already exist: save new parameters                             */
/****************************************************************************/
    if (strlen(all_param[i].parameters)+1 >= size_p)
      strcpy(all_param[i].parameters, parameter);
    else
    {
      if (buffer_used + size_p >= buffer_size)
      {
	buffer_used = 0;
	buffer_size = N_CHAR;
	buffer = MEM_ALLOC(N_CHAR, char);
      }

      all_param[i].parameters = buffer+buffer_used;
      buffer_used += size_p;
      total_buffer_size += size_p;
      strcpy(all_param[i].parameters, parameter);
    }
    all_param[i].filename = actfile;
    all_param[i].funcname = fname;
    all_param[i].line_no = n_line;

    return;
  }

  if (n_param == max_param)
  {
    all_param = MEM_REALLOC(all_param, max_param, max_param + N_KEY,
			    struct param);
    max_param += N_KEY;
  }

  for (j = n_param; j > i; j--)
    all_param[j] = all_param[j-1];

  if (buffer_used + size_k >= buffer_size)
  {
    if (buffer_used + size_p < buffer_size)
    {
      all_param[i].parameters = buffer+buffer_used;
      buffer_used += size_p;
      total_buffer_size += size_p;
      size_p = 0;
      strcpy(all_param[i].parameters, parameter);
    }
    buffer_used = 0;
    buffer_size = N_CHAR;
    buffer = MEM_ALLOC(N_CHAR, char);
  }

  all_param[i].key = buffer+buffer_used;
  buffer_used += size_k;
  total_buffer_size += size_k;
  strcpy(all_param[i].key, key);

  if (size_p)
  {
    if (buffer_used + size_p >= buffer_size)
    {
      buffer_used = 0;
      buffer_size = N_CHAR;
      buffer = MEM_ALLOC(N_CHAR, char);
    }
    
    all_param[i].parameters = buffer+buffer_used;
    buffer_used += size_p;
    total_buffer_size += size_p;
    strcpy(all_param[i].parameters, parameter);
  }
  all_param[i].filename = actfile;
  all_param[i].funcname = fname;
  all_param[i].line_no = n_line;

  n_param++;

  return;
}

static void read_parameters(const char *filename)
{
  FUNCNAME("read_parameters");
  FILE        *init_file;
  char        line[LENGTH];
  int         n_line = 0;
  char        *key, *parameter;
  const char  *actfile = nil;

  if (!(init_file = fopen(filename, "r")))  return;

  if (!cpp_read)
  {
    INFO(param_info,2)("reading from file %s\n", filename);
    actfile = get_actfile(filename);
  }

  while (fgets(line, LENGTH, init_file))
  {
    n_line++;
    if (strlen(line) == LENGTH-1)
    {
      ERROR("line %d of file %s too long; skipping it\n", n_line, filename);
      continue;
    }

    key = get_key(line, n_line, filename);
    
    if (key == nil)
      continue;
    
    if (cpp_read  &&  *key == '#')
    {
      char  *s;

      sscanf(key, "#%d", &n_line);
      n_line--;
      while (*key  &&  *key++ != '\"');
      s = key;
      while (*s  &&  *s != '\"') s++;
      *s = '\0';
      actfile = get_actfile(key);

      continue;
    }

    parameter = get_par(key, line, &n_line, filename, init_file);
    if (!parameter)  continue;

    add_param(key, parameter, actfile, n_line, nil);
  }

  fclose(init_file);

  return;
}

static void p_swap(struct param *param, int i, int j)
{
  struct param tmp;
  tmp = param[i];
  param[i] = param[j];
  param[j] = tmp;
  return;
}

static void p_qsort(struct param *param, int left, int right)
{
  int     i, last;

  if (left >= right)  return;

  p_swap(param, left, (left+right)/2);
  last = left;
  for (i = left+1; i <= right; i++)
    if (strcmp(param[i].key, param[left].key) < 0)
      p_swap(param, ++last, i);

  p_swap(param, left, last);
  p_qsort(param, left, last-1);
  p_qsort(param, last+1, right);
  return;
}

void print_parameters(void)
{
  FUNCNAME("print_parameters");
  int   i;

  for (i = 0; i < n_param; i++)
  {
    MSG("%s:  %s\n", all_param[i].key, all_param[i].parameters);
    if (all_param[i].funcname)
    {
      MSG("initialized by %s() on line %3d of file \"%s\"\n", 
	  all_param[i].funcname, all_param[i].line_no, all_param[i].filename);
    }
    else if (all_param[i].filename)
    {
      MSG("initialized on line %2d of file \"%s\"\n",all_param[i].line_no, 
	  all_param[i].filename);
    }
    else
    {
      MSG("can not locate initialization location\n");
    }
  }
}

void save_parameters(const char *file, int info)
{
  FILE  *fp;
  int   i;

  if (!(fp = fopen(file, "w")))
    return;

  for (i = 0; i < n_param; i++)
  {
    fprintf(fp, "%s:  %s\n", all_param[i].key, all_param[i].parameters);

    if (info)
    {
      if (all_param[i].funcname)
      {
	fprintf(fp, "%%initialized by %s() on line %3d of file \"%s\"\n", 
		all_param[i].funcname, all_param[i].line_no,
		all_param[i].filename);
      }
      else if (all_param[i].filename)
      {
	fprintf(fp, "%%initialized on line %2d of file \"%s\"\n",
		all_param[i].line_no, all_param[i].filename);
      }
    }
  }
  fclose(fp);
}

void init_parameters(int p, const char *file_name)
{
  FUNCNAME("init_parameters");

  if (!file_name)
    return;

  read_parameters(file_name);

  GET_PARAMETER(0, "level of information", "%d", &msg_info);
  GET_PARAMETER(0, "WAIT", "%d", &msg_wait);
  GET_PARAMETER(0, "parameter information", "%d", &param_info);
  if (!msg_info)  param_info = 0;

  if (p  &&  msg_info) print_parameters();

  return;
}  

void add_parameter(int p, const char *key, const char *par)
{
  Add_parameter(p, key, par, nil, nil, 0);
}

void Add_parameter(int p, const char *key, const char *par,
		   const char *fname, const char *file, int line)
{
  FUNCNAME("add_parameter");

  if (!key  ||  !par)
    return;

  add_param(key, par, file, line, fname);

  GET_PARAMETER(0, "level of information", "%d", &msg_info);
  GET_PARAMETER(0, "WAIT", "%d", &msg_wait);
  GET_PARAMETER(0, "parameter information", "%d", &param_info);
  if (!msg_info)  param_info = 0;

  if (p  &&  msg_info) print_parameters();

  return;
}  

static const char *get_filename(const char *path_file)
{
  const char *s = path_file;
  if (s)
  {
    while (*s) s++;  /* goto end of string */
    while (s > path_file && *s != '/') s--;
    if (*s == '/') s++;
  }
  return(s);
}

static int sys_test(const char *command, const char *file)
{
  char  line[1024];

  if (!file || !command) return(0);
  sprintf(line, "%s %s > /dev/null", command, file);
  return(!system(line));
}

#include <time.h>

void init_parameters_cpp(int p, const char *fn, const char *flags)
{
  FUNCNAME("init_parameters_cpp");
  int      val;
  char    tmp_file[LENGTH];
  char    cpp_flags[3*LENGTH];
  char    call_cpp[3*LENGTH];
  FILE    *fp;
  time_t  act_time;

  if (!fn)
    return;

  if (!(fp = fopen(fn, "r")))
  {
    ERROR("can not read from file %s\n", fn);
    return;
  }
  if (fp)  fclose(fp);
#ifndef CPP
  ERROR("no cpp available; reading file %s without help of cpp\n", fn);
  init_parameters(p, fn);

  return;
#else
  fp = nil;

  time(&act_time);
  if (sys_test("test -w", "/tmp"))               /* you may write to /tmp  */
  {
    const char  *file = get_filename(fn);
    sprintf(tmp_file, "/tmp/%s.cpp", file);
    if (sys_test("test -f", tmp_file))           /* file exists :-( */
    {
      sprintf(tmp_file, "/tmp/%s.cpp.%d", file, (int) act_time);
      if (sys_test("test -f", tmp_file))            /* file exists :-( */
	*tmp_file = 0;
    }
  }

  if (*tmp_file == 0)
  {
    sprintf(tmp_file, "%s.cpp", fn);
    if (sys_test("test -f", tmp_file))
    {
      sprintf(tmp_file, "%s.cpp.%d", fn, (int) act_time);
      if (sys_test("test -f", tmp_file))
	*tmp_file = 0;
    }
  }

  if (*tmp_file  && (fp = fopen(tmp_file, "w")))
    fclose(fp);
  else
  {
    ERROR("could not open temporary file for CPP\n");
    ERROR("can not write to /tmp and files %s.cpp and\n", fn);
    ERROR_EXIT("%s.cpp.%d either exist or are not writable\n", fn, act_time);
  }

  sprintf(cpp_flags, "-DDIM=%d -DDIM_OF_WORLD=%d ", DIM, DIM_OF_WORLD);
  if (flags)
    strcat(cpp_flags, flags);

  sprintf(call_cpp, "%s %s %s > %s", CPP, cpp_flags, fn, tmp_file);

/****************************************************************************/
/*  invoke cpp:                                                             */
/****************************************************************************/

  val = system(call_cpp);
  if (val)
  {
    MSG("val = %d\n", val);
    ERROR("error during cpp call; reading file %s without help of cpp\n", fn);
    WAIT;
    init_parameters(p, fn);
    return;
  }

/****************************************************************************/
/*  read parameters from temp file                                          */
/****************************************************************************/

  cpp_read = 1;
  read_parameters(tmp_file);
  cpp_read = 0;
/****************************************************************************/
/*  remove temp file                                                        */
/****************************************************************************/

  sprintf(call_cpp, "/bin/rm %s", tmp_file);
  val = system(call_cpp);

  GET_PARAMETER(0, "level of information", "%d", &msg_info);
  GET_PARAMETER(0, "WAIT", "%d", &msg_wait);
  GET_PARAMETER(0, "parameter information", "%d", &param_info);
  if (!msg_info)  param_info = 0;
  if (p && msg_info) print_parameters();

  return;
#endif
}  


static char *get_next_word(char **s)
{
  static char  Val[2*LENGTH];
  char         *val = Val, *line;
  
  for (line = *s; *line  &&  *line == ' '; line++);
  for (; *line  &&  (*val = *line) != ' '; val++, line++);

  *s = line;
  *val = '\0';
  return(Val);
}

static const char *param_call_fct = nil, *param_call_file = nil;
static int        param_call_line = 0;

int init_param_func_name(const char *call_fct, const char *call_file,
			 int call_line)
{
  param_call_fct = call_fct;
  param_call_file = call_file;
  param_call_line = call_line;

  return(1);
}

int get_parameter(int info, const char *key, const char *format, ...)
{
  const char   *funcName, *func_name = "get_parameter";
  int          count, i, *ival;
  char         *sval, *s, *word, *cval;
  const char   *p;
  REAL         *rval;
  U_CHAR       *Uval;
  S_CHAR       *Sval;
  double       dval;

  va_list   arg;

  funcName = param_call_fct != nil ? param_call_fct : func_name;

  if (param_info)
  {
    if (param_info > 1)
      info = MAX(info, param_info-1);
  }
  else
    info = 0;

  if (!all_param)
  {
    INFO(info,1)("no parameters defined\n");
    return(0);
  }

  i = bin_search(all_param, key, n_param);

  if (i  < 0)
  {
    if (funcName != func_name)
    {
      INFO(info,1)("initialization of parameter `%s'\n", key);
      INFO(info,1)("fails on line %d of file %s\n", param_call_line,
		   param_call_file);
    }
    else
      INFO(info,1)("initialization of parameter `%s' fails\n", key);

    param_call_fct = nil;
    return(0);
  }

  va_start(arg, format);
  count = 0;
  
  INFO(info,2)("parameter `%s' initialized with: ", key);

  s = all_param[i].parameters;
 
  for (p = format; *p; p++)
  {
    if (*p != '%')  continue;

    word = get_next_word(&s);
    if (!*word)
    {
      PRINT_INFO(info,2)("\n");
      if (all_param[i].funcname)
      {
	INFO(info,4)("parameter initialized by %s()\n",
		     all_param[i].funcname);
	INFO(info,4)("on line %d of file \"%s\"\n", all_param[i].line_no,
		     all_param[i].filename);
      }
      else if (all_param[i].filename)
      {
	INFO(info,4)
	  ("parameter initialized on line %2d of init file \"%s\"\n",
	   all_param[i].line_no, all_param[i].filename);
      }
      else
      {
	INFO(info,4)("location of initialization unknown\n");
      }
      va_end(arg);
      param_call_fct = nil;
      return(count);
    }
    count++;

    switch(*++p)
    {
    case 'S':
      Sval = va_arg(arg, S_CHAR *);
      *Sval = (S_CHAR) atoi(word);
      PRINT_INFO(info,2)("%d ", *Sval);
      break;
    case 'U':
      Uval = va_arg(arg, U_CHAR *);
      *Uval = (U_CHAR) atoi(word);
      PRINT_INFO(info,2)("%d ", *Uval);
      break;
    case 'c':
      cval = va_arg(arg, char *);
      *cval = *word;
      PRINT_INFO(info,2)("%c ", *cval);
      break;
    case 's':
      sval = va_arg(arg, char *);
      strcpy(sval, word);
      PRINT_INFO(info,2)("%s ", sval);
      break;
    case 'd':
      ival = va_arg(arg, int *);
      *ival = atoi(word);
      PRINT_INFO(info,2)("%d ", *ival);
      break;
    case 'e':
    case 'f':
    case 'g':
      rval = va_arg(arg, REAL *);
      *rval = dval = atof(word);
      PRINT_INFO(info,2)("%lg ", dval);
      break;
    case '*':
      break;
    default: 
      PRINT_INFO(info,2)("\n");
      INFO(info,2)
	("unknow format specifier `%%%c', skipping initialization of %s\n",
	 *p, key);
    }
  }

  PRINT_INFO(info,2)("\n");
  if (funcName != func_name)
    INFO(info,2)("on line %d of file %s\n", param_call_line, param_call_file);

  if (all_param[i].funcname)
  {
    INFO(info,4)("parameter initialized by %s()\n", all_param[i].funcname);
    INFO(info,4)("on line %d of file \"%s\"\n", all_param[i].line_no,
		 all_param[i].filename);
  }
  else if (all_param[i].filename)
  {
    INFO(info,4)
      ("parameter initialized on line %2d of init file \"%s\"\n",
       all_param[i].line_no, all_param[i].filename);
  }
  else
  {
    INFO(info,4)("location of initialization unknown\n");
  }
  
  va_end(arg);
  param_call_fct = nil;
  return(count);
}



const char *generate_filename(const char *path, const char *fn, int ntime)
{
  static char  name[256];
  char         *cp;
  
  if (path == nil || path[0] == '\0')
  {
    sprintf(name, "./%s", fn);
  }
  else
  {
    const char *cp = path;
    while (*cp)
      cp++;
    cp--;
    if (*cp == '/')
      sprintf(name, "%s%s", path, fn);
    else
      sprintf(name, "%s/%s", path, fn);
  }
  cp = name;
  while (*cp)
    cp++;
/*   sprintf(cp, "%d", 100000 + ntime); */
/*   if (ntime < 100000)  *cp = '0'; */
  sprintf(cp, "%06d", ntime);

  return((const char *) name);
}

/****************************************************************************/
/*  handling of general lists                                               */
/****************************************************************************/

static VOID_LIST_ELEMENT *first_element = nil;

VOID_LIST_ELEMENT *get_void_list_element()
{
  VOID_LIST_ELEMENT  *list;
  const int          n_elements = 20;
  int                i;
  
  if (!first_element)
  {
    first_element = MEM_ALLOC(n_elements, VOID_LIST_ELEMENT);
    for (i = 0; i < n_elements-1; i++)
    {
      first_element[i].data = nil;
      first_element[i].next = first_element+(i+1);
    }
    first_element[n_elements-1].data = nil;
    first_element[n_elements-1].next = nil;
  }
  list = first_element;
  first_element = list->next;

  list->next = nil;
  return(list);
}

void free_void_list_element(VOID_LIST_ELEMENT *list)
{
  list->data = nil;
  list->next = first_element;
  first_element = list;
  return;
}
