/****************************************************************************/
/* ALBERT:   an Adaptive multi Level finite element toolbox using           */
/*           Bisectioning refinement and Error control by Residual          */
/*           Techniques                                                     */
/*                                                                          */
/* file:     1d/macro.c                                                     */
/*           !!! includes Common/macro_common.c  !!!                        */
/*                                                                          */
/* description: tools for reading and writing macro triangulations          */
/*                                                                          */
/*                                                                          */
/* history:                                                                 */
/*                                                                          */
/*                                                                          */
/****************************************************************************/
/*                                                                          */
/*        authors: Alfred Schmidt and Kunibert G. Siebert                   */
/*                 Institut f"ur Angewandte Mathematik                      */
/*                 Albert-Ludwigs-Universit"at Freiburg                     */
/*                 Hermann-Herder-Str. 10                                   */
/*                 79104 Freiburg                                           */
/*                 Germany                                                  */
/*                                                                          */
/*        email:   alfred,kunibert@mathematik.uni-freiburg.de               */
/*                                                                          */
/*     (c) by A. Schmidt und K.G. Siebert (1996)                            */
/*                                                                          */
/****************************************************************************/
/****************************************************************************/
/*  modifications: vector oriented storage of macro triangulation data      */
/*  in structure MACRO_DATA; data is then converted into ALBERT-macro       */
/*  Author: Daniel Koester (2001)                                           */
/****************************************************************************/

#include <ctype.h>
#include <rpc/types.h>
#include <rpc/xdr.h>
#include "albert.h"

/****************************************************************************/
/* these definitions are necessary at the moment for the macros "MEM_ALLOC" */
/* and "MEM_FREE" to work properly                                          */
/****************************************************************************/

typedef int INT_NVERT[N_VERTICES];
typedef int INT_NNEIGH[N_NEIGH];
typedef S_CHAR SCHAR_NNEIGH[N_NEIGH];
typedef U_CHAR UCHAR_NNEIGH[N_NEIGH];

/****************************************************************************/
/* read data->bound into mel[].bound[]                                      */
/****************************************************************************/

static void fill_bound_info(MESH *mesh, MACRO_DATA *data,
                            const BOUNDARY *(*bdry)(MESH *, int))
{
  int i;
  MACRO_EL   *mel = mesh->first_macro_el;

  for(i = 0; i < data->n_macro_elements; i++) {
    mel[i].bound[1] = data->boundary[i][0];
    mel[i].bound[0] = data->boundary[i][1];
  }

  return;
}


/****************************************************************************/
/* orientation(): checks and corrects whether the element is oriented from  */
/* left to right.                                                           */
/****************************************************************************/

static U_CHAR orientation(MACRO_DATA *data)
{
  int      i, vert_buffer, neigh_buffer;
  S_CHAR   bound_buffer;
  U_CHAR   result = false;

  for(i = 0; i < data->n_macro_elements; i++) {
    if(data->coords[data->mel_vertices[i][0]][0] >
       data->coords[data->mel_vertices[i][1]][0]) {
      result = true;

      vert_buffer = data->mel_vertices[i][0];
      data->mel_vertices[i][0] = data->mel_vertices[i][1];
      data->mel_vertices[i][1] = vert_buffer;

      neigh_buffer = data->neigh[i][0];
      data->neigh[i][0] = data->neigh[i][1];
      data->neigh[i][1] = neigh_buffer;

      bound_buffer = data->boundary[i][0];
      data->boundary[i][0] = data->boundary[i][1];
      data->boundary[i][1] = bound_buffer;
    }
  }

  return(result);
}


/****************************************************************************/
/* macro_test(): In 1D only element orientation is checked (and corrected). */
/****************************************************************************/

void macro_test(MACRO_DATA *data, const char *new_name)
{
  FUNCNAME("macro_test");
 
  U_CHAR error_found = false;

#if DIM_OF_WORLD==1
  if(orientation(data)) {
    error_found = true;
    WARNING("Element orientation was corrected for some elements.\n");
  }
#endif

  if (error_found && new_name) {
    MSG("Attempting to write corrected macro data to file %s...\n", new_name);
    write_macro_data(data, new_name);
  }

  return;
}


#if 0   /* might be needed, if more DOFs are needed per element!!!!   */

/****************************************************************************/
/*  boundary_more_boundary_dofs():                                          */
/*  adds dof's at each edge, counts the edges                               */
/*  function pointer bdry only used in 3d!                                  */
/****************************************************************************/

static void fill_more_boundary_dofs(MESH *mesh,
                                    const BOUNDARY *(*bdry)(MESH *, int ))
{
  int        i, n_edges = 0, node = mesh->node[EDGE];
  MACRO_EL   *mel;
  DOF        *dof;

  for (mel = mesh->first_macro_el; mel; mel = mel->next)
  {
    for (i = 0; i < N_NEIGH; i++)
    {
      if (!mel->neigh[i] || (mel->neigh[i]->index < mel->index))
      {
	n_edges++;
	if (mesh->n_dof[EDGE])
	{
	  dof = mel->el->dof[node+i] = get_dof(mesh, EDGE);
	  if (mel->neigh[i])
	    mel->neigh[i]->el->dof[node+mel->opp_vertex[i]] = dof;
	}
      }  
    }
  }
  
  mesh->n_edges = n_edges;

  return;
}

#endif

#include "macro_common.c"

























