/****************************************************************************/
/* ALBERT:   an Adaptive multi Level finite element toolbox using           */
/*           Bisectioning refinement and Error control by Residual          */
/*           Techniques                                                     */
/*                                                                          */
/* file:     traverse_r.c                                                   */
/*                                                                          */
/*                                                                          */
/* description:                                                             */
/*           recursive mesh traversal - 1d routines:                        */
/*           fill_macro_info(), fill_elinfo()                               */
/*                                                                          */
/* history:                                                                 */
/*                                                                          */
/****************************************************************************/
/*                                                                          */
/*        authors: Alfred Schmidt and Kunibert G. Siebert                   */
/*                 Institut f"ur Angewandte Mathematik                      */
/*                 Albert-Ludwigs-Universit"at Freiburg                     */
/*                 Hermann-Herder-Str. 10                                   */
/*                 79104 Freiburg                                           */
/*                 Germany                                                  */
/*                                                                          */
/*        email:   alfred,kunibert@mathematik.uni-freiburg.de               */
/*                                                                          */
/*     (c) by A. Schmidt and K.G. Siebert (1996-99)                         */
/*                                                                          */
/****************************************************************************/

#include "albert.h"
#include "traverse_r_common.c"

/*--------------------------------------------------------------------------*/
/*   fill_macro_info:                                		            */
/*   ----------------                                   		    */
/*   fill EL_INFO structure for a macro element         		    */
/*--------------------------------------------------------------------------*/

void fill_macro_info(const MACRO_EL *mel, EL_INFO *el_info)
{
  FUNCNAME("fill_macro_info");
  EL       *nb;
  MACRO_EL *mnb;
  int      i, j;

  el_info->macro_el = mel;
  el_info->el       = mel->el;
  el_info->parent   = nil;
  el_info->level    = 0;

  if (el_info->fill_flag & FILL_COORDS)
  {
    for (i = 0; i < N_VERTICES; i++) 
    {
      if (mel->coord[i]) 
      {
	for (j = 0; j < DIM_OF_WORLD; j++)
	  el_info->coord[i][j] = mel->coord[i][j];
      }
      else
      {
	TEST_EXIT(0)("no mel->coord[%d]\n",i);
      }
    }
  }

  if (el_info->fill_flag & (FILL_NEIGH|FILL_OPP_COORDS))
  {
    REAL_D opp_coord;

    for (i = 0; i < N_NEIGH; i++) 
    {
      nb = nil;
      if ((mnb = mel->neigh[i])) 
      {
	if (el_info->fill_flag & FILL_OPP_COORDS)
	{ 
	  for (j = 0; j < DIM_OF_WORLD; j++)
	    opp_coord[j] = mnb->coord[1-i][j];
	}

	nb = mnb->el;

	while (nb->child[0])  /*--- make nb nearest element             ---*/
	{
	  if (el_info->fill_flag & FILL_OPP_COORDS)
	  { 
	    for (j = 0; j < DIM_OF_WORLD; j++)
	      opp_coord[j] = 0.5*(mel->coord[1-i][j] + opp_coord[j]);
	  }
	  nb = nb->child[i];
	}

	if (el_info->fill_flag & FILL_OPP_COORDS)
	{ 
	  for (j = 0; j < DIM_OF_WORLD; j++)
	    el_info->opp_coord[i][j] = opp_coord[j];
	}
      }
#if !NEIGH_IN_EL
      el_info->neigh[i] = nb;
      el_info->opp_vertex[i] = nb ? 1-i : -1;
#endif
    }
  }

  if (el_info->fill_flag & FILL_BOUND) 
  {
    for (i = 0; i < N_VERTICES; i++)
      el_info->bound[i] = mel->bound[i];
  }
  return;
}

/*--------------------------------------------------------------------------*/
/*   fill_el_info:                                    		            */
/*   ------------                                       		    */
/*   fill EL_INFO structure for one child of an element   		    */
/*--------------------------------------------------------------------------*/

void fill_elinfo(int ichild, const EL_INFO *el_info_old, EL_INFO *el_info)
{
  FUNCNAME("fill_elinfo");
  int     i, j;
  EL      *nb;
  EL      *el = el_info_old->el;
  FLAGS   fill_flag = el_info_old->fill_flag;

  TEST_EXIT(el->child[0])("no children?\n");
  TEST_EXIT((el_info->el = el->child[ichild]))("missing child %d?\n", ichild);

  el_info->macro_el  = el_info_old->macro_el;
  el_info->fill_flag = fill_flag;
  el_info->mesh      = el_info_old->mesh;
  el_info->parent    = el;
  el_info->level     = el_info_old->level + 1;

  if (fill_flag & FILL_COORDS)
  {
    const REAL_D *coord = el_info_old->coord;

    for (j = 0; j < DIM_OF_WORLD; j++)
    {
      el_info->coord[ichild][j] = coord[ichild][j];
      el_info->coord[1-ichild][j] = 0.5*(coord[0][j] + coord[1][j]);
    }
  }

  if (fill_flag & (FILL_NEIGH|FILL_OPP_COORDS))
  {
    REAL_D opp_coord;

    TEST_EXIT(fill_flag&FILL_COORDS)
      ("FILL_OPP_COORDS only with FILL_COORDS\n");

    for (i = 0; i < N_NEIGH; i++) 
    {
      if (i == ichild)
      {
	nb = el->child[1-ichild];
	if (nb && el_info->fill_flag & FILL_OPP_COORDS)
	{
	  for (j = 0; j < DIM_OF_WORLD; j++)
	    opp_coord[j] = el_info_old->coord[1-i][j];
	}
      }
      else
      {
	nb = NEIGH(el_info_old->el, el_info_old)[i];

	if (nb && el_info->fill_flag & FILL_OPP_COORDS)
	{
	  for (j = 0; j < DIM_OF_WORLD; j++)
	    opp_coord[j] = el_info_old->opp_coord[i][j];
	}
      }

      if (nb)
      {
	while (nb->child[0])  /*--- make nb nearest element              ---*/
	{
	  if (el_info->fill_flag & FILL_OPP_COORDS)
	  { 
	    for (j = 0; j < DIM_OF_WORLD; j++)
	      opp_coord[j] = 0.5*(el_info->coord[1-i][j] + opp_coord[j]);
	  }
	  nb = nb->child[i];
	}

	if (el_info->fill_flag & FILL_OPP_COORDS)
	{ 
	  for (j = 0; j < DIM_OF_WORLD; j++)
	    el_info->opp_coord[i][j] = opp_coord[j];
	}
      }
#if !NEIGH_IN_EL
      el_info->neigh[i] = nb;
      el_info->opp_vertex[i] = nb ? 1-i : -1;
#endif
    }
  }

  if (fill_flag & FILL_BOUND)
  {
    el_info->bound[ichild] = el_info_old->bound[ichild];
    el_info->bound[1-ichild] = INTERIOR;
  }
  return;
}



