/****************************************************************************/
/* ALBERT:   an Adaptive multi Level finite element toolbox using           */
/*           Bisectioning refinement and Error control by Residual          */
/*           Techniques                                                     */
/*                                                                          */
/* file:     Common/adapt.c                                                 */
/*                                                                          */
/*                                                                          */
/* description:  adaptive procedures for stationary and instationary        */
/*               problems                                                   */
/*                                                                          */
/* history:                                                                 */
/*                                                                          */
/*                                                                          */
/****************************************************************************/
/*                                                                          */
/*        authors: Alfred Schmidt and Kunibert G. Siebert                   */
/*                 Institut f"ur Angewandte Mathematik                      */
/*                 Albert-Ludwigs-Universit"at Freiburg                     */
/*                 Hermann-Herder-Str. 10                                   */
/*                 79104 Freiburg                                           */
/*                 Germany                                                  */
/*                                                                          */
/*        email:   alfred,kunibert@mathematik.uni-freiburg.de               */
/*                                                                          */
/*     (c) by A. Schmidt and K.G. Siebert (1996-1999)                       */
/*                                                                          */
/****************************************************************************/
#include "albert.h"

#include "time.h"
#define TIME_USED(f,s) ((double)((s)-(f))/(double)CLOCKS_PER_SEC)

/****************************************************************************/

/****************************************************************************/
/* adapt_mesh:   call marking(), refine(), and coarsen()                    */
/****************************************************************************/

static U_CHAR adapt_mesh(MESH *mesh, ADAPT_STAT *adapt)
{
  FUNCNAME("adapt_mesh");
  U_CHAR   flag = 0;
  U_CHAR   mark_flag;
  int      n_elements, iadmin;
  clock_t  first = clock();
  
  TEST_EXIT(adapt)("no ADAPT_STAT\n");

  if (adapt->marking)
    mark_flag = adapt->marking(mesh, adapt);
  else
    mark_flag = marking(mesh, adapt);

  if ((!adapt->coarsen_allowed))
    mark_flag &= MESH_REFINED;                    /* use refine mark only */

  if (adapt->build_before_refine)
    adapt->build_before_refine(mesh, mark_flag);

  n_elements = mesh->n_elements;

  if (mark_flag & MESH_REFINED)
    flag = refine(mesh);

  if (flag & MESH_REFINED)
  {
    n_elements = mesh->n_elements - n_elements;
    INFO(adapt->info,8)
      ("%d element%s refined, giving %d element%s\n",
       n_elements, n_elements > 1 ? "s" : "",
       mesh->n_elements, mesh->n_elements > 1 ? "s" : "");
    for (iadmin = 0; iadmin < mesh->n_dof_admin; iadmin++)
      INFO(adapt->info,7)("%d DOFs of admin <%s>\n", 
			  mesh->dof_admin[iadmin]->used_count, 
			  NAME(mesh->dof_admin[iadmin]));
  }
  else
    INFO(adapt->info,8)("no element refined\n");


  if (adapt->build_before_coarsen)
    adapt->build_before_coarsen(mesh, mark_flag);

  n_elements = mesh->n_elements;

  if (mark_flag & MESH_COARSENED)
    flag |= coarsen(mesh);

  if (flag & MESH_COARSENED)
  {
    n_elements -= mesh->n_elements;
    INFO(adapt->info,8)
      ("%d element%s coarsened, giving %d element%s\n",
       n_elements, n_elements > 1 ? "s" : "",
       mesh->n_elements, mesh->n_elements > 1 ? "s" : "");
    for (iadmin = 0; iadmin < mesh->n_dof_admin; iadmin++)
      INFO(adapt->info,7)("%d DOFs of dof_admin <%s>\n", 
			  mesh->dof_admin[iadmin]->used_count,
			  NAME(mesh->dof_admin[iadmin]));
  }
  else
    INFO(adapt->info,8)("no element coarsened\n");


  if (adapt->build_after_coarsen)
    adapt->build_after_coarsen(mesh, flag);

  INFO(adapt->info,6)("adapting mesh and build needed %.5lf seconds\n",
		      TIME_USED(first,clock()));

  return(flag);
}

/****************************************************************************/
/* adapt_method_stat: adapt for stationary problems                         */
/****************************************************************************/


void adapt_method_stat(MESH *mesh, ADAPT_STAT *adapt)
{
  FUNCNAME("adapt_method_stat");
  int      iter;
  REAL     est;
  clock_t  first;

  TEST_EXIT(mesh)("no MESH\n");
  TEST_EXIT(adapt)("no ADAPT_STAT\n");

  /* get solution on initial mesh */
  if (adapt->build_before_refine)  adapt->build_before_refine(mesh, 0);
  if (adapt->build_before_coarsen) adapt->build_before_coarsen(mesh, 0);
  if (adapt->build_after_coarsen)  adapt->build_after_coarsen(mesh, 0);
  if (adapt->solve)
  {
    first = clock();
    adapt->solve(mesh);
    INFO(adapt->info,8)("solution of discrete system needed %.5lf seconds\n",
			TIME_USED(first,clock()));
  }

  first = clock();
  est = adapt->estimate ? adapt->estimate(mesh, adapt) : 0.0;
  INFO(adapt->info,8)("estimation of the error needed %.5lf seconds\n",
		      TIME_USED(first,clock()));

  for (iter = 0;
       (est > adapt->tolerance) &&
	 ((adapt->max_iteration <= 0) || (iter < adapt->max_iteration));
       iter++)
  {
    if (adapt_mesh(mesh, adapt))
    {
      if (adapt->solve)
      {
	first = clock();
	adapt->solve(mesh);
	INFO(adapt->info,8)
	  ("solution of discrete system needed %.5lf seconds\n", 
	   TIME_USED(first,clock()));
      }
      first = clock();
      est = adapt->estimate ? adapt->estimate(mesh, adapt) : 0.0;
      INFO(adapt->info,8)("estimation of the error needed %.5lf seconds\n",
			  TIME_USED(first,clock()));
    }
    else
    {
      ERROR("no mesh adaption, but estimate above tolerance ???\n");
      break;
    }
    INFO(adapt->info, 4)("iterations: %d", iter);
    PRINT_INFO(adapt->info, 4)(", tolerance  %le", adapt->tolerance);
    PRINT_INFO(adapt->info, 4)(", estimate %le\n", est);
  }

  if ((iter > adapt->max_iteration) && (adapt->max_iteration > 0))
  {
    MSG("max_iterations REACHED: %d\n", adapt->max_iteration);
    MSG("prescribed tolerance    %le\n", adapt->tolerance);
    MSG("finished with estimate  %le\n", est);
  }
  else
  {
    INFO(adapt->info, 2)("no of iterations:       %d\n", iter);
    INFO(adapt->info, 2)("prescribed tolerance    %le\n", adapt->tolerance);
    INFO(adapt->info, 2)("finished with estimate  %le\n", est);
  }
  return;
}

/****************************************************************************/
/* adapt_method_instat():  adapt for timedependent problems                 */
/****************************************************************************/

static void explicit_time_strategy(MESH *mesh, ADAPT_INSTAT *adapt)
{
  FUNCNAME("explicit_time_strategy");
  ADAPT_STAT *adapt_s = adapt->adapt_space;

  /* first estimate must be computed before adapt_mesh() is called */
  if (adapt->time <= adapt->start_time) {
    if (adapt_s->estimate)
      adapt_s->estimate(mesh, adapt_s);
  }

  adapt->time += adapt->timestep;
  if (adapt->set_time)
    adapt->set_time(mesh, adapt);

  INFO(adapt->info,6)("time = %.6lf, timestep = %.6lf\n",
		      adapt->time, adapt->timestep);

  adapt_mesh(mesh, adapt_s);

  if (adapt_s->solve)
    adapt_s->solve(mesh);

  if (adapt_s->estimate)
    adapt_s->estimate(mesh, adapt_s);           /* return value is not used */

  return;
}


static void implicit_time_strategy(MESH *mesh, ADAPT_INSTAT *adapt)
{
  FUNCNAME("implicit_time_strategy");
  int     iter, iter_s;
  ADAPT_STAT *adapt_s = adapt->adapt_space;
  REAL    err_space, err_time, space_err_limit, time_err_limit, time_err_low;

  space_err_limit = adapt->tolerance * adapt->rel_space_error;
  time_err_limit  = adapt->tolerance * adapt->rel_time_error;
  time_err_low    = adapt->tolerance * adapt->rel_time_error
                     * adapt->time_theta_2;
  iter = iter_s = 0;
  err_time = 0.0;

  do
  {
    adapt->time += adapt->timestep;
    if (adapt->set_time)
      adapt->set_time(mesh, adapt);

    INFO(adapt->info,6)("time = %.6lf, try timestep = %.6lf\n",
			adapt->time, adapt->timestep);

    if (adapt_s->build_before_refine)  adapt_s->build_before_refine(mesh, 0);
    if (adapt_s->build_before_coarsen) adapt_s->build_before_coarsen(mesh, 0);
    if (adapt_s->build_after_coarsen)  adapt_s->build_after_coarsen(mesh, 0);
    if (adapt_s->solve)                adapt_s->solve(mesh);
    err_space = adapt_s->estimate ? adapt_s->estimate(mesh, adapt_s) : 0.0;
    if (adapt->get_time_est)
      err_time  = adapt->get_time_est(mesh, adapt);

    iter++;
    if (iter > adapt->max_iteration) break;

    if (err_time > time_err_limit)
    {
      adapt->time -= adapt->timestep;
      adapt->timestep *= adapt->time_delta_1;
      continue;
    }

    do
    {
      if (adapt_mesh(mesh, adapt_s))
      {
	adapt_s->solve(mesh);
	err_space = adapt_s->estimate ? adapt_s->estimate(mesh, adapt_s) : 0.0;
	if (adapt->get_time_est) {
	  err_time  = adapt->get_time_est(mesh, adapt);
	  if (err_time > time_err_limit)
	  {
	    adapt->time -= adapt->timestep;
	    adapt->timestep *= adapt->time_delta_1;
	    break;
	  }
	}
      }

      iter_s++;
      if (iter_s > adapt_s->max_iteration) break;

    } while(err_space > space_err_limit);

  } while(err_time > time_err_limit);

  if (adapt->get_time_est)
    if (err_time <= time_err_low)
    {
      adapt->timestep *= adapt->time_delta_2;
    }
  
  return;
}


static void one_timestep(MESH *mesh, ADAPT_INSTAT *adapt)
{
  FUNCNAME("one_timestep");

  switch(adapt->strategy)
  {
  case 0:
    explicit_time_strategy(mesh, adapt);
    break;

  case 1:
    implicit_time_strategy(mesh, adapt);
    break;

  default:
    MSG("unknown adapt->strategy = %d; use explicit strategy\n");
    explicit_time_strategy(mesh, adapt);
  }

  return;
}

/****************************************************************************/

void adapt_method_instat(MESH *mesh, ADAPT_INSTAT *adapt)
{
  FUNCNAME("adapt_method_instat");

  TEST_EXIT(adapt)("no ADAPT_INSTAT\n");

/****************************************************************************/
/*  adaptation of the initial grid: done by adapt_method_stat               */
/****************************************************************************/

  adapt->time = adapt->start_time;
  if (adapt->set_time) adapt->set_time(mesh, adapt);

  adapt->adapt_initial->tolerance
    = adapt->tolerance * adapt->rel_initial_error;
  adapt->adapt_space->tolerance
    = adapt->tolerance * adapt->rel_space_error;

  adapt_method_stat(mesh, adapt->adapt_initial);
  if (adapt->close_timestep)
    adapt->close_timestep(mesh, adapt);

/****************************************************************************/
/*  adaptation of timestepsize and mesh: done by one_timestep               */
/****************************************************************************/

  while (adapt->time < adapt->end_time)
  {
    if (adapt->init_timestep)
      adapt->init_timestep(mesh, adapt);

    if (adapt->one_timestep)
      adapt->one_timestep(mesh, adapt);
    else
      one_timestep(mesh, adapt);

    if (adapt->close_timestep)
      adapt->close_timestep(mesh, adapt);
  }
}

/****************************************************************************/
/****************************************************************************/
/****************************************************************************/
/*   marking functions!!!                                                   */
/****************************************************************************/
/****************************************************************************/
/****************************************************************************/

static REAL       (*get_el_est)(EL *el);
static REAL       (*get_el_estc)(EL *el);
static int        el_mark, el_mark_c;

/****************************************************************************/
/****************************************************************************/
/****  marking functions for several adaptive strategies                 ****/
/****************************************************************************/
/****************************************************************************/

static S_CHAR   g_mark_refine = 0;
static S_CHAR   g_mark_coarse = 0;
static REAL     err_max, err_sum;
static int      mark_flag;
static REAL     mark_r_limit, mark_c_limit;


static void marking_fct_1(const EL_INFO *el_info)
{
  el_info->el->mark = g_mark_refine;
  mark_flag = 1;
  el_mark++;

  return;
}


static void marking_fct_2(const EL_INFO *el_info)
{
  REAL   error = get_el_est(el_info->el);

  if (error > mark_r_limit)
  {
    el_info->el->mark = g_mark_refine;
    mark_flag = 1;
    el_mark++;
  }
  else if (error <= mark_c_limit)
  {
    if (!get_el_estc || (error + get_el_estc(el_info->el) <= mark_c_limit))
    {
      el_info->el->mark = g_mark_coarse;
      mark_flag = 1;
      el_mark_c++;
    }
  }
  return;
}


static void marking_fct_3(const EL_INFO *el_info)
{
  REAL   error = get_el_est(el_info->el);

  if (error > mark_r_limit)
  {
    el_info->el->mark = g_mark_refine;
    mark_flag = 1;
    el_mark++;
  }
  else if (error <= mark_c_limit)
  {
    if (!get_el_estc || (error + get_el_estc(el_info->el) <= mark_c_limit))
    {
      el_info->el->mark = g_mark_coarse;
      mark_flag = 1;
      el_mark_c++;
    }
  }
  return;
}



static REAL GERS_sum;

static void marking_fct_4(const EL_INFO *el_info)
{
  REAL   error = get_el_est(el_info->el);

  if (error > mark_r_limit)
  {
    GERS_sum += error;
    el_info->el->mark = g_mark_refine;
    mark_flag = 1;
    el_mark++;
  }
  return;
}

static void marking_fct_4c(const EL_INFO *el_info)
{
  REAL   error;

  if (el_info->el->mark <= 0)
  {
    error = get_el_est(el_info->el);
    if (get_el_estc) error += get_el_estc(el_info->el);

    if (error <= mark_c_limit)
    {
      GERS_sum += error;
      el_info->el->mark = g_mark_coarse;
      mark_flag = 1;
      el_mark_c++;
    }
    else
      el_info->el->mark = 0;
   }

  return;
}

/****************************************************************************/

int marking(MESH *mesh, ADAPT_STAT *adapt)
{
  FUNCNAME("marking");
  static REAL old_err_sum = 0.0;
  REAL        improv, Ltheta, redfac, wanted;
  REAL        eps_p, ES_theta_p,ES_theta_c_p;
  REAL        MS_gamma_p, MS_gamma_c_p, GERS_gamma;

  TEST_EXIT(adapt)("no adapt_stat\n");
  TEST_EXIT((get_el_est = adapt->get_el_est))("no adapt->get_el_est\n");
  TEST_EXIT(adapt->p >= 1.0)("ADAPT_STAT->p < 1\n");
  
  get_el_estc = adapt->get_el_estc;
  g_mark_refine = adapt->refine_bisections;
  g_mark_coarse = -adapt->coarse_bisections;

  mark_flag = 0;
  el_mark = el_mark_c = 0;

  eps_p = pow(adapt->tolerance, adapt->p);
  err_sum = pow(adapt->err_sum, adapt->p);

  err_max = adapt->err_max;

  switch(adapt->strategy)
  {
  case 1:
    mesh_traverse(mesh, -1, CALL_LEAF_EL, marking_fct_1);
    break;

  case 2:
    MS_gamma_p = pow(adapt->MS_gamma, adapt->p);
    if (adapt->coarsen_allowed)
      MS_gamma_c_p = pow(adapt->MS_gamma_c, adapt->p);
    else
      MS_gamma_c_p = -1.0;

    mark_r_limit = MS_gamma_p * err_max;
    if (adapt->coarsen_allowed)
      mark_c_limit = MS_gamma_c_p * err_max;
   
    mesh_traverse(mesh, -1, CALL_LEAF_EL, marking_fct_2);
    break;

  case 3:
    ES_theta_p = pow(adapt->ES_theta, adapt->p);
    mark_r_limit = ES_theta_p * eps_p / mesh->n_elements;
    if (adapt->coarsen_allowed)
    {
      ES_theta_c_p = pow(adapt->ES_theta_c, adapt->p);
      mark_c_limit = ES_theta_c_p * eps_p / mesh->n_elements;
    }
    else
      mark_c_limit = -1.0;

    INFO(adapt->info, 2)
      ("start mark_limits: %.3le %.3le nt=%d\n",
       mark_r_limit, mark_c_limit, mesh->n_elements);

    mesh_traverse(mesh, -1, CALL_LEAF_EL, marking_fct_3);
    break;

  case 4:
/****************************************************************************/
/*  Graranteed error reduction strategy                                     */
/*  using extrapolation for theta                                           */
/****************************************************************************/

    Ltheta= pow(1.0 - adapt->GERS_theta_star, adapt->p);

    if (err_sum < old_err_sum) 
    {
      improv = err_sum/old_err_sum;
      wanted = 0.8 * eps_p /err_sum;
      redfac = MIN((1.0-wanted)/(1.0-improv),1.0);
      redfac = MAX(redfac, 0.0);

      if (redfac < 1.0) 
      {
	Ltheta = Ltheta*redfac;
	INFO(adapt->info, 1)
	  ("GERS: use extrapolated theta_star = %lf\n",
	   pow(Ltheta, 1.0/adapt->p));
      }
    }
    old_err_sum = err_sum;

    GERS_gamma = 1.0;

    if (Ltheta > 0)
    {
      do
      {
	GERS_sum = 0.0;
	GERS_gamma -= adapt->GERS_nu;
	mark_r_limit = GERS_gamma * err_max;
	mesh_traverse(mesh, -1, CALL_LEAF_EL, marking_fct_4);
      }
      while ((GERS_gamma > 0) && (GERS_sum < Ltheta * err_sum));
    }

    INFO(adapt->info, 2)
      ("GERS refinement with gamma = %.3lf\n", GERS_gamma);

    if (adapt->coarsen_allowed)
    {
      GERS_gamma = 0.3;
      Ltheta = adapt->GERS_theta_c * eps_p;

      do
      {
	GERS_sum = 0.0;
	GERS_gamma -= adapt->GERS_nu;
	mark_c_limit = GERS_gamma * err_max;
	mesh_traverse(mesh, -1, CALL_LEAF_EL, marking_fct_4c);

	INFO(adapt->info, 6)
	  ("coarse loop: gamma=%.3e, sum=%.3e, limit=%.3e\n",
	   GERS_gamma, GERS_sum, Ltheta);
      }
      while (GERS_sum > Ltheta);

      INFO(adapt->info, 2)
	("GERS coarsening with gamma = %.3lf\n", GERS_gamma);
    }
    break;
  }

  INFO(adapt->info, 4)
    ("%d elements marked for refinement\n", el_mark);
  INFO(adapt->info, 4)
    ("%d elements marked for coarsening\n", el_mark_c);

  mark_flag = 0;
  if (el_mark) mark_flag = 1;
  if (el_mark_c) mark_flag |= 2;

  return(mark_flag); 
}

/****************************************************************************/


/****************************************************************************/
/*  get_adapt_stat(name, prefix, info, adapt_stat):                         */
/*                                                                          */
/*  if adapt_stat is non nil:                                               */
/*  allocates a new adpat_stat structure adapt and predefines the           */
/*  parameter entries of this structure;                                    */
/*  if name is non nil adpat->name = name                                   */
/*                                                                          */
/*  Now, either for adapt_stat or the newly allocated adapt_stat the        */
/*  parameter entries of the structures are initialized by GET_PARAMETER()  */
/*  if prefix is non nil                                                    */
/*  the info parameter is the the first argument of GET_PARAMETER()         */
/*                                                                          */
/*  the return value is a pointer to the initialized adapt_stat structure   */
/*                                                                          */
/*  for the initialization of entry we use the call                         */
/*    GET_PARAMETER(info, "prefix->entry", "%{d,f}", &adapt->entry          */
/*  !! You do not have to define all parameters in your init file !!        */
/*     entries that are not defined in the init file have so standard value */
/*                                                                          */
/* !!!  before a call of get_adapt_stat() with non nil second argument  !!! */
/* !!!  you have to initialize your parameters via init_parameters()    !!! */
/****************************************************************************/


ADAPT_STAT *get_adapt_stat(const char *name, const char *prefix, int info,
			   ADAPT_STAT *adapt_stat)
{
  FUNCNAME("get_adapt_stat");
  ADAPT_STAT  adapt_stand = {nil, 1.0, 2, 30, 2, nil, nil, nil, nil,
                             nil, 0.0, 0.0, nil, nil, nil, nil,
                             DIM, 0, DIM, 1, 0.5, 0.1, 0.9, 0.2,
			     0.6, 0.1, 0.1 };
  char  key[128], *append;
  ADAPT_STAT *adapt;

  if (adapt_stat)
  {
    adapt = adapt_stat;
  }
  else
  {
    adapt = MEM_ALLOC(1, ADAPT_STAT);
    *adapt = adapt_stand;
    if (name)
      adapt->name = strdup(name);
    if (!adapt->name  && prefix)
      adapt->name = strdup(prefix);
  }

  if (!prefix)
    return(adapt);

  strcpy(key, prefix);
  append = key;
  while (*append) append++;

  strcpy(append, "->tolerance");
  GET_PARAMETER(info, key, "%f", &adapt->tolerance);
  strcpy(append, "->p");
  GET_PARAMETER(info, key, "%f", &adapt->p);
  strcpy(append, "->max_iteration");
  GET_PARAMETER(info, key, "%d", &adapt->max_iteration);
  strcpy(append, "->info");
  GET_PARAMETER(info, key, "%d", &adapt->info);

  strcpy(append, "->refine_bisections");
  GET_PARAMETER(info, key, "%d", &adapt->refine_bisections);
  strcpy(append, "->coarsen_allowed");
  GET_PARAMETER(info, key, "%d", &adapt->coarsen_allowed);
  strcpy(append, "->coarse_bisections");
  GET_PARAMETER(info, key, "%d", &adapt->coarse_bisections);

  strcpy(append, "->strategy");
  GET_PARAMETER(info, key, "%d", &adapt->strategy);
  if (adapt->strategy == 2) 
  {
    strcpy(append, "->MS_gamma");
    GET_PARAMETER(info, key, "%f", &adapt->MS_gamma);
    strcpy(append, "->MS_gamma_c");
    GET_PARAMETER(info, key, "%f", &adapt->MS_gamma_c);
  }
  if (adapt->strategy == 3) 
  {
    strcpy(append, "->ES_theta");
    GET_PARAMETER(info, key, "%f", &adapt->ES_theta);
    strcpy(append, "->ES_theta_c");
    GET_PARAMETER(info, key, "%f", &adapt->ES_theta_c);
  }
  if (adapt->strategy == 4)
  {
    strcpy(append, "->GERS_theta_star");
    GET_PARAMETER(info, key, "%f", &adapt->GERS_theta_star);
    strcpy(append, "->GERS_nu");
    GET_PARAMETER(info, key, "%f", &adapt->GERS_nu);
    strcpy(append, "->GERS_theta_c");
    GET_PARAMETER(info, key, "%f", &adapt->GERS_theta_c);
  }

  return(adapt);
}

/****************************************************************************/
/*  get_adapt_instat(name, prefix, info, adapt_instat):                     */
/*                                                                          */
/*  if adapt_instat is non nil:                                             */
/*  allocates a new adpat_instat structure adapt and predefines the         */
/*  parameter entries of this structure;                                    */
/*  if name is non nil adpat->name = name                                   */
/*                                                                          */
/*  Now, either for adapt_instat or the newly allocated adapt_instat the    */
/*  parameter entries of the structures are initialized by GET_PARAMETER()  */
/*  if prefix is non nil                                                    */
/*  the info parameter is the the first argument of GET_PARAMETER()         */
/*                                                                          */
/*  the return value is a pointer to the initialized adapt_stat structure   */
/*                                                                          */
/*  for the initialization of entry we use the call                         */
/*    GET_PARAMETER(info, "prefix->entry", "%{d,f}", &adapt->entry)         */
/*  sub-structures adapt_initial/adapt_space are initialized by             */
/*    get_adapt_stat(nil, "prefix->initial", info, adapt->adapt_initial);   */
/*    get_adapt_stat(nil, "prefix->space", info, adapt->adapt_space);       */
/*                                                                          */
/*  !! You do not have to define all parameters in your init file !!        */
/*     entries that are not defined in the init file have so standard value */
/*                                                                          */
/* !!!  before a call of get_adapt_stat() with non nil second argument  !!! */
/* !!!  you have to initialize your parameters via init_parameters()    !!! */
/****************************************************************************/

ADAPT_INSTAT *get_adapt_instat(const char *name, const char *prefix, int info,
			       ADAPT_INSTAT *adapt_instat)
{
  FUNCNAME("get_adapt_instat");
  ADAPT_INSTAT adapt_stand = {nil,
			      {{nil,
			       1.0, 2.0, 1, 2, nil, nil, nil, nil,
			       nil, 0.0, 0.0, nil, nil, nil, nil,
			       DIM, 0, DIM, 2, 0.5, 0.1, 0.9, 0.2,
			       0.6, 0.1, 0.1}},
			      {{nil,
			       1.0, 2.0, 1, 2, nil, nil, nil, nil,
			       nil, 0.0, 0.0, nil, nil, nil, nil,
			       DIM, 1, DIM, 2, 0.5, 0.1, 0.9, 0.2,
			       0.6, 0.1, 0.1}},
			      0.0, 0.0, 1.0, 0.01, nil, nil, nil, nil, nil,
			      0, 0, 1.0, 0.1, 0.4, 0.4, 1.0, 0.3, 
			      0.7071, 1.4142, 8};
  char  key[128], *append;
  ADAPT_INSTAT *adapt;

  if (adapt_instat)
  {
    adapt = adapt_instat;
  }
  else
  {
    adapt = MEM_ALLOC(1, ADAPT_INSTAT);
    *adapt = adapt_stand;
    if (name)
      adapt->name = strdup(name);
    if (!adapt->name  && prefix)
      adapt->name = strdup(prefix);
  }

  if (!prefix)
    return(adapt);

  sprintf(key, "%s initial", adapt->name);
  adapt->adapt_initial->name = strdup(key);
  sprintf(key, "%s space", adapt->name);
  adapt->adapt_space->name = strdup(key);

  strcpy(key, prefix);
  append = key;
  while (*append) append++;

  strcpy(append, "->initial");
  get_adapt_stat(nil, key, info, adapt->adapt_initial);
  strcpy(append, "->space");
  get_adapt_stat(nil, key, info, adapt->adapt_space);

  strcpy(append, "->start_time");
  GET_PARAMETER(info, key, "%f", &adapt->start_time);
  adapt->time = adapt->start_time;
  strcpy(append, "->end_time");
  GET_PARAMETER(info, key, "%f", &adapt->end_time);
  strcpy(append, "->timestep");
  GET_PARAMETER(info, key, "%f", &adapt->timestep);
  strcpy(append, "->strategy");
  GET_PARAMETER(info, key, "%d", &adapt->strategy);
  strcpy(append, "->max_iteration");
  GET_PARAMETER(info, key, "%d", &adapt->max_iteration);
  strcpy(append, "->tolerance");
  GET_PARAMETER(info, key, "%f", &adapt->tolerance);
  strcpy(append, "->rel_initial_error");
  GET_PARAMETER(info, key, "%f", &adapt->rel_initial_error);
  strcpy(append, "->rel_space_error");
  GET_PARAMETER(info, key, "%f", &adapt->rel_space_error);
  strcpy(append, "->rel_time_error");
  GET_PARAMETER(info, key, "%f", &adapt->rel_time_error);
  strcpy(append, "->time_theta_1");
  GET_PARAMETER(info, key, "%f", &adapt->time_theta_1);
  strcpy(append, "->time_theta_2");
  GET_PARAMETER(info, key, "%f", &adapt->time_theta_2);
  strcpy(append, "->time_delta_1");
  GET_PARAMETER(info, key, "%f", &adapt->time_delta_1);
  strcpy(append, "->time_delta_2");
  GET_PARAMETER(info, key, "%f", &adapt->time_delta_2);
  strcpy(append, "->info");
  GET_PARAMETER(info, key, "%d", &adapt->info);

  return(adapt);
}
