#include "albert.h"

/*--------------------------------------------------------------------------*/
/*  matrix-vector multiplication, standard preconditioning and standard     */
/*  solver interface to OEM-library for scalar valued problems              */
/*                                                                          */
/*  A u = f, where                                                          */
/*  A is stored in a DOF_MATRIX, u and f in DOF_REAL_VECs                   */
/*--------------------------------------------------------------------------*/

struct data
{
  const FE_SPACE      *fe_space;
  const DOF_ADMIN     *admin;
  const DOF_MATRIX    *matrix;
  const DOF_SCHAR_VEC *bound;

  struct data         *next;
};

/*--------------------------------------------------------------------------*/
/*  diaoganal preconditioning for scalar problems                           */
/*--------------------------------------------------------------------------*/

static void diag_precon_s(void *ud, int dim, REAL *r)
{
  FUNCNAME("diag_precon_s");
  struct data  *data = ud;
  MATRIX_ROW   **row = data->matrix->matrix_row;
  const S_CHAR *b = data->bound ? data->bound->vec : nil;

  TEST_EXIT(dim == data->admin->size_used)
    ("argument dim != data->admin->size_used\n");

  if (b)
  {
    FOR_ALL_DOFS(data->admin, 
		 if (b[dof] <= 0  && row[dof]) r[dof] /= row[dof]->entry[0]);
  }
  else
  {
    FOR_ALL_DOFS(data->admin, 
		 if (row[dof]) r[dof] /= row[dof]->entry[0]);
  }
  return;
}

/*--------------------------------------------------------------------------*/
/*   y <- A x                                                               */
/*--------------------------------------------------------------------------*/

int mat_vec_s(void *ud, int dim, const REAL *x, REAL *y)
{
  FUNCNAME("mat_vec_s");
  DOF_REAL_VEC  dof_x = {nil, nil, "x", 0, nil, nil, nil};
  DOF_REAL_VEC  dof_y = {nil, nil, "y", 0, nil, nil, nil};
  struct data  *data = ud;
  const S_CHAR *b = data->bound ? data->bound->vec : nil;

  TEST_EXIT(dim == data->admin->size_used)
    ("argument dim != data->admin->size_used\n");

  dof_x.fe_space = dof_y.fe_space = data->fe_space;
  dof_x.size = dof_y.size = dim;
  dof_x.vec = (REAL *) x;
  dof_y.vec = y;

  dof_mv(NoTranspose, data->matrix, &dof_x, &dof_y);

  if (b)
  {
    FOR_ALL_DOFS(data->admin, if (b[dof] > 0)  y[dof] = x[dof]);
  }
  return(0);
}

/*--------------------------------------------------------------------------*/
/*   y <- A^T x                                                             */
/*--------------------------------------------------------------------------*/

static int mat_T_vec_s(void *ud, int dim, const REAL *x, REAL *y)
{
  FUNCNAME("mat_T_vec_s");
  DOF_REAL_VEC  dof_x = {nil, nil, "x", 0, nil, nil, nil};
  DOF_REAL_VEC  dof_y = {nil, nil, "y", 0, nil, nil, nil};

  struct data  *data = ud;
  const S_CHAR *b = data->bound ? data->bound->vec : nil;

  TEST_EXIT(dim == data->admin->size_used)
    ("argument dim != data->admin->size_used\n");

  dof_x.fe_space = dof_y.fe_space = data->fe_space;
  dof_x.size = dof_y.size = dim;
  dof_x.vec = (REAL *) x;
  dof_y.vec = y;

  dof_mv(Transpose, data->matrix, &dof_x, &dof_y);

  if (b)
  {
    ERROR_EXIT("don't know how to handle bound vector???\n");
  }

  return(0);
}

/*--------------------------------------------------------------------------*/
/*---  interface to ofm solver for scalar valued problems  -----------------*/
/*--------------------------------------------------------------------------*/
#include <oem.h>

const PRECON *get_BPX_precon_s(void);

int oem_solve_s(const DOF_MATRIX *A, const DOF_REAL_VEC *f, 
		const DOF_SCHAR_VEC *bound, DOF_REAL_VEC *u, 
		OEM_SOLVER solver, REAL tol, int icon,
		int restart, int max_iter, int info)
{
  FUNCNAME("oem_solve_s");
  int          dim, iter = -1;
  size_t       size;
  OEM_DATA     oem = {nil};
  struct data  data;
  const PRECON *precon = nil;
  WORKSPACE    ws = {0, nil};

  dim = A->fe_space->admin->size_used;
  FOR_ALL_FREE_DOFS(A->fe_space->admin, 
		    if (dof < dim)
		      u->vec[dof] = f->vec[dof] = 0.0);

  data.fe_space = A->fe_space;
  data.admin    = A->fe_space->admin;
  data.matrix   = A;
  data.bound    = bound;
  
  oem.mat_vec = mat_vec_s;
  oem.mat_vec_data = &data;
  switch(icon)
  {
  case 0: 
    oem.left_precon = nil;
    break;
  case 2:
    precon = get_HB_precon_s();
    break;
  case 3:
    precon = get_BPX_precon_s();
    break;
  default:
    oem.left_precon = diag_precon_s;
    oem.left_precon_data = &data;
    break;
  }

  if (precon)
  {
    
    if ((oem.left_precon_data = precon->init_precon(A, bound, info)))
      oem.left_precon = precon->precon;
    else
    {
      oem.left_precon = diag_precon_s;
      oem.left_precon_data = &data;
      precon = nil;
    }
  }

  oem.ws = &ws;

  oem.tolerance = tol;
  oem.restart   = 0;
  oem.max_iter  = max_iter;
  oem.info      = info;

  switch (solver)
  {
  case BiCGStab:
    REALLOC_WORKSPACE(&ws, 5*dim*sizeof(REAL));
    iter = oem_bicgstab(&oem, dim, f->vec, u->vec);
    break;
  case CG:
    REALLOC_WORKSPACE(&ws, 3*dim*sizeof(REAL));
    iter = oem_cg(&oem, dim, f->vec, u->vec);
    break;
  case GMRes:
    oem.restart = restart = MAX(0, MIN(restart, dim));
    size = ((restart+2)*dim + restart*(restart+4))*sizeof(REAL);
    REALLOC_WORKSPACE(&ws, size);
    iter = oem_gmres(&oem, dim, f->vec, u->vec);
    break;
  case ODir:
    REALLOC_WORKSPACE(&ws, 6*dim*sizeof(REAL));
    iter = oem_odir(&oem, dim, f->vec, u->vec);
    break;
  case ORes:
    REALLOC_WORKSPACE(&ws, 7*dim*sizeof(REAL));
    iter = oem_ores(&oem, dim, f->vec, u->vec);
    break;
  default:
    ERROR_EXIT("unknown OEM solver %d\n", (int) solver);
  }

  CLEAR_WORKSPACE(&ws);

  if (precon) precon->exit_precon(oem.left_precon_data);

  return(iter);
}

void *init_mat_vec_s(const DOF_MATRIX *A, const DOF_SCHAR_VEC *bound)
{
  FUNCNAME("init_mat_vec_s");
  static struct data *first = nil;
  struct data        *data;

  for (data = first; data; data = data->next)
    if (data->matrix == A  &&  data->bound == bound)  break;

  if (!data) 
  {
    data = MEM_ALLOC(1, struct data);

    data->fe_space = A->fe_space;
    data->admin    = A->fe_space->admin;
    data->matrix   = A;
    data->bound    = bound;

    data->next     = first;
    first          = data;
  }

  return(data);
}


/*--------------------------------------------------------------------------*/
/*---  initialize diagonal preconditioning  --------------------------------*/
/*--------------------------------------------------------------------------*/

static void *init_diag_precon_s(const DOF_MATRIX *A, 
				const DOF_SCHAR_VEC *bound, int info)
{
  FUNCNAME("init_diag_precon_s");
  static struct data *first = nil;
  struct data        *data;

  for (data = first; data; data = data->next)
    if (data->matrix == A  &&  data->bound == bound)  break;

  if (!data) 
  {
    data = MEM_ALLOC(1, struct data);

    data->fe_space = A->fe_space;
    data->admin    = A->fe_space->admin;
    data->matrix   = A;
    data->bound    = bound;

    data->next     = first;
    first          = data;
  }
  return(data);
}

static void exit_diag_precon_s(void *data){}

const PRECON *get_diag_precon_s(void)
{
  static const PRECON precon = {init_diag_precon_s,
				diag_precon_s,
				exit_diag_precon_s};
  return(&precon);
}
