/****************************************************************************/
/* ALBERT:   an Adaptive multi Level finite element toolbox using           */
/*           Bisectioning refinement and Error control by Residual          */
/*           Techniques                                                     */
/****************************************************************************/
/*        authors: Alfred Schmidt and Kunibert G. Siebert                   */
/*                 Institut f"ur Angewandte Mathematik                      */
/*                 Albert-Ludwigs-Universit"at Freiburg                     */
/*                 Hermann-Herder-Str. 10                                   */
/*                 79104 Freiburg                                           */
/*                 Germany                                                  */
/*                                                                          */
/*        email:   alfred,kunibert@mathematik.uni-freiburg.de               */
/*                                                                          */
/*     (c) by A. Schmidt and K.G. Siebert (1996-2000)                       */
/****************************************************************************/

/****************************************************************************/
/*  solver for elliptic model problem                                       */
/*                           -Delta u = f                                   */
/****************************************************************************/
#include <albert.h>

/****************************************************************************/
/*  function for displaying mesh, discrete solution, and/or estimate        */
/*  defined in graphics.c                                                   */
/****************************************************************************/
void graphics(MESH *mesh, DOF_REAL_VEC *u_h, REAL (*get_est)(EL *el));

/****************************************************************************/
/* global variables: finite element space, discrete solution                */
/*                   load vector and system matrix                          */
/****************************************************************************/

static const FE_SPACE *fe_space;         /* initialized by main()           */
static DOF_REAL_VEC   *u_h = nil;        /* initialized by main()           */
static DOF_REAL_VEC   *f_h = nil;        /* initialized by main()           */
static DOF_MATRIX     *matrix = nil;     /* initialized by main()           */

/****************************************************************************/
/* For test purposes: exact solution and its gradient (optional)            */
/****************************************************************************/

static REAL u(const REAL_D x)
{
  return(exp(-10.0*SCP_DOW(x,x)));
}

static const REAL *grd_u(const REAL_D x)
{
  static REAL_D grd;
  REAL          ux = exp(-10.0*SCP_DOW(x,x));

  grd[0] = -20.0*x[0]*ux;
  grd[1] = -20.0*x[1]*ux;
#if DIM_OF_WORLD == 3
  grd[2] = -20.0*x[2]*ux;
#endif
  return(grd);
}

/****************************************************************************/
/* problem data: right hand side, boundary values                           */
/****************************************************************************/

static REAL g(const REAL_D x)              /* boundary values, not optional */
{
  return(u(x));
}

static REAL f(const REAL_D x)              /* -Delta u, not optional        */
{
  REAL  r2 = SCP_DOW(x,x), ux  = exp(-10.0*r2);
  return(-(400.0*r2 - 20.0*DIM)*ux);
}


/****************************************************************************/
/* build(): assemblage of the linear system: matrix, load vector,           */
/*          boundary values, called by adapt_method_stat()                  */
/*          on the first call initialize u_h, f_h, matrix and information   */
/*          for assembling the system matrix                                */
/*                                                                          */
/* struct op_info: structure for passing information from init_element() to */
/*                 LALt()                                                   */
/* init_element(): initialization on the element; calculates the            */
/*                 coordinates and |det DF_S| used by LALt; passes these    */
/*                 values to LALt via user_data,                            */
/*                 called on each element by update_matrix()                */
/* LALt():         implementation of -Lambda id Lambda^t for -Delta u,      */
/*                 called by update_matrix() after init_element()           */
/****************************************************************************/

struct op_info
{
  REAL_D  Lambda[DIM+1];    /*  the gradient of the barycentric coordinates */
  REAL    det;              /*  |det D F_S|                                 */
};

static void init_element(const EL_INFO *el_info, const QUAD *quad[3], void *ud)
{
  struct op_info *info = ud;

  info->det = el_grd_lambda(el_info, info->Lambda);
  return;
}

static const REAL (*LALt(const EL_INFO *el_info, const QUAD *quad, 
			 int iq, void *ud))[DIM+1]
{
  struct op_info *info = ud;
  int            i, j, k;
  static REAL    LALt[DIM+1][DIM+1];

  for (i = 0; i <= DIM; i++)
    for (j = i; j <= DIM; j++)
    {
      for (LALt[i][j] = k = 0; k < DIM_OF_WORLD; k++)
	LALt[i][j] += info->Lambda[i][k]*info->Lambda[j][k];
      LALt[i][j] *= info->det;
      LALt[j][i] = LALt[i][j];
    }

  return((const REAL (*)[DIM+1]) LALt);
}

static void build(MESH *mesh, U_CHAR flag)
{
  FUNCNAME("build");
  static const EL_MATRIX_INFO *matrix_info = nil;
  const QUAD                  *quad;

  dof_compress(mesh);
  MSG("%d DOFs for %s\n", fe_space->admin->size_used, fe_space->name);

  if (!matrix_info)           /* information for matrix assembling         */
  {
    OPERATOR_INFO  o_info = {nil};

    o_info.row_fe_space   = o_info.col_fe_space = fe_space;
    o_info.init_element   = init_element;
    o_info.LALt           = LALt;
    o_info.LALt_pw_const  = true;        /* pw const. assemblage is faster */
    o_info.LALt_symmetric = true;        /* symmetric assemblage is faster */
    o_info.use_get_bound  = true;        /* Dirichlet boundary conditions! */
    o_info.user_data = MEM_ALLOC(1, struct op_info);         /* user data! */
    o_info.fill_flag = CALL_LEAF_EL|FILL_COORDS;

    matrix_info = fill_matrix_info(&o_info, nil);
  }

  clear_dof_matrix(matrix);                /* assembling of matrix         */
  update_matrix(matrix, matrix_info);

  dof_set(0.0, f_h);                       /* assembling of load vector    */
  quad = get_quadrature(DIM, 2*fe_space->bas_fcts->degree - 2);
  L2scp_fct_bas(f, quad, f_h);

  dirichlet_bound(g, f_h, u_h, nil);           /*  boundary values         */
  return;
}

/****************************************************************************/
/* solve(): solve the linear system, called by adapt_method_stat()          */
/****************************************************************************/

static void solve(MESH *mesh)
{
  FUNCNAME("solve");
  static REAL       tol = 1.e-8;
  static int        miter = 1000, info = 2, icon = 1, restart = 0;
  static OEM_SOLVER solver = NoSolver;

  if (solver == NoSolver)
  {
    tol = 1.e-8;
    GET_PARAMETER(1, "solver", "%d", &solver);
    GET_PARAMETER(1, "solver tolerance", "%f", &tol);
    GET_PARAMETER(1, "solver precon", "%d", &icon);
    GET_PARAMETER(1, "solver max iteration", "%d", &miter);
    GET_PARAMETER(1, "solver info", "%d", &info);
    if (solver == GMRes)
      GET_PARAMETER(1, "solver restart", "%d", &restart);
  }
  oem_solve_s(matrix, f_h, nil, u_h, solver, tol, icon, restart, miter, info);

  graphics(mesh, u_h, nil);

  return;
}

/****************************************************************************/
/* Functions for error estimate:                                            */
/* estimate():   calculates error estimate via ellipt_est()                 */
/*               calculates exact error also (only for test purpose),       */
/*               called by adapt_method_stat()                              */
/* r():          calculates the lower order terms of the element residual   */
/*               on each element at the quadrature node iq of quad          */
/*               argument to ellipt_est() and called by ellipt_est()        */
/****************************************************************************/

static REAL r(const EL_INFO *el_info, const QUAD *quad, int iq, REAL uh_iq, 
              const REAL_D grd_uh_iq)
{
  REAL_D      x;
  coord_to_world(el_info, quad->lambda[iq], x);
  return(-f(x));
}

#define EOC(e,eo) log(eo/MAX(e,1.0e-15))/M_LN2
static REAL *(*rw_el_est)(EL *) = nil;          /* initialized by main()     */
static REAL (*get_el_est)(EL *) = nil;          /* initialized by main()     */

static REAL estimate(MESH *mesh, ADAPT_STAT *adapt)
{
  FUNCNAME("estimate");
  static int     degree, norm = -1;
  static REAL    C[3] = {1.0, 1.0, 0.0};
  static REAL    est, est_old = -1.0, err, err_old = -1.0, max_e;
  static const REAL_DD A = {{1.0,0.0}, {0.0,1.0}};
  static FLAGS r_flag = 0;  /* = (INIT_UH | INIT_GRD_UH),  if needed by r() */

  if (norm < 0)
  {
    norm = H1_NORM;
    GET_PARAMETER(1, "error norm", "%d", &norm);
    GET_PARAMETER(1, "estimator C0", "%f", &C[0]);
    GET_PARAMETER(1, "estimator C1", "%f", &C[1]);
    GET_PARAMETER(1, "estimator C2", "%f", &C[2]);
  }
  degree = 2*u_h->fe_space->bas_fcts->degree;
  est = ellipt_est(u_h, adapt, rw_el_est, nil, degree, norm, C, A, r, r_flag);

  MSG("estimate   = %.8le", est);
  if (est_old >= 0)
    print_msg(", EOC: %.2lf\n", EOC(est,est_old));
  else
    print_msg("\n");
  est_old = est;

  if (norm == L2_NORM)
    err = L2_err(u, u_h, nil, 0, nil, &max_e);
  else
    err = H1_err(grd_u, u_h, nil, 0, nil, &max_e);

  MSG("||u-uh||%s = %.8le", norm == L2_NORM ? "L2" : "H1", err);
  if (err_old >= 0)
    print_msg(", EOC: %.2lf\n", EOC(err,err_old));
  else
    print_msg("\n");
  err_old = err;
  MSG("||u-uh||%s/estimate = %.2lf\n", norm == L2_NORM ? "L2" : "H1",
      err/MAX(est,1.e-15));

  graphics(mesh, nil, get_el_est);
  return(adapt->err_sum);
}

/****************************************************************************/
/* main program                                                             */
/****************************************************************************/

int main(int argc, char **argv)
{
  FUNCNAME("main");
  const DEFAULT_APPL *appl;
  int                n_refine = 0, degree;
  ADAPT_STAT         *adapt = nil;
  char               filename[100];

/****************************************************************************/
/*  first of all, init parameters of the init file                          */
/****************************************************************************/

  init_parameters(0, "INIT/ellipt.dat");
  
/****************************************************************************/
/*  get a mesh, and read the macro triangulation from file                  */
/****************************************************************************/
  GET_PARAMETER(1, "polynomial degree", "%d", &degree);
  GET_PARAMETER(1, "macro file name", "%s", filename);
  GET_PARAMETER(1, "global refinements", "%d", &n_refine);

/***************************************************************************/
/*  init one finite element space with lagrange elements of order degree   */
/*  and one REAL on each element for the element indicator                 */
/***************************************************************************/

  appl = init_default_appl("ALBERT", degree, true);
  fe_space   = appl->fe_space;
  u_h        = appl->uh;  
  f_h        = appl->fh;
  matrix     = appl->A;
  rw_el_est  = appl->rw_el_estimate;
  get_el_est = appl->get_el_estimate;

  read_macro(fe_space->mesh, filename, nil);
  global_refine(fe_space->mesh, n_refine*DIM);
  graphics(fe_space->mesh, nil, nil);

/****************************************************************************/
/*  init adapt structure and start adaptive method                          */
/****************************************************************************/

  adapt = get_adapt_stat("ellipt", "adapt", 1, nil);
  adapt->estimate = estimate;
  adapt->get_el_est = get_el_est;
  adapt->build_after_coarsen = build;
  adapt->solve = solve;

  adapt_method_stat(fe_space->mesh, adapt);
  WAIT_REALLY;
  return(0);
}
