#include "info.h"

/*--------------------------------------------------------------------------*/
/*  Solving a linear system with "OrhtoDir"                                 */
/*  Classification: OrthoDir_d(A,C,A)                                       */
/*  Variant:        sigma= d.Ad/d.Ad_old                                    */
/*  Matrix:         symmetric and positve                                   */
/*  Workspace:      5*dim                                                   */
/*                                                                          */
/*  C-version of OFM-lib by Willy Doerfler                                  */
/*     http://www.mathematik.uni-kl.de/~doerfler/ofmlib.ps                  */
/*                                                                          */
/*  author: Kunibert G. Siebert                                             */
/*          Institut f"ur Angewandte Mathematik                             */
/*          Albert-Ludwigs-Universit"at Freiburg                            */
/*          Hermann-Herder-Str. 10                                          */
/*          79104 Freiburg                                                  */
/*          Germany                                                         */
/*                                                                          */
/*          http://www.mathematik.uni-freiburg.de/IAM/homepages/kunibert    */
/*                                                                          */
/* (c) by K.G. Siebert (2000)                                               */
/*--------------------------------------------------------------------------*/

int oem_odir(OEM_DATA *oem, int dim, const REAL *b, REAL *x)
{
  FUNCNAME("oem_odir");
  REAL  *Ad, *d, *d_old, *r, *h;
  REAL  res, old_res = -1.0;
  REAL  dad, dad_old, rd, alpha, gamma, sigma;
  int   iter;

  int   (*mv)(void *, int, const REAL *, REAL *) = oem->mat_vec;
  void  *mvd = oem->mat_vec_data;
  void  (*precon)(void *, int, REAL *) = oem->left_precon;
  void  *pd = oem->left_precon_data;

  const REAL TOL = 1.e-30;
  WORKSPACE  *ws = CHECK_WORKSPACE(6*dim, oem->ws);

/*--------------------------------------------------------------------------*/
/*---  partitioning of the workspace  --------------------------------------*/
/*--------------------------------------------------------------------------*/
  Ad    = ws->work;
  d     = Ad+dim;
  d_old = d+dim;
  r     = d_old+dim;
  h     = r+dim;

/*--------------------------------------------------------------------------*/
/*---  Initalization  ------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

  if (dnrm2(dim, b, 1) < TOL)
  {
    INFO(oem->info,2)("b == 0, x = 0 is the solution of the linear system\n");
    dset(dim, 0.0, x, 1);
    oem->initial_residual = oem->residual = 0.0;
    return(0);
  }

  (*mv)(mvd, dim, x, r);
  daxpy(dim, -1.0, b, 1, r, 1);
  dcopy(dim, r, 1, h, 1);
  if (precon) (*precon)(pd, dim, h);

  dcopy(dim, h, 1, d, 1);
  dset(dim, 0.0, d_old, 1);
  dad_old = 1.0;

/*--------------------------------------------------------------------------*/
/*---  check initial residual  ---------------------------------------------*/
/*--------------------------------------------------------------------------*/

  oem->initial_residual = res = dnrm2(dim, r, 1);

  START_INFO(oem);
  if (SOLVE_INFO(oem, 0, res, &old_res, ws))
    return(0);

/*--------------------------------------------------------------------------*/
/*---  Iteration  ----------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

  for (iter = 1; iter <= oem->max_iter; iter++)
  {
/*--------------------------------------------------------------------------*/
/*---  calculate r.d, A.d and d.A.d  ---------------------------------------*/
/*--------------------------------------------------------------------------*/
    rd = ddot(dim, r, 1, d, 1);
    (*mv)(mvd, dim, d, Ad);
    dad = ddot(dim, d, 1, Ad, 1);

    if (ABS(dad) < TOL)
    {
      BREAK_INFO(oem, "(Ad,d)_2 = 0", iter, res, &old_res, ws);
      return(iter);
    }

/*--------------------------------------------------------------------------*/
/*---   update x and calculate new r  --------------------------------------*/
/*--------------------------------------------------------------------------*/
    alpha = rd/dad;
    daxpy(dim, -alpha, d, 1, x, 1);
    daxpy(dim, -alpha, Ad, 1, r, 1);

    res = dnrm2(dim, r, 1);
    if (SOLVE_INFO(oem, iter, res, &old_res, ws))
      return(iter);

    dcopy(dim, Ad, 1, h, 1);
    if (precon) (*precon)(pd, dim, h);

/*--------------------------------------------------------------------------*/
/*---  calculate gamma and sigma  ------------------------------------------*/
/*--------------------------------------------------------------------------*/

      gamma = ddot(dim, h, 1, Ad, 1)/dad;
      sigma = dad/dad_old;

/*--------------------------------------------------------------------------*/
/*---  compute new d and save old values  ----------------------------------*/
/*--------------------------------------------------------------------------*/

      dswap(dim, d, 1, d_old, 1);
      dscal(dim, -sigma, d, 1);
      daxpy(dim, -gamma, d_old, 1, d, 1);
      dexpy(dim, h, 1, d, 1);
      dad_old= dad;
  }
  return(0);    /*  statement never reached!!!                              */
}
