#include "info.h"

/*--------------------------------------------------------------------------*/
/*  Solving a linear system with method of orthogonal directions            */
/*  Classification: OrthoDir_d(A^*A,C,A)                                    */
/*  Variant:        sigma= Ad.Ad/Ad_old.Ad_old                              */
/*  Matrix:         symmetric                                               */
/*  Workspace:      7*dim                                                   */
/*                                                                          */
/*  C-version of OFM-lib by Willy Doerfler                                  */
/*     http://www.mathematik.uni-kl.de/~doerfler/ofmlib.ps                  */
/*                                                                          */
/*  author: Kunibert G. Siebert                                             */
/*          Institut f"ur Angewandte Mathematik                             */
/*          Albert-Ludwigs-Universit"at Freiburg                            */
/*          Hermann-Herder-Str. 10                                          */
/*          79104 Freiburg                                                  */
/*          Germany                                                         */
/*                                                                          */
/*          http://www.mathematik.uni-freiburg.de/IAM/homepages/kunibert    */
/*                                                                          */
/* (c) by K.G. Siebert (2000)                                               */
/*--------------------------------------------------------------------------*/

int oem_ores(OEM_DATA *oem, int dim, const REAL *b, REAL *x)
{
  FUNCNAME("oem_ores");
  REAL  *Ad, *Ad_old, *Ah, *d, *d_old, *h, *r;
  REAL  alpha, gamma, sigma;
  REAL  daad, daad_old, rad, res, old_res = -1.0;
  int   iter, dim2;

  int   (*mv)(void *, int, const REAL *, REAL *) = oem->mat_vec;
  void  *mvd = oem->mat_vec_data;
  void  (*precon)(void *, int, REAL *) = oem->left_precon;
  void  *pd = oem->left_precon_data;

  WORKSPACE  *ws = CHECK_WORKSPACE(7*dim, oem->ws);

/*--------------------------------------------------------------------------*/
/*---  partitioning of the workspace  --------------------------------------*/
/*--------------------------------------------------------------------------*/

  d       = ws->work;
  Ad      = d+dim;
  d_old   = Ad+dim;
  Ad_old  = d_old+dim;
  Ah      = Ad_old+dim;
  r       = Ah+dim;
  h       = r+dim;

/*--------------------------------------------------------------------------*/
/*---  Initalization  ------------------------------------------------------*/
/*--------------------------------------------------------------------------*/

  dim2 = 2*dim; /* Important: sequence of vectors in memory for use of dim2!*/

  (*mv)(mvd, dim, x, r);
  daxpy(dim, -1.0, b, 1, r, 1);
  dcopy(dim, r, 1, h, 1);

  if (precon) (*precon)(pd, dim, h);

  dcopy(dim, h, 1, d, 1);
  dset(dim2, 0.0, d_old, 1);  /*--- d_old = 0 and Ad_old = 0 ---------------*/
  (*mv)(mvd, dim, d, Ad);
  daad_old = 1.0;

/*--------------------------------------------------------------------------*/
/*---  check initial residual  ---------------------------------------------*/
/*--------------------------------------------------------------------------*/
  
  oem->initial_residual = res = dnrm2(dim, r, 1);

  START_INFO(oem);
  if (SOLVE_INFO(oem, 0, res, &old_res, ws))
    return(0);

  for (iter = 1; iter <= oem->max_iter; iter++)
  {
/*--- compute r.Ad and d.A.A.d  ---------------------------------------------*/
    rad = ddot(dim, r, 1, Ad, 1);
    daad = ddot(dim, Ad, 1, Ad, 1);

    if (daad <= 1.e-30)
    {
      BREAK_INFO(oem, "(Ad,d)_2 = 0", iter, res, &old_res, ws);
      return(iter);
    }

/*--- update u and r  ------------------------------------------------------*/
    
    alpha= rad/daad;
    daxpy(dim, -alpha, d, 1, x, 1);
    daxpy(dim, -alpha, Ad, 1, r, 1);

    res = dnrm2(dim, r, 1);

    if (SOLVE_INFO(oem, iter, res, &old_res, ws))
      return(iter);

    dcopy(dim, Ad, 1, h, 1);
    if (precon) (*precon)(pd, dim, h);
    
    (*mv)(mvd, dim, h, Ah);

    gamma = ddot(dim, Ah, 1, Ad, 1)/daad;
    sigma = daad/daad_old;

/*--- compute new d.Ad, save old values!  ----------------------------------*/
    dswap(dim2, d, 1, d_old, 1);         /*--- d_old = d, Ad_old = Ad ------*/
    dscal(dim2, -sigma, d, 1);           /*--- d,Ad *= -sigma --------------*/
    daxpy(dim2, -gamma, d_old, 1, d, 1); /*--- d,Ad -= gamma*d_old,Ad_old --*/
    daxpy(dim, 1.0, h, 1, d, 1);
    daxpy(dim, 1.0, Ah, 1, Ad, 1);
    daad_old = daad;
  }
  
  return(0);  /*--- statement never reached!!!  ----------------------------*/
}
